//========================================================================================
//  XEiJ.java
//    en:Main class
//    ja:メインクラス
//  Copyright (C) 2003-2022 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

package xeij;

import java.awt.*;  //BasicStroke,BorderLayout,BoxLayout,Color,Component,Container,Cursor,Desktop,Dimension,Font,FlowLayout,Frame,Graphics,Graphics2D,GraphicsDevice,GraphicsEnvironment,GridLayout,Image,Insets,Paint,Point,Rectangle,RenderingHints,Robot,Shape,Stroke,TexturePaint,Toolkit
import java.awt.datatransfer.*;  //Clipboard,DataFlavor,FlavorEvent,FlavorListener,Transferable,UnsupportedFlavorException
import java.awt.dnd.*;  //DnDConstants,DropTarget,DropTargetAdapter,DropTargetDragEvent
import java.awt.event.*;  //ActionEvent,ActionListener,ComponentAdapter,ComponentEvent,ComponentListener,FocusAdapter,FocusEvent,FocusListener,InputEvent,KeyAdapter,KeyEvent,KeyListener,MouseAdapter,MouseEvent,MouseListener,MouseMotionAdapter,MouseWheelEvent,WindowAdapter,WindowEvent,WindowListener,WindowStateListener
import java.awt.font.*;  //FontRenderContext,LineMetrics,TextLayout
import java.awt.geom.*;  //AffineTransform,GeneralPath,Point2D,Rectangle2D
import java.awt.im.*;  //InputContext
import java.awt.image.*;  //BufferedImage,DataBuffer,DataBufferByte,DataBufferInt,IndexColorModel
import java.beans.*;  //PropertyChangeListener,XMLDecoder,XMLEncoder
import java.io.*;  //BufferedInputStream,BufferedOutputStream,BufferedReader,BufferedWriter,ByteArrayOutputStream,File,FileInputStream,FileNotFoundException,FileReader,InputStream,InputStreamReader,IOException,OutputStreamWriter,RandomAccessFile,UnsupportedEncodingException
import java.lang.*;  //Boolean,Character,Class,Comparable,Double,Exception,Float,IllegalArgumentException,Integer,InterruptedException,Long,Math,Number,Object,Runnable,SecurityException,String,StringBuilder,System
import java.math.*;  //BigDecimal,BigInteger,MathContext,RoundingMode
import java.net.*;  //MalformedURLException,URI,URL
import java.nio.*;  //ByteBuffer,ByteOrder
import java.nio.charset.*;  //Charset
import java.text.*;  //DecimalFormat,NumberFormat,ParseException
import java.util.*;  //ArrayList,Arrays,Calendar,GregorianCalendar,HashMap,Map,Map.Entry,LinkedList,TimeZone,Timer,TimerTask,TreeMap
import java.util.function.*;  //IntConsumer,IntSupplier
import java.util.regex.*;  //Matcher,Pattern
import java.util.zip.*;  //CRC32,Deflater,GZIPInputStream,GZIPOutputStream,ZipEntry,ZipInputStream
import javax.accessibility.*;  //AccessibleContext
import javax.imageio.*;  //ImageIO
import javax.imageio.stream.*;  //ImageOutputStream
import javax.sound.sampled.*;  //AudioFormat,AudioSystem,DataLine,LineUnavailableException,SourceDataLine
import javax.swing.*;  //AbstractButton,AbstractSpinnerModel,Box,ButtonGroup,DefaultListModel,ImageIcon,JButton,JCheckBox,JCheckBoxMenuItem,JComponent,JDialog,JFileChooser,JFrame,JLabel,JList,JMenu,JMenuBar,JMenuItem,JPanel,JRadioButton,JScrollPane,JSpinner,JTextArea,JTextField,JTextPane,JViewport,ScrollPaneConstants,SpinnerListModel,SpinnerNumberModel,SwingConstants,SwingUtilities,UIManager,UIDefaults,UnsupportedLookAndFeelException
import javax.swing.border.*;  //Border,CompoundBorder,EmptyBorder,EtchedBorder,LineBorder,TitledBorder
import javax.swing.event.*;  //CaretEvent,CaretListener,ChangeEvent,ChangeListener,DocumentEvent,DocumentListener,ListSelectionListener
import javax.swing.filechooser.*;  //FileFilter,FileNameExtensionFilter
import javax.swing.plaf.*;  //ColorUIResource,FontUIResource,IconUIResource,InsetsUIResource
import javax.swing.plaf.basic.*;  //BasicTextUI
import javax.swing.plaf.metal.*;  //MetalLookAndFeel,MetalTheme,OceanTheme
import javax.swing.text.*;  //AbstractDocument,BadLocationException,DefaultCaret,Document,DocumentFilter,JTextComponent,ParagraphView,Style,StyleConstants,StyleContext,StyledDocument
import netscape.javascript.*;  //JSException,JSObject。jfxrt.jarではなくplugin.jarを使うこと

public class XEiJ {

  //名前とバージョン
  public static final String PRG_TITLE = "XEiJ (X68000 Emulator in Java)";  //タイトル
  public static final String PRG_VERSION = "0.22.01.30";  //バージョン
  public static final String PRG_AUTHOR = "Makoto Kamada";  //作者
  public static final String PRG_WEBPAGE = "https://stdkmd.net/xeij/";  //ウェブページ

  public static final String PRG_JAVA_VENDOR = "Oracle Corporation";  //動作を確認しているJavaのベンダー
  public static final String PRG_JAVA_VERSION = "17.0.2";  //動作を確認しているJavaのバージョン
  public static final String PRG_OS_ARCH = "64";  //動作を確認しているアーキテクチャ

  //全体の設定
  //  bit0..3のテストにシフトを使う
  //    TEST_BIT_0_SHIFT ? a << 31 != 0 : (a & 1) != 0
  //    TEST_BIT_1_SHIFT ? a << 30 < 0 : (a & 2) != 0
  //    TEST_BIT_2_SHIFT ? a << 29 < 0 : (a & 4) != 0
  //    TEST_BIT_3_SHIFT ? a << 28 < 0 : (a & 8) != 0
  public static final boolean TEST_BIT_0_SHIFT = false;  //true=bit0のテストにシフトを使う
  public static final boolean TEST_BIT_1_SHIFT = false;  //true=bit1のテストにシフトを使う
  public static final boolean TEST_BIT_2_SHIFT = true;  //true=bit2のテストにシフトを使う
  public static final boolean TEST_BIT_3_SHIFT = true;  //true=bit3のテストにシフトを使う
  //  shortの飽和処理にキャストを使う
  //    x = SHORT_SATURATION_CAST ? (short) x == x ? x : x >> 31 ^ 32767 : Math.max (-32768, Math.min (32767, x));
  //    m = SHORT_SATURATION_CAST ? (short) m == m ? m : m >> 31 ^ 32767 : Math.max (-32768, Math.min (32767, m));
  //    l = SHORT_SATURATION_CAST ? (short) l == l ? l : l >> 31 ^ 32767 : Math.max (-32768, Math.min (32767, l));
  //    r = SHORT_SATURATION_CAST ? (short) r == r ? r : r >> 31 ^ 32767 : Math.max (-32768, Math.min (32767, r));
  //  または
  //    if (SHORT_SATURATION_CAST) {
  //      if ((short) x != x) {
  //        x = x >> 31 ^ 32767;
  //      }
  //    } else {
  //      if (x < -32768) {
  //        x = -32768;
  //      } else if (x > 32767) {
  //        x = 32767;
  //      }
  //    }
  public static final boolean SHORT_SATURATION_CAST = false;  //shortの飽和処理にキャストを使う

  //バイナリデータの埋め込み
  //  byte[]の場合
  //    Javaはbyteの定数配列をstatic final byte[] XXX={～}で直接記述しにくい
  //      bit7がセットされているデータをいちいち(byte)でキャストしなければならない
  //      初期化コードが巨大化してコンパイラを通らなくなる
  //    Stringに詰め込んで起動時にString.getBytes(Charset)を使ってbyte[]に展開する
  //    ISO-8859-1はすべてのJava実行環境で実装しなければならないことになっているので環境依存にはならない
  //    static final int[] XXX={～}で書いておいてPerlスクリプトで文字列に変換する
  //    final int[]をbyte[]に変更すると動作が遅くなる場合があることに注意する
  //    perl misc/itob.pl xeij/???.java XXX
  public static final Charset ISO_8859_1 = Charset.forName ("ISO-8859-1");
  static {
    if (false) {
      //ISO-8859-1が8bitバイナリデータを素通りさせるかどうかのテスト
      StringBuilder sb = new StringBuilder ();
      for (int i = 0; i < 256; i++) {
        sb.append ((char) i);
      }
      byte[] bb = sb.toString ().getBytes (ISO_8859_1);
      for (int i = 0; i < 256; i++) {
        System.out.printf ("%02x %02x %s\n", i, bb[i] & 255, i == (bb[i] & 255) ? "OK" : "ERROR");
      }
    }
  }
  //  char[]の場合
  //    byte[]の場合と同様にStringに詰め込んで起動時にString.toCharArray()を使ってchar[]に展開する
  //    static final int[] XXX={～}で書いておいてPerlスクリプトで文字列に変換する
  //    final int[]をchar[]に変更すると動作が遅くなる場合があることに注意する
  //    perl misc/itoc.pl xeij/???.java XXX



  //========================================================================================
  //$$PRG プログラムの入り口と出口

  //動作環境
  public static String prgJavaVendor;  //動作環境のJavaのベンダー
  public static String prgJavaVersion;  //動作環境のJavaのバージョン
  public static String prgOsArch;  //動作環境のアーキテクチャ
  public static String prgOsName;  //OSの名前
  public static boolean prgIsMac;  //true=Mac

  public static boolean prgCaseIgnored;  //true=ファイル名の大文字と小文字が区別されない

  public static boolean prgVerbose;

  public static String[] prgArgs;

  //main (args)
  //  開始
  public static void main (String[] args) {

    prgArgs = args;

    //起動する
    SwingUtilities.invokeLater (new Runnable () {
      @Override public void run () {
        new XEiJ ();
      }
    });

  }  //main(String[])

  //XEiJ ()
  //  コンストラクタ
  public XEiJ () {

    prgJavaVendor = System.getProperty ("java.vendor");
    prgJavaVersion = System.getProperty ("java.version");
    prgOsArch = System.getProperty ("os.arch");  //os.archはJREのアーキテクチャであってOSのアーキテクチャではないことに注意
    prgOsName = System.getProperty ("os.name");
    prgIsMac = 0 <= prgOsName.indexOf ("Mac");

    System.out.print ("\n" +
                      "-------------------------------------------------\n" +
                      PRG_TITLE + " version " + PRG_VERSION + "\n" +
                      "-------------------------------------------------\n");

    //初期化
    //  この段階でコンポーネントを参照してはならない
    //  メニューの初期値に必要な情報があればここで作っておく
    prgCaseIgnored = new File ("A").equals (new File ("a"));  //ファイル名の大文字と小文字が区別されるか
    fmtInit ();  //BCD 10進数変換
    Multilingual.mlnInit ();  //MLN 多言語化  sgsInit()よりも前

    System.out.println (Multilingual.mlnJapanese ? "java.vendor は " + prgJavaVendor + " です" :
                        "java.vendor is " + prgJavaVendor);
    System.out.println (Multilingual.mlnJapanese ? "java.version は " + prgJavaVersion + " です" :
                        "java.version is " + prgJavaVersion);
    System.out.println (Multilingual.mlnJapanese ? "os.arch は " + prgOsArch + " です" :
                        "os.arch is " + prgOsArch);
    System.out.println (Multilingual.mlnJapanese ? "os.name は " + prgOsName + " です" :
                        "os.name is " + prgOsName);

    rbtInit ();  //RBT ロボット

    Settings.sgsInit ();  //SGS 設定  mlnInit()よりも後、他の～Init()よりも前
    LnF.lnfInit ();  //Look&Feel

    RS232CTerminal.trmInit ();  //TRM ターミナルウインドウ

    CharacterCode.chrInit ();  //CHR 文字コード

    TickerQueue.tkqInit ();  //TKQ ティッカーキュー

    xt3Init ();  //XT3 Xellent30
    mdlInit ();  //MDL 機種。sgsInit()よりも後

    ROM.romInit ();  //ROM ROM。mpuInit()よりも前

    if (InstructionBreakPoint.IBP_ON) {
      InstructionBreakPoint.ibpInit ();  //IBP 命令ブレークポイント
    }
    if (DataBreakPoint.DBP_ON) {
      DataBreakPoint.dbpInit ();  //DBP データブレークポイント
    }
    busInit ();  //BUS バスコントローラ  ibpInit()とdbpInit()よりも後
    MainMemory.mmrInit ();  //MMR メインメモリ
    graInit ();  //GRA グラフィック画面
    txtInit ();  //TXT テキスト画面
    CRTC.crtInit ();  //CRT CRTコントローラ
    VideoController.vcnInit ();  //VCN ビデオコントローラ
    HD63450.dmaInit ();  //DMA DMAコントローラ
    svsInit ();  //SVS スーパーバイザ領域設定
    MC68901.mfpInit ();  //MFP MFP
    RP5C15.rtcInit ();  //RTC RTC
    sysInit ();  //SYS システムポート
    YM2151.opmInit ();  //OPM FM音源
    ADPCM.pcmInit ();  //PCM ADPCM音源
    FDC.fdcInit ();  //FDC FDコントローラ
    HDC.hdcInit ();  //HDC SASI HDコントローラ
    SPC.spcInit ();  //SPC SCSIプロトコルコントローラ
    Z8530.sccInit ();  //SCC SCC  rbtInit()よりも後
    IOInterrupt.ioiInit ();  //IOI I/O割り込み
    SpriteScreen.sprInit ();  //SPR スプライト画面
    bnkInit ();  //BNK バンクメモリ
    smrInit ();  //SMR SRAM

    PPI.ppiInit ();  //PPI PPI
    PrinterPort.prnInit ();  //PRN プリンタポート
    Indicator.indInit ();  //IND インジケータ

    SlowdownTest.sdtInit ();  //SDT 鈍化テスト
    Keyboard.kbdInit ();  //KBD キーボード  keydlyとkeyrepはSRAMが先なのでsmrInit()よりも後
    CONDevice.conInit ();  //CON CONデバイス制御
    musInit ();  //MUS マウス
    pnlInit ();  //PNL パネル
    frmInit ();  //FRM フレーム

    dbgInit ();  //DBG デバッガ共通コンポーネント
    RegisterList.drpInit ();  //DRP レジスタ
    DisassembleList.ddpInit ();  //DDP 逆アセンブルリスト
    MemoryDumpList.dmpInit ();  //DMP メモリダンプリスト
    LogicalSpaceMonitor.atwInit ();  //ATW アドレス変換ウインドウ
    PhysicalSpaceMonitor.paaInit ();  //PAA 物理アドレス空間ウインドウ
    DebugConsole.dgtInit ();  //DGT デバッグコンソール
    if (BranchLog.BLG_ON) {
      BranchLog.blgInit ();  //BLG 分岐ログ
    }
    if (ProgramFlowVisualizer.PFV_ON) {
      ProgramFlowVisualizer.pfvInit ();  //PFV プログラムフロービジュアライザ
    }
    if (RasterBreakPoint.RBP_ON) {
      RasterBreakPoint.rbpInit ();  //RBP ラスタブレークポイント
    }
    if (ScreenModeTest.SMT_ON) {
      ScreenModeTest.smtInit ();  //SMT 表示モードテスト
    }
    if (RootPointerList.RTL_ON) {
      RootPointerList.rtlInit ();  //RTL ルートポインタリスト
    }
    if (SpritePatternViewer.SPV_ON) {
      SpritePatternViewer.spvInit ();  //SPV スプライトパターンビュア
    }
    if (ATCMonitor.ACM_ON) {
      ATCMonitor.acmInit ();  //ACM アドレス変換キャッシュモニタ
    }

    SoundSource.sndInit ();  //SND サウンド
    FEFunction.fpkInit ();  //FPK FEファンクション
    mpuInit ();  //MPU MPU。romInit()よりも後
    MC68060.mmuInit ();  //MMU メモリ管理ユニット
    SoundMonitor.smnInit ();  //SMN 音声モニタ
    HFS.hfsInit ();  //HFS ホストファイルシステムインタフェイス

    //コンポーネントを作る
    //  他のコンポーネントを参照するときは順序に注意する
    Settings.sgsMakeMenu ();  //SGS 設定  mnbMakeMenu()よりも前
    mdlMakeMenu ();  //MDL 機種  mnbMakeMenu(),mpuMakeMenu()よりも前
    FDC.fdcMakeMenu ();  //FDC FDコントローラ  mnbMakeMenu()よりも前
    HDC.hdcMakeMenu ();  //HDC SASI HDコントローラ  mdlMakeMenu()よりも後,mnbMakeMenu()よりも前
    SPC.spcMakeMenu ();  //SPC SCSIプロトコルコントローラ  mdlMakeMenu()よりも後,mnbMakeMenu()よりも前
    mpuMakeMenu ();  //MPU MPU  mdlMakeMenu()よりも後,mnbMakeMenu()よりも前
    smrMakeMenu ();  //SMR SRAM  mnbMakeMenu()よりも前
    mnbMakeMenu ();  //MNB メニューバー
    pnlMake ();  //PNL パネル
    frmMake ();  //FRM フレーム
    clpMake ();  //CLP クリップボード
    dbgMakePopup ();  //DBG デバッガ共通コンポーネント

    //デバッグフラグを消し忘れないようにする
    final String flags = (
      "" +
      (EFPBox.CIR_DEBUG_TRACE ? " EFPBox.CIR_DEBUG_TRACE" : "") +
      (FDC.FDC_DEBUG_TRACE ? " FDC.FDC_DEBUG_TRACE" : "") +
      (FEFunction.FPK_DEBUG_TRACE ? " FEFunction.FPK_DEBUG_TRACE" : "") +
      (HD63450.DMA_DEBUG_TRACE != 0 ? " HD63450.DMA_DEBUG_TRACE" : "") +
      (HDC.HDC_DEBUG_TRACE ? " HDC.HDC_DEBUG_TRACE" : "") +
      (HDC.HDC_DEBUG_COMMAND ? " HDC.HDC_DEBUG_COMMAND" : "") +
      (HFS.HFS_DEBUG_TRACE ? " HFS.HFS_DEBUG_TRACE" : "") +
      (HFS.HFS_DEBUG_FILE_INFO ? " HFS.HFS_DEBUG_FILE_INFO" : "") +
      (HFS.HFS_COMMAND_TRACE ? " HFS.HFS_COMMAND_TRACE" : "") +
      (HFS.HFS_BUFFER_TRACE ? " HFS.HFS_BUFFER_TRACE" : "") +
      (IOInterrupt.IOI_DEBUG_TRACE ? " IOInterrupt.IOI_DEBUG_TRACE" : "") +
      (Keyboard.KBD_DEBUG_LED ? " Keyboard.KBD_DEBUG_LED" : "") +
      (M68kException.M6E_DEBUG_ERROR ? " M68kException.M6E_DEBUG_ERROR" : "") +
      (MC68060.MMU_DEBUG_COMMAND ? " MC68060.MMU_DEBUG_COMMAND" : "") +
      (MC68060.MMU_DEBUG_TRANSLATION ? " MC68060.MMU_DEBUG_TRANSLATION" : "") +
      (MC68060.MMU_NOT_ALLOCATE_CACHE ? " MC68060.MMU_NOT_ALLOCATE_CACHE" : "") +
      (PrinterPort.PRN_DEBUG_TRACE ? " PrinterPort.PRN_DEBUG_TRACE" : "") +
      (RP5C15.RTC_DEBUG_TRACE ? " RP5C15.RTC_DEBUG_TRACE" : "") +
      (SPC.SPC_DEBUG_TRACE ? " SPC.SPC_DEBUG_TRACE" : "") +
      (SPC.SPC_DEBUG_SCSIROM ? " SPC.SPC_DEBUG_SCSIROM" : "") +
      (Z8530.SCC_DEBUG_TRACE ? " Z8530.SCC_DEBUG_TRACE" : "")
      );
    if (!"".equals (flags)) {
      JOptionPane.showMessageDialog (null, "debug flags:" + flags);
    }

    //動作を開始する
    //  イベントリスナーを設定する
    //  タイマーを起動する
    tmrStart ();  //TMR タイマー

    Keyboard.kbdStart ();  //KBD キーボード
    musStart ();  //MUS マウス
    pnlStart ();  //PNL パネル
    frmStart ();  //FRM フレーム
    SoundSource.sndStart ();  //SND サウンド

    if (DataBreakPoint.DBP_ON) {
      DataBreakPoint.dbpStart ();  //DBP データブレークポイント
    }
    if (RasterBreakPoint.RBP_ON) {
      RasterBreakPoint.rbpStart ();  //RBP ラスタブレークポイント
    }
    if (ScreenModeTest.SMT_ON) {
      ScreenModeTest.smtStart ();  //SMT 表示モードテスト
    }
    SoundMonitor.smnStart ();  //SMN 音声モニタ
    RS232CTerminal.trmStart ();  //TRM ターミナルウインドウ
    PPI.ppiStart ();  //PPI PPI
    PrinterPort.prnStart ();  //PRN プリンタポート
    if (BranchLog.BLG_ON) {
      BranchLog.blgStart ();  //BLG 分岐ログ
    }
    if (ProgramFlowVisualizer.PFV_ON) {
      ProgramFlowVisualizer.pfvStart ();  //PFV プログラムフロービジュアライザ
    }
    RegisterList.drpStart ();  //DRP レジスタ
    DisassembleList.ddpStart ();  //DDP 逆アセンブルリスト
    MemoryDumpList.dmpStart ();  //DMP メモリダンプリスト
    LogicalSpaceMonitor.atwStart ();  //ATW アドレス変換ウインドウ
    PhysicalSpaceMonitor.paaStart ();  //PAA 物理アドレス空間ウインドウ
    DebugConsole.dgtStart ();  //DGT コンソール
    if (RootPointerList.RTL_ON) {
      RootPointerList.rtlStart ();  //RTL ルートポインタリスト
    }
    if (SpritePatternViewer.SPV_ON) {
      SpritePatternViewer.spvStart ();  //SPV スプライトパターンビュア
    }
    if (ATCMonitor.ACM_ON) {
      ATCMonitor.acmStart ();  //ACM アドレス変換キャッシュモニタ
    }

    if (Settings.sgsOpmtestOn) {  //OPMテスト
      OPMTest.otsTest ();
      prgTini ();
      return;
    }

    if (Settings.sgsSaveiconValue != null) {
      String[] a = Settings.sgsSaveiconValue.split (",");
      if (0 < a.length) {
        saveIcon (a[0], LnF.LNF_ICON_IMAGES);
        if (1 < a.length) {
          saveImage (LnF.LNF_ICON_IMAGE_16, a[1]);
          if (2 < a.length) {
            saveImage (LnF.LNF_ICON_IMAGE_32, a[2]);
            if (3 < a.length) {
              saveImage (LnF.LNF_ICON_IMAGE_48, a[3]);
            }
          }
        }
      }
      prgTini ();
      return;
    }

    //コアを起動する
    mpuReset (-1, -1);

  }  //コンストラクタ

  //prgTini ()
  //  プログラムの後始末
  public static void prgTini () {
    try {
      SoundSource.sndTini ();  //SND サウンド
      Keyboard.kbdTini ();  //KBD キーボード
      PPI.ppiTini ();  //PPI PPI
      FDC.fdcTini ();  //FDC FDコントローラ
      HDC.hdcTini ();  //HDC SASI HDコントローラ
      SPC.spcTini ();  //SPC SCSIプロトコルコントローラ
      HFS.hfsTini ();  //HFS ホストファイルシステムインタフェイス
      pnlTini ();  //PNL パネル
      bnkTini ();  //BNK バンクメモリ
      ROM.romTini ();  //ROM
      smrTini ();  //SMR SRAM
      tmrTini ();  //TMR タイマー
      Settings.sgsTini ();  //SGS 設定
    } catch (Exception e) {  //終了時に予期しないエラーが発生すると終了できなくなってしまうので、すべてのExceptionをcatchする
      e.printStackTrace ();
    }
    System.exit (0);
  }  //prgTini()

  //prgOpenJavaDialog ()
  //  Java実行環境の情報
  public static void prgOpenJavaDialog () {
    JOptionPane.showMessageDialog (
      frmFrame,
      ComponentFactory.createGridPanel (
        3, 5, "paddingLeft=6,paddingRight=6", "italic,right;left;left", "italic,center;colSpan=3,widen", "",
        null,  //(0,0)
        Multilingual.mlnJapanese ? "実行中" : "Running",  //(1,0)
        Multilingual.mlnJapanese ? "推奨" : "Recommended",  //(2,0)
        ComponentFactory.createHorizontalSeparator (),  //(0,1)
        Multilingual.mlnJapanese ? "Java のベンダー" : "Java Vendor",  //(0,2)
        prgJavaVendor,  //(1,2)
        PRG_JAVA_VENDOR,  //(2,2)
        Multilingual.mlnJapanese ? "Java のバージョン" : "Java Version",  //(0,3)
        prgJavaVersion,  //(1,3)
        PRG_JAVA_VERSION,  //(2,3)
        Multilingual.mlnJapanese ? "アーキテクチャ" : "Architecture",  //(0,4)
        (prgOsArch != null && prgOsArch.endsWith ("64") ? "64" : "32") + (Multilingual.mlnJapanese ? " ビット" : "-bit"),  //(1,4)
        PRG_OS_ARCH + (Multilingual.mlnJapanese ? " ビット" : "-bit")  //(2,4)
        ),
      Multilingual.mlnJapanese ? "Java 実行環境の情報" : "Java runtime environment information",
      JOptionPane.PLAIN_MESSAGE);
  }  //prgOpenJavaDialog()

  //prgOpenAboutDialog ()
  //  バージョン情報
  public static void prgOpenAboutDialog () {
    JOptionPane.showMessageDialog (
      frmFrame,
      ComponentFactory.createGridPanel (
        2, 4, "paddingLeft=6,paddingRight=6", "italic,right;left", "", "",
        Multilingual.mlnJapanese ? "タイトル" : "Title"  ,  //(0,0)
        PRG_TITLE,  //(1,0)
        Multilingual.mlnJapanese ? "バージョン" : "Version",  //(0,1)
        PRG_VERSION,  //(1,1)
        Multilingual.mlnJapanese ? "作者" : "Author" ,  //(0,2)
        PRG_AUTHOR,  //(1,2)
        Multilingual.mlnJapanese ? "ウェブページ" : "Webpage",  //(0,3)
        PRG_WEBPAGE  //(1,3)
        ),
      Multilingual.mlnJapanese ? "バージョン情報" : "Version information",
      JOptionPane.PLAIN_MESSAGE);
  }  //prgOpenAboutDialog()

  //prgOpenXEiJLicenseDialog ()
  //  XEiJ使用許諾条件ダイアログ
  public static void prgOpenXEiJLicenseDialog () {
    JOptionPane.showMessageDialog (
      frmFrame,
      ComponentFactory.createScrollTextPane (rscGetResourceText ("license_XEiJ.txt"), 550, 300),
      Multilingual.mlnJapanese ? "XEiJ 使用許諾条件" : "XEiJ License",
      JOptionPane.PLAIN_MESSAGE);
  }  //prgOpenXEiJLicenseDialog()

  //prgOpenSHARPLicenseDialog ()
  //  許諾条件.txtダイアログ
  public static void prgOpenSHARPLicenseDialog () {
    JOptionPane.showMessageDialog (
      frmFrame,
      ComponentFactory.createScrollTextPane (rscGetResourceText ("license_FSHARP.txt", "Shift_JIS"), 550, 300),
      Multilingual.mlnJapanese ? "無償公開された X68000 の基本ソフトウェア製品の許諾条件" : "License of the basic software products for X68000 that were distributed free of charge",
      JOptionPane.PLAIN_MESSAGE);
  }  //prgOpenSHARPLicenseDialog()

  //prgOpenMAMELicenseDialog ()
  //  license.txtダイアログ
  public static void prgOpenMAMELicenseDialog () {
    JOptionPane.showMessageDialog (
      frmFrame,
      ComponentFactory.createScrollTextPane (rscGetResourceText ("license_MAME.txt"), 550, 300),
      "MAME License",
      JOptionPane.PLAIN_MESSAGE);
  }  //prgOpenMAMELicenseDialog()

  //prgPrintClass (o)
  //  オブジェクトを表示する
  public static void prgPrintClass (Object o) {
    System.out.println (o.toString ());
    //スーパークラスを遡る
    try {
      Stack<Class<?>> s = new Stack<Class<?>> ();
      for (Class<?> c = o.getClass (); c != null; c = c.getSuperclass ()) {
        s.push (c);
      }
      for (int i = 0; !s.empty (); i++) {
        for (int j = 0; j < i; j++) {
          System.out.print ("  ");
        }
        System.out.println (s.pop ().getName ());
      }
    } catch (Exception e) {
    }
  }  //prgPrintClass(Object)

  //prgPrintStackTrace ()
  //  スタックトレースを表示する
  //  メソッドがどのような経路で呼び出されたか確認したいときに使う
  public static void prgPrintStackTrace () {
    Exception e = new Exception ();
    e.fillInStackTrace ();
    prgPrintStackTraceOf (e);
  }  //prgPrintStackTrace()
  public static void prgPrintStackTraceOf (Exception e) {
    //e.printStackTrace ();
    System.out.println ("------------------------------------------------");
    System.out.println (e.toString ());
    System.out.println ("\t" + e.getMessage ());
    for (StackTraceElement ste : e.getStackTrace ()) {
      System.out.println ("\tat " + ste.toString ());
    }
    System.out.println ("------------------------------------------------");
  }  //prgPrintStackTraceOf()

  //prgStopOnce ()
  //  1回目だけ停止する。2回目以降は何もしない
  //  特定の条件で止めて近くにブレークポイントを仕掛けたいときに使う
  public static boolean prgStopDone = false;
  public static void prgStopOnce () {
    if (!prgStopDone) {
      prgStopDone = true;
      mpuStop (null);
    }
  }  //prgStopOnce()



  //========================================================================================
  //$$TMR タイマ
  //  tmrTimerは1つだけ存在する
  //  1つのタイマにスケジュールされたタスクはオーバーラップしない
  //  固定遅延実行
  //    tmrTimer.schedule (task, delay, interval)
  //    次回の実行開始予定時刻=max(今回の実行終了時刻,今回の実行開始時刻+interval)
  //  固定頻度実行
  //    tmrTimer.scheduleAtFixedRate (task, delay, interval)
  //    次回の実行開始予定時刻=max(今回の実行終了時刻,初回の実行開始時刻+interval*今回までの実行回数)

  //時刻の周波数
  //  mpuClockTimeなどのカウンタが1秒間に進む数
  //  10^10のとき
  //    1周期   0.1nanosecond
  //    2^31-1  0.2second
  //    2^53-1  10day
  //    2^63-1  29year
  //  10^11のとき
  //    1周期   10picosecond
  //    2^31-1  21millisecond
  //    2^53-1  1day
  //    2^63-1  2.9year
  //  10^12のとき
  //    1周期   1picosecond
  //    2^31-1  2.1millisecond
  //    2^53-1  2.5hour
  //    2^63-1  3.5month
  public static final long TMR_FREQ = 1000000000000L;  //10^12Hz。1ps

  //メインタイマ
  public static final long TMR_DELAY = 10L;  //ms
  public static final long TMR_INTERVAL = 10L;  //ms

  //タイマ
  public static java.util.Timer tmrTimer;  //Timerだけだとjavax.swing.Timerと紛らわしい

  //tmrStart ()
  //  タイマを開始する
  public static void tmrStart () {
    tmrTimer = new java.util.Timer ();  //Timerだけだとjavax.swing.Timerと紛らわしい
  }  //tmrStart()

  //tmrTini ()
  //  タイマの後始末
  public static void tmrTini () {
    if (tmrTimer != null) {
      tmrTimer.cancel ();
    }
  }  //tmrTini()



  //========================================================================================
  //$$PNL パネル
  //
  //  固定倍率のとき
  //    パネルの最小サイズはスクリーンのサイズに固定倍率を掛けて切り上げた値
  //    スクリーンの表示サイズはスクリーンのサイズに固定倍率を掛けて丸めた値
  //  ウインドウに合わせるとき
  //    倍率は
  //      パネルの幅をスクリーンの幅で割った結果
  //      パネルの高さからキーボードの高さを引いてスクリーンの高さで割った結果
  //    のどちらか小さい方
  //
  //  スクリーンの大きさに固定倍率を掛けて丸めた値から倍率を逆算すると固定倍率よりも小さくなってしまう場合がある
  //
  //  全画面表示
  //    キーボードと合わせてパネルにちょうど入り切るようにスクリーンの拡大率と表示位置を計算する
  //  ウインドウに合わせる
  //    可能ならばユーザがパネルの大きさを変更できるようにする
  //    パネルの大きさが最小倍率で入り切らないとき
  //      プログラムがパネルの大きさを変更できるとき
  //        パネルの大きさを最小倍率でちょうど入り切る大きさに変更する
  //      パネルの中央に表示する
  //    パネルの大きさが最小倍率で入り切るとき
  //      パネルの大きさに合わせてスクリーンを拡大縮小する
  //  固定倍率
  //    可能ならばユーザがパネルの大きさを変更できないようにする
  //    プログラムがパネルの大きさを変更できるとき
  //      パネルの大きさを固定倍率でちょうど入り切る大きさに変更する
  //    スクリーンを拡大縮小してキーボードと一緒にパネルの中央に表示する


  //ビットマップのサイズ
  public static final int PNL_BM_OFFSET_BITS = 10;
  public static final int PNL_BM_WIDTH = 1 << PNL_BM_OFFSET_BITS;
  public static final int PNL_BM_HEIGHT = 1024;

  //水平方向の拡大率
  //  [CRTC.crtHRLCurr<<2|CRTC.crtHResoCurr]
  public static final float[] PNL_STRETCH_BASE = {
    8.0F / 3.0F,  //HRL=0,width=256
    4.0F / 3.0F,  //HRL=0,width=512
    1.0F,         //HRL=0,width=768
    1.0F,         //HRL=0,width=768
    4.0F,         //HRL=1,width=256
    2.0F,         //HRL=1,width=512
    3.0F / 2.0F,  //HRL=1,width=768
    3.0F / 2.0F,  //HRL=1,width=768
  };

  //サイズと位置
  public static int pnlScreenWidth;  //X68000から見た表示領域のサイズ。幅は常に8の倍数
  public static int pnlScreenHeight;
  public static float pnlStretchMode;  //水平方向の拡大率。PNL_STRETCH_BASE[CRTC.crtHRLCurr<<2|CRTC.crtHResoCurr]
  public static int pnlStretchWidth;  //ピクセルの縦横比に合わせて伸縮された表示領域の幅。Math.round((float)pnlScreenWidth*pnlStretchMode)
  public static int pnlZoomWidth;  //描画サイズ。pnlStretchWidth,pnlScreenHeightを同じ比率で拡大
  public static int pnlZoomHeight;
  public static int pnlZoomRatioOut;  //65536*pnlZoomHeight/pnlScreenHeight
  public static int pnlZoomRatioInX;  //65536*pnlScreenWidth/pnlZoomWidth
  public static int pnlZoomRatioInY;  //65536*pnlScreenHeight/pnlZoomHeight
  public static int pnlWidth;  //パネルのサイズ
  public static int pnlHeight;
  public static Dimension pnlSize;  //パネルの推奨サイズ。pnlWidth,pnlHeight
  public static int pnlScreenX;  //スクリーンの表示位置。pnlWidth-pnlStretchWidth>>1,pnlHeight-pnlScreenHeight>>1
  public static int pnlScreenY;
  public static int pnlKeyboardX;  //キーボードの表示位置。pnlUpdateArrangement()が設定する
  public static int pnlKeyboardY;
  public static int pnlMinimumWidth;  //パネルの最小サイズ
  public static int pnlMinimumHeight;
  public static int pnlGlobalX;  //画面上の表示位置
  public static int pnlGlobalY;

  //モード
  public static final boolean PNL_FILL_BACKGROUND = true;  //true=常に背景を塗り潰してから描画する
  public static boolean pnlFillBackgroundRequest;  //true=次回のpaintで背景を塗り潰す
  public static boolean pnlIsFullscreenSupported;  //true=全画面表示に移行できる
  public static boolean pnlFullscreenOn;  //true=全画面表示
  public static boolean pnlIsFitInWindowSupported;  //true=ウインドウに合わせられる
  public static boolean pnlFitInWindowOn;  //true=ウインドウに合わせる
  public static boolean pnlPrevFitInWindowOn;  //全画面表示にする前のpnlFitInWindowOn

  //補間アルゴリズム
  //  RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR  軽い
  //  RenderingHints.VALUE_INTERPOLATION_BILINEAR          頑張る
  //  RenderingHints.VALUE_INTERPOLATION_BICUBIC           重すぎる？
  public static Object pnlInterpolation;  //補間アルゴリズム

  //イメージとビットマップ
  public static BufferedImage pnlScreenImage;  //イメージ。サイズは1024x1024に固定
  public static int[] pnlBM;  //ビットマップ

  //パネル
  public static JPanel pnlPanel;  //パネル

  //メニュー
  public static int pnlFixedScale;
  public static SpinnerNumberModel pnlFixedModel;
  public static JSpinner pnlFixedSpinner;

  //pnlInit ()
  //  パネルのフィールドを初期化する
  public static void pnlInit () {

    //設定
    //  画面スケーリングモード
    switch (Settings.sgsGetString ("scaling").toLowerCase ()) {
    case "fullscreen":  //全画面表示
      pnlFullscreenOn = true;
      pnlFitInWindowOn = false;
      break;
    case "fitinwindow":  //ウインドウに合わせる
      pnlFullscreenOn = false;
      pnlFitInWindowOn = true;
      break;
    case "fixedscale":  //固定倍率
      pnlFullscreenOn = false;
      pnlFitInWindowOn = false;
      break;
    default:
      pnlFullscreenOn = false;
      pnlFitInWindowOn = true;
    }
    //  固定倍率
    pnlFixedScale = Math.max (10, Math.min (1000, Settings.sgsGetInt ("fixedscale")));
    //  補間アルゴリズム
    switch (Settings.sgsGetString ("interpolation").toLowerCase ()) {
    case "nearest":  //最近傍補間
      pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR;
      break;
    case "bilinear":  //線形補間
      pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_BILINEAR;
      break;
    case "bicubic":  //三次補間
      pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_BICUBIC;
      break;
    default:
      pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_BILINEAR;
    }

    pnlPrevFitInWindowOn = pnlFitInWindowOn;

    //サイズと位置
    pnlScreenWidth = 768;
    pnlScreenHeight = 512;
    pnlStretchMode = 1.0F;
    pnlStretchWidth = Math.round ((float) pnlScreenWidth * pnlStretchMode);
    pnlZoomWidth = pnlStretchWidth;
    pnlZoomHeight = pnlScreenHeight;
    pnlZoomRatioOut = (pnlZoomHeight << 16) / pnlScreenHeight;
    pnlZoomRatioInX = (pnlScreenWidth << 16) / pnlZoomWidth;
    pnlZoomRatioInY = (pnlScreenHeight << 16) / pnlZoomHeight;
    pnlWidth = Math.max (pnlZoomWidth, Keyboard.kbdWidth);
    pnlHeight = pnlZoomHeight + Keyboard.kbdHeight;
    pnlSize = new Dimension (pnlWidth, pnlHeight);
    pnlScreenX = pnlWidth - pnlStretchWidth >> 1;
    pnlScreenY = 0;
    pnlKeyboardX = pnlWidth - Keyboard.kbdWidth >> 1;
    pnlKeyboardY = pnlZoomHeight;
    pnlMinimumWidth = Math.max (256, Keyboard.kbdWidth);
    pnlMinimumHeight = 64 + Keyboard.kbdHeight;
    pnlGlobalX = 0;
    pnlGlobalY = 0;

    //モード
    if (!PNL_FILL_BACKGROUND) {
      pnlFillBackgroundRequest = true;
    }

    //イメージとビットマップ
    pnlScreenImage = new BufferedImage (PNL_BM_WIDTH, PNL_BM_HEIGHT, BufferedImage.TYPE_INT_RGB);
    pnlBM = ((DataBufferInt) pnlScreenImage.getRaster ().getDataBuffer ()).getData ();
    //Arrays.fill (bm, 0xff330000);

    //メニュー
    pnlFixedModel = new SpinnerNumberModel (pnlFixedScale, 10, 1000, 1);
    pnlFixedSpinner = ComponentFactory.createNumberSpinner (pnlFixedModel, 4, new ChangeListener () {
      @Override public void stateChanged (ChangeEvent ce) {
        mnbFixedScaleMenuItem.setSelected (true);
        pnlSetFitInWindowOn (false);
        pnlSetFullscreenOn (false);
        pnlUpdateArrangement ();
      }
    });

  }  //pnlInit()

  //  後始末
  public static void pnlTini () {
    //設定
    //  画面スケーリングモード
    Settings.sgsPutString ("scaling",
                           pnlFullscreenOn ? "fullscreen" :
                           pnlFitInWindowOn ? "fitinwindow" :
                           "fixedscale");
    //  固定倍率
    Settings.sgsPutInt ("fixedscale", pnlFixedScale);
    //  補間アルゴリズム
    Settings.sgsPutString ("interpolation",
                           pnlInterpolation == RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR ? "nearest" :
                           pnlInterpolation == RenderingHints.VALUE_INTERPOLATION_BILINEAR ? "bilinear" :
                           pnlInterpolation == RenderingHints.VALUE_INTERPOLATION_BICUBIC ? "bicubic" :
                           "bilinear");
  }

  //pnlMake ()
  //  パネルを作る
  public static void pnlMake () {

    //パネル
    pnlPanel = new JPanel () {
      @Override protected void paintComponent (Graphics g) {
        Graphics2D g2 = (Graphics2D) g;
        if (PNL_FILL_BACKGROUND || pnlFillBackgroundRequest) {  //背景を塗り潰す
          if (!PNL_FILL_BACKGROUND) {
            pnlFillBackgroundRequest = false;
          }
          g2.setColor (Color.black);
          g2.fillRect (0, 0, pnlWidth, pnlHeight);
        }
        g2.setRenderingHint (RenderingHints.KEY_INTERPOLATION, pnlInterpolation);
        g2.drawImage (pnlScreenImage,
                      pnlScreenX, pnlScreenY, pnlScreenX + pnlZoomWidth, pnlScreenY + pnlZoomHeight,
                      0, 0, pnlScreenWidth, pnlScreenHeight,
                      null);
        g2.drawImage (Keyboard.kbdImage, pnlKeyboardX, pnlKeyboardY, null);  //Graphics.drawImage()はimage==nullのとき何もしない
      }
      @Override protected void paintBorder (Graphics g) {
      }
      @Override protected void paintChildren (Graphics g) {
      }
    };
    pnlPanel.setBackground (Color.black);
    pnlPanel.setOpaque (true);
    pnlPanel.setPreferredSize (pnlSize);

    if (musCursorAvailable) {
      pnlPanel.setCursor (musCursorArray[1]);  //矢印カーソルを表示する
    }

  }  //pnlMake()

  //pnlStart ()
  //  パネルの動作を開始する
  //  イベントリスナーを設定する
  public static void pnlStart () {

    //コンポーネントリスナー
    ComponentFactory.addListener (
      pnlPanel,
      new ComponentAdapter () {
        @Override public void componentResized (ComponentEvent ce) {
          pnlWidth = pnlPanel.getWidth ();
          pnlHeight = pnlPanel.getHeight ();
          pnlUpdateArrangement ();
        }
      });

  }  //pnlStart()

  //pnlSetFullscreenOn (on)
  //  全画面表示を設定する
  public static void pnlSetFullscreenOn (boolean on) {
    if (pnlIsFullscreenSupported && pnlFullscreenOn != on) {
      if (on) {  //OFF→ON
        pnlFullscreenOn = true;
        mnbFullscreenMenuItem.setSelected (true);
        frmSetFullscreenOn (true);  //全画面表示に移行する
        pnlPrevFitInWindowOn = pnlFitInWindowOn;  //全画面表示に移行する前にウインドウに合わせていたか
        pnlSetFitInWindowOn (true);  //ウインドウに合わせる
      } else {  //ON→OFF
        pnlFullscreenOn = false;
        if (!pnlPrevFitInWindowOn) {  //全画面表示に移行する前はウインドウに合わせていなかった
          pnlSetFitInWindowOn (false);  //ウインドウに合わせない
        }
        frmSetFullscreenOn (false);  //全画面表示を解除する
      }
    }
  }  //pnlSetFullscreenOn(boolean)

  //pnlSetFitInWindowOn (mode)
  //  ウインドウに合わせるモードを設定する
  //  ウインドウに合わせるモードには全画面表示が含まれる
  public static void pnlSetFitInWindowOn (boolean on) {
    if (pnlIsFitInWindowSupported && pnlFitInWindowOn != on) {
      pnlFitInWindowOn = on;
      pnlUpdateArrangement ();  //スクリーンとキーボードの配置を再計算する
      if (!pnlFullscreenOn) {  //全画面表示でないとき
        if (on) {
          mnbFitInWindowMenuItem.setSelected (true);
        } else {
          mnbFixedScaleMenuItem.setSelected (true);
        }
      }
    }
  }  //pnlSetFitInWindowOn(boolean)

  //pnlUpdateArrangement ()
  //  スクリーンとキーボードの配置を再計算する
  //    リサイズ、最大化、全画面表示などの操作でパネルの大きさが変わったとき
  //    ウインドウに合わせるかどうかが変わったとき
  //    ウインドウに合わせないが固定倍率が変わったとき
  //    キーボードの有無または種類が変わったとき
  //    X68000の画面モードが変更されてスクリーンの大きさが変わったとき
  public static void pnlUpdateArrangement () {
    pnlFixedScale = pnlFixedModel.getNumber ().intValue ();  //固定倍率
    //スクリーンとキーボードの配置を決める
    if (!pnlFitInWindowOn) {  //固定倍率のとき
      //配置の計算
      //perl optdiv.pl 32768 100
      //  x/100==x*5243>>>19 (0<=x<=43698) [32768*5243==171802624]
      //pnlZoomWidth = (pnlStretchWidth * pnlFixedScale + 50) / 100;
      //pnlZoomHeight = (pnlScreenHeight * pnlFixedScale + 50) / 100;
      pnlZoomWidth = (pnlStretchWidth * pnlFixedScale + 50) * 5243 >>> 19;
      pnlZoomHeight = (pnlScreenHeight * pnlFixedScale + 50) * 5243 >>> 19;
      int width = Math.max (Math.max (256, pnlZoomWidth), Keyboard.kbdWidth);
      int height = Math.max (64, pnlZoomHeight) + Keyboard.kbdHeight;
      pnlScreenX = width - pnlZoomWidth >> 1;
      pnlScreenY = height - pnlZoomHeight - Keyboard.kbdHeight >> 1;
      if (pnlWidth != width || pnlHeight != height) {  //パネルの大きさが合っていないとき
        pnlWidth = width;
        pnlHeight = height;
        pnlMinimumWidth = width;  //固定倍率では使わないがウインドウに合わせるモードに移行したとき最小サイズが変化したことを検知できるようにする
        pnlMinimumHeight = height;
        pnlSize.setSize (width, height);
        //最小サイズと最大サイズ
        frmMinimumSize.setSize (width + frmMarginWidth, height + frmMarginHeight);
        frmFrame.setMinimumSize (frmMinimumSize);
        frmFrame.setMaximumSize (frmMinimumSize);
        frmFrame.setPreferredSize (frmMinimumSize);
        frmFrame.setResizable (false);
        pnlPanel.setMinimumSize (pnlSize);
        pnlPanel.setMaximumSize (pnlSize);
        pnlPanel.setPreferredSize (pnlSize);
        frmFrame.pack ();
      }
    } else {  //ウインドウに合わせるとき
      //配置の計算
      if (pnlWidth * pnlScreenHeight >= (pnlHeight - Keyboard.kbdHeight) * pnlStretchWidth) {  //ウインドウに合わせると上下に隙間ができないとき
        //パネルの下端にキーボード配置して残った部分にスクリーンを目一杯拡大する
        //    pnlScreenX                                          pnlScreenX            pnlScreenX
        //    |pnlZoomWidth|    |pnlZoomWidth|  |pnlZoomWidth|    |pnlZoomWidth|        |pnlZoomWidth|
        //  +-+------------+-+  +------------+  +------------+  +-+------------+-+  +---+------------+---+ --
        //  | |            | |  |            |  |            |  | |            | |  |   |            |   |
        //  | |   screen   | |  |   screen   |  |   screen   |  | |   screen   | |  |   |   screen   |   | pnlZoomHeight
        //  | |            | |  |            |  |            |  | |            | |  |   |            |   |
        //  | +-+--------+-+ |  +-+--------+-+  +------------+  +-+------------+-+  | +-+------------+-+ | -- pnlKeyboardY
        //  |   |keyboard|   |  | |keyboard| |  |  keyboard  |  |    keyboard    |  | |    keyboard    | | kbdHeight
        //  +---+--------+---+  +-+--------+-+  +------------+  +----------------+  +-+----------------+-+ --
        //      |                 |                             |    kbdWidth    |    |    kbdWidth    |
        //      pnlKeyboardX      pnlKeyboardX                                        pnlKeyboardX
        pnlZoomHeight = pnlHeight - Keyboard.kbdHeight;
        pnlZoomWidth = (pnlZoomHeight * pnlStretchWidth + (pnlScreenHeight >> 1)) / pnlScreenHeight;
        pnlScreenX = pnlWidth - pnlZoomWidth >> 1;
        pnlScreenY = 0;
      } else {  //ウインドウに合わせると上下に隙間ができるとき
        //左右が先につっかえたのだからスクリーンの幅がキーボードの幅よりも狭いということはない
        //  スクリーンの幅がキーボードの幅よりも狭かったらスクリーンの上と左右の両方に隙間があることになってしまうのでウインドウに合っていない
        //  |pnlZoomWidth|  |pnlZoomWidth|
        //  +------------+  +------------+
        //  |            |  |            |
        //  +------------+  +------------+ -- pnlScreenY
        //  |            |  |            |
        //  |   screen   |  |   screen   | pnlZoomHeight
        //  |            |  |            |
        //  +-+--------+-+  +------------+ -- pnlKeyboardY
        //  | |keyboard| |  |  keyboard  | kbdHeight
        //  | +--------+ |  +------------+ --
        //  |            |  |            |
        //  +------------+  +------------+
        //    |kbdWidth|    |  kbdWidth  |
        //    pnlKeyboardX
        pnlZoomWidth = pnlWidth;
        pnlZoomHeight = (pnlZoomWidth * pnlScreenHeight + (pnlStretchWidth >> 1)) / pnlStretchWidth;
        pnlScreenX = 0;
        pnlScreenY = pnlHeight - pnlZoomHeight - Keyboard.kbdHeight >> 1;
      }
      //最小サイズと最大サイズ
      int minimumWidth = Math.max (256, Keyboard.kbdWidth);
      int minimumHeight = 64 + Keyboard.kbdHeight;
      if (pnlMinimumWidth != minimumWidth || pnlMinimumHeight != minimumHeight) {  //最小サイズが変化した。ウインドウに合わせるモードに移行したかキーボードの有無または種類が変わった
        pnlMinimumWidth = minimumWidth;
        pnlMinimumHeight = minimumHeight;
        frmMinimumSize.setSize (minimumWidth + frmMarginWidth, minimumHeight + frmMarginHeight);
        frmFrame.setMinimumSize (frmMinimumSize);
        frmFrame.setMaximumSize (null);
        frmFrame.setResizable (true);
      }
    }
    pnlKeyboardX = pnlWidth - Keyboard.kbdWidth >> 1;
    pnlKeyboardY = pnlScreenY + pnlZoomHeight;
    pnlZoomRatioOut = (pnlZoomHeight << 16) / pnlScreenHeight;
    pnlZoomRatioInX = (pnlScreenWidth << 16) / pnlZoomWidth;
    pnlZoomRatioInY = (pnlScreenHeight << 16) / pnlZoomHeight;
    Z8530.scc0UpdateRatio ();
    if (!PNL_FILL_BACKGROUND) {
      pnlFillBackgroundRequest = true;
    }
    pnlPanel.repaint ();
  }  //pnlUpdateArrangement()



  //========================================================================================
  //$$RBT ロボット

  public static Robot rbtRobot;  //ロボット

  //rbtInit ()
  public static void rbtInit () {

    //ロボット
    rbtRobot = null;
    try {
      rbtRobot = new Robot ();  //プライマリスクリーンのみを対象とする
    } catch (Exception e) {
    }

  }  //rbtInit()



  //========================================================================================
  //$$MUS マウス
  //  パネルのマウスカーソルを変更する
  //    X68000のマウスカーソルが表示されているときはホストのマウスカーソルを消す
  //    X68000のマウスカーソルが表示されていないときはホストのマウスカーソルをX68000と同じサイズにする
  //  SCCから要求があったときマウスデータを作る
  //    シームレスモードのときは逆アクセラレーションを行う
  //      X68000のマウスカーソルがホストのマウスカーソルの真下に来るようにマウスの移動データを作る
  //  キーボードが操作されたときはキーボードにデータを渡す

  //逆アクセラレーション
  //  ROMのアクセラレーション処理のコード
  //    ;<d0.w:移動量
  //    ;>d0.w:移動量
  //    ;?d1
  //    accelerate:
  //        clr.w   -(sp)
  //        tst.w   d0
  //        bgt.s   1f
  //        addq.w  #1,(sp)
  //        neg.w   d0
  //    1:  move.w  d0,d1   ;  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25
  //        lsr.w   #3,d1   ;  0  0  0  0  0  0  0  0  1  1  1  1  1  1  1  1  2  2  2  2  2  2  2  2  3  3
  //        bne.s   2f
  //        move.w  #1,d1   ;  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1  2  2  2  2  2  2  2  2  3  3
  //    2:  mulu.w  d1,d0   ;  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 32 34 36 38 40 42 44 46 72 75
  //        move.w  d0,d1
  //        lsr.w   #2,d1   ;  0  0  0  0  1  1  1  1  2  2  2  2  3  3  3  3  8  8  9  9 10 10 11 11 18 18
  //        add.w   d1,d0   ;  0  1  2  3  5  6  7  8 10 11 12 13 15 16 17 18 40 42 45 47 50 52 55 57 90 93
  //        tst.w   (sp)+
  //        beq.s   3f
  //        neg.w   d0
  //    3:  rts
  //  アクセラレーションテーブル
  //    変位:移動距離
  //      0:   0   1:   1   2:   2   3:   3   4:   5   5:   6   6:   7   7:   8   8:  10   9:  11
  //     10:  12  11:  13  12:  15  13:  16  14:  17  15:  18  16:  40  17:  42  18:  45  19:  47
  //     20:  50  21:  52  22:  55  23:  57  24:  90  25:  93  26:  97  27: 101  28: 105  29: 108
  //     30: 112  31: 116  32: 160  33: 165  34: 170  35: 175  36: 180  37: 185  38: 190  39: 195
  //     40: 250  41: 256  42: 262  43: 268  44: 275  45: 281  46: 287  47: 293  48: 360  49: 367
  //     50: 375  51: 382  52: 390  53: 397  54: 405  55: 412  56: 490  57: 498  58: 507  59: 516
  //     60: 525  61: 533  62: 542  63: 551  64: 640  65: 650  66: 660  67: 670  68: 680  69: 690
  //     70: 700  71: 710  72: 810  73: 821  74: 832  75: 843  76: 855  77: 866  78: 877  79: 888
  //     80:1000  81:1012 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //    ~~~~~~~~~~~~~~~~~~ 82:1025  83:1037  84:1050  85:1062  86:1075  87:1087  88:1210  89:1223
  //     90:1237  91:1251  92:1265  93:1278  94:1292  95:1306  96:1440  97:1455  98:1470  99:1485
  //    100:1500 101:1515 102:1530 103:1545 104:1690 105:1706 106:1722 107:1738 108:1755 109:1771
  //    110:1787 111:1803 112:1960 113:1977 114:1995 115:2012 116:2030 117:2047 118:2065 119:2082
  //    120:2250 121:2268 122:2287 123:2306 124:2325 125:2343 126:2362 127:2381 128:2560
  //  逆アクセラレーション
  //    現在のマウスカーソルの位置とX68000のマウス座標の距離を超えない最大の移動距離をアクセラレーションテーブルから探してその変位に符号を付けて返す
  //    例えば距離が639ピクセルのとき63(639-551=88),23(88-57=31),15(31-18=13),11(13-13=0)の4回で移動が完了する
  public static final int[] MUS_DEACCELERATION_TABLE = new int[1025];  //逆アクセラレーションテーブル

  //マウスカーソル
  public static final String[][] MUS_CURSOR_PATTERN = {
    {
    },
    {
      "00.........",
      "010........",
      "0110.......",
      "01110......",
      "011110.....",
      "0111110....",
      "01111110...",
      "011111110..",
      "0111111110.",
      "01111100000",
      "0110110....",
      "010.0110...",
      "00..0110...",
      ".....0110..",
      ".....0110..",
      "......00...",
    },
  };

  //モード
  //  シームレス
  //    X68000のマウスカーソルが動作環境のマウスポインタを追いかける
  //  エクスクルーシブ
  //    動作環境のマウスポインタを独占して相対座標を利用できるようにする
  public static boolean musSeamlessOn;  //true=シームレス,false=エクスクルーシブ
  public static boolean musExclusiveStart;  //true=エクスクルーシブに切り替えた直後
  public static boolean musEdgeAccelerationOn;  //true=シームレスのとき縁部加速を行う
  public static boolean musHostsPixelUnitsOn;  //true=エクスクルーシブのときマウスはホストの画素単位で動く,false=X68000の画素単位で動く

  //マウスの状態
  public static boolean musButtonLeft;  //マウスの左ボタンの状態。true=押されている。マウスカーソルがスクリーン上になくても有効
  public static boolean musButtonRight;  //マウスの右ボタンの状態。true=押されている。マウスカーソルがスクリーン上になくても有効
  public static int musData;  //マウスのボタンのデータ。0=ボタンが押されていない,1=左ボタンだけ押されている,2=右ボタンだけ押されている,3=左右のボタンが押されている。マウスカーソルがスクリーン上にあるときだけセットされる
  public static int musExtraData;  //マウスのボタンの延長データ。0=ボタンが押されていない,1=左ボタンだけ押されている,2=右ボタンだけ押されている,3=左右のボタンが押されている。マウスカーソルがスクリーン上にあるときだけセットされる。ボタンが押されてからSCCのポートBが読み出されるまで押されたままになる
  public static int musPanelX;  //パネル座標
  public static int musPanelY;
  public static int musScreenX;  //スクリーン座標。スクリーン上にあるとは限らない
  public static int musScreenY;
  public static boolean musOnScreen;  //true=マウスカーソルがスクリーン上にある
  public static boolean musOnKeyboard;  //true=マウスカーソルがキーボード上にある

  //マウスカーソル
  public static boolean musCursorAvailable;  //true=カスタムカーソルを利用できる
  public static int musCursorNumber;  //表示されているマウスカーソルの番号
  public static Cursor[] musCursorArray;  //マウスカーソルの配列

  //マウスのボタン
  public static boolean musOutputButtonStatus;  //true=ボタンの状態を出力する
  public static int musNumberOfButtons;  //ボタンの数
  public static int musLastModifiersEx;  //前回のmodifiersEx
  public static boolean musCtrlRightOn;  //Ctrlキー＋左ボタンを右ボタンとみなす

  public static final int SHIFT_MASK          = 0x00000001;
  public static final int CTRL_MASK           = 0x00000002;
  public static final int META_MASK           = 0x00000004;
  public static final int BUTTON3_MASK        = 0x00000004;
  public static final int ALT_MASK            = 0x00000008;
  public static final int BUTTON2_MASK        = 0x00000008;
  public static final int BUTTON1_MASK        = 0x00000010;
  public static final int ALT_GRAPH_MASK      = 0x00000020;

  public static final int SHIFT_DOWN_MASK     = 0x00000040;
  public static final int CTRL_DOWN_MASK      = 0x00000080;
  public static final int META_DOWN_MASK      = 0x00000100;
  public static final int ALT_DOWN_MASK       = 0x00000200;
  public static final int BUTTON1_DOWN_MASK   = 0x00000400;
  public static final int BUTTON2_DOWN_MASK   = 0x00000800;
  public static final int BUTTON3_DOWN_MASK   = 0x00001000;
  public static final int ALT_GRAPH_DOWN_MASK = 0x00002000;
  public static final int BUTTON4_DOWN_MASK   = 0x00004000;
  public static final int BUTTON5_DOWN_MASK   = 0x00008000;

  //マウスリスナー、マウスモーションリスナー、マウスホイールリスナー
  //
  //  マウスの操作とイベント
  //                                  modifiers   modifiersEx
  //    button1  left        pressed  0x00000010  0x00000400
  //                        released  0x00000010  0x00000000
  //                         clicked  0x00000010  0x00000000
  //    button2  wheel       pressed  0x00000008  0x00000800
  //                        released  0x00000008  0x00000200  Altがセットされる。clickedは発生しない
  //    button3  right       pressed  0x00000004  0x00001000
  //                        released  0x00000004  0x00000100  Metaがセットされる
  //                         clicked  0x00000004  0x00000100  Metaがセットされる
  //    button4  back        pressed  0x00000000  0x00004000
  //                        released  0x00000000  0x00000000  clickedは発生しない
  //    button5  forward     pressed  0x00000000  0x00008000
  //                        released  0x00000000  0x00000000  clickedは発生しない
  //    wheel    pull     wheelmoved  preciseWheelRotation = 1.000000
  //                                  scrollAmount = 3
  //                                  scrollType = 0
  //                                  unitsToScroll = 3
  //                                  wheelRotation = 1
  //             push     wheelmoved  preciseWheelRotation = -1.000000
  //                                  scrollAmount = 3
  //                                  scrollType = 0
  //                                  unitsToScroll = -3
  //                                  wheelRotation = -1
  //             left     左スクロールではイベントが発生しない
  //             right    右スクロールではイベントが発生しない
  //    ダブルクリックはleftを2回クリックしているだけ
  //    ズームインとズームアウトはkeyCode=0x11=VK_CONTROL,keyLocation=2=KEY_LOCATION_LEFTをpressしてホイールを回している
  //
  //  Macbookのマルチタッチトラックパッドでクリックする方法
  //    左クリック
  //      「リンゴ」→「システム環境設定」→「トラックパッド」の「タップでクリック」をONにする
  //      1本指でタップする
  //    右クリック
  //      「リンゴ」→「システム環境設定」→「トラックパッド」の「副ボタンのクリック」をONにする
  //      「2本指でクリックまたはタップ」を選択する
  //      2本指でタップする
  //    メモ
  //      control+タップはCtrlがセットされた左クリックでしかない
  //        個々のアプリケーションがこの操作に右クリックと同じ機能を割り当てなければ右クリックの代わりにならない
  //      有効になっていない2本指タップやcontrol+タップでもmousePressedとmouseReleasedが発生する
  //        右クリックとして機能しないのはmousePressedとmouseReleasedの間隔が短すぎるためかも知れない
  //
  //        Button1  左ボタン
  //   Ctrl+Button1  右ボタン
  //        Button2  マウスモード切り替え。mouseClickedは発生しない。modifiersExにAltがセットされる。Alt+Button2は無効
  //        Button3  右ボタン。modifiersExにMetaがセットされるでMeta+Button3は無効
  //        Button4  停止・再開
  //        Button5  トレース1回
  //  Shift+Button5  ステップ1回
  //

  //musInit ()
  //  マウスを初期化する
  public static void musInit () {
    //musSeamlessOn = true;
    musExclusiveStart = false;
    //musEdgeAccelerationOn = false;
    //musHostsPixelUnitsOn = false;
    musButtonLeft = false;
    musButtonRight = false;
    musData = 0;
    musExtraData = 0;
    musPanelX = 0;
    musPanelY = 0;
    musScreenX = 0;
    musScreenY = 0;
    musOnScreen = false;
    musOnKeyboard = false;
    //逆アクセラレーション
    {
      int index = 0;
      for (int delta = 0; delta <= 81; delta++) {
        int next = delta + 1;
        if (next >= 8) {
          next *= next >> 3;
        }
        next += next >> 2;
        while (index < next) {  //delta==81のときnext==1025
          MUS_DEACCELERATION_TABLE[index++] = delta;
        }
      }
    }
    //マウスカーソル
    musCursorAvailable = false;
    try {
      Toolkit toolkit = Toolkit.getDefaultToolkit ();
      Dimension bestCursorSize = toolkit.getBestCursorSize (16, 16);
      int width = bestCursorSize.width;
      int height = bestCursorSize.height;
      if (width >= 16 && height >= 16) {  //カスタムカーソルを利用できるとき
        BufferedImage cursorImage = new BufferedImage (width, height, BufferedImage.TYPE_INT_ARGB);
        int[] cursorBitmap = ((DataBufferInt) cursorImage.getRaster ().getDataBuffer ()).getData ();
        Point point = new Point (0, 0);
        musCursorArray = new Cursor[MUS_CURSOR_PATTERN.length];
        for (int i = 0; i < MUS_CURSOR_PATTERN.length; i++) {
          String[] ss = MUS_CURSOR_PATTERN[i];
          int h = ss.length;
          for (int y = 0; y < height; y++) {
            String s = y < h ? ss[y] : "";
            int w = s.length ();
            for (int x = 0; x < width; x++) {
              char c = x < w ? s.charAt (x) : '.';
              cursorBitmap[x + width * y] = 0xff000000 & ('.' - c) | -(c & 1);
            }
          }
          musCursorArray[i] = toolkit.createCustomCursor (cursorImage, point, "XEiJ_" + i);
        }
        musCursorAvailable = true;
        musCursorNumber = 1;
      }
    } catch (Exception e) {
    }
    //マウスのボタン
    musOutputButtonStatus = false;
    musNumberOfButtons = -1;
    try {
      musNumberOfButtons = MouseInfo.getNumberOfButtons ();  //手元では5だった
    } catch (Exception e) {
    }
    musLastModifiersEx = 0;
    //musCtrlRightOn = false;
  }  //musInit()

  //musStart ()
  //  マウスの動作を開始する
  public static void musStart () {

    ComponentFactory.addListener (
      pnlPanel,
      new MouseAdapter () {
        @Override public void mouseClicked (MouseEvent me) {
          if (musOutputButtonStatus) {
            int modifiersEx = me.getModifiersEx ();
            if ((modifiersEx & BUTTON1_DOWN_MASK) != 0) {
              System.out.println (String.format ("mouse button %d/%d was clicked. (0x%08x)",
                                                 1, musNumberOfButtons, modifiersEx));
            }
            if ((modifiersEx & BUTTON2_DOWN_MASK) != 0) {
              System.out.println (String.format ("mouse button %d/%d was clicked. (0x%08x)",
                                                 2, musNumberOfButtons, modifiersEx));
            }
            if ((modifiersEx & BUTTON3_DOWN_MASK) != 0) {
              System.out.println (String.format ("mouse button %d/%d was clicked. (0x%08x)",
                                                 3, musNumberOfButtons, modifiersEx));
            }
            if ((modifiersEx & BUTTON4_DOWN_MASK) != 0) {
              System.out.println (String.format ("mouse button %d/%d was clicked. (0x%08x)",
                                                 4, musNumberOfButtons, modifiersEx));
            }
            if ((modifiersEx & BUTTON5_DOWN_MASK) != 0) {
              System.out.println (String.format ("mouse button %d/%d was clicked. (0x%08x)",
                                                 5, musNumberOfButtons, modifiersEx));
            }
            if ((modifiersEx & (BUTTON1_DOWN_MASK |
                                BUTTON2_DOWN_MASK |
                                BUTTON3_DOWN_MASK |
                                BUTTON4_DOWN_MASK |
                                BUTTON5_DOWN_MASK)) == 0) {
              System.out.println (String.format ("mouse button ?/%d was clicked. (0x%08x)",
                                                 musNumberOfButtons, modifiersEx));
            }
          }
          if (!pnlPanel.isFocusOwner ()) {
            pnlPanel.requestFocusInWindow ();
          }
        }
        //@Override public void mouseEntered (MouseEvent me) {
        //}
        @Override public void mouseExited (MouseEvent me) {
          if (musOnScreen) {  //スクリーンから出た
            musOnScreen = false;
          }
          if (musOnKeyboard) {  //キーボードから出た
            musOnKeyboard = false;
            if (Keyboard.kbdPointedIndex >= 0) {  //ポイントされているキーがある
              Keyboard.kbdHover (0, 0);  //ポイントを解除する
            }
          }
        }
        @Override public void mousePressed (MouseEvent me) {
          musPressedOrReleased (me, true);  //マウスのボタンが操作された
        }
        @Override public void mouseReleased (MouseEvent me) {
          musPressedOrReleased (me, false);  //マウスのボタンが操作された
        }
        @Override public void mouseDragged (MouseEvent me) {
          musDraggedOrMoved (me);  //マウスが動いた
        }
        @Override public void mouseMoved (MouseEvent me) {
          musDraggedOrMoved (me);  //マウスが動いた
        }
        @Override public void mouseWheelMoved (MouseWheelEvent mwe) {
          if (musOutputButtonStatus) {
            int modifiersEx = mwe.getModifiersEx ();
            double preciseWheelRotation = mwe.getPreciseWheelRotation ();
            int scrollAmount = mwe.getScrollAmount ();
            int scrollType = mwe.getScrollType ();
            int unitsToScroll = mwe.getUnitsToScroll ();
            int wheelRotation = mwe.getWheelRotation ();
            System.out.println (String.format ("mouse wheel moved (0x%08x)", modifiersEx));
            System.out.println (String.format ("  preciseWheelRotation = %f", preciseWheelRotation));
            System.out.println (String.format ("  scrollAmount = %d", scrollAmount));
            System.out.println (String.format ("  scrollType = %d", scrollType));
            System.out.println (String.format ("  unitsToScroll = %d", unitsToScroll));
            System.out.println (String.format ("  wheelRotation = %d", wheelRotation));
          }
          //ホイールでトレースしてみる
          int wheelRotation = mwe.getWheelRotation ();
          if (0 < wheelRotation) {  //pull
            mpuAdvance (mwe.isControlDown () ? 100 : mwe.isShiftDown () ? 10 : 1);  //トレース
          } else if (wheelRotation < 0) {  //push。高解像度マウスのとき端数が蓄積されるまで0が返る
            if (mwe.isAltDown ()) {  //Alt+push
              mpuStepUntilReturn ();  //ステップアンティルリターン
            } else {  //push
              mpuStep (mwe.isControlDown () ? 100 : mwe.isShiftDown () ? 10 : 1);  //ステップ
            }
          }
          //マウスホイールイベントを消費する
          mwe.consume ();
        }
      });

  }  //musStart()

  //musPressedOrReleased (me, pressed)
  //  マウスのボタンが操作された
  public static void musPressedOrReleased (MouseEvent me, boolean pressed) {
    //  InputEvent.getModifiers()は変化したものだけ返す
    //  InputEvent.getModifiersEx()は変化していないものも含めて現在の状態を返す
    int modifiersEx = me.getModifiersEx ();
    int pressedMask = ~musLastModifiersEx & modifiersEx;  //0→1
    int releasedMask = musLastModifiersEx & ~modifiersEx;  //1→0
    musLastModifiersEx = modifiersEx;
    if (musOutputButtonStatus) {
      if ((pressedMask & BUTTON1_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was pressed. (0x%08x)",
                                           1, musNumberOfButtons, modifiersEx));
      } else if ((releasedMask & BUTTON1_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was released. (0x%08x)",
                                           1, musNumberOfButtons, modifiersEx));
      }
      if ((pressedMask & BUTTON2_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was pressed. (0x%08x)",
                                           2, musNumberOfButtons, modifiersEx));
      } else if ((releasedMask & BUTTON2_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was released. (0x%08x)",
                                           2, musNumberOfButtons, modifiersEx));
      }
      if ((pressedMask & BUTTON3_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was pressed. (0x%08x)",
                                           3, musNumberOfButtons, modifiersEx));
      } else if ((releasedMask & BUTTON3_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was released. (0x%08x)",
                                           3, musNumberOfButtons, modifiersEx));
      }
      if ((pressedMask & BUTTON4_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was pressed. (0x%08x)",
                                           4, musNumberOfButtons, modifiersEx));
      } else if ((releasedMask & BUTTON4_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was released. (0x%08x)",
                                           4, musNumberOfButtons, modifiersEx));
      }
      if ((pressedMask & BUTTON5_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was pressed. (0x%08x)",
                                           5, musNumberOfButtons, modifiersEx));
      } else if ((releasedMask & BUTTON5_DOWN_MASK) != 0) {
        System.out.println (String.format ("mouse button %d/%d was released. (0x%08x)",
                                           5, musNumberOfButtons, modifiersEx));
      }
      if (((pressedMask | releasedMask) & (BUTTON1_DOWN_MASK |
                                           BUTTON2_DOWN_MASK |
                                           BUTTON3_DOWN_MASK |
                                           BUTTON4_DOWN_MASK |
                                           BUTTON5_DOWN_MASK)) == 0) {
        System.out.println (String.format ("mouse button ?/%d was %s. (0x%08x)",
                                           musNumberOfButtons, pressed ? "pressed" : "released", modifiersEx));
      }
    }
    if (musCtrlRightOn &&  //Ctrlキー＋左ボタンを右ボタンとみなす
        (modifiersEx & MouseEvent.CTRL_DOWN_MASK) != 0) {  //Ctrlキーが押されている
      if ((pressedMask & MouseEvent.BUTTON1_DOWN_MASK) != 0) {  //左ボタンが押された
        pressedMask = ((pressedMask & ~MouseEvent.BUTTON1_DOWN_MASK) |  //左ボタンをクリアして
                       MouseEvent.BUTTON3_DOWN_MASK);  //右ボタンをセットする
      }
      if ((releasedMask & MouseEvent.BUTTON1_DOWN_MASK) != 0) {  //左ボタンが離された
        releasedMask = ((releasedMask & ~MouseEvent.BUTTON1_DOWN_MASK) |  //左ボタンをクリアして
                        MouseEvent.BUTTON3_DOWN_MASK);  //右ボタンをセットする
      }
    }
    if ((pressedMask & MouseEvent.BUTTON1_DOWN_MASK) != 0) {  //左ボタンが押された
      musButtonLeft = true;
      if (musOnScreen) {  //マウスデータはスクリーン上で押されたときだけON
        musData |= 1;
        musExtraData |= 1;
      } else {
        musData &= ~1;
      }
    } else if ((releasedMask & MouseEvent.BUTTON1_DOWN_MASK) != 0) {  //左ボタンが離された
      musButtonLeft = false;
      musData &= ~1;
    }
    if (musNumberOfButtons < 3) {  //2ボタンマウスのとき
      if ((pressedMask & (MouseEvent.BUTTON2_DOWN_MASK |
                          MouseEvent.BUTTON3_DOWN_MASK)) != 0) {  //右ボタンが押された
        if ((modifiersEx & MouseEvent.ALT_DOWN_MASK) != 0) {  //Altキーが押されている
          musSetSeamlessOn (!musSeamlessOn);  //シームレス/エクスクルーシブを切り替える
        } else {  //Altキーが押されていない
          musButtonRight = true;
          if (musOnScreen) {  //マウスデータはスクリーン上で押されたときだけON
            musData |= 2;
            musExtraData |= 2;
          } else {
            musData &= ~2;
          }
        }
      } else if ((releasedMask & (MouseEvent.BUTTON2_DOWN_MASK |
                                  MouseEvent.BUTTON3_DOWN_MASK)) != 0) {  //右ボタンが離された
        musButtonRight = false;
        musData &= ~2;
      }
    } else {  //3ボタンマウスのとき
      if ((pressedMask & MouseEvent.BUTTON2_DOWN_MASK) != 0) {  //中ボタンが押された
        musSetSeamlessOn (!musSeamlessOn);  //シームレス/エクスクルーシブを切り替える
      }
      if ((pressedMask & MouseEvent.BUTTON3_DOWN_MASK) != 0) {  //右ボタンが押された
        musButtonRight = true;
        if (musOnScreen) {  //マウスデータはスクリーン上で押されたときだけON
          musData |= 2;
          musExtraData |= 2;
        } else {
          musData &= ~2;
        }
      } else if ((releasedMask & MouseEvent.BUTTON3_DOWN_MASK) != 0) {  //右ボタンが離された
        musButtonRight = false;
        musData &= ~2;
      }
      if (5 <= musNumberOfButtons) {  //5ボタンマウスのとき
        if ((pressedMask & BUTTON4_DOWN_MASK) != 0) {  //ボタン4が押された
          if (mpuTask == null) {  //Run
            mpuStart ();
          } else {  //Stop
            if (RootPointerList.RTL_ON) {
              if (RootPointerList.rtlCurrentSupervisorTaskIsStoppable ||
                  RootPointerList.rtlCurrentUserTaskIsStoppable) {
                mpuStop (null);
              }
            } else {
              mpuStop (null);
            }
          }
        }
/*
        if ((pressedMask & BUTTON5_DOWN_MASK) != 0) {  //ボタン5が押された
          kbdDoCapture ();
        }
*/
      }
    }
    musDraggedOrMoved (me);
  }  //musPressedOrReleased(MouseEvent,boolean)

  //musDraggedOrMoved (me)
  //  マウスが動いた
  public static void musDraggedOrMoved (MouseEvent me) {
    int x = musPanelX = me.getX ();
    int y = musPanelY = me.getY ();
    if (pnlScreenX <= x && x < pnlScreenX + pnlZoomWidth &&
        pnlScreenY <= y && y < pnlScreenY + pnlZoomHeight) {  //スクリーン上にある
      musOnScreen = true;  //スクリーンに入った
      musScreenX = (x - pnlScreenX) * pnlZoomRatioInX >> 16;  //端数は切り捨てる
      musScreenY = (y - pnlScreenY) * pnlZoomRatioInY >> 16;
    } else {  //スクリーン上にない
      musOnScreen = false;  //スクリーンから出た
    }
    if (pnlKeyboardX <= x && x < pnlKeyboardX + Keyboard.kbdWidth &&
        pnlKeyboardY <= y && y < pnlKeyboardY + Keyboard.kbdHeight) {  //キーボード上にある
      musOnKeyboard = true;  //キーボードに入った
      Keyboard.kbdHover (x - pnlKeyboardX, y - pnlKeyboardY);
    } else {  //キーボード上にない
      if (musOnKeyboard) {  //キーボードから出た
        musOnKeyboard = false;
        if (Keyboard.kbdPointedIndex >= 0) {  //ポイントされているキーがあった
          Keyboard.kbdHover (0, 0);  //ポイントを解除する
        }
      }
    }
  }  //musDraggedOrMoved(MouseEvent)

  //musSetSeamlessOn (on)
  //  シームレス/エクスクルーシブを切り替える
  public static void musSetSeamlessOn (boolean on) {
    if (rbtRobot == null) {  //ロボットが使えないときは切り替えない(シームレスのみ)
      return;
    }
    if (musSeamlessOn != on) {
      musSeamlessOn = on;
      if (on) {  //エクスクルーシブ→シームレス
        musShow ();
        //ホストのマウスカーソルをX68000のマウスカーソルの位置に移動させる
        int x, y;
        if (XEiJ.mpuCoreType < 6) {  //MMUなし
          if (Z8530.SCC_FSX_MOUSE &&
              Z8530.sccFSXMouseHook != 0 &&  //FSX.Xが常駐している
              MainMemory.mmrRls (0x0938) == Z8530.sccFSXMouseHook) {  //マウス受信データ処理ルーチンがFSX.Xを指している。SX-Windowが動作中
            int xy = MainMemory.mmrRls (Z8530.sccFSXMouseWork + 0x0a);
            x = (xy >> 16) - CRTC.crtR10TxXPort;  //SX-Windowのマウスカーソルの見かけのX座標
            y = (short) xy - CRTC.crtR11TxYPort;  //SX-Windowのマウスカーソルの見かけのY座標
          } else {  //SX-Windowが動作中ではない
            int xy = MainMemory.mmrRls (0x0ace);
            x = xy >> 16;  //IOCSのマウスカーソルのX座標
            y = (short) xy;  //IOCSのマウスカーソルのY座標
          }
        } else {  //MMUあり
          if (Z8530.SCC_FSX_MOUSE &&
              Z8530.sccFSXMouseHook != 0 &&  //FSX.Xが常駐している
              MC68060.mmuPeekLongData (0x0938, 1) == Z8530.sccFSXMouseHook) {  //マウス受信データ処理ルーチンがFSX.Xを指している。SX-Windowが動作中
            int xy = MC68060.mmuPeekLongData (Z8530.sccFSXMouseWork + 0x0a, 1);
            x = (xy >> 16) - CRTC.crtR10TxXPort;  //SX-Windowのマウスカーソルの見かけのX座標
            y = (short) xy - CRTC.crtR11TxYPort;  //SX-Windowのマウスカーソルの見かけのY座標
          } else {  //SX-Windowが動作中ではない
            int xy = MC68060.mmuPeekLongData (0x0ace, 1);
            x = xy >> 16;  //IOCSのマウスカーソルのX座標
            y = (short) xy;  //IOCSのマウスカーソルのY座標
          }
        }
        XEiJ.rbtRobot.mouseMove (x * pnlZoomWidth / pnlScreenWidth + pnlScreenX + pnlGlobalX,
                                 y * pnlZoomHeight / pnlScreenHeight + pnlScreenY + pnlGlobalY);
      } else {  //シームレス→エクスクルーシブ
        musHide ();
        Point point = pnlPanel.getLocationOnScreen ();
        pnlGlobalX = point.x;
        pnlGlobalY = point.y;
        musExclusiveStart = true;  //エクスクルーシブに切り替えた直後
      }
    }
    if (mnbSeamlessMenuItem.isSelected () != on) {
      mnbSeamlessMenuItem.setSelected (on);
    }
  }  //musSetSeamlessOn(boolean)

  //musHide ()
  //  マウスカーソルを消す
  public static void musHide () {
    if (musCursorNumber != 0 && musCursorAvailable) {
      musCursorNumber = 0;
      pnlPanel.setCursor (musCursorArray[0]);
    }
  }  //musHide()

  //musShow ()
  //  マウスカーソルを表示する
  public static void musShow () {
    if (musCursorNumber != 1 && musCursorAvailable) {
      musCursorNumber = 1;
      pnlPanel.setCursor (musCursorArray[1]);
    }
  }  //musShow()

  //musSetEdgeAccelerationOn (on)
  //  縁部加速
  public static void musSetEdgeAccelerationOn (boolean on) {
    musEdgeAccelerationOn = on;
  }  //musSetEdgeAccelerationOn(boolean)

  //musSetHostsPixelUnitsOn (on)
  //  true=エクスクルーシブのときマウスはホストの画素単位で動く,false=X68000の画素単位で動く
  public static void musSetHostsPixelUnitsOn (boolean on) {
    musHostsPixelUnitsOn = on;
    Z8530.scc0UpdateRatio ();
  }  //musSetHostsPixelUnitsOn(boolean)



  //========================================================================================
  //$$MNB メニューバー
  //  メニューバーの高さは23px

  public static final int MNB_MODIFIERS = InputEvent.ALT_DOWN_MASK | InputEvent.SHIFT_DOWN_MASK;  //アクセラレータのモディファイヤ

  //メニューバー
  public static JMenuBar mnbMenuBar;  //メニューバー

  //メニュー
  public static JMenu mnbFileMenu;  //ファイル
  public static JMenu mnbDisplayMenu;  //画面
  public static JMenu mnbSoundMenu;  //音声
  public static JMenu mnbInputMenu;  //入力
  public static JMenu mnbConfigMenu;  //設定
  public static JMenu mnbLanguageMenu;  //言語

  //メニューアイテム
  //  チェックボックスなどの変更内容はアクションイベントから取り出せるので個々のメニューアイテムに名前を付ける必要はない
  //  メニュー以外の方法で変更できるアイテムと一時的に変更できなくなるアイテムに名前をつけておく
  //  最初から最後まで選択できないメニューアイテムはメニューバーを作る時点で無効化するか、表示しない
  public static JMenuItem mnbQuitMenuItem;  //終了
  public static JRadioButtonMenuItem mnbFullscreenMenuItem;  //全画面表示
  public static JRadioButtonMenuItem mnbFitInWindowMenuItem;  //ウインドウに合わせる
  public static JRadioButtonMenuItem mnbFixedScaleMenuItem;  //固定倍率
  public static JCheckBoxMenuItem mnbPlayMenuItem;  //音声出力
  public static JMenuItem mnbPasteMenuItem;  //貼り付け
  public static JCheckBoxMenuItem mnbSeamlessMenuItem;  //シームレスマウス
  public static JRadioButtonMenuItem mnbStandardKeyboardMenuItem;  //標準キーボード
  public static JRadioButtonMenuItem mnbCompactKeyboardMenuItem;  //コンパクトキーボード
  public static JRadioButtonMenuItem mnbNoKeyboardMenuItem;  //キーボードなし
  public static JLabel mnbVolumeLabel;  //音量

  //mnbMakeMenu ()
  //  メニューバーを作る
  //  メニューバーの幅は狭くしたいがメニューの幅が狭すぎると隣のメニューに流れやすくなって操作しにくいのでメニューの数を必要最小限にする
  public static void mnbMakeMenu () {

    //アクションリスナー
    ActionListener listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {

          //ファイルメニュー
        case "Quit":  //終了
          prgTini ();
          break;

          //画面メニュー
        case "Full screen":  //全画面表示
          pnlSetFullscreenOn (true);
          break;
        case "Fit in window":  //ウインドウに合わせる
          if (pnlFullscreenOn) {  //全画面表示になっているとき
            pnlPrevFitInWindowOn = true;
            pnlSetFullscreenOn (false);
          } else {  //全画面表示になっていないとき
            pnlSetFitInWindowOn (true);
          }
          break;
        case "Fixed scale":  //固定倍率
          if (pnlFullscreenOn) {  //全画面表示になっているとき
            pnlPrevFitInWindowOn = false;
            pnlSetFullscreenOn (false);
          } else {  //全画面表示になっていないとき
            pnlSetFitInWindowOn (false);
          }
          break;
        case "50%":
        case "75%":
        case "100%":
        case "150%":
        case "200%":
          pnlFixedModel.setValue (Integer.valueOf (Integer.parseInt (command.substring (0, command.length () - 1))));  //pnlFixedSpinnerのChangeEventが発動する
          break;
        case "Nearest neighbor":  //最近傍補間
          pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR;
          pnlPanel.repaint ();
          break;
        case "Bilinear":  //線形補間
          pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_BILINEAR;
          pnlPanel.repaint ();
          break;
        case "Bicubic":  //三次補間
          pnlInterpolation = RenderingHints.VALUE_INTERPOLATION_BICUBIC;
          pnlPanel.repaint ();
          break;
        case "Draw all changed pictures":  //変化した画像をすべて描画する
          if (CRTC.CRT_ENABLE_INTERMITTENT) {
            CRTC.crtIntermittentInterval = 0;
          }
          break;
        case "Draw a changed picture once every two times":  //変化した画像を 2 回に 1 回描画する
          if (CRTC.CRT_ENABLE_INTERMITTENT) {
            CRTC.crtIntermittentInterval = 1;
          }
          break;
        case "Draw a changed picture once every three times":  //変化した画像を 3 回に 1 回描画する
          if (CRTC.CRT_ENABLE_INTERMITTENT) {
            CRTC.crtIntermittentInterval = 2;
          }
          break;
        case "Draw a changed picture once every four times":  //変化した画像を 4 回に 1 回描画する
          if (CRTC.CRT_ENABLE_INTERMITTENT) {
            CRTC.crtIntermittentInterval = 3;
          }
          break;
        case "Draw a changed picture once every five times":  //変化した画像を 5 回に 1 回描画する
          if (CRTC.CRT_ENABLE_INTERMITTENT) {
            CRTC.crtIntermittentInterval = 4;
          }
          break;
        case "Extended graphic screen":  //拡張グラフィック画面
          CRTC.crtExtendedGraphicRequest = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Sprite pattern viewer":  //スプライトパターンビュア
          if (SpritePatternViewer.SPV_ON) {
            SpritePatternViewer.spvOpen ();
          }
          break;
        case "Screen mode test":  //表示モードテスト
          if (ScreenModeTest.SMT_ON) {
            ScreenModeTest.smtOpen ();
          }
          break;

          //音声メニュー
        case "Play":  //音声出力
          SoundSource.sndSetPlayOn (((JCheckBoxMenuItem) source).isSelected ());
          break;
        case "OPM output":  //OPM出力
          YM2151.opmSetOutputOn (((JCheckBoxMenuItem) source).isSelected ());
          break;
        case "PCM output":  //PCM出力
          ADPCM.pcmSetOutputOn (((JCheckBoxMenuItem) source).isSelected ());
          break;
        case "Sound thinning":  //音声 間引き
          SoundSource.sndRateConverter = SoundSource.SND_CHANNELS == 1 ? SoundSource.SNDRateConverter.THINNING_MONO : SoundSource.SNDRateConverter.THINNING_STEREO;
          break;
        case "Sound linear interpolation":  //音声 線形補間
          SoundSource.sndRateConverter = SoundSource.SND_CHANNELS == 1 ? SoundSource.SNDRateConverter.LINEAR_MONO : SoundSource.SNDRateConverter.LINEAR_STEREO;
          break;
        case "Sound piecewise-constant area interpolation":  //音声 区分定数面積補間
          SoundSource.sndRateConverter = SoundSource.SNDRateConverter.CONSTANT_AREA_STEREO_48000;
          break;
        case "Sound linear area interpolation":  //音声 線形面積補間
          SoundSource.sndRateConverter = SoundSource.SNDRateConverter.LINEAR_AREA_STEREO_48000;
          break;
        case "Sound monitor":  //音声モニタ
          SoundMonitor.smnOpen ();
          break;
        case "PCM piecewise-constant interpolation":  //PCM 区分定数補間
          ADPCM.pcmSetInterpolationAlgorithm (ADPCM.PCM_INTERPOLATION_CONSTANT);
          break;
        case "PCM linear interpolation":  //PCM 線形補間
          ADPCM.pcmSetInterpolationAlgorithm (ADPCM.PCM_INTERPOLATION_LINEAR);
          break;
        case "PCM hermite interpolation":  //PCM エルミート補間
          ADPCM.pcmSetInterpolationAlgorithm (ADPCM.PCM_INTERPOLATION_HERMITE);
          break;
        case "PCM 8MHz/4MHz":
          ADPCM.pcmOSCFreqRequest = 0;
          break;
        case "PCM 8MHz/16MHz":
          ADPCM.pcmOSCFreqRequest = 1;
          break;

          //入力メニュー
        case "Paste":  //貼り付け
          CONDevice.conDoPaste ();
          break;
        case "No keyboard":  //キーボードなし
          Keyboard.kbdSetOn (false);
          break;
        case "Standard keyboard":  //標準キーボード
          Keyboard.kbdSetType (Keyboard.KBD_STANDARD_TYPE);
          Keyboard.kbdSetOn (true);
          break;
        case "Compact keyboard":  //コンパクトキーボード
          Keyboard.kbdSetType (Keyboard.KBD_COMPACT_TYPE);
          Keyboard.kbdSetOn (true);
          break;
        case "Key assignments":  //キー割り当て
          Keyboard.kbdOpen ();
          break;
        case "F11 key to toggle fullscreen":  //F11 キーで全画面表示
          Keyboard.kbdF11Mode = Keyboard.KBD_F11_FULLSCREEN;
          break;
        case "F11 key to take a screenshot":  //F11 キーでスクリーンショット
          Keyboard.kbdF11Mode = Keyboard.KBD_F11_SCREENSHOT;
          break;
        case "F11 key to stop and start":  //F11 キーで停止と再開
          Keyboard.kbdF11Mode = Keyboard.KBD_F11_STOPANDSTART;
          break;
        case "Default delay":  //既定の開始
          smrRepeatDelay = -1;
          Keyboard.kbdSetRepeatDelay (MainMemory.mmrRbs (0x00ed003a));
          break;
        case (200 + 100 *  0) + "ms":
        case (200 + 100 *  1) + "ms":
        case (200 + 100 *  2) + "ms":
        case (200 + 100 *  3) + "ms":
        case (200 + 100 *  4) + "ms":
        case (200 + 100 *  5) + "ms":
        case (200 + 100 *  6) + "ms":
        case (200 + 100 *  7) + "ms":
        case (200 + 100 *  8) + "ms":
        case (200 + 100 *  9) + "ms":
        case (200 + 100 * 10) + "ms":
        case (200 + 100 * 11) + "ms":
        case (200 + 100 * 12) + "ms":
        case (200 + 100 * 13) + "ms":
        case (200 + 100 * 14) + "ms":
        case (200 + 100 * 15) + "ms":
          {
            int ms = Integer.parseInt (command.substring (0, command.length () - 2));  //200..1700
            //perl optdiv.pl 1500 100
            //  x/100==x*1311>>>17 (0<=x<=4698) [1500*1311==1966500]
            smrRepeatDelay = (ms - 200) * 1311 >>> 17;  //0..15
            Keyboard.kbdSetRepeatDelay (smrRepeatDelay);
            MainMemory.mmrWb (0x00ed003a, smrRepeatDelay);
          }
          break;
        case "Default interval":  //既定の間隔
          smrRepeatInterval =  -1;
          Keyboard.kbdSetRepeatInterval (MainMemory.mmrRbs (0x00ed003b));
          break;
        case (30 + 5 *  0 *  0) + "ms":
        case (30 + 5 *  1 *  1) + "ms":
        case (30 + 5 *  2 *  2) + "ms":
        case (30 + 5 *  3 *  3) + "ms":
        case (30 + 5 *  4 *  4) + "ms":
        case (30 + 5 *  5 *  5) + "ms":
        case (30 + 5 *  6 *  6) + "ms":
        case (30 + 5 *  7 *  7) + "ms":
        case (30 + 5 *  8 *  8) + "ms":
        case (30 + 5 *  9 *  9) + "ms":
        case (30 + 5 * 10 * 10) + "ms":
        case (30 + 5 * 11 * 11) + "ms":
        case (30 + 5 * 12 * 12) + "ms":
        case (30 + 5 * 13 * 13) + "ms":
        case (30 + 5 * 14 * 14) + "ms":
        case (30 + 5 * 15 * 15) + "ms":
          {
            int ms = Integer.parseInt (command.substring (0, command.length () - 2));  //30..1155
            //perl optdiv.pl 1125 5
            //  x/5==x*1639>>>13 (0<=x<=2733) [1125*1639==1843875]
            smrRepeatInterval = (int) Math.sqrt ((double) ((ms - 30) * 1639 >>> 13));  //0..15
            Keyboard.kbdSetRepeatInterval (smrRepeatInterval);
            MainMemory.mmrWb (0x00ed003b, smrRepeatInterval);
          }
          break;
        case "Seamless mouse":  //シームレスマウス
          musSetSeamlessOn (((JCheckBoxMenuItem) source).isSelected ());
          break;
        case "Ctrl-key + left-button = right-button":  //Ctrl キー＋左ボタン＝右ボタン
          musCtrlRightOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Edge acceleration":  //縁部加速
          musSetEdgeAccelerationOn (((JCheckBoxMenuItem) source).isSelected ());
          break;
        case "Use host's pixel units":  //ホストの画素単位を使う
          musSetHostsPixelUnitsOn (((JCheckBoxMenuItem) source).isSelected ());
          break;
        case "Joystick port":  //ジョイスティックポート
          PPI.ppiOpen ();
          break;

          //設定メニュー
        case "Terminal":  //ターミナル
          RS232CTerminal.trmOpen ();
          break;
          //  デバッグ
        case "Console":  //コンソール
          DebugConsole.dgtOpen ();
          break;
        case "Register list":  //レジスタ
          RegisterList.drpOpen ();
          break;
        case "Disassemble list":  //逆アセンブルリスト
          DisassembleList.ddpOpen (-1, -1, true);
          break;
        case "Memory dump list":  //メモリダンプリスト
          MemoryDumpList.dmpOpen (-1, -1, true);
          break;
        case "Logical space monitor":  //論理空間モニタ
          LogicalSpaceMonitor.atwOpen ();
          break;
        case "Physical space monitor":  //物理空間モニタ
          PhysicalSpaceMonitor.paaOpen ();
          break;
        case "Address translation caches monitor":  //アドレス変換キャッシュモニタ
          if (ATCMonitor.ACM_ON) {
            ATCMonitor.acmOpen ();
          }
          break;
        case "Branch log":  //分岐ログ
          if (BranchLog.BLG_ON) {
            BranchLog.blgOpen (BranchLog.BLG_SELECT_NONE);
          }
          break;
        case "Program flow visualizer":  //プログラムフロービジュアライザ
          if (ProgramFlowVisualizer.PFV_ON) {
            ProgramFlowVisualizer.pfvOpen ();
          }
          break;
        case "Raster break point":  //ラスタブレークポイント
          if (RasterBreakPoint.RBP_ON) {
            RasterBreakPoint.rbpOpen ();
          }
          break;
        case "Data break point":  //データブレークポイント
          if (DataBreakPoint.DBP_ON) {
            DataBreakPoint.dbpOpen ();
          }
          break;
        case "Root pointer list":  //ルートポインタリスト
          if (RootPointerList.RTL_ON) {
            RootPointerList.rtlOpen ();
          }
          break;
        case "ROM debugger start flag":  //ROM デバッガ起動フラグ
          smrROMDBOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
          //  起動デバイス
          //  RTC
        case "Adjust clock to host":  //時計をホストに合わせる
          RP5C15.rtcSetByHost ();
          break;
          //  SRAM
          //  設定ファイル
          //
        case "Printer":  //プリンタ
          PrinterPort.prnOpen ();
          break;
          //
        case "Create XEIJROM.DAT":  //XEIJROM.DAT の作成
          ROM.romOpenXEIJROMDialog ();
          break;
        case "Create XEIJROM30.DAT":  //XEIJROM30.DAT の作成
          ROM.romOpenXEIJROM30Dialog ();
          break;
        case "Create XEIJIPLROM30.DAT":  //XEIJIPLROM30.DAT の作成
          ROM.romOpenXEIJIPLROM30Dialog ();
          break;
        case "Create XEIJSCSIINROM.DAT":  //XEIJSCSIINROM.DAT の作成
          ROM.romOpenXEIJSCSIINROMDialog ();
          break;
        case "Create XEIJSCSIEXROM.DAT":  //XEIJSCSIEXROM.DAT の作成
          ROM.romOpenXEIJSCSIEXROMDialog ();
          break;
        case "Create EVEN030.BIN":  //EVEN030.BIN の作成
          ROM.romOpenEVEN030Dialog ();
          break;
        case "Create ODD030.BIN":  //ODD030.BIN の作成
          ROM.romOpenODD030Dialog ();
          break;
        case "Create EVENXVI.BIN":  //EVENXVI.BIN の作成
          ROM.romOpenEVENXVIDialog ();
          break;
        case "Create ODDXVI.BIN":  //ODDXVI.BIN の作成
          ROM.romOpenODDXVIDialog ();
          break;
        case "Mouse button status":  //マウスのボタンの状態
          musOutputButtonStatus = ((JCheckBoxMenuItem) source).isSelected ();
          break;
          //
        case "Java runtime environment information":
          prgOpenJavaDialog ();
          break;
        case "Version information":
          prgOpenAboutDialog ();
          break;
        case "XEiJ License":
          prgOpenXEiJLicenseDialog ();
          break;
        case "FSHARP License":
          prgOpenSHARPLicenseDialog ();
          break;
        case "MAME License":
          prgOpenMAMELicenseDialog ();
          break;

          //言語メニュー
        case "English":
          Multilingual.mlnChange ("en");
          break;
        case "日本語":
          Multilingual.mlnChange ("ja");
          break;

        }
      }
    };

    //  メインメモリ
    ActionListener mainMemoryListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {
        case "1MB":
          MainMemory.mmrMemorySizeRequest = 0x00100000;
          break;
        case "2MB":
          MainMemory.mmrMemorySizeRequest = 0x00200000;
          break;
        case "4MB":
          MainMemory.mmrMemorySizeRequest = 0x00400000;
          break;
        case "6MB":
          MainMemory.mmrMemorySizeRequest = 0x00600000;
          break;
        case "8MB":
          MainMemory.mmrMemorySizeRequest = 0x00800000;
          break;
        case "10MB":
          MainMemory.mmrMemorySizeRequest = 0x00a00000;
          break;
        case "12MB":
          MainMemory.mmrMemorySizeRequest = 0x00c00000;
          break;
        case "Save contents on exit":  //終了時に内容を保存する
          MainMemory.mmrMemorySaveOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        }
      }
    };

    //  X68030/Xellent30 のハイメモリ
    ActionListener highMemoryListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {
        case "None":  //なし
          busHighMemorySize = 0 << 20;
          break;
        case "16MB":
          busHighMemorySize = 16 << 20;
          break;
        case "Save contents on exit":  //終了時に内容を保存する
          busHighMemorySaveOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        }
      }
    };

    //  060turbo のローカルメモリ
    ActionListener localMemoryListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {
        case "None":  //なし
          busLocalMemorySize = 0 << 20;
          break;
        case "16MB":
          busLocalMemorySize = 16 << 20;
          break;
        case "32MB":
          busLocalMemorySize = 32 << 20;
          break;
        case "64MB":
          busLocalMemorySize = 64 << 20;
          break;
        case "128MB":
          busLocalMemorySize = 128 << 20;
          break;
        case "256MB":
          busLocalMemorySize = 256 << 20;
          break;
        case "Save contents on exit":  //終了時に内容を保存する
          busLocalMemorySaveOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        }
      }
    };

    ButtonGroup unitGroup = new ButtonGroup ();
    ButtonGroup frameGroup = new ButtonGroup ();
    ButtonGroup fullGroup = new ButtonGroup ();
    ButtonGroup hintGroup = new ButtonGroup ();
    ButtonGroup intermittentGroup = new ButtonGroup ();
    ButtonGroup soundInterpolationGroup = new ButtonGroup ();
    ButtonGroup adpcmInterpolationGroup = new ButtonGroup ();
    ButtonGroup adpcmOSCFreqGroup = new ButtonGroup ();
    ButtonGroup keyboardGroup = new ButtonGroup ();
    ButtonGroup keydlyGroup = new ButtonGroup ();
    ButtonGroup keyrepGroup = new ButtonGroup ();
    ButtonGroup memoryGroup = new ButtonGroup ();
    ButtonGroup highGroup = new ButtonGroup ();
    ButtonGroup localGroup = new ButtonGroup ();
    ButtonGroup languageGroup = new ButtonGroup ();
    ButtonGroup f11Group = new ButtonGroup ();

    //メニューバー
    mnbMenuBar = ComponentFactory.createMenuBar (

      //ファイルメニュー
      mnbFileMenu = Multilingual.mlnText (
        ComponentFactory.createMenu (
          "File", 'F',
          //FDDメニュー
          FDC.fdcMenu,
          //SASI HDDメニュー
          HDC.hdcMenu,
          //内蔵 SCSI HDDメニュー
          SPC.spcMenu,
          //HFSメニュー
          HFS.hfsMenu,
          ComponentFactory.createHorizontalSeparator (),
          mnbQuitMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Quit", 'Q', MNB_MODIFIERS, listener), "ja", "終了")
          ),
        "ja", "ファイル"),

      //MPUメニュー
      mpuMenu,

      //画面メニュー
      mnbDisplayMenu = Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Display", 'D',
          mnbFullscreenMenuItem = ComponentFactory.setEnabled (
            Multilingual.mlnText (
              ComponentFactory.createRadioButtonMenuItem (
                fullGroup, pnlFullscreenOn,
                "Full screen", KeyEvent.VK_F11, listener), "ja", "全画面表示"),
            pnlIsFullscreenSupported),  //全画面表示に移行できるとき
          mnbFitInWindowMenuItem = ComponentFactory.setEnabled (
            Multilingual.mlnText (
              ComponentFactory.createRadioButtonMenuItem (
                fullGroup, !pnlFullscreenOn && pnlFitInWindowOn,
                "Fit in window", 'W', MNB_MODIFIERS, listener), "ja", "ウインドウに合わせる"),
            pnlIsFitInWindowSupported),  //ウインドウに合わせられるとき
          mnbFixedScaleMenuItem = Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (
              fullGroup, !pnlFullscreenOn && !pnlFitInWindowOn,
              "Fixed scale", 'X', MNB_MODIFIERS, listener), "ja", "固定倍率"),
          ComponentFactory.createHorizontalBox (
            Box.createHorizontalStrut (20),
            pnlFixedSpinner,
            ComponentFactory.createLabel ("%"),
            Box.createHorizontalGlue ()
            ),
          ComponentFactory.createMenuItem ("50%", listener),
          ComponentFactory.createMenuItem ("75%", listener),
          ComponentFactory.createMenuItem ("100%", listener),
          ComponentFactory.createMenuItem ("150%", listener),
          ComponentFactory.createMenuItem ("200%", listener),
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (
              hintGroup, pnlInterpolation == RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR,
              "Nearest neighbor", listener),
            "ja", "最近傍補間"),
          Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (
              hintGroup, pnlInterpolation == RenderingHints.VALUE_INTERPOLATION_BILINEAR,
              "Bilinear", listener),
            "ja", "線形補間"),
          Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (
              hintGroup, pnlInterpolation == RenderingHints.VALUE_INTERPOLATION_BICUBIC,
              "Bicubic", listener),
            "ja", "三次補間"),
          !CRTC.CRT_ENABLE_INTERMITTENT ? null : ComponentFactory.createHorizontalSeparator (),
          !CRTC.CRT_ENABLE_INTERMITTENT ? null : Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Intermittent drawing",
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (
                  intermittentGroup, CRTC.crtIntermittentInterval == 0, "Draw all changed pictures", listener),
                "ja", "変化した画像をすべて描画する"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (
                  intermittentGroup, CRTC.crtIntermittentInterval == 1, "Draw a changed picture once every two times", listener),
                "ja", "変化した画像を 2 回に 1 回描画する"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (
                  intermittentGroup, CRTC.crtIntermittentInterval == 2, "Draw a changed picture once every three times", listener),
                "ja", "変化した画像を 3 回に 1 回描画する"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (
                  intermittentGroup, CRTC.crtIntermittentInterval == 3, "Draw a changed picture once every four times", listener),
                "ja", "変化した画像を 4 回に 1 回描画する"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (
                  intermittentGroup, CRTC.crtIntermittentInterval == 4, "Draw a changed picture once every five times", listener),
                "ja", "変化した画像を 5 回に 1 回描画する")
              ),
            "ja", "間欠描画"),
          !CRTC.CRT_EXTENDED_GRAPHIC ? null : ComponentFactory.createHorizontalSeparator (),
          !CRTC.CRT_EXTENDED_GRAPHIC ? null : Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (CRTC.crtExtendedGraphicRequest, "Extended graphic screen", listener), "ja", "拡張グラフィック画面"),
          ComponentFactory.createHorizontalSeparator (),
          SpritePatternViewer.SPV_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Sprite pattern viewer", listener), "ja", "スプライトパターンビュア") : null,
          ScreenModeTest.SMT_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Screen mode test", listener), "ja", "表示モードテスト") : null
          ),
        "ja", "画面"),

      //音声メニュー
      mnbSoundMenu = ComponentFactory.setEnabled (
        Multilingual.mlnText (
          ComponentFactory.createMenu (
            "Sound", 'S',
            mnbPlayMenuItem = ComponentFactory.setEnabled (
              Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (SoundSource.sndPlayOn, "Play", 'P', MNB_MODIFIERS, listener), "ja", "音声出力"),
              SoundSource.sndLine != null),
            //ボリュームのラベル
            //  JLabelのalignmentでセンタリングしようとするとチェックボックスのサイズの分だけ右に寄ってしまう
            //  Boxで囲み、左右にglueを置くことでセンタリングする
            ComponentFactory.createHorizontalBox (
              Box.createHorizontalGlue (),
              Multilingual.mlnText (ComponentFactory.createLabel ("Volume "), "ja", "音量"),
              mnbVolumeLabel = ComponentFactory.createLabel (String.valueOf (SoundSource.sndVolume)),
              Box.createHorizontalGlue ()
              ),
            //ボリュームスライダ
            //  デフォルトのサイズだと間延びした感じになるので幅を狭くする
            //  高さの43pxはpreferredSizeを指定しなかった場合と同じ
            ComponentFactory.setPreferredSize (
              ComponentFactory.createHorizontalSlider (
                0,
                SoundSource.SND_VOLUME_MAX,
                SoundSource.sndVolume,
                SoundSource.SND_VOLUME_STEP,
                1,
                new ChangeListener () {
                  @Override public void stateChanged (ChangeEvent ce) {
                    SoundSource.sndSetVolume (((JSlider) ce.getSource ()).getValue ());
                  }
                }),
              214, 43),
            Multilingual.mlnText (
              ComponentFactory.createMenu (
                "Sound interpolation",
                Multilingual.mlnText (
                  ComponentFactory.createRadioButtonMenuItem (
                    soundInterpolationGroup, SoundSource.sndRateConverter == SoundSource.SNDRateConverter.THINNING_STEREO,
                    "Sound thinning", listener),
                  "ja", "音声 間引き"),
                Multilingual.mlnText (
                  ComponentFactory.createRadioButtonMenuItem (
                    soundInterpolationGroup, SoundSource.sndRateConverter == SoundSource.SNDRateConverter.LINEAR_STEREO,
                    "Sound linear interpolation", listener),
                  "ja", "音声 線形補間"),
                ComponentFactory.setEnabled (
                  Multilingual.mlnText (
                    ComponentFactory.createRadioButtonMenuItem (
                      soundInterpolationGroup, SoundSource.sndRateConverter == SoundSource.SNDRateConverter.CONSTANT_AREA_STEREO_48000,
                      "Sound piecewise-constant area interpolation", listener),
                    "ja", "音声 区分定数面積補間"),
                  SoundSource.SND_CHANNELS == 2 && SoundSource.SND_SAMPLE_FREQ == 48000),
                ComponentFactory.setEnabled (
                  Multilingual.mlnText (
                    ComponentFactory.createRadioButtonMenuItem (
                      soundInterpolationGroup, SoundSource.sndRateConverter == SoundSource.SNDRateConverter.LINEAR_AREA_STEREO_48000,
                      "Sound linear area interpolation", listener),
                    "ja", "音声 線形面積補間"),
                  SoundSource.SND_CHANNELS == 2 && SoundSource.SND_SAMPLE_FREQ == 48000)
                ),
              "ja", "音声補間"),
            Multilingual.mlnText (ComponentFactory.createMenuItem ("Sound monitor", listener), "ja", "音声モニタ"),
            ComponentFactory.createHorizontalSeparator (),
            ComponentFactory.setEnabled (
              Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (YM2151.opmOutputMask != 0, "OPM output", listener), "ja", "OPM 出力"),
              SoundSource.sndLine != null),
            ComponentFactory.createHorizontalSeparator (),
            ComponentFactory.setEnabled (
              Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (ADPCM.pcmOutputOn, "PCM output", listener), "ja", "PCM 出力"),
              SoundSource.sndLine != null),
            Multilingual.mlnText (
              ComponentFactory.createMenu (
                "PCM interpolation",
                Multilingual.mlnText (
                  ComponentFactory.createRadioButtonMenuItem (
                    adpcmInterpolationGroup, ADPCM.pcmInterpolationAlgorithm == ADPCM.PCM_INTERPOLATION_CONSTANT,
                    "PCM piecewise-constant interpolation", listener),
                  "ja", "PCM 区分定数補間"),
                Multilingual.mlnText (
                  ComponentFactory.createRadioButtonMenuItem (
                    adpcmInterpolationGroup, ADPCM.pcmInterpolationAlgorithm == ADPCM.PCM_INTERPOLATION_LINEAR,
                    "PCM linear interpolation", listener),
                  "ja", "PCM 線形補間"),
                Multilingual.mlnText (
                  ComponentFactory.createRadioButtonMenuItem (
                    adpcmInterpolationGroup, ADPCM.pcmInterpolationAlgorithm == ADPCM.PCM_INTERPOLATION_HERMITE,
                    "PCM hermite interpolation", listener),
                  "ja", "PCM エルミート補間")
                ),
              "ja", "PCM 補間"),
            Multilingual.mlnText (
              ComponentFactory.createMenu (
                "PCM source oscillator frequency",
                ComponentFactory.createRadioButtonMenuItem (adpcmOSCFreqGroup, ADPCM.pcmOSCFreqRequest == 0, "PCM 8MHz/4MHz", listener),
                ComponentFactory.createRadioButtonMenuItem (adpcmOSCFreqGroup, ADPCM.pcmOSCFreqRequest == 1, "PCM 8MHz/16MHz", listener)
                ),
              "ja", "PCM 原発振周波数")
            ),
          "ja", "音声"),
        SoundSource.sndLine != null),

      //入力メニュー
      mnbInputMenu = Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Input", 'I',
          mnbPasteMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Paste", 'V', MNB_MODIFIERS, listener), "ja", "貼り付け"),
          ComponentFactory.createHorizontalSeparator (),
          mnbNoKeyboardMenuItem = Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (keyboardGroup, !Keyboard.kbdOn, "No keyboard", 'K', MNB_MODIFIERS, listener),
            "ja", "キーボードなし"),
          mnbStandardKeyboardMenuItem = Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (keyboardGroup, Keyboard.kbdOn && Keyboard.kbdType == Keyboard.KBD_STANDARD_TYPE, "Standard keyboard", listener),
            "ja", "標準キーボード"),
          mnbCompactKeyboardMenuItem = Multilingual.mlnText (
            ComponentFactory.createRadioButtonMenuItem (keyboardGroup, Keyboard.kbdOn && Keyboard.kbdType == Keyboard.KBD_COMPACT_TYPE, "Compact keyboard", listener),
            "ja", "コンパクトキーボード"),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Key assignments", listener), "ja", "キー割り当て"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Repeat delay",
              Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == -1, "Default delay", listener), "ja", "既定の開始"),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  0, (200 + 100 *  0) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  1, (200 + 100 *  1) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  2, (200 + 100 *  2) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  3, (200 + 100 *  3) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  4, (200 + 100 *  4) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  5, (200 + 100 *  5) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  6, (200 + 100 *  6) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  7, (200 + 100 *  7) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  8, (200 + 100 *  8) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay ==  9, (200 + 100 *  9) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == 10, (200 + 100 * 10) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == 11, (200 + 100 * 11) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == 12, (200 + 100 * 12) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == 13, (200 + 100 * 13) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == 14, (200 + 100 * 14) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keydlyGroup, smrRepeatDelay == 15, (200 + 100 * 15) + "ms", listener)
              ),
            "ja", "リピート開始"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Repeat interval",
              Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == -1, "Default interval", listener), "ja", "既定の間隔"),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  0, (30 + 5 *  0 *  0) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  1, (30 + 5 *  1 *  1) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  2, (30 + 5 *  2 *  2) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  3, (30 + 5 *  3 *  3) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  4, (30 + 5 *  4 *  4) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  5, (30 + 5 *  5 *  5) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  6, (30 + 5 *  6 *  6) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  7, (30 + 5 *  7 *  7) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  8, (30 + 5 *  8 *  8) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval ==  9, (30 + 5 *  9 *  9) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == 10, (30 + 5 * 10 * 10) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == 11, (30 + 5 * 11 * 11) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == 12, (30 + 5 * 12 * 12) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == 13, (30 + 5 * 13 * 13) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == 14, (30 + 5 * 14 * 14) + "ms", listener),
              ComponentFactory.createRadioButtonMenuItem (keyrepGroup, smrRepeatInterval == 15, (30 + 5 * 15 * 15) + "ms", listener)
              ),
            "ja", "リピート間隔"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "F11 key",
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (f11Group, Keyboard.kbdF11Mode == Keyboard.KBD_F11_FULLSCREEN, "F11 key to toggle fullscreen", listener),
                "ja", "F11 キーで全画面表示"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (f11Group, Keyboard.kbdF11Mode == Keyboard.KBD_F11_SCREENSHOT, "F11 key to take a screenshot", listener),
                "ja", "F11 キーでスクリーンショット"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (f11Group, Keyboard.kbdF11Mode == Keyboard.KBD_F11_STOPANDSTART, "F11 key to stop and start", listener),
                "ja", "F11 キーで停止と再開")),
            "ja", "F11 キー"),
          ComponentFactory.createHorizontalSeparator (),
          mnbSeamlessMenuItem = ComponentFactory.setEnabled (
            Multilingual.mlnText (
              ComponentFactory.createCheckBoxMenuItem (musSeamlessOn, "Seamless mouse", KeyEvent.VK_F12, listener),
              "ja", "シームレスマウス"),
            rbtRobot != null),
          Multilingual.mlnText (
            ComponentFactory.createCheckBoxMenuItem (
              musCtrlRightOn,
              "Ctrl-key + left-button = right-button",
              listener),
            "ja", "Ctrl キー＋左ボタン＝右ボタン"),
          Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (musEdgeAccelerationOn, "Edge acceleration", listener), "ja", "縁部加速"),
          ComponentFactory.createHorizontalBox (
            Box.createHorizontalGlue (),
            Multilingual.mlnText (ComponentFactory.createLabel ("Mouse cursor speed "), "ja", "マウスカーソルの速度 "),
            Z8530.scc0Label,
            Box.createHorizontalGlue ()
            ),
          Z8530.scc0Slider,
          ComponentFactory.setEnabled (
            Multilingual.mlnText (
              ComponentFactory.createCheckBoxMenuItem (musHostsPixelUnitsOn, "Use host's pixel units", listener),
              "ja", "ホストの画素単位を使う"),
            rbtRobot != null),
          Z8530.SCC_DEBUG_TRACE ? Z8530.sccTraceMenuItem : null,
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Joystick port", listener), "ja", "ジョイスティックポート")
          ),
        "ja", "入力"),

      //設定メニュー
      mnbConfigMenu = Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Config", 'C',
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Terminal", listener), "ja", "ターミナル"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Debug",
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Console", listener), "ja", "コンソール"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Register list", listener), "ja", "レジスタリスト"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Disassemble list", listener), "ja", "逆アセンブルリスト"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Memory dump list", listener), "ja", "メモリダンプリスト"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Logical space monitor", listener), "ja", "論理空間モニタ"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Physical space monitor", listener), "ja", "物理空間モニタ"),
              ATCMonitor.ACM_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Address translation caches monitor", listener), "ja", "アドレス変換キャッシュモニタ") : null,
              BranchLog.BLG_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Branch log", listener), "ja", "分岐ログ") : null,
              ProgramFlowVisualizer.PFV_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Program flow visualizer", listener), "ja", "プログラムフロービジュアライザ") : null,
              RasterBreakPoint.RBP_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Raster break point", listener), "ja", "ラスタブレークポイント") : null,
              DataBreakPoint.DBP_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Data break point", listener), "ja", "データブレークポイント") : null,
              RootPointerList.RTL_ON ? Multilingual.mlnText (ComponentFactory.createMenuItem ("Root pointer list", listener), "ja", "ルートポインタリスト") : null,
              ComponentFactory.createHorizontalSeparator (),
              Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (smrROMDBOn, "ROM debugger start flag", listener), "ja", "ROM デバッガ起動フラグ")
              ),
            "ja", "デバッグ"),
          smrBootMenu,
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Main memory",
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00100000, "1MB", mainMemoryListener),
                "ja", "1MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00200000, "2MB", mainMemoryListener),
                "ja", "2MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00400000, "4MB", mainMemoryListener),
                "ja", "4MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00600000, "6MB", mainMemoryListener),
                "ja", "6MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00800000, "8MB", mainMemoryListener),
                "ja", "8MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00a00000, "10MB", mainMemoryListener),
                "ja", "10MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (memoryGroup, MainMemory.mmrMemorySizeRequest == 0x00c00000, "12MB", mainMemoryListener),
                "ja", "12MB"),
              ComponentFactory.createHorizontalSeparator (),
              Multilingual.mlnText (
                ComponentFactory.createCheckBoxMenuItem (MainMemory.mmrMemorySaveOn, "Save contents on exit", mainMemoryListener),
                "ja", "終了時に内容を保存する")
              ),
            "ja", "メインメモリ"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "High memory on X68030",
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (highGroup, busHighMemorySize == 0 << 20, "None", highMemoryListener),
                "ja", "なし"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (highGroup, busHighMemorySize == 16 << 20, "16MB", highMemoryListener),
                "ja", "16MB"),
              ComponentFactory.createHorizontalSeparator (),
              Multilingual.mlnText (
                ComponentFactory.createCheckBoxMenuItem (busHighMemorySaveOn, "Save contents on exit", highMemoryListener),
                "ja", "終了時に内容を保存する")
              ),
            "ja", "X68030/Xellent30 のハイメモリ"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Local memory on 060turbo",
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (localGroup, busLocalMemorySize == 0 << 20, "None", localMemoryListener),
                "ja", "なし"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (localGroup, busLocalMemorySize == 16 << 20, "16MB", localMemoryListener),
                "ja", "16MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (localGroup, busLocalMemorySize == 32 << 20, "32MB", localMemoryListener),
                "ja", "32MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (localGroup, busLocalMemorySize == 64 << 20, "64MB", localMemoryListener),
                "ja", "64MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (localGroup, busLocalMemorySize == 128 << 20, "128MB", localMemoryListener),
                "ja", "128MB"),
              Multilingual.mlnText (
                ComponentFactory.createRadioButtonMenuItem (localGroup, busLocalMemorySize == 256 << 20, "256MB", localMemoryListener),
                "ja", "256MB"),
              ComponentFactory.createHorizontalSeparator (),
              Multilingual.mlnText (
                ComponentFactory.createCheckBoxMenuItem (busLocalMemorySaveOn, "Save contents on exit", localMemoryListener),
                "ja", "終了時に内容を保存する")
              ),
            "ja", "060turbo のローカルメモリ"),
          ComponentFactory.createMenu (
            "RTC",
            Multilingual.mlnText (
              ComponentFactory.createMenuItem ("Adjust clock to host", listener),
              "ja", "時計をホストに合わせる")
            ),
          smrMenu,
          Settings.sgsMenu,
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Printer", listener), "ja", "プリンタ"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Create ROM file",
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create XEIJROM.DAT", listener), "ja", "XEIJROM.DAT の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create XEIJROM30.DAT", listener), "ja", "XEIJROM30.DAT の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create XEIJIPLROM30.DAT", listener), "ja", "XEIJIPLROM30.DAT の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create XEIJSCSIINROM.DAT", listener), "ja", "XEIJSCSIINROM.DAT の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create XEIJSCSIEXROM.DAT", listener), "ja", "XEIJSCSIEXROM.DAT の作成"),
              ComponentFactory.createHorizontalSeparator (),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create EVEN030.BIN", listener), "ja", "EVEN030.BIN の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create ODD030.BIN", listener), "ja", "ODD030.BIN の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create EVENXVI.BIN", listener), "ja", "EVENXVI.BIN の作成"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("Create ODDXVI.BIN", listener), "ja", "ODDXVI.BIN の作成")
              ),
            "ja", "ROM ファイルの作成"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Miscellaneous",
              SlowdownTest.sdtCheckBoxMenuItem,
              SlowdownTest.sdtBox,
              Multilingual.mlnText (
                ComponentFactory.createCheckBoxMenuItem (musOutputButtonStatus, "Mouse button status", listener),
                "ja", "マウスのボタンの状態")
              ),
            "ja", "その他"),
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (
            ComponentFactory.createMenuItem ("Java runtime environment information", listener),
            "ja", "Java 実行環境の情報"),
          Multilingual.mlnText (
            ComponentFactory.createMenuItem ("Version information", listener),
            "ja", "バージョン情報"),
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "License",
              Multilingual.mlnText (ComponentFactory.createMenuItem ("XEiJ License", listener), "ja", "XEiJ 使用許諾条件"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("FSHARP License", listener), "ja", "FSHARP 許諾条件"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("MAME License", listener), "ja", "MAME License")
              ),
            "ja", "使用許諾条件")
          ),
        "ja", "設定"),

      //インジケータ
      Box.createHorizontalGlue (),  //インジケータをセンタリングする
      ComponentFactory.createVerticalBox (
        Box.createVerticalGlue (),
        Indicator.indBox,
        Box.createVerticalGlue ()
        ),
      Box.createHorizontalGlue (),  //言語メニューを右に寄せる

      //言語メニュー
      //  言語選択メニューはテキストで書く
      //    国旗アイコンは対象とするマーケットを選択させるときに使うものであり、表示言語を切り替えるだけのメニューに国旗を用いるのは不適切
      //  言語メニューのメニューアイテムは多言語化しない
      //  言語メニュー自体を多言語化しないという方法もある
      //    簡潔な日本語メニューの並びに幅を食うLanguageメニューが混在しているのはあまり美しくない
      mnbLanguageMenu = Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Language", 'L',
          ComponentFactory.createRadioButtonMenuItem (languageGroup, Multilingual.mlnEnglish, "English", listener),
          ComponentFactory.createRadioButtonMenuItem (languageGroup, Multilingual.mlnJapanese, "日本語", listener)
          ),
        "ja", "言語")

      );
  }  //mnbMakeMenu()



  //========================================================================================
  //$$FRM フレーム

  //モード
  public static boolean frmIsActive;  //true=フォーカスがある

  //フレーム
  public static JFrame frmFrame;  //フレーム
  public static int frmMarginWidth;  //パネルからフレームまでのマージン
  public static int frmMarginHeight;
  public static Dimension frmMinimumSize;  //pnlMinimumWidth+frmMarginWidth,pnlMinimumHeight+frmMarginHeight フレームの最小サイズ

  //スクリーンデバイス
  public static GraphicsDevice frmScreenDevice;  //スクリーンデバイス

  //ドラッグアンドドロップ
  public static DropTarget frmDropTarget;

  //frmInit ()
  //  フレームを初期化する
  public static void frmInit () {
    frmIsActive = false;
    frmScreenDevice = GraphicsEnvironment.getLocalGraphicsEnvironment ().getDefaultScreenDevice ();  //スクリーンデバイス
    pnlIsFullscreenSupported = frmScreenDevice.isFullScreenSupported ();  //全画面表示に移行できるかどうか
    pnlIsFitInWindowSupported = true;  //ウインドウに合わせられるかどうか
  }  //frmInit()

  //frmMake ()
  //  フレームを作る
  public static void frmMake () {

    //フレーム
    frmFrame = ComponentFactory.createRestorableFrame (
      Settings.SGS_FRM_FRAME_KEY,
      PRG_TITLE + " version " + PRG_VERSION,
      mnbMenuBar,
      pnlPanel);
    frmFrame.setIconImage (LnF.LNF_ICON_IMAGE_16);  //タスクバーのアイコンを変更する
    frmFrame.setDefaultCloseOperation (WindowConstants.DISPOSE_ON_CLOSE);
    //frmFrame.setResizable (false);  //リサイズ不可

    //パネルからフレームまでのマージンを確認する
    frmMarginWidth = frmFrame.getWidth () - pnlWidth;
    frmMarginHeight = frmFrame.getHeight () - pnlHeight;
    frmMinimumSize = new Dimension (pnlMinimumWidth + frmMarginWidth, pnlMinimumHeight + frmMarginHeight);
    frmFrame.setMinimumSize (frmMinimumSize);

    //ドラッグアンドドロップ
    //  FDイメージが放り込まれたらそこから再起動する
    frmDropTarget = new DropTarget (pnlPanel, DnDConstants.ACTION_COPY, new DropTargetAdapter () {
      @Override public void dragOver (DropTargetDragEvent dtde) {
        if (dtde.isDataFlavorSupported (DataFlavor.javaFileListFlavor)) {
          dtde.acceptDrag (DnDConstants.ACTION_COPY);
          return;
        }
        dtde.rejectDrag ();
      }
      @Override public void drop (DropTargetDropEvent dtde) {
        try {
          if (dtde.isDataFlavorSupported (DataFlavor.javaFileListFlavor)) {
            dtde.acceptDrop (DnDConstants.ACTION_COPY);
            boolean reset = true;
            int fdu0 = -1;
            int fdu = 0;
            int hdu0 = -1;
            int hdu = 0;
            int scu0 = -1;
            int scu = 0;
            int hfu0 = -1;
            int hfu = 0;
            for (Object o : (java.util.List) dtde.getTransferable ().getTransferData (DataFlavor.javaFileListFlavor)) {
              if (o instanceof File) {
                File file = (File) o;
                if (file.isFile ()) {  //ファイルのとき。HFS以外のフィルタがディレクトリを受け入れてしまうのでディレクトリを除外する
                  if (FDC.fdcFileFilter.accept (file)) {  //FDのイメージファイルのとき
                    if (fdu < FDC.FDC_MAX_UNITS && FDC.fdcUnitArray[fdu].insert (file.getPath ())) {  //挿入できた
                      if (fdu0 < 0) {
                        fdu0 = fdu;
                      }
                      fdu++;
                      continue;
                    }
                  }
                  if (HDC.hdcFileFilter.accept (file)) {  //SASIハードディスクのイメージファイルのとき
                    if (hdu < 16 && HDC.hdcUnitArray[hdu].insert (file.getPath ())) {  //挿入できた
                      if (hdu0 < 0) {
                        hdu0 = hdu;
                      }
                      hdu++;
                      continue;
                    }
                  }
                  if (SPC.spcOpenFileFilter.accept (file)) {  //SCSIハードディスク/CD-ROMのイメージファイルのとき
                    if (scu < 16 && SPC.spcUnitArray[scu].insert (file.getPath ())) {  //挿入できた
                      if (scu0 < 0) {
                        scu0 = scu;
                      }
                      scu++;
                      continue;
                    }
                  }
                }
                if (HFS.hfsOpenFileFilter.accept (file)) {  //ディレクトリまたはHUMAN.SYSのとき
                  if (hfu < 16 && HFS.hfsUnitArray[hfu].insert (file.getPath ())) {  //挿入できた
                    if (hfu0 < 0) {
                      hfu0 = hfu;
                    }
                    hfu++;
                    continue;
                  }
                }
              }
              reset = false;  //挿入できないファイルがあったときはリセットをキャンセルする
            }
            dtde.dropComplete (true);
            if (reset) {
              if (fdu0 >= 0) {
                mpuReset (0x9070 | fdu0 << 8, -1);
              } else if (hdu0 >= 0) {
                mpuReset (0x8000 | hdu0 << 8, -1);
              } else if (scu0 >= 0) {
                mpuReset (0xa000, SPC.SPC_HANDLE_EX + (scu0 << 2));  //拡張SCSIがなければ内蔵SCSIに読み替えられる
              } else if (hfu0 >= 0) {
                HFS.hfsBootUnit = hfu0;
                mpuReset (0xa000, HFS.HFS_BOOT_HANDLE);
              }
            }
            return;
          }
        } catch (UnsupportedFlavorException ufe) {
          //ufe.printStackTrace ();
        } catch (IOException ioe) {
          //ioe.printStackTrace ();
        }
        dtde.rejectDrop();
      }
    });

  }  //frmMake()

  //frmStart ()
  //  フレームのイベントリスナーを設定して動作を開始する
  public static void frmStart () {

    //ウインドウリスナー
    //  ウインドウを開いたとき  activated,opened
    //  フォーカスを失ったとき  deactivated
    //  フォーカスを取得したとき  activated
    //  ウインドウをアイコン化したとき  iconified,[deactivated]
    //  ウインドウを元のサイズに戻したとき  deiconified,activated
    //  ウインドウを閉じたとき  closing,[deactivated],closed
    ComponentFactory.addListener (
      frmFrame,
      new WindowAdapter () {
        @Override public void windowActivated (WindowEvent we) {
          frmIsActive = true;
        }
        @Override public void windowClosing (WindowEvent we) {
          prgTini ();
        }
        @Override public void windowDeactivated (WindowEvent we) {
          frmIsActive = false;
          musSetSeamlessOn (true);  //フォーカスを失ったときはシームレスに切り替える
        }
      });

    //コンポーネントリスナー
    //  エクスクルーシブマウスモードのときに使うパネルの座標を得る
    //  全画面表示のON/OFFを行ったときcomponentMovedが呼ばれないことがあるのでcomponentResizedでもパネルの座標を得る
    ComponentFactory.addListener (
      frmFrame,
      new ComponentAdapter () {
        @Override public void componentMoved (ComponentEvent ce) {
          Point p = pnlPanel.getLocationOnScreen ();
          pnlGlobalX = p.x;
          pnlGlobalY = p.y;
        }
        @Override public void componentResized (ComponentEvent ce) {
          Point p = pnlPanel.getLocationOnScreen ();
          pnlGlobalX = p.x;
          pnlGlobalY = p.y;
        }
      });

  }  //frmStart()

  //frmSetFullscreenOn (on)
  //  全画面表示を設定する
  //  メニューアイテムは更新しないので必要ならば呼び出し側で更新すること
  public static void frmSetFullscreenOn (boolean on) {
    pnlFullscreenOn = on;
    //全画面表示を変更する
    if (on) {
      if (frmScreenDevice.getFullScreenWindow () != frmFrame) {  //自分が全画面表示でないとき
        frmFrame.getRootPane().setWindowDecorationStyle (JRootPane.NONE);  //飾り枠を消す
        frmScreenDevice.setFullScreenWindow (frmFrame);  //全画面表示に移行する
      }
    } else {
      if (frmScreenDevice.getFullScreenWindow () == frmFrame) {  //自分が全画面表示のとき
        frmScreenDevice.setFullScreenWindow (null);  //全画面表示を解除する
        frmFrame.getRootPane().setWindowDecorationStyle (JRootPane.FRAME);  //飾り枠を描く
      }
    }
    //全画面表示を変更するとコンポーネントイベントが発生する
    //コンポーネントリスナーで画面とキーボードの配置を再計算する
  }  //frmSetFullscreenOn(boolean)



  //========================================================================================
  //$$CLP クリップボード

  public static BufferedImage clpClipboardImage;  //コピーされるイメージ
  public static String clpClipboardString;  //コピーされる文字列
  public static Clipboard clpClipboard;  //クリップボード
  public static Transferable clpImageContents;  //イメージをコピーするときに渡すデータ
  public static Transferable clpStringContents;  //文字列をコピーするときに渡すデータ
  public static ClipboardOwner clpClipboardOwner;  //クリップボードオーナー。コピーするときにデータに付ける情報
  public static boolean clpIsClipboardOwner;  //true=クリップボードに入っているデータは自分がコピーした

  //clpMake ()
  //  クリップボードを作る
  public static void clpMake () {
    Toolkit toolkit = Toolkit.getDefaultToolkit ();
    clpClipboard = null;
    try {
      clpClipboard = toolkit.getSystemClipboard ();  //クリップボード
    } catch (Exception e) {
      return;
    }
    clpClipboardImage = null;  //コピーされるイメージ
    clpClipboardString = null;  //コピーされる文字列
    clpImageContents = new Transferable () {
      public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException {
        if (flavor == DataFlavor.imageFlavor) {
          return clpClipboardImage;
        } else {
          throw new UnsupportedFlavorException (flavor);
        }
      }
      public DataFlavor[] getTransferDataFlavors () {
        return new DataFlavor[] { DataFlavor.imageFlavor };
      }
      public boolean isDataFlavorSupported (DataFlavor flavor) {
        return flavor == DataFlavor.imageFlavor;
      }
    };
    clpStringContents = new Transferable () {
      public Object getTransferData (DataFlavor flavor) throws UnsupportedFlavorException {
        if (flavor == DataFlavor.stringFlavor) {
          return clpClipboardString;
        } else {
          throw new UnsupportedFlavorException (flavor);
        }
      }
      public DataFlavor[] getTransferDataFlavors () {
        return new DataFlavor[] { DataFlavor.stringFlavor };
      }
      public boolean isDataFlavorSupported (DataFlavor flavor) {
        return flavor == DataFlavor.stringFlavor;
      }
    };
    clpIsClipboardOwner = false;  //自分はまだコピーしていない
    //クリップボードオーナー
    //  lostOwnership   クリップボードの所有者でなくなったとき
    clpClipboardOwner = new ClipboardOwner () {
      @Override public void lostOwnership (Clipboard clipboard, Transferable contents) {
        clpIsClipboardOwner = false;
      }
    };
    //フレーバーリスナー
    //  flavorsChanged  クリップボードのDataFlavorが変化したとき
    clpClipboard.addFlavorListener (new FlavorListener () {
      @Override public void flavorsChanged (FlavorEvent fe) {
        boolean available = false;
        try {
          available = clpClipboard.isDataFlavorAvailable (DataFlavor.stringFlavor);
        } catch (IllegalStateException ise) {
        }
        mnbPasteMenuItem.setEnabled (available);  //文字列ならば貼り付けできる
      }
    });
    if (!clpClipboard.isDataFlavorAvailable (DataFlavor.stringFlavor)) {  //文字列がコピーされていない
      mnbPasteMenuItem.setEnabled (false);  //貼り付け選択不可
    }
  }  //clpMake



  //========================================================================================
  //$$XT3 Xellent30

  //ポートのアドレス
  //  +$4000まで同じワードポートが繰り返される
  public static final int XT3_PORT_ADDRESS = 0x00ecc000;

  //MPU
  public static boolean xt3Accelerated;  //false=MC68000,true=MC68EC030
  public static int xt3SavedPC;  //MC68000のPC
  public static int[] xt3SavedRn;  //MC68000のRn

  //SRAM
  public static boolean xt3MemoryEnabled;  //false=SRAMが無効,true=SRAMが有効
  public static int xt3MemoryAddress;  //SRAMの開始アドレス。($00C00000または$01000000)-xt3MemorySize
  public static int xt3MemorySize;  //SRAMのサイズ。$00040000(256KB)または$00100000(1MB)
  public static byte[] xt3MemoryArray;  //SRAMの内容の配列
  public static boolean xt3MemorySave;  //true=SRAMの内容を保存する

  //初期化
  public static void xt3Init () {
    xt3Accelerated = false;  //MC68000
    xt3SavedPC = 0;
    xt3SavedRn = new int[16];
    xt3MemoryEnabled = false;  //SRAMが無効
    xt3MemoryAddress = 0x00bc0000;
    xt3MemorySize = 0x00040000;
    xt3MemoryArray = new byte[xt3MemorySize];
    xt3MemorySave = false;
  }

  public static int xt3PortRead () {
    return (xt3Accelerated ? 4 : 0) | (xt3MemoryEnabled ? 2 : 0) | (0x00c00000 <= xt3MemoryAddress ? 1 : 0);
  }

  public static void xt3PortWrite (int d) {
    boolean accelerated = (d & 4) != 0;
    boolean memoryEnabled = (d & 2) != 0;
    int memoryAddress = ((d & 1) != 0 ? 0x01000000 : 0x00c00000) - xt3MemorySize;
    if (xt3MemoryAddress != memoryAddress) {  //SRAMのアドレスが変わる
      xt3MemoryAddress = memoryAddress;
      if (xt3MemoryEnabled) {  //SRAMが有効
        if (memoryAddress < 0x00c00000) {  //$00Fxxxxx→$00Bxxxxx
          //$00FxxxxxをROMに戻す
          busSuper (MemoryMappedDevice.MMD_ROM, 0x01000000 - xt3MemorySize, 0x01000000);
          //$00BxxxxxをSRAMにする
          busUser (MemoryMappedDevice.MMD_XTM, 0x00c00000 - xt3MemorySize, 0x00c00000);
        } else {  //$00Bxxxxx→$00Fxxxxx
          if (MainMemory.mmrMemorySizeCurrent < 0x00c00000) {  //12MBフル実装ではない
            //$00Bxxxxxを空きに戻す
            busSuper (MemoryMappedDevice.MMD_NUL, 0x00c00000 - xt3MemorySize, 0x00c00000);
          } else {  //12MBフル実装
            //$00Bxxxxxをメインメモリに戻す
            busSuper (MemoryMappedDevice.MMD_MMR, 0x00c00000 - xt3MemorySize, 0x00c00000);
          }
          //$00FxxxxxをSRAMにする
          busSuper (MemoryMappedDevice.MMD_XTM, 0x01000000 - xt3MemorySize, 0x01000000);
        }
      }
    }
    if (xt3MemoryEnabled != memoryEnabled) {  //SRAMが有効かどうかが変わる
      xt3MemoryEnabled = memoryEnabled;
      if (memoryEnabled) {  //SRAMが有効になる
        if (xt3MemoryAddress < 0x00c00000) {  //$00Bxxxxx
          //$00BxxxxxをSRAMにする
          busUser (MemoryMappedDevice.MMD_XTM, 0x00c00000 - xt3MemorySize, 0x00c00000);
        } else {  //$00Fxxxxx
          //$00FxxxxxをSRAMにする
          busSuper (MemoryMappedDevice.MMD_XTM, 0x01000000 - xt3MemorySize, 0x01000000);
        }
      } else {  //SRAMが無効になる
        if (xt3MemoryAddress < 0x00c00000) {  //$00Bxxxxx
          if (MainMemory.mmrMemorySizeCurrent < 0x00c00000) {  //12MBフル実装ではない
            //$00Bxxxxxを空きに戻す
            busSuper (MemoryMappedDevice.MMD_NUL, 0x00c00000 - xt3MemorySize, 0x00c00000);
          } else {  //12MBフル実装
            //$00Bxxxxxをメインメモリに戻す
            busSuper (MemoryMappedDevice.MMD_MMR, 0x00c00000 - xt3MemorySize, 0x00c00000);
          }
        } else {  //$00Fxxxxx
          //$00FxxxxxをROMに戻す
          busSuper (MemoryMappedDevice.MMD_ROM, 0x01000000 - xt3MemorySize, 0x01000000);
        }
      }
    }
    if (xt3Accelerated != accelerated) {  //MPUが変わる
      xt3Accelerated = accelerated;

      if (accelerated) {  //MC68000→MC68EC030

        //HALTしたときのPCとRnを保存する
        xt3SavedPC = regPC;
        System.arraycopy (regRn, 0, xt3SavedRn, 0, 16);

        //MC68000を停止する
        if (mpuTask != null) {
          mpuClockLimit = 0L;
          System.out.println (Multilingual.mlnJapanese ? "MC68000 を停止します" : "MC68000 stops");
          mpuTask.cancel ();
          mpuTask = null;
        }

        //MC68EC030を起動する
        mdlRequestCore (3);
        mdlClockRequest = mpuClockMHz * 2.0;
        if (MC68EC030.M30_DIV_ZERO_V_FLAG) {
          MC68EC030.m30DivZeroVFlag = false;
        }
        tmrTimer.schedule (new TimerTask () {
          @Override public void run () {
            //MPUの切り替え
            if (mpuCoreType != mdlCoreRequest) {
              mpuCoreType = mdlCoreRequest;
            }
            //クロックの切り替え
            if (0.0 < mdlClockRequest) {  //機種の指定によるクロックの指定がまだ反映されていない
              if (mdlClockRequest == 10.0) {
                mpuClock10MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 50.0 / 3.0) {  //16.7MHz
                mpuClock16MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 25.0) {
                mpuClock25MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 100.0 / 3.0) {  //33.3MHz
                mpuClock33MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 50.0) {
                mpuClock50MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              }
              mpuClockMHz = mdlClockRequest;
              mdlClockRequest = -1.0;  //反映済み。クロックを調整してもリセットする度に元に戻ってしまうのは嫌なので2回目からは変更しない
            }
            mpuSetClockMHz (mpuClockMHz);  //mpuCoreTypeが変更されたときは再計算が必要なのでクロックが変わっていなくても省略できない
            RegisterList.drpSetMPU ();
            mpuSFC = mpuDFC = mpuCACR = mpuBUSCR = mpuUSP = mpuVBR = mpuCAAR = mpuMSP = mpuISP = 0;
            mpuPCR = 0x04300500 | MPU_060_REV << 8;
            MC68060.mmuReset ();  //TCR,ITT0,ITT1,DTT0,DTT1,URP,SRP。060→000/030のときアドレス変換をOFFにする必要がある
            //MC68EC030
            mpuIgnoreAddressError = true;
            fpuOn = fpuMode != 0;
            fpuBox = fpuMotherboardCoprocessor;
            fpuBox.epbSetMC68882 ();
            if (fpuMode == 1) {
              fpuBox.epbSetExtended ();
            } else if (fpuMode == 2) {
              fpuBox.epbSetTriple ();
            }
            fpuBox.epbReset ();
            fpuFPn = fpuBox.epbFPn;
            mpuSetROMWait ((mpuCACR & 0x00000101) != 0);
            //
            regSRT1 = regSRT0 = 0;
            regSRS = REG_SR_S;
            regSRM = 0;
            regSRI = REG_SR_I;
            regCCR = 0;
            Arrays.fill (regRn, 0);
            //PCとSSPをRAMから読み出す
            regRn[15] = MainMemory.mmrRls (0x00000000);
            regPC = MainMemory.mmrRls (0x00000004);
            //割り込み
            mpuIMR = 0;
            mpuIRR = 0;
            if (MC68901.MFP_DELAYED_INTERRUPT) {
              mpuDIRR = 0;
            }
            mpuISR = 0;
            //動作開始
            mpuStart ();
          }
        }, TMR_DELAY);

      } else {  //MC68EC030→MC68000

        //MC68EC030を停止する
        if (mpuTask != null) {
          mpuClockLimit = 0L;
          System.out.println (Multilingual.mlnJapanese ? "MC68EC030 を停止します" : "MC68EC030 stops");
          mpuTask.cancel ();
          mpuTask = null;
        }

        //MC68000を起動する
        mdlRequestCore (0);
        mdlClockRequest = mpuClockMHz * 0.5;
        tmrTimer.schedule (new TimerTask () {
          @Override public void run () {
            //MPUの切り替え
            if (mpuCoreType != mdlCoreRequest) {
              mpuCoreType = mdlCoreRequest;
            }
            //クロックの切り替え
            if (0.0 < mdlClockRequest) {  //機種の指定によるクロックの指定がまだ反映されていない
              if (mdlClockRequest == 10.0) {
                mpuClock10MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 50.0 / 3.0) {  //16.7MHz
                mpuClock16MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 25.0) {
                mpuClock25MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 100.0 / 3.0) {  //33.3MHz
                mpuClock33MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              } else if (mdlClockRequest == 50.0) {
                mpuClock50MenuItem.setSelected (true);
                mpuArbFreqOn = false;
                mpuUtilOn = false;
              }
              mpuClockMHz = mdlClockRequest;
              mdlClockRequest = -1.0;  //反映済み。クロックを調整してもリセットする度に元に戻ってしまうのは嫌なので2回目からは変更しない
            }
            mpuSetClockMHz (mpuClockMHz);  //mpuCoreTypeが変更されたときは再計算が必要なのでクロックが変わっていなくても省略できない
            RegisterList.drpSetMPU ();
            mpuSFC = mpuDFC = mpuCACR = mpuBUSCR = mpuUSP = mpuVBR = mpuCAAR = mpuMSP = mpuISP = 0;
            mpuPCR = 0x04300500 | MPU_060_REV << 8;
            MC68060.mmuReset ();  //TCR,ITT0,ITT1,DTT0,DTT1,URP,SRP。060→000/030のときアドレス変換をOFFにする必要がある
            //MC68000
            mpuIgnoreAddressError = false;
            fpuOn = false;
            mpuSetROMWait (false);
            //
            regSRT1 = regSRT0 = 0;
            regSRS = REG_SR_S;
            regSRM = 0;
            regSRI = REG_SR_I;
            regCCR = 0;
            //HALTしたときのPCとRnを復元する
            regPC = xt3SavedPC;
            System.arraycopy (xt3SavedRn, 0, regRn, 0, 16);
            //割り込み
            mpuIMR = 0;
            mpuIRR = 0;
            if (MC68901.MFP_DELAYED_INTERRUPT) {
              mpuDIRR = 0;
            }
            mpuISR = 0;
            //動作開始
            mpuStart ();
          }
        }, TMR_DELAY);

      }
    }
  }



  //========================================================================================
  //$$MDL 機種
  //  MPUとIPLROMのバージョン

  //                                            MPU        clock    IPLROM      内蔵HD
  public static final int MDL_EXPERT    = 0;  //MC68000    10MHz    IPLROM 1.0  SASI
  public static final int MDL_SUPER     = 1;  //MC68000    10MHz    IPLROM 1.0  SCSI
  public static final int MDL_XVI       = 2;  //MC68000    16.7MHz  IPLROM 1.1  SCSI
  public static final int MDL_XELLENT30 = 3;  //MC68000    16.7MHz  IPLROM 1.1  SCSI
  //                                            MC68EC030  33.3MHz  IPLROM 1.1  SCSI
  public static final int MDL_COMPACT   = 4;  //MC68000    25MHz    IPLROM 1.2  SCSI
  public static final int MDL_HYBRID    = 5;  //MC68000    33.3MHz  IPLROM 1.6  SCSI
  public static final int MDL_X68030    = 6;  //MC68EC030  25MHz    IPLROM 1.3  SCSI
  public static final int MDL_060TURBO  = 7;  //MC68060    33.3MHz  IPLROM 1.6  SCSI

  public static JMenu mdlMenu;  //メニュー

  public static JRadioButtonMenuItem mdlEXPERTMenuItem;
  public static JRadioButtonMenuItem mdlSUPERMenuItem;
  public static JRadioButtonMenuItem mdlXVIMenuItem;
  public static JRadioButtonMenuItem mdlXellent30MenuItem;
  public static JRadioButtonMenuItem mdlCompactMenuItem;
  public static JRadioButtonMenuItem mdlHybridMenuItem;
  public static JRadioButtonMenuItem mdlX68030MenuItem;
  public static JRadioButtonMenuItem mdl060turboMenuItem;
  public static JCheckBoxMenuItem mdlMC68010MenuItem;

  public static JRadioButtonMenuItem fpuMenuItem0;
  public static JRadioButtonMenuItem fpuMenuItem1;
  public static JRadioButtonMenuItem fpuMenuItem2;
  public static JCheckBoxMenuItem fpuMenuItem3;

  public static int mdlModel;  //機種の指定
  public static int mdlBodyRequest;  //本体の指定。0=X68000,3=X68030
  public static int mdlCurrentBody;
  public static int mdlCoreRequest;  //MPUの指定。0/1/3/6
  public static double mdlClockRequest;  //クロックの指定。-1.0=反映済み
  public static boolean mdlSCSIINRequest;  //SCSIINの指定
  public static int fpuMode;  //0=FPUなし,1=拡張精度,2=三倍精度
  public static boolean fpuOn;  //true=FPUあり(2<=mpuCoreType&&fpuMode!=0)
  public static boolean fpuFullSpec;  //true=フルスペックFPU。MC68060のときすべての浮動小数点命令を実装する

  //mdlInit ()
  //  機種の指定を読み取る
  public static void mdlInit () {

    //設定
    //  機種
    String paramModel = Settings.sgsGetString ("model");
    switch (paramModel.toLowerCase ()) {
    case "expert":
      mdlModel = MDL_EXPERT;
      break;
    case "super":
      mdlModel = MDL_SUPER;
      break;
    case "xvi":
      mdlModel = MDL_XVI;
      break;
    case "xellent30":
      mdlModel = MDL_XELLENT30;
      break;
    case "compact":
      mdlModel = MDL_COMPACT;
      break;
    case "x68030":
      mdlModel = MDL_X68030;
      break;
    case "hybrid":
      mdlModel = MDL_HYBRID;
      break;
    case "060turbo":
      mdlModel = MDL_060TURBO;
      break;
    case "":
      mdlModel = MDL_HYBRID;
      break;
    default:
      System.out.println (Multilingual.mlnJapanese ?
                          paramModel + " は不明な機種です" :
                          paramModel + " is unknown model");
      mdlModel = MDL_HYBRID;
    }
    //  本体
    switch (mdlModel) {
    case MDL_EXPERT:
    case MDL_SUPER:
    case MDL_XVI:
    case MDL_XELLENT30:
    case MDL_COMPACT:
    case MDL_HYBRID:
      mdlBodyRequest = 0;
      break;
    case MDL_X68030:
    case MDL_060TURBO:
      mdlBodyRequest = 3;
      break;
    default:
      mdlBodyRequest = 0;
    }
    //  MPU
    String paramMpu = Settings.sgsGetString ("mpu");
    switch (paramMpu.toLowerCase ()) {
    case "0":
    case "68000":
    case "mc68000":
    case "expert":
    case "super":
    case "xvi":
    case "compact":
    case "hybrid":
      mdlCoreRequest = 0;
      break;
    case "1":
    case "68010":
    case "mc68010":
      mdlCoreRequest = 1;
      break;
    case "3":
    case "68ec030":
    case "mc68ec030":
    case "x68030":
      mdlCoreRequest = 3;
      break;
    case "6":
    case "68060":
    case "mc68060":
    case "060turbo":
      mdlCoreRequest = 6;
      break;
    case "":
    case "-1":
      mdlCoreRequest = -1;
      break;
    default:
      System.out.println (Multilingual.mlnJapanese ?
                          paramMpu + " は不明な MPU です" :
                          paramMpu + " is unknown MPU");
    }
    if (mdlCoreRequest == -1) {
      switch (mdlModel) {
      case MDL_EXPERT:
      case MDL_SUPER:
      case MDL_XVI:
      case MDL_XELLENT30:
      case MDL_COMPACT:
      case MDL_HYBRID:
        mdlCoreRequest = 0;
        break;
      case MDL_X68030:
        mdlCoreRequest = 3;
        break;
      case MDL_060TURBO:
        mdlCoreRequest = 6;
        break;
      default:
        mdlCoreRequest = 0;
      }
    }
    //  任意の負荷率
    mpuUtilOn = Settings.sgsGetOnOff ("util");  //任意の負荷率
    mpuUtilRatio = fmtParseInt (Settings.sgsGetString ("ratio"), 0, 1, 100, 90);  //任意の負荷率
    //  クロック
    String paramClock = Settings.sgsGetString ("clock");
    switch (paramClock.toLowerCase ()) {
    case "expert":
    case "super":
      mdlClockRequest = 10.0;  //10MHz
      break;
    case "xvi":
    case "xellent30":
      mdlClockRequest = 50.0 / 3.0;  //16.7MHz
      break;
    case "compact":
    case "x68030":
      mdlClockRequest = 25.0;  //25MHz
      break;
    case "hybrid":
    case "060turbo":
      mdlClockRequest = 100.0 / 3.0;  //33.3MHz
      break;
    case "":
    case "-1":
      mdlClockRequest = -1.0;
      break;
    default:
      mdlClockRequest = -1.0;
      if (paramClock.matches ("^(?:" +
                              "[-+]?" +  //符号
                              "(?:[0-9]+(?:\\.[0-9]*)?|\\.[0-9]+)" +  //仮数部
                              "(?:[Ee][-+]?[0-9]+)?" +  //指数部
                              ")$")) {
        double d = Double.parseDouble (paramClock);
        if (1.0 <= d && d <= 1000.0) {
          mdlClockRequest = d;
          mpuUtilOn = false;  //クロックが数字で指定されたら任意の負荷率を無効にする
        }
      }
      if (mdlClockRequest == -1.0) {
        System.out.println (Multilingual.mlnJapanese ?
                            paramClock + " は不明な動作周波数です" :
                            paramClock + " is unknown clock frequency");
      }
    }
    if (mdlClockRequest == -1.0) {
      switch (mdlModel) {
      case MDL_EXPERT:
      case MDL_SUPER:
        mdlClockRequest = 10.0;  //10MHz
        break;
      case MDL_XVI:
      case MDL_XELLENT30:
        mdlClockRequest = 50.0 / 3.0;  //16.7MHz
        break;
      case MDL_COMPACT:
      case MDL_X68030:
        mdlClockRequest = 25.0;  //25MHz
        break;
      case MDL_HYBRID:
      case MDL_060TURBO:
        mdlClockRequest = 100.0 / 3.0;  //33.3MHz
        break;
      default:
        mdlClockRequest = 10.0;  //10MHz
      }
    }
    //ここでmdlClockRequest==-1.0は不可
    //  任意の周波数
    mpuArbFreqMHz = fmtParseInt (Settings.sgsGetString ("mhz"), 0, 1, 1000, 100);
    if (mpuUtilOn) {
      mpuArbFreqOn = false;
    } else {
      mpuArbFreqOn = !(mdlClockRequest == 10.0 ||
                       mdlClockRequest == 50.0 / 3.0 ||  //16.7MHz
                       mdlClockRequest == 25.0 ||
                       mdlClockRequest == 100.0 / 3.0 ||  //33.3MHz
                       mdlClockRequest == 50.0);
      if (mpuArbFreqOn) {
        mpuArbFreqMHz = (int) mdlClockRequest;
      }
    }
    //  FPU
    fpuMode = fmtParseInt (Settings.sgsGetString ("fpumode"), 0, 0, 2, 1);  //FPUモード
    fpuOn = 2 <= mdlCoreRequest && fpuMode != 0;
    fpuFullSpec = Settings.sgsGetOnOff ("fullspecfpu");  //フルスペックFPU
    //  SCSIIN
    switch (mdlModel) {
    case MDL_EXPERT:
      mdlSCSIINRequest = false;
      break;
    case MDL_SUPER:
    case MDL_XVI:
    case MDL_XELLENT30:
    case MDL_COMPACT:
    case MDL_HYBRID:
    case MDL_X68030:
    case MDL_060TURBO:
      mdlSCSIINRequest = true;
      break;
    default:
      mdlSCSIINRequest = true;
    }

  }  //mdlInit()

  public static void mdlMakeMenu () {

    //アクションリスナー
    ActionListener listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        switch (ae.getActionCommand ()) {
        case "X68000 EXPERT (10MHz)":
          mdlRequestModel (MDL_EXPERT, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "X68000 SUPER (10MHz)":
          mdlRequestModel (MDL_SUPER, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "X68000 XVI (16.7MHz)":
          mdlRequestModel (MDL_XVI, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "X68000 Compact (25MHz)":
          mdlRequestModel (MDL_COMPACT, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "X68000 Hybrid (33.3MHz)":
          mdlRequestModel (MDL_HYBRID, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "Xellent30 (33.3MHz)":
          mdlRequestModel (MDL_XELLENT30, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "X68030 (25MHz)":
          mdlRequestModel (MDL_X68030, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "060turbo (33.3MHz)":
          mdlRequestModel (MDL_060TURBO, true);
          //リセット
          mpuReset (-1, -1);
          break;
        case "MC68010":
          mdlRequestCore (((JCheckBoxMenuItem) source).isSelected () ? 1 : 0);
          break;

        case "No FPU":  //FPUなし
          fpuMode = 0;
          fpuOn = false;
          break;
        case "Extended precision (19 digit)":  //拡張精度 (19 桁)
          fpuMode = 1;
          fpuOn = 2 <= mpuCoreType;
          break;
        case "Triple precision (24 digit)":  //三倍精度 (24 桁)
          fpuMode = 2;
          fpuOn = 2 <= mpuCoreType;
          break;
        case "Full specification FPU":  //フルスペック FPU
          fpuFullSpec = ((JCheckBoxMenuItem) source).isSelected ();
          break;

        case "Built-in SCSI port":  //内蔵 SCSI ポート
          mdlSCSIINRequest = ((JCheckBoxMenuItem) source).isSelected ();
          break;

        }
      }
    };

    //メニュー
    ButtonGroup modelGroup = new ButtonGroup ();
    mdlMenu = Multilingual.mlnText (
      ComponentFactory.createMenu (
        "Change the model and reset",
        mdlEXPERTMenuItem    = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_EXPERT   , "X68000 EXPERT (10MHz)",   listener),
        mdlSUPERMenuItem     = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_SUPER    , "X68000 SUPER (10MHz)",    listener),
        mdlXVIMenuItem       = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_XVI      , "X68000 XVI (16.7MHz)",    listener),
        mdlXellent30MenuItem = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_XELLENT30, "Xellent30 (33.3MHz)",     listener),
        mdlCompactMenuItem   = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_COMPACT  , "X68000 Compact (25MHz)",  listener),
        mdlHybridMenuItem    = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_HYBRID   , "X68000 Hybrid (33.3MHz)", listener),
        ComponentFactory.createHorizontalSeparator (),
        mdlX68030MenuItem    = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_X68030   , "X68030 (25MHz)",          listener),
        mdl060turboMenuItem  = ComponentFactory.createRadioButtonMenuItem (modelGroup, mdlModel == MDL_060TURBO , "060turbo (33.3MHz)",      listener),
        ComponentFactory.createHorizontalSeparator (),
        mdlMC68010MenuItem   = ComponentFactory.createCheckBoxMenuItem (mdlCoreRequest == 1, "MC68010", listener)
        ),
      "ja", "機種を変更してリセット");

    ButtonGroup fpuGroup = new ButtonGroup ();
    fpuMenuItem0 = ComponentFactory.setEnabled (
      Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (fpuGroup, fpuMode == 0, "No FPU" , listener),
                            "ja", "FPU なし"),
      2 <= mdlCoreRequest);
    fpuMenuItem1 = ComponentFactory.setEnabled (
      Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (fpuGroup, fpuMode == 1, "Extended precision (19 digit)" , listener),
                            "ja", "拡張精度 (19 桁)"),
      2 <= mdlCoreRequest);
    fpuMenuItem2 = ComponentFactory.setEnabled (
      Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (fpuGroup, fpuMode == 2, "Triple precision (24 digit)" , listener),
                            "ja", "三倍精度 (24 桁)"),
      2 <= mdlCoreRequest);
    fpuMenuItem3 = ComponentFactory.setEnabled (
      Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (fpuFullSpec, "Full specification FPU", listener), "ja", "フルスペック FPU"),
      4 <= mdlCoreRequest);

    //機種
    //mdlRequestModel (mdlModel, true);
    //mdlRequestModel (mdlModel, false);  //SGSでmdlClockRequestを設定済み

  }  //mdlMakeMenu

  //mdlRequestModel (model, all)
  //  機種の指定をリクエストに変換する
  //  all  動的に変更できる項目も初期化する
  public static void mdlRequestModel (int model, boolean all) {
    mdlModel = model;
    switch (model) {
    case MDL_EXPERT:
      mdlBodyRequest = 0;
      mdlEXPERTMenuItem.setSelected (true);
      mdlRequestCore (0);
      mdlRequestSCSIIN (false);
      if (all) {
        mdlClockRequest = 10.0;
      }
      break;
    case MDL_SUPER:
      mdlBodyRequest = 0;
      mdlSUPERMenuItem.setSelected (true);
      mdlRequestCore (0);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 10.0;
      }
      break;
    case MDL_XVI:
      mdlBodyRequest = 0;
      mdlXVIMenuItem.setSelected (true);
      mdlRequestCore (0);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 50.0 / 3.0;  //16.7MHz
      }
      break;
    case MDL_XELLENT30:
      mdlBodyRequest = 0;
      mdlXellent30MenuItem.setSelected (true);
      mdlRequestCore (0);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 50.0 / 3.0;  //16.7MHz
      }
      break;
    case MDL_COMPACT:
      mdlBodyRequest = 0;
      mdlCompactMenuItem.setSelected (true);
      mdlRequestCore (0);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 25.0;
      }
      break;
    case MDL_HYBRID:
      mdlBodyRequest = 0;
      mdlHybridMenuItem.setSelected (true);
      mdlRequestCore (mdlCoreRequest == 0 || mdlCoreRequest == 1 ? mdlCoreRequest : 0);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 100.0 / 3.0;  //33.3MHz
      }
      break;
    case MDL_X68030:
      mdlBodyRequest = 3;
      mdlX68030MenuItem.setSelected (true);
      mdlRequestCore (3);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 25.0;
      }
      break;
    case MDL_060TURBO:
      mdlBodyRequest = 3;
      mdl060turboMenuItem.setSelected (true);
      mdlRequestCore (6);
      mdlRequestSCSIIN (true);
      if (all) {
        mdlClockRequest = 100.0 / 3.0;  //33.3MHz
      }
      break;
    }
  }  //mdlRequestModel(model)

  public static void mdlRequestCore (int core) {
    mdlCoreRequest = core;
    if (core == 0 || core == 1) {
      mdlMC68010MenuItem.setSelected (core == 1);
    }
    fpuMenuItem0.setEnabled (2 <= core);
    fpuMenuItem1.setEnabled (2 <= core);
    fpuMenuItem2.setEnabled (2 <= core);
    fpuMenuItem3.setEnabled (4 <= core);
  }

  public static void mdlRequestSCSIIN (boolean request) {
    mdlSCSIINRequest = request;
    HDC.hdcSASIMenuItem.setSelected (!request);
    SPC.spcSCSIINMenuItem.setSelected (request);
  }  //mdlRequestSCSIIN()



  //========================================================================================
  //$$MPU MPU

  //コンパイルスイッチ
  public static final boolean MPU_INLINE_EXCEPTION = true;  //true=例外処理をインライン展開する。速くなる
  public static final boolean MPU_COMPOUND_POSTINCREMENT = false;  //true=(pc)+をbusRbs((pc+=2)-1),busRws((pc+=2)-2),busRls((pc+=4)-4)のように書く。見た目はスマートだが最適化しにくくなる？

  public static final boolean MPU_SWITCH_MISC_OPCODE = false;  //true=RTSなどのswitchのキーはオペコード全体,false=下位6bit
  public static final boolean MPU_SWITCH_BCC_CONDITION = false;  //true=オペコードのswitchでBccをccで分類する
  public static final boolean MPU_SWITCH_BCC_OFFSET = false;  //true=オペコードのswitchでBRA/BSR/Bccを8bitオフセットの上位2bitで分類する
  public static final boolean MPU_SWITCH_SCC_CONDITION = true;  //true=オペコードのswitchでScc/DBRA/DBcc/TRAPccをccで分類する

  public static final boolean MPU_OMIT_EXTRA_READ = false;  //true=余分なリードを省略する。MC68000のBRA.S/BSR.S/Bcc.S/CLR/SF
  public static final boolean MPU_OMIT_OFFSET_READ = false;  //true=条件分岐命令で分岐しないときオフセットのリードを省略する

  //TMR_FREQ単位のタイマカウンタで到達し得ない時刻を表す定数
  //  TMR_FREQ=10^12のとき到達するのに3.5ヶ月かかる
  //  3.5ヶ月間動かしっぱなしにすると破綻する
  public static final long FAR_FUTURE = 0x7fffffffffffffffL;

  //ステータスレジスタ
  //  トレース
  //     srT1    srT0
  //    0x0000  0x0000  トレースなし
  //    0x0000  0x4000  フローの変化をトレース
  //    0x8000  0x0000  すべての命令をトレース
  //    0x8000  0x4000  未定義
  public static final int REG_SR_T1  = 0b10000000_00000000;
  public static final int REG_SR_T0  = 0b01000000_00000000;  //(020/030/040)
  //  モード
  //      srS     srM
  //    0x0000  0x0000  ユーザモード(USPを使用)
  //    0x0000  0x1000  ユーザモード(USPを使用)
  //    0x2000  0x0000  スーパーバイザ割り込みモード(ISPを使用)
  //    0x2000  0x1000  スーパーバイザマスタモード(MSPを使用)
  public static final int REG_SR_S   = 0b00100000_00000000;
  public static final int REG_SR_M   = 0b00010000_00000000;  //(020/030/040/060)
  //  割り込み
  public static final int REG_SR_I   = 0b00000111_00000000;

  //コンディションコードレジスタ
  public static final int REG_CCR_X  = 0b00000000_00010000;
  public static final int REG_CCR_N  = 0b00000000_00001000;
  public static final int REG_CCR_Z  = 0b00000000_00000100;
  public static final int REG_CCR_V  = 0b00000000_00000010;
  public static final int REG_CCR_C  = 0b00000000_00000001;
  public static final int REG_CCR_MASK = REG_CCR_X | REG_CCR_N | REG_CCR_Z | REG_CCR_V | REG_CCR_C;  //CCRの有効なビット

  public static char[] REG_CCRXMAP = "00000000000000001111111111111111".toCharArray ();
  public static char[] REG_CCRNMAP = "00000000111111110000000011111111".toCharArray ();
  public static char[] REG_CCRZMAP = "00001111000011110000111100001111".toCharArray ();
  public static char[] REG_CCRVMAP = "00110011001100110011001100110011".toCharArray ();
  public static char[] REG_CCRCMAP = "01010101010101010101010101010101".toCharArray ();

  //割り込みレベル
  //  順序の変更やレベルの追加はコードの変更が必要
  public static final int MPU_IOI_INTERRUPT_LEVEL = 1;
  public static final int MPU_EB2_INTERRUPT_LEVEL = 2;
  public static final int MPU_DMA_INTERRUPT_LEVEL = 3;
  public static final int MPU_SCC_INTERRUPT_LEVEL = 5;
  public static final int MPU_MFP_INTERRUPT_LEVEL = 6;
  public static final int MPU_SYS_INTERRUPT_LEVEL = 7;
  public static final int MPU_IOI_INTERRUPT_MASK = 0x80 >> MPU_IOI_INTERRUPT_LEVEL;  //0x40
  public static final int MPU_EB2_INTERRUPT_MASK = 0x80 >> MPU_EB2_INTERRUPT_LEVEL;  //0x20
  public static final int MPU_DMA_INTERRUPT_MASK = 0x80 >> MPU_DMA_INTERRUPT_LEVEL;  //0x10
  public static final int MPU_SCC_INTERRUPT_MASK = 0x80 >> MPU_SCC_INTERRUPT_LEVEL;  //0x04
  public static final int MPU_MFP_INTERRUPT_MASK = 0x80 >> MPU_MFP_INTERRUPT_LEVEL;  //0x02
  public static final int MPU_SYS_INTERRUPT_MASK = 0x80 >> MPU_SYS_INTERRUPT_LEVEL;  //0x01

  public static final boolean MPU_INTERRUPT_SWITCH = true;  //true=最上位の割り込みをswitchで判別する

  //コンディションコード
  public static final boolean T = true;
  public static final boolean F = false;
  //  cccc==CCCC_cc                               cccc  cc
  public static final int CCCC_T  = 0b0000;  //0000  T       1                always true
  public static final int CCCC_F  = 0b0001;  //0001  F       0                always false
  public static final int CCCC_HI = 0b0010;  //0010  HI      ~C&~Z            high
  public static final int CCCC_LS = 0b0011;  //0011  LS      C|Z              low or same
  public static final int CCCC_CC = 0b0100;  //0100  CC(HS)  ~C               carry clear (high or same)
  public static final int CCCC_CS = 0b0101;  //0101  CS(LO)  C                carry set (low)
  public static final int CCCC_NE = 0b0110;  //0110  NE      ~Z               not equal
  public static final int CCCC_EQ = 0b0111;  //0111  EQ      Z                equal
  public static final int CCCC_VC = 0b1000;  //1000  VC      ~V               overflow clear
  public static final int CCCC_VS = 0b1001;  //1001  VS      V                overflow set
  public static final int CCCC_PL = 0b1010;  //1010  PL      ~N               plus
  public static final int CCCC_MI = 0b1011;  //1011  MI      N                minus
  public static final int CCCC_GE = 0b1100;  //1100  GE      N&V|~N&~V        greater or equal
  public static final int CCCC_LT = 0b1101;  //1101  LT      N&~V|~N&V        less than
  public static final int CCCC_GT = 0b1110;  //1110  GT      N&V&~Z|~N&~V&~Z  greater than
  public static final int CCCC_LE = 0b1111;  //1111  LE      Z|N&~V|~N&V      less or equal
  //F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //X
  //F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,  //N
  //F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //Z
  //F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,  //V
  //F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //C
  //  BCCMAP[CCCC_cc<<5|ccr]==trueならば条件成立
  public static final boolean[] BCCMAP = {
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //T       NF
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //F       NT
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //HI      NLS
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //LS      NHI
    T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,  //CC(HS)  NCS(NLO)
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //CS(LO)  NCC(NHS)
    T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,  //NE(NZ)  NEQ(NZE)
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //EQ(ZE)  NNE(NNZ)
    T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,  //VC      NVS
    F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,F,F,T,T,  //VS      NVC
    T,T,T,T,T,T,T,T,F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,F,F,F,F,F,F,F,F,  //PL      NMI
    F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,  //MI      NPL
    T,T,F,F,T,T,F,F,F,F,T,T,F,F,T,T,T,T,F,F,T,T,F,F,F,F,T,T,F,F,T,T,  //GE      NLT
    F,F,T,T,F,F,T,T,T,T,F,F,T,T,F,F,F,F,T,T,F,F,T,T,T,T,F,F,T,T,F,F,  //LT      NGE
    T,T,F,F,F,F,F,F,F,F,T,T,F,F,F,F,T,T,F,F,F,F,F,F,F,F,T,T,F,F,F,F,  //GT      NLE
    F,F,T,T,T,T,T,T,T,T,F,F,T,T,T,T,F,F,T,T,T,T,T,T,T,T,F,F,T,T,T,T,  //LE      NGT
  };

  //  MPU_CCCMAP[cccc<<5|ccr]==(BCCMAP[cccc<<5|ccr]?'1':'0')
  public static final char[] MPU_CCCMAP = (
    "11111111111111111111111111111111" +
    "00000000000000000000000000000000" +
    "10100000101000001010000010100000" +
    "01011111010111110101111101011111" +
    "10101010101010101010101010101010" +
    "01010101010101010101010101010101" +
    "11110000111100001111000011110000" +
    "00001111000011110000111100001111" +
    "11001100110011001100110011001100" +
    "00110011001100110011001100110011" +
    "11111111000000001111111100000000" +
    "00000000111111110000000011111111" +
    "11001100001100111100110000110011" +
    "00110011110011000011001111001100" +
    "11000000001100001100000000110000" +
    "00111111110011110011111111001111").toCharArray ();

  //  (MPU_CC_cc<<ccr<0)==trueならば条件成立
  //  (MPU_CC_cc<<ccr<0)==BCCMAP[CCCC_cc<<5|ccr]
  public static final int MPU_CC_T  = 0b11111111111111111111111111111111;  //T
  public static final int MPU_CC_F  = 0b00000000000000000000000000000000;  //F
  public static final int MPU_CC_HI = 0b10100000101000001010000010100000;  //HI
  public static final int MPU_CC_LS = 0b01011111010111110101111101011111;  //LS
  public static final int MPU_CC_HS = 0b10101010101010101010101010101010;  //HS
  public static final int MPU_CC_LO = 0b01010101010101010101010101010101;  //LO
  public static final int MPU_CC_NE = 0b11110000111100001111000011110000;  //NE
  public static final int MPU_CC_EQ = 0b00001111000011110000111100001111;  //EQ
  public static final int MPU_CC_VC = 0b11001100110011001100110011001100;  //VC
  public static final int MPU_CC_VS = 0b00110011001100110011001100110011;  //VS
  public static final int MPU_CC_PL = 0b11111111000000001111111100000000;  //PL
  public static final int MPU_CC_MI = 0b00000000111111110000000011111111;  //MI
  public static final int MPU_CC_GE = 0b11001100001100111100110000110011;  //GE
  public static final int MPU_CC_LT = 0b00110011110011000011001111001100;  //LT
  public static final int MPU_CC_GT = 0b11000000001100001100000000110000;  //GT
  public static final int MPU_CC_LE = 0b00111111110011110011111111001111;  //LE

  //TST.Bのテーブル
  //  z=255&(～);ccr=ccr&CCR_X|MPU_TSTB_TABLE[z]をz=～;ccr=ccr&CCR_X|MPU_TSTB_TABLE[255&z]にすると速くなることがある
  //  インデックスが明示的にマスクされていると最適化しやすいのだろう
/*
  public static final byte[] MPU_TSTB_TABLE = new byte[256];
  static {
    for (int z = 0; z < 256; z++) {
      MPU_TSTB_TABLE[z] = (byte) (z >> 7 << 3 | z - 1 >> 6 & CCR_Z);
    }
  }  //static
*/
/*
  public static final byte[] MPU_TSTB_TABLE = {
    REG_CCR_Z, 0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    0,         0,         0,         0,         0,         0,         0,         0,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
    REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N, REG_CCR_N,
  };
*/
  //  perl misc/itob.pl xeij/XEiJ.java MPU_TSTB_TABLE
  public static final byte[] MPU_TSTB_TABLE = "\4\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b".getBytes (XEiJ.ISO_8859_1);

  //BITREVのテーブル
  //  初期化コードが大きくなりすぎるので最初から展開しておくにはクラスを分ける必要がある
  public static final int[] MPU_BITREV_TABLE_0 = new int[2048];
  public static final int[] MPU_BITREV_TABLE_1 = new int[2048];
  public static final int[] MPU_BITREV_TABLE_2 = new int[2048];
  static {
    for (int i = 0; i < 2048; i++) {
      MPU_BITREV_TABLE_2[i] = (MPU_BITREV_TABLE_1[i] = (MPU_BITREV_TABLE_0[i] = Integer.reverse (i)) >>> 11) >>> 11;
    }
  }

  //アドレッシングモード
  //                                                              data  memory  control  alterable
  public static final int EA_DR = 0b000_000;  //D  Dr             x                      x
  public static final int EA_AR = 0b001_000;  //A  Ar                                    x
  public static final int EA_MM = 0b010_000;  //M  (Ar)           x     x       x        x
  public static final int EA_MP = 0b011_000;  //+  (Ar)+          x     x                x
  public static final int EA_MN = 0b100_000;  //-  -(Ar)          x     x                x
  public static final int EA_MW = 0b101_000;  //W  (d16,Ar)       x     x       x        x
  public static final int EA_MX = 0b110_000;  //X  (d8,Ar,Rn.wl)  x     x       x        x
  public static final int EA_ZW = 0b111_000;  //Z  (xxx).W        x     x       x        x
  public static final int EA_ZL = 0b111_001;  //Z  (xxx).L        x     x       x        x
  public static final int EA_PW = 0b111_010;  //P  (d16,PC)       x     x       x
  public static final int EA_PX = 0b111_011;  //P  (d8,PC,Rn.wl)  x     x       x
  public static final int EA_IM = 0b111_100;  //I  #<data>        x
  public static final int MMM_DR = EA_DR >> 3;
  public static final int MMM_AR = EA_AR >> 3;
  public static final int MMM_MM = EA_MM >> 3;
  public static final int MMM_MP = EA_MP >> 3;
  public static final int MMM_MN = EA_MN >> 3;
  public static final int MMM_MW = EA_MW >> 3;
  public static final int MMM_MX = EA_MX >> 3;
  public static final long EAM_DR = 0xff00000000000000L >>> EA_DR;
  public static final long EAM_AR = 0xff00000000000000L >>> EA_AR;
  public static final long EAM_MM = 0xff00000000000000L >>> EA_MM;
  public static final long EAM_MP = 0xff00000000000000L >>> EA_MP;
  public static final long EAM_MN = 0xff00000000000000L >>> EA_MN;
  public static final long EAM_MW = 0xff00000000000000L >>> EA_MW;
  public static final long EAM_MX = 0xff00000000000000L >>> EA_MX;
  public static final long EAM_ZW = 0x8000000000000000L >>> EA_ZW;
  public static final long EAM_ZL = 0x8000000000000000L >>> EA_ZL;
  public static final long EAM_PW = 0x8000000000000000L >>> EA_PW;
  public static final long EAM_PX = 0x8000000000000000L >>> EA_PX;
  public static final long EAM_IM = 0x8000000000000000L >>> EA_IM;
  public static final long EAM_ALL = EAM_DR|EAM_AR|EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX|EAM_IM;  //|DAM+-WXZPI|すべて
  public static final long EAM_ALT = EAM_DR|EAM_AR|EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL                     ;  //|DAM+-WXZ  |可変
  public static final long EAM_DAT = EAM_DR       |EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX|EAM_IM;  //|D M+-WXZPI|データ
  public static final long EAM_DME = EAM_DR       |EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX       ;  //|D M+-WXZP |データレジスタ直接またはメモリ
  public static final long EAM_DLT = EAM_DR       |EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL                     ;  //|D M+-WXZ  |データ可変
  public static final long EAM_DCN = EAM_DR       |EAM_MM              |EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX       ;  //|D M  WXZP |データレジスタ直接または制御
  public static final long EAM_DCL = EAM_DR       |EAM_MM              |EAM_MW|EAM_MX|EAM_ZW|EAM_ZL                     ;  //|D M  WXZ  |データレジスタ直接または制御可変
  public static final long EAM_ANY =               EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX|EAM_IM;  //|  M+-WXZPI|レジスタ以外
  public static final long EAM_MEM =               EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX       ;  //|  M+-WXZP |メモリ
  public static final long EAM_MLT =               EAM_MM|EAM_MP|EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL                     ;  //|  M+-WXZ  |メモリ可変
  public static final long EAM_RDL =               EAM_MM|EAM_MP       |EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX       ;  //|  M+ WXZP |リードリスト
  public static final long EAM_WTL =               EAM_MM       |EAM_MN|EAM_MW|EAM_MX|EAM_ZW|EAM_ZL                     ;  //|  M -WXZ  |ライトリスト
  public static final long EAM_CNT =               EAM_MM              |EAM_MW|EAM_MX|EAM_ZW|EAM_ZL|EAM_PW|EAM_PX       ;  //|  M  WXZP |制御
  public static final long EAM_CLT =               EAM_MM              |EAM_MW|EAM_MX|EAM_ZW|EAM_ZL                     ;  //|  M  WXZ  |制御可変

  //汎用レジスタ
  //  r[15]は現在のスタックポインタ
  //                              ユーザモード      スーパーバイザモード
  //                                            マスタモード  割り込みモード
  //     ユーザスタックポインタ     regRn[15]       mpuUSP         mpuUSP
  //     マスタスタックポインタ      mpuMSP       regRn[15]        mpuMSP
  //    割り込みスタックポインタ     mpuISP        mpuISP        regRn[15]
  //  モードによってmpuUSP,mpuMSP,mpuISPのいずれかをregRn[15]にコピーして使う
  //  他のモードに切り替えるときはregRn[15]をコピー元のmpuUSP,mpuMSP,mpuISPのいずれかに書き戻してから切り替える
  //  例えばユーザモードから割り込みモードへ移行するときはregRn[15]→mpuUSP,mpuISP→regRn[15]とする
  public static final int[] regRn = new int[16 + 1];  //汎用レジスタ。regRn[16]はテンポラリ。レジスタの更新を回避するとき条件分岐する代わりにregRn[16]に書き込む

  //プログラムカウンタ
  public static int regPC;  //プログラムカウンタ
  public static int regPC0;  //実行中の命令の先頭アドレス

  //オペコードレジスタ
  public static int regOC;  //オペコード。16bit、0拡張

  //ステータスレジスタ
  public static int regSRT1;  //ステータスレジスタのT1。0=トレースなし,REG_SR_T1=すべての命令をトレース
  public static int regSRT0;  //ステータスレジスタのT0(020/030/040)。0=トレースなし,REG_SR_T0=フローの変化をトレース
  public static int mpuTraceFlag;  //トレースフラグ。命令実行前にregSRT1をコピーし、分岐命令を実行するときregSRT0をorする。例外が発生したとき0にする。命令実行後に0でなければトレース例外を発動する
  public static int regSRS;  //ステータスレジスタのS。0=ユーザモード,REG_SR_S=スーパーバイザモード
  public static int regSRM;  //ステータスレジスタのM(020/030/040/060)。0=割り込みモード,REG_SR_M=マスタモード。割り込みを開始するときクリアされる
  public static int regSRI;  //ステータスレジスタのI2,I1,I0

  //コンディションコードレジスタ
  public static int regCCR;  //コンディションコードレジスタ。000XNZVC。0のビットは0に固定

  //mpuIMR
  //  インタラプトマスクレジスタ
  //  ビットの並び順は01234567
  //  現在の割り込みマスクレベルよりも高いビットをセットしたマスク
  //    割り込みマスクレベルがnのときレベル1..nの割り込みが禁止されてレベルn+1..7の割り込みが許可される
  //    ただし、割り込みマスクレベルが7のときはレベル7割り込みの処理中でなければレベル7割り込みが許可される
  //  mpuIMR&mpuIRRで最下位の1のビットの割り込みを受け付ける
  //  irpSetSRでmpuIMR&mpuISRで1のビットの割り込みを終了する
  //  mpuIMR=0x7f>>(srI>>8)|~mpuISR&1
  //      srI     mpuIMR    受け付ける割り込みのレベルと順序
  //    0x0700  0b00000000  なし  レベル7割り込みの処理中のとき
  //            0b00000001  7     レベル7割り込みの処理中ではないとき
  //    0x0600  0b00000001  7
  //    0x0500  0b00000011  7 6
  //    0x0400  0b00000111  7 6 5
  //    0x0300  0b00001111  7 6 5 4
  //    0x0200  0b00011111  7 6 5 4 3
  //    0x0100  0b00111111  7 6 5 4 3 2
  //    0x0000  0b01111111  7 6 5 4 3 2 1
  public static int mpuIMR;

  //mpuIRR
  //  インタラプトリクエストレジスタ
  //  ビットの並び順は01234567
  //  デバイスが割り込みを要求するときレベルに対応するビットをセットする
  //  コアが割り込みを受け付けるときレベルに対応するビットをクリアしてデバイスのacknowledge()を呼び出す
  public static int mpuIRR;
  public static int mpuDIRR;  //遅延割り込み要求

  //mpuISR
  //  インタラプトインサービスレジスタ
  //  ビットの並び順は01234567
  //  コアが割り込み処理を開始するときレベルに対応するビットをセットする
  //  割り込みマスクレベルが下がったとき新しいレベルよりも高いレベルの割り込み処理が終了したものとみなし、
  //  レベルに対応するビットをクリアしてデバイスのdone()を呼び出す
  //  done()が呼び出された時点でまだ処理されていない割り込みが残っているデバイスは再度割り込みを要求する
  public static int mpuISR;

  //制御レジスタ
  public static int mpuSFC;    //000  -12346  SFC
  public static int mpuDFC;    //001  -12346  DFC
  public static int mpuCACR;   //002  --2346  CACR
  //protected static int mpuTC;     //003  ----46  TC                        030MMUのTC
  //protected static int mpuITT0;   //004  ----46  ITT0   IACR0 @ MC68EC040  030MMUのTT0
  //protected static int mpuITT1;   //005  ----46  ITT1   IACR1 @ MC68EC040  030MMUのTT1
  //protected static int mpuDTT0;   //006  ----46  DTT0   DACR0 @ MC68EC040
  //protected static int mpuDTT1;   //007  ----46  DTT1   DACR1 @ MC68EC040
  public static int mpuBUSCR;  //008  -----6  BUSCR
  public static int mpuUSP;    //800  -12346  USP    隠れたユーザスタックポインタ。ユーザモードのときはr[15]を参照すること
  public static int mpuVBR;    //801  -12346  VBR    ベクタベースレジスタ
  public static int mpuCAAR;   //802  --23--  CAAR
  public static int mpuMSP;    //803  --234-  MSP    隠れたマスタスタックポインタ。マスタモードのときはr[15]を参照すること
  public static int mpuISP;    //804  --234-  ISP    隠れた割り込みスタックポインタ。割り込みモードのときはr[15]を参照すること
  //protected static int mpuMMUSR;  //805  ----4-  MMUSR                     030MMUのMMUSR
  //protected static int mpuURP;    //806  ----46  URP                       030MMUのCRPの下位32bit
  //protected static int mpuSRP;    //807  ----46  SRP                       030MMUのSRPの下位32bit
  public static int mpuPCR;    //808  -----6  PCR
  //protected static int mpuHCRP;   //                                       030MMUのCRPの上位32bit
  //protected static int mpuHSRP;   //                                       030MMUのSRPの上位32bit

  public static final int MPU_060_REV = 7;  //MC68060のリビジョンナンバー。1=F43G,5=G65V,6=E41J

  //クロック
  //  時刻は開始からの経過時間
  public static long mpuClockTime;  //時刻(TMR_FREQ単位)
  public static long mpuClockLimit;  //タスクの終了時刻
  public static double mpuClockMHz;  //動作周波数の設定値(MHz)
  public static double mpuCurrentMHz;  //動作周波数の現在値(MHz)
  public static int mpuCycleCount;  //命令のサイクル数。実行アドレス計算のサイクル数を含む
  public static long mpuCycleUnit;  //周波数の表示に使用する1サイクルあたりの時間(TMR_FREQ単位)
  public static long mpuModifiedUnit;  //mpuCycleCountの1サイクルあたりの時間(TMR_FREQ単位)。MC68030のときmpuCycleUnit*3/5

  public static long dmaCycleUnit;  //DMAの1サイクルの時間(TMR_FREQ単位)
  public static long busCycleUnit;  //ROMとSRAMのウェイトに使用する1サイクルの時間(TMR_FREQ単位)。通常はmpuCycleUnitでDMACが一時的にdmaCycleUnitに変更する

  //タイマ
  //  mpuTaskはコアをスケジュールしなおすときに毎回作り直す
  public static TimerTask mpuTask;  //null=停止中。null以外=動作中

  //その他
  public static int mpuBootDevice;  //起動デバイス。-1=指定なし
  public static int mpuROMBootHandle;  //ROM起動ハンドル。-1=指定なし
  public static int mpuSavedBootDevice;  //保存されている起動デバイス。-1=保存されていない
  public static int mpuSavedROMBootHandle;  //保存されているROM起動ハンドル。-1=保存されていない

  //コア
  public static int mpuCoreType;  //0=MC68000,1=MC68010,3=MC68EC030,6=MC68060
  public static int mpuRomWait;  //ROMとSRAMのアクセスウェイト。X68000のとき1、X68030のとき0。これがないとクロックの表示がずれる
  public static boolean mpuIgnoreAddressError;  //true=アドレスエラーを無視する

  //任意の周波数の指定
  public static boolean mpuArbFreqOn;  //true=任意の周波数の指定がある。mpuArbFreqOn&&mpuUtilOnは不可
  public static int mpuArbFreqMHz;  //任意の周波数の指定(MHz)。1～1000
  public static SpinnerNumberModel mpuArbFreqModel;
  public static JSpinner mpuArbFreqSpinner;
  public static JRadioButtonMenuItem mpuArbFreqRadioButtonMenuItem;

  //任意の負荷率の指定
  public static boolean mpuUtilOn;  //true=任意の負荷率の指定がある。mpuArbFreqOn&&mpuUtilOnは不可
  public static int mpuUtilRatio;  //任意の負荷率の指定(%)。1～100
  public static SpinnerNumberModel mpuUtilModel;
  public static JSpinner mpuUtilSpinner;
  public static JRadioButtonMenuItem mpuUtilRadioButtonMenuItem;

  //メニュー
  public static JMenu mpuMenu;
  public static JMenuItem mpuResetMenuItem;
  public static JMenuItem mpuOpt1ResetMenuItem;
  public static JRadioButtonMenuItem mpuClock10MenuItem;
  public static JRadioButtonMenuItem mpuClock16MenuItem;
  public static JRadioButtonMenuItem mpuClock25MenuItem;
  public static JRadioButtonMenuItem mpuClock33MenuItem;
  public static JRadioButtonMenuItem mpuClock50MenuItem;

  //デバッグ
  public static ActionListener mpuDebugActionListener;  //デバッグアクションリスナー
  public static ArrayList<AbstractButton> mpuButtonsRunning;  //MPUが動作中のときだけ有効なボタン
  public static ArrayList<AbstractButton> mpuButtonsStopped;  //MPUが停止中のときだけ有効なボタン
  public static int mpuAdvanceCount;  //トレース実行のカウンタ。0以外=トレース実行中
  public static int mpuStepCount;  //ステップ実行のカウンタ。1以上=ステップ実行中。-1=ステップアンティルリターン実行中
  public static boolean mpuContinue;  //自然に終了したときは変化しない。インスタント命令ブレークポイントで停止したときはtrue、それ以外で停止したときはfalse
  public static int mpuUntilReturnSRS;  //ステップアンティルリターン開始時のregSRS
  public static int mpuUntilReturnRP;  //ステップアンティルリターン開始時のmmuSRPまたはmmuURP
  public static int mpuUntilReturnPC0;  //ステップアンティルリターン開始時のregPC0
  public static int mpuUntilReturnSP;  //ステップアンティルリターン開始時のregRn[15]

  //SX-Window ver3.1のバグ対策
  //
  //  無償公開されたSXWIN311.XDFから起動したときリソースファイルから読み込まれるコードに問題がある
  //  マウスカーソルが指しているメニューの項目の番号を返すサブルーチンでd1レジスタの上位ワードが不定のままdivu.w #$0010,d1を実行している
  //  このサブルーチンの引数ではないd1レジスタの上位ワードを0にしてから呼び出さないとメニューを選択できない
  //  実機で露見しないのは直前に呼び出されるサブルーチンが使っているmovem.w <ea>,<list>命令がd1レジスタの上位ワードをほぼ0にしているため
  //  XEiJで露見したのはこの命令の処理が間違っていてd1レジスタが符号拡張されず上位ワードが0になっていなかったため
  //
  //  問題のサブルーチン
  //    00BFCC74  48E77000              movem.l d1-d3,-(sp)                 H輛.
  //    00BFCC78  2600                  move.l  d0,d3                       &.
  //    00BFCC7A  2F00                  move.l  d0,-(sp)                    /.
  //    00BFCC7C  486DFF80              pea.l   $FF80(a5)                   Hm..
  //    00BFCC80  A156                  SXCALL  __GMPtInRect                ｡V
  //    00BFCC82  508F                  addq.l  #$08,sp                     P夙
  //    00BFCC84  6738                  beq.s   $00BFCCBE                   g8
  //    00BFCC86  2003                  move.l  d3,d0                        .
  //    00BFCC88  3200                  move.w  d0,d1                       2.
  //    00BFCC8A  4840                  swap.w  d0                          H@
  //    00BFCC8C  906DFF80              sub.w   $FF80(a5),d0                仁..
  //    00BFCC90  6B2C                  bmi.s   $00BFCCBE                   k,
  //    00BFCC92  B06DFF92              cmp.w   $FF92(a5),d0                ｰm.地
  //    00BFCC96  6E26                  bgt.s   $00BFCCBE                   n&
  //    00BFCC98  926DFF82              sub.w   $FF82(a5),d1                知.Ｌ
  //    00BFCC9C  6B20                  bmi.s   $00BFCCBE                   k 
  //    00BFCC9E  342DFF9C              move.w  $FF9C(a5),d2                4-.愼
  //    00BFCCA2  C4FC0010              mulu.w  #$0010,d2                   ﾄ...
  //    00BFCCA6  B242                  cmp.w   d2,d1                       ｲB
  //    00BFCCA8  6E14                  bgt.s   $00BFCCBE                   n.
  //    00BFCCAA  82FC0010              divu.w  #$0010,d1                   ※..    ←d1レジスタの上位ワードが不定のままdivu.wを実行している
  //    00BFCCAE  3001                  move.w  d1,d0                       0.
  //    00BFCCB0  4841                  swap.w  d1                          HA
  //    00BFCCB2  4A41                  tst.w   d1                          JA
  //    00BFCCB4  6702                  beq.s   $00BFCCB8                   g.
  //    00BFCCB6  5240                  addq.w  #$01,d0                     R@
  //    00BFCCB8  4CDF000E              movem.l (sp)+,d1-d3                 Lﾟ..
  //    00BFCCBC  4E75                  rts                                 Nu
  //    00BFCCBE  7000                  moveq.l #$00,d0                     p.
  //    00BFCCC0  60F6                  bra.s   $00BFCCB8                   `.
  //
  //  問題のサブルーチンの呼び出し元
  //    00BFC9D6  4E56FFF0              link.w  a6,#$FFF0                   NV..
  //    00BFC9DA  6100FEF4              bsr.w   $00BFC8D0                   a...    ←直前に呼び出されるサブルーチン
  //    00BFC9DE  202C000A              move.l  $000A(a4),d0                 ,..
  //    00BFC9E2  61000290              bsr.w   $00BFCC74                   a...    ←問題のサブルーチン
  //    00BFC9E6  3200                  move.w  d0,d1                       2.
  //    00BFC9E8  3600                  move.w  d0,d3                       6.
  //    (以下略)
  //
  //  直前に呼び出されるサブルーチン
  //    00BFC8D0  426DFFA2              clr.w   $FFA2(a5)                   Bm.｢
  //    00BFC8D4  206C0006              movea.l $0006(a4),a0                 l..
  //    00BFC8D8  4C90000F              movem.w (a0),d0-d3                  L...
  //    00BFC8DC  5240                  addq.w  #$01,d0                     R@
  //    00BFC8DE  5241                  addq.w  #$01,d1                     RA
  //    00BFC8E0  5342                  subq.w  #$01,d2                     SB
  //    00BFC8E2  5343                  subq.w  #$01,d3                     SC
  //    00BFC8E4  48AD000FFF80          movem.w d0-d3,$FF80(a5)             0/0
  //    00BFC8EA  206C0002              movea.l $0002(a4),a0                 l..
  //    00BFC8EE  2050                  movea.l (a0),a0                      P
  //    00BFC8F0  30280012              move.w  $0012(a0),d0                0(..
  //    00BFC8F4  5240                  addq.w  #$01,d0                     R@
  //    00BFC8F6  3B40FF90              move.w  d0,$FF90(a5)                ;@.食
  //    00BFC8FA  486DFFA4              pea.l   $FFA4(a5)                   Hm.､
  //    00BFC8FE  A432                  SXCALL  __SXGetDispRect             ､2
  //    00BFC900  588F                  addq.l  #$04,sp                     X臭
  //    00BFC902  4CAD000FFFA4          movem.w $FFA4(a5),d0-d3             Lｭ...､  ←ここでd1レジスタの上位ワードがほぼ0になる
  //    00BFC908  9641                  sub.w   d1,d3                       泡
  //    00BFC90A  48C3                  ext.l   d3                          Hﾃ      ←このサブルーチンはext.lで符号拡張してからdivu.wを使っているが、符号拡張しておいて符号なし除算というのもおかしい。movem.wでd3も符号拡張されているのでこのext.lは不要で、0<=d1.w<=d3.wでないときもext.lは役に立たない
  //    00BFC90C  86FC0010              divu.w  #$0010,d3                   ※..
  //    (以下略)
  //
  public static final boolean MPU_SXMENU = false;  //true=対策を施す。movem.w <ea>,<list>を修正して実害がなくなったので外しておく

  //mpuInit ()
  //  MPUを初期化する
  public static void mpuInit () {
    //コア
    mpuCoreType = mdlCoreRequest;
    mpuIgnoreAddressError = false;
    //レジスタ
    //r = new int[16];
    //FPU
    fpuInit ();  //FPU FPU
    //クロック
    mpuClockTime = 0L;
    mpuClockLimit = 0L;
    mpuClockMHz = mdlClockRequest;
    mpuSetClockMHz (mpuClockMHz);
    mpuCycleCount = 0;
    //タイマ
    mpuTask = null;
    //例外処理
    M68kException.m6eSignal = new M68kException ();
    M68kException.m6eNumber = 0;
    M68kException.m6eAddress = 0;
    M68kException.m6eDirection = MPU_WR_WRITE;
    M68kException.m6eSize = MPU_SS_BYTE;
    //その他
    mpuBootDevice = -1;
    mpuROMBootHandle = -1;
    mpuSavedBootDevice = -1;
    mpuSavedROMBootHandle = -1;
    //任意の周波数の指定
    //mpuArbFreqOn = !(clockMHz == 10.0 ||
    //                 clockMHz == 50.0 / 3.0 ||  //16.7MHz
    //                 clockMHz == 25.0 ||
    //                 clockMHz == 100.0 / 3.0 ||  //33.3MHz
    //                 clockMHz == 50.0);
    //mpuArbFreqMHz = mpuArbFreqOn ? (int) clockMHz : 100;
    //任意の負荷率の指定
    //mpuUtilOn = false;
    //mpuUtilRatio = 100;

    mpuButtonsRunning = new ArrayList<AbstractButton> ();
    mpuButtonsStopped = new ArrayList<AbstractButton> ();
    mpuAdvanceCount = 0;
    mpuStepCount = 0;
    mpuContinue = false;
    mpuUntilReturnSRS = 0;
    mpuUntilReturnRP = 0;
    mpuUntilReturnPC0 = 0;
    mpuUntilReturnSP = 0;

    //デバッグアクションリスナー
    mpuDebugActionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        switch (ae.getActionCommand ()) {
        case "Stop":  //停止
          if (RootPointerList.RTL_ON) {
            if (RootPointerList.rtlCurrentSupervisorTaskIsStoppable ||
                RootPointerList.rtlCurrentUserTaskIsStoppable) {
              mpuStop (null);  //"Stop Button"
            }
          } else {
            mpuStop (null);  //"Stop Button"
          }
          break;
        case "Trace":  //トレース
          mpuAdvance (1);
          break;
        case "Trace 10 times":  //トレース 10 回
          mpuAdvance (10);
          break;
        case "Trace 100 times":  //トレース 100 回
          mpuAdvance (100);
          break;
        case "Step":  //ステップ
          mpuStep (1);
          break;
        case "Step 10 times":  //ステップ 10 回
          mpuStep (10);
          break;
        case "Step 100 times":  //ステップ 100 回
          mpuStep (100);
          break;
        case "Step until return":  //ステップアンティルリターン
          mpuStepUntilReturn ();
          break;
        case "Run":  //実行
          mpuStart ();
          break;
        }
      }  //actionPerformed(ActionEvent)
    };  //mpuDebugActionListener

  }  //mpuInit()

  //mpuMakeMenu ()
  public static void mpuMakeMenu () {
    //メニュー
    ButtonGroup unitGroup = new ButtonGroup ();
    ActionListener listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        switch (ae.getActionCommand ()) {
        case "Reset":  //リセット
          mpuReset (-1, -1);
          break;
        case "Hold down OPT.1 and reset":  //OPT.1 を押しながらリセット
          mpuReset (0, -1);
          break;
        case "Interrupt":  //インタラプト
          sysInterrupt ();  //インタラプトスイッチが押された
          break;
        case "10MHz":
          mpuArbFreqOn = false;
          mpuUtilOn = false;
          mpuClockMHz = 10.0;
          mpuSetClockMHz (mpuClockMHz);
          break;
        case "16.7MHz":
          mpuArbFreqOn = false;
          mpuUtilOn = false;
          mpuClockMHz = 50.0 / 3.0;  //16.7MHz
          mpuSetClockMHz (mpuClockMHz);
          break;
        case "25MHz":
          mpuArbFreqOn = false;
          mpuUtilOn = false;
          mpuClockMHz = 25.0;
          mpuSetClockMHz (mpuClockMHz);
          break;
        case "33.3MHz":
          mpuArbFreqOn = false;
          mpuUtilOn = false;
          mpuClockMHz = 100.0 / 3.0;  //33.3MHz
          mpuSetClockMHz (mpuClockMHz);
          break;
        case "50MHz":
          mpuArbFreqOn = false;
          mpuUtilOn = false;
          mpuClockMHz = 50.0;
          mpuSetClockMHz (mpuClockMHz);
          break;
        case "Arbitrary frequency":  //任意の周波数
          mpuArbFreqOn = true;
          mpuUtilOn = false;
          mpuClockMHz = (double) mpuArbFreqMHz;
          mpuSetClockMHz (mpuClockMHz);
          break;
        case "Arbitrary load factor":  //任意の負荷率
          mpuArbFreqOn = false;
          mpuUtilOn = true;
          break;
        case "FE function instruction":  //FE ファンクション命令
          FEFunction.fpkOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Reject FLOATn.X":  //FLOATn.X を組み込まない
          FEFunction.fpkRejectFloatOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Cut FC2 pin":  //FC2 ピンをカットする
          busRequestCutFC2Pin = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Use IPLROM 1.6":  //IPLROM 1.6 を使う
          ROM.romIPLROM16On = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Increase IPLROM to 256KB":  //IPLROM を 256KB に増やす
          ROM.romIPLROM256KOn = ((JCheckBoxMenuItem) source).isSelected ();
          break;

        case "Run / Stop":  //実行／停止
          if (((JCheckBox) source).isSelected ()) {  //Run
            mpuStart ();
          } else {  //Stop
            if (RootPointerList.RTL_ON) {
              if (RootPointerList.rtlCurrentSupervisorTaskIsStoppable ||
                  RootPointerList.rtlCurrentUserTaskIsStoppable) {
                mpuStop (null);  //"Stop Button"
              }
            } else {
              mpuStop (null);  //"Stop Button"
            }
          }
          pnlPanel.requestFocusInWindow ();  //パネルにフォーカスを戻す。戻さないとキー入力が入らない
          break;
        }
      }
    };
    mpuMenu = ComponentFactory.createMenu (
      "MPU", 'M',
      mpuResetMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Reset", 'R', MNB_MODIFIERS, listener), "ja", "リセット"),
      mpuOpt1ResetMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Hold down OPT.1 and reset", 'O', MNB_MODIFIERS, listener), "ja", "OPT.1 を押しながらリセット"),
      Multilingual.mlnText (ComponentFactory.createMenuItem ("Interrupt", listener), "ja", "インタラプト"),
      ComponentFactory.createHorizontalSeparator (),
      mdlMenu,
      ComponentFactory.createHorizontalSeparator (),
      mpuClock10MenuItem =
      ComponentFactory.createRadioButtonMenuItem (unitGroup, !mpuArbFreqOn && !mpuUtilOn && mpuClockMHz == 10.0       , "10MHz"  , listener),
      mpuClock16MenuItem =
      ComponentFactory.createRadioButtonMenuItem (unitGroup, !mpuArbFreqOn && !mpuUtilOn && mpuClockMHz == 50.0 / 3.0 , "16.7MHz", listener),
      mpuClock25MenuItem =
      ComponentFactory.createRadioButtonMenuItem (unitGroup, !mpuArbFreqOn && !mpuUtilOn && mpuClockMHz == 25.0       , "25MHz"  , listener),
      mpuClock33MenuItem =
      ComponentFactory.createRadioButtonMenuItem (unitGroup, !mpuArbFreqOn && !mpuUtilOn && mpuClockMHz == 100.0 / 3.0, "33.3MHz", listener),
      mpuClock50MenuItem =
      ComponentFactory.createRadioButtonMenuItem (unitGroup, !mpuArbFreqOn && !mpuUtilOn && mpuClockMHz == 50.0       , "50MHz"  , listener),
      mpuArbFreqRadioButtonMenuItem =
      Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (unitGroup, mpuArbFreqOn, "Arbitrary frequency", listener), "ja", "任意の周波数"),
      ComponentFactory.createHorizontalBox (
        Box.createHorizontalStrut (20),
        mpuArbFreqSpinner =
        ComponentFactory.createNumberSpinner (
          mpuArbFreqModel = new SpinnerNumberModel (mpuArbFreqMHz, 1, 1000, 1),
          4,
          new ChangeListener () {
            @Override public void stateChanged (ChangeEvent ce) {
              //mpuArbFreqRadioButtonMenuItem.setSelected (true);
              mpuArbFreqMHz = mpuArbFreqModel.getNumber ().intValue ();
              if (mpuArbFreqOn) {
                mpuClockMHz = (double) mpuArbFreqMHz;
                mpuSetClockMHz (mpuClockMHz);
              }
            }
          }
          ),
        ComponentFactory.createLabel ("MHz"),
        Box.createHorizontalGlue ()
        ),
      mpuUtilRadioButtonMenuItem =
      Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (unitGroup, mpuUtilOn, "Arbitrary load factor", listener), "ja", "任意の負荷率"),
      ComponentFactory.createHorizontalBox (
        Box.createHorizontalStrut (20),
        mpuUtilSpinner =
        ComponentFactory.createNumberSpinner (
          mpuUtilModel = new SpinnerNumberModel (mpuUtilRatio, 1, 100, 1),
          4,
          new ChangeListener () {
            @Override public void stateChanged (ChangeEvent ce) {
              //mpuUtilRadioButtonMenuItem.setSelected (true);
              mpuUtilRatio = mpuUtilModel.getNumber ().intValue ();
            }
          }
          ),
        ComponentFactory.createLabel ("%"),
        Box.createHorizontalGlue ()
        ),
      ComponentFactory.createHorizontalSeparator (),
      fpuMenuItem0,
      fpuMenuItem1,
      fpuMenuItem2,
      fpuMenuItem3,
      Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (FEFunction.fpkOn, "FE function instruction", listener), "ja", "FE ファンクション命令"),
      Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (FEFunction.fpkRejectFloatOn, "Reject FLOATn.X", listener), "ja", "FLOATn.X を組み込まない"),
      ComponentFactory.createHorizontalSeparator (),
      Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (busRequestCutFC2Pin, "Cut FC2 pin", listener), "ja", "FC2 ピンをカットする"),
      ComponentFactory.createHorizontalSeparator (),
      Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (ROM.romIPLROM16On, "Use IPLROM 1.6", listener), "ja", "IPLROM 1.6 を使う"),
      Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (ROM.romIPLROM256KOn, "Increase IPLROM to 256KB", listener), "ja", "IPLROM を 256KB に増やす")
      );
  }  //mpuMakeMenu()

  //mpuSetClockMHz (mhz)
  //  mpuCurrentMHz,mpuCycleUnit,mpuModifiedUnitを設定する
  //  TMR_FREQ=10^12のとき
  //               mpuModifiedUnitと1サイクルの時間
  //  mpuClockMHz     000/060         030
  //    10.0MHz    100000(100ns)  60000(60ns)
  //    16.7MHz     60000 (60ns)  36000(36ns)
  //    25.0MHz     40000 (40ns)  24000(24ns)
  //    33.3MHz     30000 (30ns)  18000(18ns)
  //    50.0MHz     20000 (20ns)  12000(12ns)
  public static void mpuSetClockMHz (double mhz) {
    mhz = Math.max (1.0, Math.min (1000.0, mhz));
    mpuCurrentMHz = mhz;
    mpuCycleUnit = (long) (((double) TMR_FREQ / 1000000.0) / mhz + 0.5);
    busCycleUnit = mpuCycleUnit;
    //  DBRA命令で比較するとMC68030は3/5、MC68040は2/5、MC68060は1/5くらい？
    mpuModifiedUnit = (mpuCoreType <= 1 ? mpuCycleUnit :
                       mpuCoreType == 3 ? (long) (((double) TMR_FREQ * 3.0 / (5.0 * 1000000.0)) / mhz + 0.5) :
                       mpuCoreType == 4 ? (long) (((double) TMR_FREQ * 2.0 / (5.0 * 1000000.0)) / mhz + 0.5) :
                       mpuCycleUnit);
  }  //mpuSetClockMHz(double)

  //mpuSetROMWait (cacheOn)
  //  本体がX68000のときROMとSRAMに1サイクルのウェイトがかかる
  //  MC68030以上でキャッシュがONのときウェイトを0にしてキャッシュが効いているように見せかける
  public static void mpuSetROMWait (boolean cacheOn) {
    mpuRomWait = mdlCurrentBody == 0 && !cacheOn ? 1 : 0;
  }

  //mpuReset (device, romHandle)
  //  MPUをリセットしてからコアのタスクを起動する
  //  コアのタスクが動いているときはそれを中断する
  //  動いていたタスクが完全に止まってから再起動する
  public static void mpuReset (int device, int romHandle) {

    mpuBootDevice = device;
    mpuROMBootHandle = romHandle;

    //mpuBootDevice
    //  -1=リセット,0=OPT.1キーを押しながらリセット、その他=ここから再起動
    //  リセットのとき
    //    保存されている起動デバイスがあるとき
    //      起動デバイスを保存されている起動デバイスに変更する
    //      保存されている起動デバイスを消す
    //  リセット以外のとき
    //    保存されている起動デバイスがないとき
    //      起動デバイスを保存する
    //  メニューで起動デバイスが既定以外に設定されたとき
    //    保存されている起動デバイスを消す
    if (mpuBootDevice == -1) {  //リセットのとき
      if (mpuSavedBootDevice != -1) {  //保存されている起動デバイスがあるとき
        mpuBootDevice = mpuSavedBootDevice;  //起動デバイスを保存されている起動デバイスに変更する
        mpuROMBootHandle = mpuSavedROMBootHandle;
        mpuSavedBootDevice = -1;  //保存されている起動デバイスを消す
        mpuSavedROMBootHandle = -1;
      }
    } else {  //リセット以外のとき
      if (mpuSavedBootDevice == -1) {  //保存されている起動デバイスがないとき
        mpuSavedBootDevice = MainMemory.mmrRwz (0x00ed0018);  //起動デバイスを保存する
        mpuSavedROMBootHandle = MainMemory.mmrRls (0x00ed000c);
      }
    }

    //Xellent30でMC68EC030のときMC68000に戻す
    if (mdlModel == MDL_XELLENT30 && xt3Accelerated) {
      xt3Accelerated = false;
      mdlRequestCore (0);
      mdlClockRequest = mpuClockMHz * 0.5;
    }

    if (MC68EC030.M30_DIV_ZERO_V_FLAG) {
      MC68EC030.m30DivZeroVFlag = false;
    }

    if (mpuTask != null) {
      mpuClockLimit = 0L;
      System.out.println (Multilingual.mlnJapanese ?
                          "MPU を停止します" :
                          "MPU stops");
      mpuTask.cancel ();
      mpuTask = null;
    }

    tmrTimer.schedule (new TimerTask () {
      @Override public void run () {

        //本体
        mdlCurrentBody = mdlBodyRequest;

        //メモリマップを再構築する
        if (mdlCoreRequest == 0 ||
            mdlCoreRequest == 1) {
          busRequestExMemoryStart = 0x10000000;
          busRequestExMemorySize = 0 << 20;
          busRequestExMemoryArray = BUS_DUMMY_MEMORY_ARRAY;
        } else if (mdlCoreRequest == 3) {
          busRequestExMemoryStart = 0x01000000;
          busRequestExMemorySize = busHighMemorySize;
          busRequestExMemoryArray = busHighMemoryArray;
        } else {
          busRequestExMemoryStart = 0x10000000;
          busRequestExMemorySize = busLocalMemorySize;
          busRequestExMemoryArray = busLocalMemoryArray;
        }
        busUpdateMemoryMap ();

        //ROMを構築する
        ROM.romReset ();

        //MPUの切り替え
        if (mpuCoreType != mdlCoreRequest) {
          mpuCoreType = mdlCoreRequest;
        }

        //クロックの切り替え
        if (0.0 < mdlClockRequest) {  //機種の指定によるクロックの指定がまだ反映されていない
          if (mdlClockRequest == 10.0) {
            mpuClock10MenuItem.setSelected (true);
            mpuArbFreqOn = false;
            mpuUtilOn = false;
          } else if (mdlClockRequest == 50.0 / 3.0) {  //16.7MHz
            mpuClock16MenuItem.setSelected (true);
            mpuArbFreqOn = false;
            mpuUtilOn = false;
          } else if (mdlClockRequest == 25.0) {
            mpuClock25MenuItem.setSelected (true);
            mpuArbFreqOn = false;
            mpuUtilOn = false;
          } else if (mdlClockRequest == 100.0 / 3.0) {  //33.3MHz
            mpuClock33MenuItem.setSelected (true);
            mpuArbFreqOn = false;
            mpuUtilOn = false;
          } else if (mdlClockRequest == 50.0) {
            mpuClock50MenuItem.setSelected (true);
            mpuArbFreqOn = false;
            mpuUtilOn = false;
          }
          mpuClockMHz = mdlClockRequest;
          mdlClockRequest = -1.0;  //反映済み。クロックを調整してもリセットする度に元に戻ってしまうのは嫌なので2回目からは変更しない
        }
        mpuSetClockMHz (mpuClockMHz);  //mpuCoreTypeが変更されたときは再計算が必要なのでクロックが変わっていなくても省略できない

        RegisterList.drpSetMPU ();

        mpuSFC = mpuDFC = mpuCACR = mpuBUSCR = mpuUSP = mpuVBR = mpuCAAR = mpuMSP = mpuISP = 0;
        mpuPCR = 0x04300500 | MPU_060_REV << 8;
        MC68060.mmuReset ();  //TCR,ITT0,ITT1,DTT0,DTT1,URP,SRP。060→000/030のときアドレス変換をOFFにする必要がある

        if (mdlCurrentBody == 0) {  //X68000
          XEiJ.dmaCycleUnit = XEiJ.TMR_FREQ / 10000000L;  //DMACは10MHz
          HD63450.dmaBurstInterval = XEiJ.dmaCycleUnit << 4 + (HD63450.dmaBT >> 2);
          HD63450.dmaBurstSpan = HD63450.dmaBurstInterval >> 1 + (HD63450.dmaBR & 3);
        } else {  //X68030
          XEiJ.dmaCycleUnit = XEiJ.TMR_FREQ / 12500000L;  //DMACは12.5MHz
          HD63450.dmaBurstInterval = XEiJ.dmaCycleUnit << 4 + (HD63450.dmaBT >> 2);
          HD63450.dmaBurstSpan = HD63450.dmaBurstInterval >> 1 + (HD63450.dmaBR & 3);
        }

        if (mpuCoreType <= 1) {

          mpuIgnoreAddressError = false;
          fpuOn = false;
          //mpuSFC = mpuDFC = mpuVBR = mpuISP = 0;
          mpuSetROMWait (false);

        } else if (mpuCoreType == 3) {

          mpuIgnoreAddressError = true;
          fpuOn = fpuMode != 0;
          fpuBox = fpuMotherboardCoprocessor;
          fpuBox.epbSetMC68882 ();
          if (fpuMode == 1) {
            fpuBox.epbSetExtended ();
          } else if (fpuMode == 2) {
            fpuBox.epbSetTriple ();
          }
          fpuBox.epbReset ();
          fpuFPn = fpuBox.epbFPn;
          //mpuCACR = mpuUSP = mpuVBR = mpuCAAR = mpuMSP = mpuISP = 0;
          mpuSetROMWait ((mpuCACR & 0x00000101) != 0);

        } else if (mpuCoreType == 6) {

          mpuIgnoreAddressError = true;
          fpuOn = fpuMode != 0;
          fpuBox = fpuOnChipFPU;
          if (fpuFullSpec) {
            fpuBox.epbSetFullSpec ();
          } else {
            fpuBox.epbSetMC68060 ();
          }
          if (fpuMode == 1) {
            fpuBox.epbSetExtended ();
          } else if (fpuMode == 2) {
            fpuBox.epbSetTriple ();
          }
          fpuBox.epbReset ();
          fpuFPn = fpuBox.epbFPn;
          //mpuSFC = mpuDFC = mpuCACR = mpuBUSCR = mpuUSP = mpuVBR = mpuURP = mpuSRP = 0;
          mpuPCR = 0x04300500 | MPU_060_REV << 8;
          mpuSetROMWait ((mpuCACR & 0x80008000) != 0);

        }

        //! SSPとPCをROMのアドレスから直接読み出している
        regSRT1 = regSRT0 = 0;
        regSRS = REG_SR_S;
        regSRM = 0;
        regSRI = REG_SR_I;
        regCCR = 0;
        Arrays.fill (regRn, 0);
        //r[14] = 0x00001000;  //ROMDB2.32のバグ対策。コードにパッチをあてることにしたので不要
        regRn[15] = MainMemory.mmrRls (0x00ff0000);
        regPC = MainMemory.mmrRls (0x00ff0004);
        //メインメモリ
        MainMemory.mmrReset ();
        //バスコントローラ
        busReset ();
        if (InstructionBreakPoint.IBP_ON) {
          InstructionBreakPoint.ibpOp1MemoryMap = InstructionBreakPoint.ibpOp1SuperMap;
          InstructionBreakPoint.ibpReset ();
        }
        if (BranchLog.BLG_ON) {
          BranchLog.blgReset ();
        }
        //割り込み
        mpuIMR = 0;
        mpuIRR = 0;
        if (MC68901.MFP_DELAYED_INTERRUPT) {
          mpuDIRR = 0;
        }
        mpuISR = 0;
        //これでリセット命令が実行されるまでメインメモリとROM以外のデバイスは動かないはず
        //動作開始
        mpuStart ();
      }
    }, TMR_DELAY);

  }  //mpuReset(int,int)

  //mpuStart ()
  //  コアのタスクを起動する
  //  コアのタスクが動いているときはそれを中断する
  //  動いていたタスクが完全に止まってから再起動する
  public static void mpuStart () {
    if (mpuTask != null) {
      mpuClockLimit = 0L;
      System.out.println (Multilingual.mlnJapanese ?
                          "MPU を停止します" :
                          "MPU stops");
      mpuTask.cancel ();
      mpuTask = null;
    }
    //停止中だけ有効なボタンを無効にする
    for (AbstractButton button : mpuButtonsStopped) {
      button.setEnabled (false);
    }
    DisassembleList.ddpStoppedBy = null;
    if (mpuCoreType == 0) {
      System.out.println (Multilingual.mlnJapanese ?
                          "MC68000 を起動します" :
                          "MC68000 starts up");
      mpuTask = new TimerTask () {
        @Override public void run () {
          mpuContinue = true;
          mpuClockLimit = mpuClockTime + TMR_FREQ * TMR_INTERVAL / 1000;
          MC68000.mpuCore ();
        }
      };
    } else if (mpuCoreType == 1) {
      System.out.println (Multilingual.mlnJapanese ?
                          "MC68010 を起動します" :
                          "MC68010 starts up");
      mpuTask = new TimerTask () {
        @Override public void run () {
          mpuContinue = true;
          mpuClockLimit = mpuClockTime + TMR_FREQ * TMR_INTERVAL / 1000;
          MC68010.mpuCore ();
        }
      };
    } else if (mpuCoreType == 3) {
      System.out.println (Multilingual.mlnJapanese ?
                          "MC68EC030 を起動します" :
                          "MC68EC030 starts up");
      mpuTask = new TimerTask () {
        @Override public void run () {
          mpuContinue = true;
          mpuClockLimit = mpuClockTime + TMR_FREQ * TMR_INTERVAL / 1000;
          MC68EC030.mpuCore ();
        }
      };
    } else {
      System.out.println (Multilingual.mlnJapanese ?
                          "MC68060 を起動します" :
                          "MC68060 starts up");
      mpuTask = new TimerTask () {
        @Override public void run () {
          mpuContinue = true;
          mpuClockLimit = mpuClockTime + TMR_FREQ * TMR_INTERVAL / 1000;
          MC68060.mpuCore ();
        }
      };
    }
    tmrTimer.scheduleAtFixedRate (mpuTask, TMR_DELAY, TMR_INTERVAL);  //固定頻度実行
    //動作中だけ有効なボタンを有効にする
    for (AbstractButton button : mpuButtonsRunning) {
      button.setEnabled (true);
    }
  }  //mpuStart()

  //mpuStop (message)
  //  コアのタスクが動いているとき、それを止める
  //  完全に止まってからデバッグダイアログを更新する
  public static void mpuStop (String message) {
    //トレース実行とステップ実行を中止する
    mpuAdvanceCount = 0;
    mpuStepCount = 0;
    mpuContinue = false;
    mpuStop1 (message);
  }  //mpuStop(String)
  public static void mpuStop1 (String message) {
    if (mpuTask == null) {  //既に停止しているか停止処理が始まっている
      return;
    }
    DisassembleList.ddpStoppedBy = message;  //停止理由
    mpuClockLimit = 0L;
    System.out.println (Multilingual.mlnJapanese ?
                        "MPU を停止します" :
                        "MPU stops");
    mpuTask.cancel ();
    mpuTask = null;
    //ステップ実行を継続する
    if (mpuStepCount != 0 && mpuContinue) {
      if (mpuStepCount == -1 || --mpuStepCount != 0) {
        mpuStep (mpuStepCount);
        return;
      }
    }
    mpuAdvanceCount = 0;
    mpuStepCount = 0;
    mpuContinue = false;
    //動作中だけ有効なボタンを無効にする
    for (AbstractButton button : mpuButtonsRunning) {
      button.setEnabled (false);
    }
    tmrTimer.schedule (new TimerTask () {
      @Override public void run () {
        mpuUpdateWindow ();
      }
    }, TMR_DELAY);
  }  //mpuStop1(message)

  //mpuAdvance (n)
  //  n回トレース実行する。0回は不可
  //  1ターンで実行するのであまり大きな値を指定しないこと
  //  終わってからデバッグダイアログを更新する
  //  コアが止まっていないときは何もしない
  public static void mpuAdvance (int n) {
    if (mpuTask != null) {  //コアが止まっていない
      return;
    }
    mpuAdvanceCount = n;
    DisassembleList.ddpStoppedBy = null;
    mpuTask = new TimerTask () {
      @Override public void run () {
        mpuContinue = true;
        do {
          mpuClockLimit = mpuClockTime + 1L;
          if (mpuCoreType == 0) {
            MC68000.mpuCore ();
          } else if (mpuCoreType == 1) {
            MC68010.mpuCore ();
          } else if (mpuCoreType == 3) {
            MC68EC030.mpuCore ();
          } else {
            MC68060.mpuCore ();
          }
        } while (mpuContinue && --mpuAdvanceCount != 0);
        mpuClockLimit = 0L;
        if (mpuTask != null) {  //最初の命令のエラーで停止したときnullになっている場合がある
          mpuTask.cancel ();
          mpuTask = null;
        }
        if (mpuStepCount != 0 && mpuContinue) {  //ステップ実行中に分岐命令を1命令実行して自然に終了した
          if (mpuStepCount == -1 || --mpuStepCount != 0) {
            mpuStep (mpuStepCount);
            return;
          }
        }
        mpuAdvanceCount = 0;
        mpuStepCount = 0;
        mpuContinue = false;
        mpuUpdateWindow ();
      }
    };
    tmrTimer.schedule (mpuTask, TMR_DELAY);
  }  //mpuAdvance(int)

  //mpuStep (n)
  //  n回ステップ実行する。0回は不可。-1はステップアンティルリターン
  //  次の命令が分岐命令のときはトレース実行と同じ
  //  次の命令が分岐命令でないときはその直後まで実行する
  //  コアが止まっていないときは何もしない
  public static void mpuStep (int n) {
    if (mpuTask != null) {  //コアが止まっていない
      return;
    }
    mpuStepCount = n;
    Disassembler.disDisassemble (new StringBuilder (), regPC, regSRS);
    if ((Disassembler.disStatus & (Disassembler.DIS_ALWAYS_BRANCH | Disassembler.DIS_SOMETIMES_BRANCH)) != 0) {
      if (mpuStepCount == -1 &&  //ステップアンティルリターン
          (Disassembler.disOC == 0x4e73 ||  //RTE
           Disassembler.disOC == 0x4e74 ||  //RTD
           Disassembler.disOC == 0x4e75 ||  //RTR
           Disassembler.disOC == 0x4e77) &&  //RTS
          mpuUntilReturnSRS == regSRS &&  //ユーザモード/スーパーバイザモードが同じ
          (mpuCoreType < 6 || mpuUntilReturnRP == (regSRS != 0 ? MC68060.mmuSRP : MC68060.mmuURP)) &&  //MC68060のときルートポインタが同じ
          mpuUntilReturnPC0 != regPC0 &&  //プログラムカウンタが違う
          Integer.compareUnsigned (mpuUntilReturnSP, regRn[15]) <= 0) {  //スタックポインタが減っていない
        mpuAdvanceCount = 0;
        mpuStepCount = 0;
        mpuContinue = false;
        mpuUpdateWindow ();
        return;
      }
      mpuAdvance (1);
    } else {
      if (InstructionBreakPoint.IBP_ON) {
        InstructionBreakPoint.ibpInstant (Disassembler.disPC, DisassembleList.ddpSupervisorMode);
        mpuStart ();
      }
    }
  }  //mpuStep(int)

  //mpuStepUntilReturn ()
  //  ステップアンティルリターン
  //  RTD,RTE,RTR,RTSを実行する直前で
  //    ステップアンティルリターン開始時と
  //      ユーザモード/スーパーバイザモードが同じ
  //      MC68060のときルートポインタが同じ
  //      プログラムカウンタが違う
  //      スタックポインタが減っていない
  //    ならば停止する
  //  コアが止まっていないときは何もしない
  public static void mpuStepUntilReturn () {
    if (mpuTask != null) {  //コアが止まっていない
      return;
    }
    mpuUntilReturnSRS = regSRS;
    mpuUntilReturnRP = regSRS != 0 ? MC68060.mmuSRP : MC68060.mmuURP;
    mpuUntilReturnPC0 = regPC0;
    mpuUntilReturnSP = regRn[15];
    mpuStep (-1);
  }  //mpuStepUntilReturn()

  //mpuUpdateWindow ()
  //  停止したときデバッグ関連ウインドウやボタンを更新する
  public static void mpuUpdateWindow () {
    if (dbgVisibleMask != 0) {  //デバッグ関連ウインドウが表示されている
      if ((dbgVisibleMask & DBG_DDP_VISIBLE_MASK) != 0) {
        DisassembleList.ddpBacktraceRecord = -1L;  //分岐レコードの選択を解除する
        DisassembleList.ddpUpdate (-1, -1, false);
      }
      if (BranchLog.BLG_ON) {
        if ((dbgVisibleMask & DBG_BLG_VISIBLE_MASK) != 0) {
          BranchLog.blgUpdate (BranchLog.BLG_SELECT_NEWEST);
        }
      }
      if (ProgramFlowVisualizer.PFV_ON) {
        if ((dbgVisibleMask & DBG_PFV_VISIBLE_MASK) != 0) {
          ProgramFlowVisualizer.pfvUpdate ();
        }
      }
      if (RasterBreakPoint.RBP_ON) {
        if ((dbgVisibleMask & DBG_RBP_VISIBLE_MASK) != 0) {
          RasterBreakPoint.rbpUpdateFrame ();
        }
      }
      if (ScreenModeTest.SMT_ON) {
        if ((dbgVisibleMask & DBG_SMT_VISIBLE_MASK) != 0) {
          ScreenModeTest.smtUpdateFrame ();
        }
      }
      if (RootPointerList.RTL_ON) {
        if ((dbgVisibleMask & DBG_RTL_VISIBLE_MASK) != 0) {
          RootPointerList.rtlUpdateFrame ();
        }
      }
      if (SpritePatternViewer.SPV_ON) {
        if ((dbgVisibleMask & DBG_SPV_VISIBLE_MASK) != 0) {
          SpritePatternViewer.spvUpdateFrame ();
        }
      }
      if (ATCMonitor.ACM_ON) {
        if ((dbgVisibleMask & DBG_ACM_VISIBLE_MASK) != 0) {
          ATCMonitor.acmUpdateFrame ();
        }
      }
    }
    //コンソールにレジスタ一覧を表示する
    if (DebugConsole.dgtRequestRegs) {
      ExpressionEvaluator.ElementType.ETY_COMMAND_REGS.etyEval (null, ExpressionEvaluator.EVM_COMMAND);
      DebugConsole.dgtRequestRegs = false;
      DebugConsole.dgtPrintPrompt ();
    }
    //動作中だけ有効なボタンを無効にする
    for (AbstractButton button : mpuButtonsRunning) {
      button.setEnabled (false);
    }
    //停止中だけ有効なボタンを有効にする
    for (AbstractButton button : mpuButtonsStopped) {
      button.setEnabled (true);
    }
  }  //mpuUpdateWindow()

  //button = mpuMakeBreakButton ()
  //  停止ボタンを作る
  public static JButton mpuMakeBreakButton () {
    return mpuAddButtonRunning (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_BREAK_IMAGE,
          LnF.LNF_BREAK_DISABLED_IMAGE,
          "Stop", mpuDebugActionListener),
        "ja", "停止")
      );
  }  //mpuMakeBreakButton()

  //button = mpuMakeTraceButton ()
  //  トレース実行ボタンを作る
  public static JButton mpuMakeTraceButton () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_TRACE_IMAGE,
          LnF.LNF_TRACE_DISABLED_IMAGE,
          "Trace", mpuDebugActionListener),
        "ja", "トレース")
      );
  }  //mpuMakeTraceButton()

  //button = mpuMakeTrace10Button ()
  //  トレース10回ボタンを作る
  public static JButton mpuMakeTrace10Button () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_TRACE_10_IMAGE,
          LnF.LNF_TRACE_10_DISABLED_IMAGE,
          "Trace 10 times", mpuDebugActionListener),
        "ja", "トレース 10 回")
      );
  }  //mpuMakeTrace10Button()

  //button = mpuMakeTrace100Button ()
  //  トレース100回ボタンを作る
  public static JButton mpuMakeTrace100Button () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_TRACE_100_IMAGE,
          LnF.LNF_TRACE_100_DISABLED_IMAGE,
          "Trace 100 times", mpuDebugActionListener),
        "ja", "トレース 100 回")
      );
  }  //mpuMakeTrace100Button()

  //button = mpuMakeStepButton ()
  //  ステップ実行ボタンを作る
  public static JButton mpuMakeStepButton () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_STEP_IMAGE,
          LnF.LNF_STEP_DISABLED_IMAGE,
          "Step", mpuDebugActionListener),
        "ja", "ステップ")
      );
  }  //mpuMakeStepButton()

  //button = mpuMakeStep10Button ()
  //  ステップ10回ボタンを作る
  public static JButton mpuMakeStep10Button () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_STEP_10_IMAGE,
          LnF.LNF_STEP_10_DISABLED_IMAGE,
          "Step 10 times", mpuDebugActionListener),
        "ja", "ステップ 10 回")
      );
  }  //mpuMakeStep10Button()

  //button = mpuMakeStep100Button ()
  //  ステップ100回ボタンを作る
  public static JButton mpuMakeStep100Button () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_STEP_100_IMAGE,
          LnF.LNF_STEP_100_DISABLED_IMAGE,
          "Step 100 times", mpuDebugActionListener),
        "ja", "ステップ 100 回")
      );
  }  //mpuMakeStep100Button()

  //button = mpuMakeReturnButton ()
  //  ステップアンティルリターンボタンを作る
  public static JButton mpuMakeReturnButton () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_STEP_UNTIL_RETURN_IMAGE,
          LnF.LNF_STEP_UNTIL_RETURN_DISABLED_IMAGE,
          "Step until return", mpuDebugActionListener),
        "ja", "ステップアンティルリターン")
      );
  }  //mpuMakeReturnButton()

  //button = mpuMakeRunButton ()
  //  実行ボタンを作る
  public static JButton mpuMakeRunButton () {
    return mpuAddButtonStopped (
      Multilingual.mlnToolTipText (
        ComponentFactory.createImageButton (
          LnF.LNF_RUN_IMAGE,
          LnF.LNF_RUN_DISABLED_IMAGE,
          "Run", mpuDebugActionListener),
        "ja", "実行")
      );
  }  //mpuMakeRunButton()

  //button = mpuAddButtonRunning (button)
  //  MPUが動作中のときだけ有効なボタンを追加する
  public static <T extends AbstractButton> T mpuAddButtonRunning (T button) {
    button.setEnabled (mpuTask != null);
    mpuButtonsRunning.add (button);
    return button;
  }

  //button = mpuAddButtonStopped (button)
  //  MPUが停止中のときだけ有効なボタンを追加する
  public static <T extends AbstractButton> T mpuAddButtonStopped (T button) {
    button.setEnabled (mpuTask == null);
    mpuButtonsStopped.add (button);
    return button;
  }



  //========================================================================================
  //$$EMX エミュレータ拡張命令
  //  下位6bitの組み合わせで64個まで追加できる
  //  オペコードが変更される場合があるので原則としてユーザは使用禁止
  //  オペコード開始位置の上位10bitを変更するときはジャンプテーブルも修正すること
  //
  //  WAIT命令
  //    現在の割り込みマスクレベルで割り込みを受け付ける以外に何もしない
  //    やっていることはBRA *相当の無限ループだが、分岐ログは変化しない
  //    1回あたりの所要時間はMPUの種類や動作周波数に関係なく0.01msとする
  //    ホストのファイルにアクセスするといった時間のかかる処理を行うとき、
  //    コアを止めてしまうと本来ならば割り込みを用いて続けられる音楽やアニメーションが止まってしまうので、
  //    時間のかかる処理は別スレッドで行い、それが終わるまでコアはWAIT命令を繰り返して待つ
  public static final int EMX_OPCODE_BASE    = 0x4e00;  //オペコード開始位置
  public static final int EMX_OPCODE_HFSBOOT = EMX_OPCODE_BASE + 0x00;
  public static final int EMX_OPCODE_HFSINST = EMX_OPCODE_BASE + 0x01;
  public static final int EMX_OPCODE_HFSSTR  = EMX_OPCODE_BASE + 0x02;
  public static final int EMX_OPCODE_HFSINT  = EMX_OPCODE_BASE + 0x03;
  public static final int EMX_OPCODE_EMXNOP  = EMX_OPCODE_BASE + 0x04;

  public static final String[] EMX_MNEMONIC_ARRAY = {
    "hfsboot",
    "hfsinst",
    "hfsstr",
    "hfsint",
    "emxnop",
  };

  //emxNop ()
  //  命令としては何もしない
  //  コードに埋め込んでパッチをあてたりするのに使う
  public static void emxNop () {
    if (MainMemory.mmrHumanVersion == 0x0302 && regPC0 == 0x00007140) {  //デバイスドライバを初期化する直前
      int head = regRn[9];  //初期化するデバイスドライバのデバイスヘッダのアドレス
      emxPatchPCM8A (head);
    } else if (MainMemory.mmrHumanVersion == 0x0302 && regPC0 == 0x0000716c) {  //デバイスドライバを初期化した直後
      int head = regRn[9];  //初期化されたデバイスドライバのデバイスヘッダのアドレス
      //----------------------------------------------------------------
      //060turbo.sysにパッチをあてる
      //  sysStat_8000::
      //  00000EC0  203C302E3536  move.l  #'0.56',d0
      //  00000EC6  227C30363054  movea.l #'060T',a1
      //  00000ECC  4E75          rts
      if (MC68060.mmuPeekLongData (head + 0x00000ec0, 1) == 0x203c302e &&
          MC68060.mmuPeekLongData (head + 0x00000ec4, 1) == 0x3536227c &&
          MC68060.mmuPeekLongData (head + 0x00000ec8, 1) == 0x30363054) {  //060turbo.sys version 0.56
        System.out.println (Multilingual.mlnJapanese ?
                            "060turbo.sys 0.56 があります" :
                            "060turbo.sys 0.56 exists");
        //SCSIコールでバスエラーが出ることがある
        //  SRAMのソフト転送フラグを確認する命令がbtstではなくbsetになっている
        //  000021E6  08F9000400ED0070  bset.b  #4,$00ED0070  →  000021E6  0839000400ED0070  btst.b  #4,$00ED0070
        int patched = 0;
        int failed = 0;
        if (MC68060.mmuPeekLongData (head + 0x000021e6, 1) == 0x08f90004 &&
            MC68060.mmuPeekLongData (head + 0x000021ea, 1) == 0x00ed0070) {
          MC68060.mmuPokeWordData (head + 0x000021e6, 0x0839, 1);
          patched++;
        } else {
          failed++;
        }
        System.out.printf (Multilingual.mlnJapanese ?
                           "060turbo.sys 0.56 にパッチをあてました (%d/%d)\n" :
                           "060turbo.sys 0.56 was patched (%d/%d)\n",
                           patched, patched + failed);
      }
      //----------------------------------------------------------------
      //FSX.Xのマウス受信データ処理ルーチンとマウスワークのアドレスを保存する
      if (Z8530.SCC_FSX_MOUSE) {
        emxCheckFSX (head);
      }
      //----------------------------------------------------------------
      //TwentyOne.xのオプションのアドレスを保存する
      if (HFS.HFS_USE_TWENTY_ONE) {
        emxCheckTwentyOne (head);
      }
      //----------------------------------------------------------------
      //ラベルをクリアする
      LabeledAddress.lblClear ();
    } else if (MainMemory.mmrHumanVersion == 0x0302 && regPC0 == 0x0000972c) {  //プロセスを起動する直前
      int head = regRn[8] + 256;  //起動するプロセスのメモリ管理テーブルのアドレス+256=プロセスの先頭
      //System.out.println (MC68060.mmuPeekStringZ (head - 60, 1));  //実行ファイル名
      //----------------------------------------------------------------
      //PCM8Aにパッチをあてる
      emxPatchPCM8A (head);
      //----------------------------------------------------------------
      //ラベルをクリアする
      LabeledAddress.lblClear ();
    } else if (MainMemory.mmrHumanVersion == 0x0302 && regPC0 == 0x0000a090) {  //プロセスが常駐した直後
      int head = regRn[8] + 256;  //常駐したプロセスのメモリ管理テーブルのアドレス+256=プロセスの先頭
      //----------------------------------------------------------------
      //FSX.Xのマウス受信データ処理ルーチンとマウスワークのアドレスを保存する
      if (Z8530.SCC_FSX_MOUSE) {
        emxCheckFSX (head);
      }
      //----------------------------------------------------------------
      //TwentyOne.xのオプションのアドレスを保存する
      if (HFS.HFS_USE_TWENTY_ONE) {
        emxCheckTwentyOne (head);
      }
    }
  }  //emxNop()

  public static final int[] emxPCM8AFFMap = {
    0x00000138, 0x000001f6, 0x00000394, 0x000011ec, 0x0000120a, 0x00001400, 0x00001814, 0x00001870, 0x00001882, 0x0000188a,
    0x00001892, 0x000018a2, 0x000018a8, 0x000018ca, 0x000018d4, 0x000018e0, 0x000018e8, 0x00001908, 0x000019e4, 0x00001afa,
    0x00001b58, 0x00001b7c, 0x00001bac, 0x00001c38, 0x00001ccc, 0x000021f8, 0x00002250, 0x00002258, 0x00002290, 0x000022a6,
    0x000022b0, 0x000022c0, 0x000022c8, 0x000022de, 0x000022ea, 0x000030c8, 0x000030de, 0x000030e6, 0x000030ea, 0x000030f6,
    0x00003112, 0x00003188, 0x0000334c, 0x0000338a, 0x000033a2, 0x000033c4, 0x000033d0, 0x0000341a, 0x00003428, 0x00003496,
    0x000034a6, 0x000034d6, 0x0000fe0e, 0x0000fec8, 0x0000feec, 0x0000ff46, 0x0000ff4e,
  };

  //emxPatchPCM8A (head)
  //  headから始まるデバイスドライバまたはプロセスがPCM8A.X v1.02ならばパッチをあてる
  public static void emxPatchPCM8A (int head) {
    if (MC68060.mmuPeekLongData (head + 0x10f8, 1) == 0x50434d38 &&  //PCM8
        MC68060.mmuPeekLongData (head + 0x10fc, 1) == 0x41313032) {  //A102
      System.out.println (Multilingual.mlnJapanese ?
                          "PCM8A.X 1.02 があります" :
                          "PCM8A.X 1.02 exists");
      int patched = 0;
      int failed = 0;
      //  I/Oポートのアドレスの上位8ビットが$FFになっているところを$00に修正します。(57箇所)
      for (int offset : emxPCM8AFFMap) {
        if (MC68060.mmuPeekByteZeroData (head + offset, 1) == 0xff) {
          MC68060.mmuPokeByteData (head + offset, 0x00, 1);
          patched++;
        } else {
          failed++;
        }
      }
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "PCM8A.X 1.02 にパッチをあてました (%d/%d)\n" :
                           "PCM8A.X 1.02 was patched (%d/%d)\n",
                           patched, patched + failed);
      }
    }
  }  //emxPatchPCM8A(int)

  //emxCheckFSX (head)
  //  headから始まるデバイスドライバまたはプロセスがFSX.Xならばマウス受信データ処理ルーチンとマウスワークのアドレスを保存する
  public static void emxCheckFSX (int head) {
    if (Z8530.SCC_FSX_MOUSE) {
      if ("\r\nSX SYSTEM for X68000  version 3.10\r\nCopyright 1990,91,92,93,94 SHARP/First Class Technology\r\n".equals (MC68060.mmuPeekStringZ (head + 0x0001ae, 5))) {
        System.out.println (Multilingual.mlnJapanese ?
                            "FSX.X 3.10 があります" :
                            "FSX.X 3.10 exists");
        Z8530.sccFSXMouseHook = head + 0x04f82a;  //マウス受信データ処理ルーチン
        Z8530.sccFSXMouseWork = head + 0x063184;  //マウスワーク
      }
    }
  }  //emxCheckFSX(int)

  //emxCheckTwentyOne (head)
  //  headから始まるデバイスドライバまたはプロセスがTwentyOne.xならばオプションのアドレスを保存する
  public static void emxCheckTwentyOne (int head) {
    if (HFS.HFS_USE_TWENTY_ONE) {
      if (MainMemory.mmrTwentyOneOptionAddress != 0 ||  //TwentyOne.xのオプションのアドレスは確認済みまたは非対応のバージョン
          MainMemory.mmrHumanVersion <= 0) {  //Human68kのバージョンが未確認または未知のバージョン
        return;
      }
      int name1 = MC68060.mmuPeekLongData (head + 14, 1);
      if (name1 == ('*' << 24 | 'T' << 16 | 'w' << 8 | 'e')) {
        int name2 = MC68060.mmuPeekLongData (head + 18, 1);
        if (name2 == ('n' << 24 | 't' << 16 | 'y' << 8 | '*')) {  //TwentyOne.x v1.10まで
          MainMemory.mmrTwentyOneOptionAddress = -1;  //非対応
        }
      } else if (name1 == ('?' << 24 | 'T' << 16 | 'w' << 8 | 'e')) {
        int name2 = MC68060.mmuPeekLongData (head + 18, 1);
        if (name2 == ('n' << 24 | 't' << 16 | 'y' << 8 | '?') ||
            name2 == ('n' << 24 | 't' << 16 | 'y' << 8 | 'E')) {  //TwentyOne.x v1.11から
          System.out.println (Multilingual.mlnJapanese ?
                              "TwentyOne.x があります" :
                              "TwentyOne.x exists");
          MainMemory.mmrTwentyOneOptionAddress = head + 22;
        }
      }
    }
  }  //emxCheckTwentyOne(int)



  //========================================================================================
  //$$IRP 命令の処理
  //
  //  変数名
  //    op                                   オペコード。iiii_qqq_nnn_mmm_rrr
  //    iiii  op >> 12                       命令の種類。ここでは定数
  //    qqq   (op >> 9) - (iiii << 3)        クイックイミディエイトまたはデータレジスタDqの番号
  //    aqq   (op >> 9) - ((iiii << 3) - 8)  アドレスレジスタAqの番号
  //    nnn   op >> 6 & 7                    デスティネーションの実効アドレスのモード
  //    ea    op & 63                        実効アドレスのモードとレジスタ
  //    mmm   ea >> 3                        実効アドレスのモード
  //    rrr   op & 7                         実効アドレスのレジスタ。DrまたはRrのときr[rrr]はr[ea]で代用できる
  //    cccc  op >> 8 & 15                   コンディションコード
  //    a                                    実効アドレス
  //    s                                    テンポラリ
  //    t                                    テンポラリ
  //    w                                    拡張ワード
  //    x                                    被演算数
  //    y                                    演算数
  //    z                                    結果
  //
  //  サイクル数
  //    mpuCycleCountにMC68000のサイクル数を加算する
  //      MC68030のサイクル数はMC68000のサイクル数の0.6倍とみなして計算する
  //      これはROM1.3で起動したときDBRA命令で計測される動作周波数の表示の辻褄を合わせるための係数であり、
  //      MC68000とMC68030のサイクル数の比の平均値ではない
  //        10MHzのMC68000と25MHzのMC68030の速度の比が25/10/0.6=4.17倍となるので何倍も外れてはいないと思われる
  //    MC68000に存在しない命令のサイクル数はMC68000に存在すると仮定した場合の推定値を用いる
  //      オペコードを含むリードとライトを1ワードあたり4サイクルとする
  //
  //  拡張命令
  //    差し障りのない範囲でいくつかの命令を追加してある
  //    エミュレータ拡張命令
  //      HFSBOOT                                         |-|012346|-|-----|-----|          |0100_111_000_000_000
  //      HFSINST                                         |-|012346|-|-----|-----|          |0100_111_000_000_001
  //      HFSSTR                                          |-|012346|-|-----|-----|          |0100_111_000_000_010
  //      HFSINT                                          |-|012346|-|-----|-----|          |0100_111_000_000_011
  //      EMXNOP                                          |-|012346|-|-----|-----|          |0100_111_000_000_100
  //    MC68000で欠番になっているオペコードに割り当てられているColdFireの命令
  //      BITREV.L Dr                                     |-|------|-|-----|-----|D         |0000_000_011_000_rrr (ISA_C)
  //      BYTEREV.L Dr                                    |-|------|-|-----|-----|D         |0000_001_011_000_rrr (ISA_C)
  //      FF1.L Dr                                        |-|------|-|-UUUU|-**00|D         |0000_010_011_000_rrr (ISA_C)
  //      MVS.B <ea>,Dq                                   |-|------|-|-UUUU|-**00|D M+-WXZPI|0111_qqq_100_mmm_rrr (ISA_B)
  //      MVS.W <ea>,Dq                                   |-|------|-|-UUUU|-**00|D M+-WXZPI|0111_qqq_101_mmm_rrr (ISA_B)
  //      MVZ.B <ea>,Dq                                   |-|------|-|-UUUU|-0*00|D M+-WXZPI|0111_qqq_110_mmm_rrr (ISA_B)
  //      MVZ.W <ea>,Dq                                   |-|------|-|-UUUU|-0*00|D M+-WXZPI|0111_qqq_111_mmm_rrr (ISA_B)
  //      SATS.L Dr                                       |-|------|-|-UUUU|-**00|D         |0100_110_010_000_rrr (ISA_B)

  public static final boolean IRP_BITREV_REVERSE = false;  //true=BITREVでInteger.reverseを使う
  public static final boolean IRP_BITREV_SHIFT = false;  //true=BITREVでシフト演算子を使う
  public static final boolean IRP_BITREV_TABLE = true;  //true=BITREVでテーブルを使う

  public static final boolean IRP_MOVEM_MAINMEMORY = true;  //true=000のときMOVEMでメインメモリを特別扱いにする
  public static final boolean IRP_MOVEM_EXPAND = false;  //true=MOVEMで16回展開する。遅くなる
  public static final boolean IRP_MOVEM_LOOP = false;  //true=MOVEMで16回ループする。コンパイラが展開する
  public static final boolean IRP_MOVEM_SHIFT_LEFT = false;  //true=MOVEMで0になるまで左にシフトする。reverseが入る分遅い
  public static final boolean IRP_MOVEM_SHIFT_RIGHT = true;  //true=MOVEMで0になるまで右にシフトする
  public static final boolean IRP_MOVEM_ZEROS = false;  //true=MOVEMでInteger.numberOfTrailingZerosを使う。ループ回数は少ないがスキップする処理が冗長になるので最速ではない

  public static void irpReset () {
    //メインメモリのmmrResetとバスコントローラのbusResetはmpuResetへ
    CRTC.crtReset ();  //CRT CRTコントローラ
    VideoController.vcnReset ();  //VCN ビデオコントローラ
    HD63450.dmaReset ();  //DMA DMAコントローラ
    MC68901.mfpReset ();  //MFP MFP
    Keyboard.kbdRePress();  //押されているキーを再入力する
    RP5C15.rtcReset ();  //RTC RTC
    PrinterPort.prnReset ();  //PRN プリンタポート
    SoundSource.sndReset ();  //SND サウンド
    YM2151.opmReset ();  //OPM FM音源
    ADPCM.pcmReset ();  //PCM ADPCM音源
    FDC.fdcReset ();  //FDC FDコントローラ
    IOInterrupt.ioiReset ();  //IOI I/O割り込み
    XEiJ.eb2Reset ();  //EB2 拡張ボードレベル2割り込み
    SPC.spcReset ();  //SPC SCSIプロトコルコントローラ
    Z8530.sccReset ();  //SCC SCC
    PPI.ppiReset ();  //PPI PPI
    HFS.hfsReset ();  //HFS ホストファイルシステムインタフェイス
    SpriteScreen.sprReset ();  //SPR スプライト画面
    //smrReset()はspcSCSIEXOnとspcSCSIINOnを使うのでSPC.spcReset()よりも後であること
    XEiJ.smrReset ();  //SMR SRAM
  }  //irpReset()

  //右シフト・ローテート命令
  //
  //  ASR
  //    X  countが0のとき変化しない。他は最後に押し出されたビット
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときクリア。他は最後に押し出されたビット
  //    ASR.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X 0
  //       1 ........................ｱｱｲｳｴｵｶｷ ｸ ｸ
  //       :
  //       7 ........................ｱｱｱｱｱｱｱｱ ｲ ｲ
  //       8 ........................ｱｱｱｱｱｱｱｱ ｱ ｱ
  //    ASR.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X 0
  //       1 ................ｱｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿ ﾀ ﾀ
  //       :
  //      15 ................ｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱ ｲ ｲ
  //      16 ................ｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱ ｱ ｱ
  //    ASR.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X 0
  //       1 ｱｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏ ﾐ ﾐ
  //       :
  //      31 ｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱ ｲ ｲ
  //      32 ｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱｱ ｱ ｱ
  //
  //  LSR
  //    X  countが0のとき変化しない。他は最後に押し出されたビット
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときクリア。他は最後に押し出されたビット
  //    LSR.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X 0
  //       1 ........................0ｱｲｳｴｵｶｷ ｸ ｸ
  //       :
  //       7 ........................0000000ｱ ｲ ｲ
  //       8 ........................00000000 ｱ ｱ
  //       9 ........................00000000 0 0
  //    LSR.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X 0
  //       1 ................0ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿ ﾀ ﾀ
  //       :
  //      15 ................000000000000000ｱ ｲ ｲ
  //      16 ................0000000000000000 ｱ ｱ
  //      17 ................0000000000000000 0 0
  //    LSR.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X 0
  //       1 0ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏ ﾐ ﾐ
  //       :
  //      31 0000000000000000000000000000000ｱ ｲ ｲ
  //      32 00000000000000000000000000000000 ｱ ｱ
  //      33 00000000000000000000000000000000 0 0
  //
  //  ROXR
  //    X  countが0のとき変化しない。他は最後に押し出されたビット
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときXのコピー。他は最後に押し出されたビット
  //    ROXR.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X X
  //       1 ........................Xｱｲｳｴｵｶｷ ｸ ｸ
  //       2 ........................ｸXｱｲｳｴｵｶ ｷ ｷ
  //       :
  //       7 ........................ｳｴｵｶｷｸXｱ ｲ ｲ
  //       8 ........................ｲｳｴｵｶｷｸX ｱ ｱ
  //       9 ........................ｱｲｳｴｵｶｷｸ X X
  //    ROXR.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X X
  //       1 ................Xｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿ ﾀ ﾀ
  //       2 ................ﾀXｱｲｳｴｵｶｷｸｹｺｻｼｽｾ ｿ ｿ
  //       :
  //      15 ................ｳｴｵｶｷｸｹｺｻｼｽｾｿﾀXｱ ｲ ｲ
  //      16 ................ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀX ｱ ｱ
  //      17 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X X
  //    ROXR.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X X
  //       1 Xｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏ ﾐ ﾐ
  //       2 ﾐXｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎ ﾏ ﾏ
  //       :
  //      31 ｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐXｱ ｲ ｲ
  //      32 ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐX ｱ ｱ
  //      33 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X X
  //
  //  ROR
  //    X  常に変化しない
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときクリア。他は結果の最上位ビット
  //    ROR.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X 0
  //       1 ........................ｸｱｲｳｴｵｶｷ X ｸ
  //       :
  //       7 ........................ｲｳｴｵｶｷｸｱ X ｲ
  //       8 ........................ｱｲｳｴｵｶｷｸ X ｱ
  //    ROR.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X 0
  //       1 ................ﾀｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿ X ﾀ
  //       :
  //      15 ................ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀｱ X ｲ
  //      16 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X ｱ
  //    ROR.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X 0
  //       1 ﾐｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏ X ﾐ
  //       :
  //      31 ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐｱ X ｲ
  //      32 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X ｱ

  //左シフト・ローテート命令
  //
  //  ASL
  //    X  countが0のとき変化しない。他は最後に押し出されたビット
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  ASRで元に戻せないときセット。他はクリア
  //    C  countが0のときクリア。他は最後に押し出されたビット
  //    ASL.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X 0
  //       1 ........................ｲｳｴｵｶｷｸ0 ｱ ｱ
  //       :
  //       7 ........................ｸ0000000 ｷ ｷ
  //       8 ........................00000000 ｸ ｸ
  //       9 ........................00000000 0 0
  //    ASL.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X 0
  //       1 ................ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ0 ｱ ｱ
  //       :
  //      15 ................ﾀ000000000000000 ｿ ｿ
  //      16 ................0000000000000000 ﾀ ﾀ
  //      17 ................0000000000000000 0 0
  //    ASL.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X 0
  //       1 ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ0 ｱ ｱ
  //       :
  //      31 ﾐ0000000000000000000000000000000 ﾏ ﾏ
  //      32 00000000000000000000000000000000 ﾐ ﾐ
  //      33 00000000000000000000000000000000 0 0
  //
  //  LSL
  //    X  countが0のとき変化しない。他は最後に押し出されたビット
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときクリア。他は最後に押し出されたビット
  //    LSL.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X 0
  //       1 ........................ｲｳｴｵｶｷｸ0 ｱ ｱ
  //       :
  //       7 ........................ｸ0000000 ｷ ｷ
  //       8 ........................00000000 ｸ ｸ
  //       9 ........................00000000 0 0
  //    LSL.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X 0
  //       1 ................ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ0 ｱ ｱ
  //       :
  //      15 ................ﾀ000000000000000 ｿ ｿ
  //      16 ................0000000000000000 ﾀ ﾀ
  //      17 ................0000000000000000 0 0
  //    LSL.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X 0
  //       1 ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ0 ｱ ｱ
  //       :
  //      31 ﾐ0000000000000000000000000000000 ﾏ ﾏ
  //      32 00000000000000000000000000000000 ﾐ ﾐ
  //      33 00000000000000000000000000000000 0 0
  //
  //  ROXL
  //    X  countが0のとき変化しない。他は最後に押し出されたビット
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときXのコピー。他は最後に押し出されたビット
  //    ROXL.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X X
  //       1 ........................ｲｳｴｵｶｷｸX ｱ ｱ
  //       2 ........................ｳｴｵｶｷｸXｱ ｲ ｲ
  //       :
  //       7 ........................ｸXｱｲｳｴｵｶ ｷ ｷ
  //       8 ........................Xｱｲｳｴｵｶｷ ｸ ｸ
  //       9 ........................ｱｲｳｴｵｶｷｸ X X
  //    ROXL.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X X
  //       1 ................ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀX ｱ ｱ
  //       2 ................ｳｴｵｶｷｸｹｺｻｼｽｾｿﾀXｱ ｲ ｲ
  //       :
  //      15 ................ﾀXｱｲｳｴｵｶｷｸｹｺｻｼｽｾ ｿ ｿ
  //      16 ................Xｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿ ﾀ ﾀ
  //      17 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X X
  //    ROXL.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X X
  //       1 ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐX ｱ ｱ
  //       2 ｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐXｱ ｲ ｲ
  //       :
  //      31 ﾐXｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎ ﾏ ﾏ
  //      32 Xｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏ ﾐ ﾐ
  //      33 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X X
  //
  //  ROL
  //    X  常に変化しない
  //    N  結果の最上位ビット
  //    Z  結果が0のときセット。他はクリア
  //    V  常にクリア
  //    C  countが0のときクリア。他は結果の最下位ビット
  //    ROL.B
  //         ........................ｱｲｳｴｵｶｷｸ X C
  //       0 ........................ｱｲｳｴｵｶｷｸ X 0
  //       1 ........................ｲｳｴｵｶｷｸｱ X ｱ
  //       :
  //       7 ........................ｸｱｲｳｴｵｶｷ X ｷ
  //       8 ........................ｱｲｳｴｵｶｷｸ X ｸ
  //    ROL.W
  //         ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X C
  //       0 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X 0
  //       1 ................ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀｱ X ｱ
  //       :
  //      15 ................ﾀｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿ X ｿ
  //      16 ................ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀ X ﾀ
  //    ROL.L
  //         ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X C
  //       0 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X 0
  //       1 ｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐｱ X ｱ
  //       :
  //      31 ﾐｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏ X ﾏ
  //      32 ｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐ X ﾐ



  //========================================================================================
  //以下のメソッドはインライン展開する
/*
  public static void pushb (int x) throws M68kException {
    wb (r[15] -= 2, x);  //ワードの上位側
  }  //pushb(int)
  public static void pushw (int x) throws M68kException {
    ww (r[15] -= 2, x);
  }  //pushw(int)
  public static void pushl (int x) throws M68kException {
    wl (r[15] -= 4, x);
  }  //pushl(int)

  public static int popbs () throws M68kException {
    return rbs ((r[15] += 2) - 2);  //ワードの上位側
  }  //popbs()
  public static int popbz () throws M68kException {
    return rbz ((r[15] += 2) - 2);  //ワードの上位側
  }  //popbz()
  public static int popws () throws M68kException {
    return rws ((r[15] += 2) - 2);
  }  //popws()
  public static int popwz () throws M68kException {
    return rwz ((r[15] += 2) - 2);
  }  //popwz()
  public static int popls () throws M68kException {
    return rls ((r[15] += 4) - 4);
  }  //popls()

  public static int pcbs () throws M68kException {
    return rbs ((pc += 2) - 1);  //ワードの下位側
  }  //pcbs()
  public static int pcbz () throws M68kException {
    return rbz ((pc += 2) - 1);  //ワードの下位側
  }  //pcbz()
  public static int pcws () throws M68kException {
    return rwse ((pc += 2) - 2);
  }  //pcws()
  public static int pcwz () throws M68kException {
    return rwze ((pc += 2) - 2);
  }  //pcwz()
  public static int pcls () throws M68kException {
    return rlse ((pc += 4) - 4);
  }  //pcls()

  public static void ccr_tst (int z) {  //Xは変化しない。VとCはクリア
    ccr = z >> 28 & CCR_N | (z == 0 ? ccr & CCR_X | CCR_Z : ccr & CCR_X);  //ccr_tst
  }  //ccr_tst(int)
  public static void ccr_btst (int z) {  //Z以外は変化しない
    ccr = (ccr & (CCR_X | CCR_N | CCR_V | CCR_C) | (z == 0 ? CCR_Z : 0));
  }  //ccr_btst(int)
  public static void ccr_clr () {  //Xは変化しない。Zはセット。NとVとCはクリア
    ccr = ccr & CCR_X | CCR_Z;  //ccr_clr
  }  //ccr_clr(int)

  //                  x-y V                                  x-y C
  //  x  y|  0   1   2   3  -4  -3  -2  -1   x  y|  0   1   2   3   4   5   6   7
  //  ----+--------------------------------  ----+--------------------------------
  //    0 |  0  -1  -2  -3   4*  3   2   1     0 |  0  -1* -2* -3* -4* -5* -6* -7*
  //    1 |  1   0  -1  -2   5*  4*  3   2     1 |  1   0  -1* -2* -3* -4* -5* -6*
  //    2 |  2   1   0  -1   6*  5*  4*  3     2 |  2   1   0  -1* -2* -3* -4* -5*
  //    3 |  3   2   1   0   7*  6*  5*  4*    3 |  3   2   1   0  -1* -2* -3* -4*
  //   -4 | -4  -5* -6* -7*  0  -1  -2  -3     4 |  4   3   2   1   0  -1* -2* -3*
  //   -3 | -3  -4  -5* -6*  1   0  -1  -2     5 |  5   4   3   2   1   0  -1* -2*
  //   -2 | -2  -3  -4  -5*  2   1   0  -1     6 |  6   5   4   3   2   1   0  -1*
  //   -1 | -1  -2  -3  -4   3   2   1   0     7 |  7   6   5   4   3   2   1   0
  //                 x-y-1 V                                x-y-1 C
  //  x  y|  0   1   2   3  -4  -3  -2  -1   x  y|  0   1   2   3   4   5   6   7
  //  ----+--------------------------------  ----+--------------------------------
  //    0 | -1  -2  -3  -4   3   2   1   0     0 | -1* -2* -3* -4* -5* -6* -7* -8*
  //    1 |  0  -1  -2  -3   4*  3   2   1     1 |  0  -1* -2* -3* -4* -5* -6* -7*
  //    2 |  1   0  -1  -2   5*  4*  3   2     2 |  1   0  -1* -2* -3* -4* -5* -6*
  //    3 |  2   1   0  -1   6*  5*  4*  3     3 |  2   1   0  -1* -2* -3* -4* -5*
  //   -4 | -5* -6* -7* -8* -1  -2  -3  -4     4 |  3   2   1   0  -1* -2* -3* -4*
  //   -3 | -4  -5* -6* -7*  0  -1  -2  -3     5 |  4   3   2   1   0  -1* -2* -3*
  //   -2 | -3  -4  -5* -6*  1   0  -1  -2     6 |  5   4   3   2   1   0  -1* -2*
  //   -1 | -2  -3  -4  -5*  2   1   0  -1     7 |  6   5   4   3   2   1   0  -1*
  //  x        y         z=x-y    v        c         z=x-y-1  v        c
  //  00000000 00001111  01111000 00001000 01111111  11110000 00000000 11111111
  //  00000000 00001111  00111100 00001100 00111111  01111000 00001000 01111111
  //  00000000 00001111  00011110 00001110 00011111  00111100 00001100 00111111
  //  00000000 00001111  00001111 00001111 00001111  00011110 00001110 00011111
  //  11111111 00001111  10000111 01110000 00000111  00001111 11110000 00001111
  //  11111111 00001111  11000011 00110000 00000011  10000111 01110000 00000111
  //  11111111 00001111  11100001 00010000 00000001  11000011 00110000 00000011
  //  11111111 00001111  11110000 00000000 00000000  11100001 00010000 00000001
  //  Vは右上と左下でxとzが異なる部分
  //    V = ((x ^ y) & (x ^ z)) < 0
  //  Cは右上全部および左上と右下でzがある部分
  //    C = (~x & y | ~(x ^ y) & z) < 0
  //    ~を使わずに書けるおそらく最短の等価式
  //    C = (x & (y ^ z) ^ (y | z)) < 0
  //      perl -e "for$x(0..1){for$y(0..1){for$z(0..1){print join(',',$x,$y,$z,(1^$x)&$y|(1^($x^$y))&$z,$x&($y^$z)^($y|$z)).chr(10);}}}"
  //      0,0,0,0,0
  //      0,0,1,1,1
  //      0,1,0,1,1
  //      0,1,1,1,1
  //      1,0,0,0,0
  //      1,0,1,0,0
  //      1,1,0,0,0
  //      1,1,1,1,1
  public static void ccr_sub (int x, int y, int z) {
    ccr = (z >> 28 & CCR_N | (z == 0 ? CCR_Z : 0) |
           ((x ^ y) & (x ^ z)) >>> 31 << 1 |
           (x & (y ^ z) ^ (y | z)) >> 31 & (CCR_X | CCR_C));
  }  //ccr_sub(int,int,int)
  public static void ccr_subx (int x, int y, int z) {  //Zは結果が0のとき変化しない
    ccr = (z >> 28 & CCR_N | (z == 0 ? ccr & CCR_Z : 0) |
           ((x ^ y) & (x ^ z)) >>> 31 << 1 |
           (x & (y ^ z) ^ (y | z)) >> 31 & (CCR_X | CCR_C));
  }  //ccr_subx(int,int,int)
  public static void ccr_subq (int x, int y, int z) {  //ccr_subを常にy>0としたもの。Vは負→正のとき1。Cは正→負のとき1
    ccr = (z >> 28 & CCR_N | (z == 0 ? CCR_Z : 0) |
           (x & ~z) >>> 31 << 1 |
           (~x & z) >> 31 & (CCR_X | CCR_C));
  }  //ccr_subq(int,int,int)
  public static void ccr_neg (int y, int z) {  //ccr_subを常にx==0としたもの。Vは-MAX→-MAXのみ1。Cは0→0以外1
    ccr = (z >> 28 & CCR_N | (z == 0 ? CCR_Z : 0) |
           (y & z) >>> 31 << 1 |
           (y | z) >> 31 & (CCR_X | CCR_C));
  }  //ccr_neg(int,int)
  public static void ccr_negx (int y, int z) {  //ccr_subxを常にx==0としたもの。Zは結果が0のとき変化しない
    ccr = (z >> 28 & CCR_N | (z == 0 ? ccr & CCR_Z : 0) |
           (y & z) >>> 31 << 1 |
           (y | z) >> 31 & (CCR_X | CCR_C));
  }  //ccr_negx(int,int)
  public static void ccr_cmp (int x, int y, int z) {  //Xは変化しない
    ccr = (z >> 28 & CCR_N | (z == 0 ? ccr & CCR_X | CCR_Z : ccr & CCR_X) |
           ((x ^ y) & (x ^ z)) >>> 31 << 1 |
           (x & (y ^ z) ^ (y | z)) >>> 31);
  }  //ccr_cmp(int,int,int)

  //                  x+y V                                  x+y C
  //  x  y|  0   1   2   3  -4  -3  -2  -1   x  y|  0   1   2   3   4   5   6   7
  //  ----+--------------------------------  ----+--------------------------------
  //    0 |  0   1   2   3  -4  -3  -2  -1     0 |  0   1   2   3   4   5   6   7
  //    1 |  1   2   3   4* -3  -2  -1   0     1 |  1   2   3   4   5   6   7   8*
  //    2 |  2   3   4*  5* -2  -1   0   1     2 |  2   3   4   5   6   7   8*  9*
  //    3 |  3   4*  5*  6* -1   0   1   2     3 |  3   4   5   6   7   8*  9* 10*
  //   -4 | -4  -3  -2  -1  -8* -7* -6* -5*    4 |  4   5   6   7   8*  9* 10* 11*
  //   -3 | -3  -2  -1   0  -7* -6* -5* -4     5 |  5   6   7   8*  9* 10* 11* 12*
  //   -2 | -2  -1   0   1  -6* -5* -4  -3     6 |  6   7   8*  9* 10* 11* 12* 13*
  //   -1 | -1   0   1   2  -5* -4  -3  -2     7 |  7   8*  9* 10* 11* 12* 13* 14*
  //                 x+y+1 V                                x+y+1 C
  //  x  y|  0   1   2   3  -4  -3  -2  -1   x  y|  0   1   2   3   4   5   6   7
  //  ----+--------------------------------  ----+--------------------------------
  //    0 |  1   2   3   4* -3  -2  -1   0     0 |  1   2   3   4   5   6   7   8*
  //    1 |  2   3   4*  5* -2  -1   0   1     1 |  2   3   4   5   6   7   8*  9*
  //    2 |  3   4*  5*  6* -1   0   1   2     2 |  3   4   5   6   7   8*  9* 10*
  //    3 |  4*  5*  6*  7*  0   1   2   3     3 |  4   5   6   7   8*  9* 10* 11*
  //   -4 | -3  -2  -1   0  -7* -6* -5* -4     4 |  5   6   7   8*  9* 10* 11* 12*
  //   -3 | -2  -1   0   1  -6* -5* -4  -3     5 |  6   7   8*  9* 10* 11* 12* 13*
  //   -2 | -1   0   1   2  -5* -4  -3  -2     6 |  7   8*  9* 10* 11* 12* 13* 14*
  //   -1 |  0   1   2   3  -4  -3  -2  -1     7 |  8*  9* 10* 11* 12* 13* 14* 15*
  //  x        y         z=x+y    v        c         z=x+y+1  v        c
  //  00000000 00001111  00001111 00000000 00000000  00011110 00010000 00000001
  //  00000000 00001111  00011110 00010000 00000001  00111100 00110000 00000011
  //  00000000 00001111  00111100 00110000 00000011  01111000 01110000 00000111
  //  00000000 00001111  01111000 01110000 00000111  11110000 11110000 00001111
  //  11111111 00001111  11110000 00001111 00001111  11100001 00001110 00011111
  //  11111111 00001111  11100001 00001110 00011111  11000011 00001100 00111111
  //  11111111 00001111  11000011 00001100 00111111  10000111 00001000 01111111
  //  11111111 00001111  10000111 00001000 01111111  00001111 00000000 11111111
  //  Vは左上と右下でxとzが異なる部分
  //    V = (~(x ^ y) & (x ^ z)) < 0
  //    ~を使わずに書けるおそらく最短の等価式
  //    V = ((x ^ z) & (y ^ z)) < 0
  //      perl -e "for$x(0..1){for$y(0..1){for$z(0..1){print join(',',$x,$y,$z,(1^($x^$y))&($x^$z),($x^$z)&($y^$z)).chr(10);}}}"
  //      0,0,0,0,0
  //      0,0,1,1,1
  //      0,1,0,0,0
  //      0,1,1,0,0
  //      1,0,0,0,0
  //      1,0,1,0,0
  //      1,1,0,1,1
  //      1,1,1,0,0
  //  Cは右下全部および右上と左下でzがない部分
  //    C = (x & y | (x ^ y) & ~z) < 0
  //    ~を使わずに書けるおそらく最短の等価式
  //    C = ((x | y) ^ (x ^ y) & z) < 0
  //      perl -e "for$x(0..1){for$y(0..1){for$z(0..1){print join(',',$x,$y,$z,$x&$y|($x^$y)&(1^$z),($x|$y)^($x^$y)&$z).chr(10);}}}"
  //      0,0,0,0,0
  //      0,0,1,0,0
  //      0,1,0,1,1
  //      0,1,1,0,0
  //      1,0,0,1,1
  //      1,0,1,0,0
  //      1,1,0,1,1
  //      1,1,1,1,1
  public static void ccr_add (int x, int y, int z) {
    ccr = (z >> 28 & CCR_N | (z == 0 ? CCR_Z : 0) |
           ((x ^ z) & (y ^ z)) >>> 31 << 1 |
           ((x | y) ^ (x ^ y) & z) >> 31 & (CCR_X | CCR_C));
  }  //ccr_add(int,int,int)
  public static void ccr_addx (int x, int y, int z) {  //Zは結果が0のとき変化しない
    ccr = (z >> 28 & CCR_N | (z == 0 ? ccr & CCR_Z : 0) |
           ((x ^ z) & (y ^ z)) >>> 31 << 1 |
           ((x | y) ^ (x ^ y) & z) >> 31 & (CCR_X | CCR_C));
  }  //ccr_addx(int,int,int)
  public static void ccr_addq (int x, int y, int z) {  //ccr_addを常にy>0としたもの。Vは正→負のとき1。Cは負→正のとき1
    ccr = (z >> 28 & CCR_N | (z == 0 ? CCR_Z : 0) |
           (~x & z) >>> 31 << 1 |
           (x & ~z) >> 31 & (CCR_X | CCR_C));
  }  //ccr_addq(int,int,int)
*/



  //========================================================================================
  //$$EFA 実効アドレス計算
  //
  //  アドレッシングモードとサイクル数
  //    DrまたはArを指定すると不当命令になる
  //    例えばArを指定できない命令ならDrだけを特別扱いにして残りをefa～に渡せばArは自動的に不当命令になる
  //    実効アドレス計算とオペランドのアクセス1回分のサイクル数がXEiJ.mpuCycleCountに加算される
  //                        mmm rrr  Any Mem Mlt Cnt Clt  Byte Word Long Quad Extd  LeaPea JmpJsr
  //         -------------  --- ---  --- --- --- --- ---  ---- ---- ---- ---- ----  ------ ------
  //      M  (Ar)           010 rrr    *   *   *   *   *     4    4    8   16   24       4      8
  //      +  (Ar)+          011 rrr    *   *   *             4    4    8   16   24
  //      -  -(Ar)          100 rrr    *   *   *             6    6   10   18   26
  //      W  (d16,Ar)       101 rrr    *   *   *   *   *     8    8   12   20   28       8     10
  //      X  (d8,Ar,Rn.wl)  110 rrr    *   *   *   *   *    10   10   14   22   30      12     14
  //      Z  (xxx).W        111 000    *   *   *   *   *     8    8   12   20   28       8     10
  //      Z  (xxx).L        111 001    *   *   *   *   *    12   12   16   24   32      12     12
  //      P  (d16,PC)       111 010    *   *       *         8    8   12   20   28       8     10
  //      P  (d8,PC,Rn.wl)  111 011    *   *       *        10   10   14   22   30      12     14
  //      I  #<data>        111 100    *                     4    4    8   16   24
  //    MoveToMemByte/MoveToMemWord/MoveToMemLongはデスティネーションが-(Aq)のとき2減らす
  //    AddToRegLong/AddaLong/AndToRegLong/OrToRegLong/SubToRegLong/SubaLongはソースがDr/Ar/#<data>のとき2増やす
  //
  //  フルフォーマットの拡張ワードの処理の冗長表現
  //      t = r[ea - (0b110_000 - 8)];  //ベースレジスタ
  //      w = rwze ((pc += 2) - 2);  //pcwz。拡張ワード
  //      x = r[w >> 12];  //インデックスレジスタ
  //      if ((w & 0x0800) == 0) {  //ワードインデックス
  //        x = (short) x;
  //      }
  //      x <<= w >> 9 & 3;  //スケールファクタ。ワードインデックスのときは符号拡張してから掛ける
  //      if ((w & 0x0100) == 0) {  //短縮フォーマット
  //        t += (byte) w + x;  //8ビットディスプレースメント
  //      } else {  //フルフォーマット
  //        if ((w & 0x0080) != 0) {  //ベースサプレス
  //          t = 0;
  //        }
  //        if ((w & 0x0040) != 0) {  //インデックスサプレス
  //          x = 0;
  //        }
  //        if ((w & 0x0020) != 0) {  //ベースディスプレースメントあり
  //          if ((w & 0x0010) == 0) {  //ワードベースディスプレースメント
  //            t += rwse ((pc += 2) - 2);  //pcws
  //          } else {  //ロングベースディスプレースメント
  //            t += rlse ((pc += 4) - 4);  //pcls
  //          }
  //        }
  //        if ((w & 0x0003) == 0) {  //メモリ間接なし
  //          t += x;
  //        } else {  //メモリ間接あり
  //          if ((w & 0x0004) == 0) {  //プリインデックス
  //            t = rls (t + x);
  //          } else {  //ポストインデックス
  //            t = rls (t) + x;
  //          }
  //          if ((w & 0x0002) != 0) {  //アウタディスプレースメントあり
  //            if ((w & 0x0001) == 0) {  //ワードアウタディスプレースメント
  //              t += rwse ((pc += 2) - 2);  //pcws
  //            } else {  //ロングアウタディスプレースメント
  //              t += rlse ((pc += 4) - 4);  //pcls
  //            }
  //          }
  //        }
  //      }
  //      return t;
  //
  //  フルフォーマットの拡張ワードのサイクル数
  //    ベースディスプレースメントとメモリ間接とアウターディスプレースメントのリード回数に応じてサイクル数を加算する
  //    ベースレジスタとインデックスレジスタとスケールファクタの有無はサイクル数に影響しないものとする
  //      fedcba9876543210  bd  []  od  計
  //      .......0........               0  (d8,～)
  //      .......1..01..00               0  (～)
  //      .......1..01..01       8       8  ([～])
  //      .......1..01..10       8   4  12  ([～],od.W)
  //      .......1..01..11       8   8  16  ([～],od.L)
  //      .......1..10..00   4           4  (bd.W,～)
  //      .......1..10..01   4   8      12  ([bd.W,～])
  //      .......1..10..10   4   8   4  16  ([bd.W,～],od.W)
  //      .......1..10..11   4   8   8  20  ([bd.W,～],od.L)
  //      .......1..11..00   8           8  (bd.L,～)
  //      .......1..11..01   8   8      16  ([bd.L,～])
  //      .......1..11..10   8   8   4  20  ([bd.L,～],od.W)
  //      .......1..11..11   8   8   8  24  ([bd.L,～],od.L)
  //    1つの式で書くこともできるが冗長になるのでテーブル参照にする
  //
  //  MC68060のサイクル数
  //    ブリーフフォーマットは0、フルフォーマットのメモリ間接なしは1、フルフォーマットのメモリ間接ありは3
  //

  //拡張ワードのサイクル数
/*
  public static final int[] EFA_EXTENSION_CLK = {                  //fedcba9876543210
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  //.......0........
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..00....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..01....
    4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  //.......1..10....
    8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  //.......1..11....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..00....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..01....
    4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  //.......1..10....
    8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  //.......1..11....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..00....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..01....
    4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  //.......1..10....
    8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  //.......1..11....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..00....
    0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  0,  8, 12, 16,  //.......1..01....
    4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  4, 12, 16, 20,  //.......1..10....
    8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  8, 16, 20, 24,  //.......1..11....
  };
*/
  //  perl misc/itob.pl xeij/XEiJ.java EFA_EXTENSION_CLK
  public static final byte[] EFA_EXTENSION_CLK = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\4\f\20\24\4\f\20\24\4\f\20\24\4\f\20\24\b\20\24\30\b\20\24\30\b\20\24\30\b\20\24\30\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\4\f\20\24\4\f\20\24\4\f\20\24\4\f\20\24\b\20\24\30\b\20\24\30\b\20\24\30\b\20\24\30\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\4\f\20\24\4\f\20\24\4\f\20\24\4\f\20\24\b\20\24\30\b\20\24\30\b\20\24\30\b\20\24\30\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\0\b\f\20\4\f\20\24\4\f\20\24\4\f\20\24\4\f\20\24\b\20\24\30\b\20\24\30\b\20\24\30\b\20\24\30".getBytes (XEiJ.ISO_8859_1);

  public static final boolean EFA_SEPARATE_AR = false;  //true=(Ar)を(A0)～(A7)に分ける



  //========================================================================================
  //$$BUS バスコントローラ

  public static final boolean BUS_SPLIT_UNALIGNED_LONG = false;  //true=4の倍数ではない偶数アドレスへのロングワードアクセスを常に分割する

  //マザーボードのアドレス空間
  public static final int BUS_MOTHER_BITS = 24;  //マザーボードのアドレス空間のビット数
  public static final int BUS_MOTHER_SIZE = BUS_MOTHER_BITS < 32 ? 1 << BUS_MOTHER_BITS : 0;  //マザーボードのアドレス空間のサイズ。1<<32は1が押し出されて0になるのではなくてシフトカウントが5bitでマスクされて1<<0=1になることに注意
  public static final int BUS_MOTHER_MASK = BUS_MOTHER_SIZE - 1;  //マザーボードのアドレスのマスク。a&BUS_MOTHER_MASK
  //ダミー4bitページ
  public static final int BUS_DUMMY_4BIT_PAGE = BUS_MOTHER_SIZE;
  public static final int BUS_DUMMY_4BIT_SIZE = 2 * 512 * 512;  //256KB
  //1024ドット256色(改造)
  public static final int BUS_MODIFIED_256_PAGE = BUS_DUMMY_4BIT_PAGE + BUS_DUMMY_4BIT_SIZE;
  public static final int BUS_MODIFIED_256_SIZE = 2 * 1024 * 1024;  //2MB。上位バイト(偶数アドレス)は0
  //1024ドット65536色(改造)
  public static final int BUS_MODIFIED_65536_PAGE = BUS_MODIFIED_256_PAGE + BUS_MODIFIED_256_SIZE;
  public static final int BUS_MODIFIED_65536_SIZE = 2 * 1024 * 1024;  //2MB
  //合計
  public static final int BUS_ARRAY_SIZE = BUS_MODIFIED_65536_PAGE + BUS_MODIFIED_65536_SIZE;

  //ページ
  public static final int BUS_PAGE_BITS = 12;  //ページのビット数。a>>>BUS_PAGE_BITS
  public static final int BUS_PAGE_SIZE = 1 << BUS_PAGE_BITS;  //ページのサイズ
  public static final int BUS_PAGE_COUNT = 1 << (32 - BUS_PAGE_BITS);  //ページの数

  //ss
  public static final int MPU_SS_BYTE = 0;
  public static final int MPU_SS_WORD = 1;
  public static final int MPU_SS_LONG = 2;

  //wr
  public static final int MPU_WR_WRITE = 0;
  public static final int MPU_WR_READ = 1;

  //us
  public static final int MPU_US_USER = 0;
  public static final int MPU_US_SUPERVISOR = 1;

  //メモリマップ
  public static final MemoryMappedDevice[] busUserMap = new MemoryMappedDevice[BUS_PAGE_COUNT];  //ユーザモード用のメモリマップ
  public static final MemoryMappedDevice[] busSuperMap = new MemoryMappedDevice[BUS_PAGE_COUNT];  //スーパーバイザモード用のメモリマップ
  public static MemoryMappedDevice[] busMemoryMap;  //現在のメモリマップ。srS==0?um:sm。DataBreakPoint.DBP_ON==trueのときは使わない

  //X68030/Xellent30のハイメモリ
  public static final int BUS_HIGH_MEMORY_START = 0x01000000;  //X68030/Xellent30のハイメモリの開始アドレス
  public static int busHighMemorySize;  //X68030/Xellent30のハイメモリのサイズ
  public static byte[] busHighMemoryArray;  //X68030/Xellent30のハイメモリの配列
  public static boolean busHighMemorySaveOn;  //true=X68030/Xellent30のハイメモリの内容を保存する

  //060turboのローカルメモリ
  public static final int BUS_LOCAL_MEMORY_START = 0x10000000;  //060turboのローカルメモリの開始アドレス
  public static int busLocalMemorySize;  //060turboのローカルメモリのサイズ
  public static byte[] busLocalMemoryArray;  //060turboのローカルメモリの配列
  public static boolean busLocalMemorySaveOn;  //true=060turboのローカルメモリの内容を保存する

  //拡張メモリ
  public static final byte[] BUS_DUMMY_MEMORY_ARRAY = new byte[0];  //X68030と060turbo以外の拡張メモリの配列
  public static int busRequestExMemoryStart;  //次回起動時の拡張メモリの開始アドレス
  public static int busRequestExMemorySize;  //次回起動時の拡張メモリの長さ
  public static byte[] busRequestExMemoryArray;  //次回起動時の拡張メモリの配列。BUS_DUMMY_MEMORY_ARRAY,busHighMemoryArray,busLocalMemoryArrayのいずれか
  public static int busExMemoryStart;  //拡張メモリの開始アドレス
  public static int busExMemorySize;  //拡張メモリの長さ
  public static byte[] busExMemoryArray;  //拡張メモリの配列。BUS_DUMMY_MEMORY_ARRAY,busHighMemoryArray,busLocalMemoryArrayのいずれか

  //FC2ピンをカットする
  public static boolean busRequestCutFC2Pin;
  public static boolean busCutFC2Pin;

  //busInit ()
  //  バスコントローラを初期化する
  public static void busInit () {
    //um = new MMD[BUS_PAGE_COUNT];
    //sm = new MMD[BUS_PAGE_COUNT];
    if (!DataBreakPoint.DBP_ON) {
      busMemoryMap = busSuperMap;
    }

    //X68030/Xellent30のハイメモリ
    int highMemorySizeMB = Settings.sgsGetInt ("highmemory");  //X68030/Xellent30のハイメモリのサイズ
    busHighMemorySize = highMemorySizeMB == 16 ? highMemorySizeMB << 20 : 0 << 20;
    if (busHighMemorySize == 0) {
        System.out.println (Multilingual.mlnJapanese ?
                            "X68030/Xellent30 のハイメモリはありません" :
                            "X68030/Xellent30 high memory does not exists");
    } else {
      System.out.printf (Multilingual.mlnJapanese ?
                         "X68030/Xellent30 のハイメモリのサイズは %dMB です\n" :
                         "X68030/Xellent30 high memory size is %dMB\n",
                         busHighMemorySize >> 20);
    }
    busHighMemoryArray = new byte[busHighMemorySize];

    busHighMemorySaveOn = Settings.sgsGetOnOff ("highmemorysave");  //X68030/Xellent30のハイメモリの内容を保存するか

    byte[] highMemoryArray = Settings.sgsGetData ("highmemorydata");  //X68030/Xellent30のハイメモリの内容(gzip+base64)
    if (busHighMemorySize != 0) {  //X68030/Xellent30のハイメモリがある
      if (highMemoryArray.length != 0) {  //復元するデータがある
        System.out.println (Multilingual.mlnJapanese ?
                            "X68030/Xellent30 のハイメモリのデータを復元します" :
                            "X68030/Xellent30 high memory data is restored");
        System.arraycopy (highMemoryArray, 0, busHighMemoryArray, 0, Math.min (highMemoryArray.length, busHighMemorySize));
      } else {
        System.out.println (Multilingual.mlnJapanese ?
                            "X68030/Xellent30 のハイメモリをゼロクリアします" :
                            "X68030/Xellent30 high memory is zero-cleared");
      }
      if (highMemoryArray.length < busHighMemorySize) {
        Arrays.fill (busHighMemoryArray, highMemoryArray.length, busHighMemorySize, (byte) 0);  //復元されなかった部分をゼロクリアする
      }
    }

    //060turboのローカルメモリ
    int localMemorySizeMB = Settings.sgsGetInt ("localmemory");  //060turboのローカルメモリのサイズ
    busLocalMemorySize = (localMemorySizeMB == 16 ||
                          localMemorySizeMB == 32 ||
                          localMemorySizeMB == 64 ||
                          localMemorySizeMB == 128 ||
                          localMemorySizeMB == 256 ? localMemorySizeMB << 20 : 128 << 20);
    if (busLocalMemorySize == 0) {
        System.out.println (Multilingual.mlnJapanese ?
                            "060turbo のローカルメモリはありません" :
                            "060turbo local memory does not exists");
    } else {
      System.out.printf (Multilingual.mlnJapanese ?
                         "060turbo のローカルメモリのサイズは %dMB です\n" :
                         "060turbo local memory size is %dMB\n",
                         busLocalMemorySize >> 20);
    }
    busLocalMemoryArray = new byte[busLocalMemorySize];

    busLocalMemorySaveOn = Settings.sgsGetOnOff ("localmemorysave");  //060turboのローカルメモリの内容を保存するか

    byte[] localMemoryArray = Settings.sgsGetData ("localmemorydata");  //060turboのローカルメモリの内容(gzip+base64)
    if (busLocalMemorySize != 0) {  //060turboのローカルメモリがある
      if (localMemoryArray.length != 0) {  //復元するデータがある
        System.out.println (Multilingual.mlnJapanese ?
                            "060turbo のローカルメモリのデータを復元します" :
                            "060turbo local memory data is restored");
        System.arraycopy (localMemoryArray, 0, busLocalMemoryArray, 0, Math.min (localMemoryArray.length, busLocalMemorySize));
      } else {
        System.out.println (Multilingual.mlnJapanese ?
                            "060turbo のローカルメモリをゼロクリアします" :
                            "060turbo local memory is zero-cleared");
      }
      if (localMemoryArray.length < busLocalMemorySize) {
        Arrays.fill (busLocalMemoryArray, localMemoryArray.length, busLocalMemorySize, (byte) 0);  //復元されなかった部分をゼロクリアする
      }
    }

    //現在の拡張メモリ
    busExMemoryStart = busRequestExMemoryStart = 0x10000000;
    busExMemorySize = busRequestExMemorySize = 0 << 20;
    busExMemoryArray = busRequestExMemoryArray = BUS_DUMMY_MEMORY_ARRAY;

    //FC2ピンをカットする
    busRequestCutFC2Pin = Settings.sgsGetOnOff ("cutfc2pin");  //FC2ピンをカットする(on/off)
    busCutFC2Pin = !busRequestCutFC2Pin;

    busUpdateMemoryMap ();

  }  //busInit()

  public static void busUpdateMemoryMap () {
    if (busExMemoryStart == busRequestExMemoryStart &&
        busExMemorySize == busRequestExMemorySize &&
        busExMemoryArray == busRequestExMemoryArray &&
        busExMemoryArray.length == busExMemorySize &&
        busCutFC2Pin == busRequestCutFC2Pin) {
      return;
    }
    //拡張メモリ
    busExMemoryStart = busRequestExMemoryStart;
    busExMemorySize = busRequestExMemorySize;
    busExMemoryArray = busRequestExMemoryArray;
    if (busExMemoryArray.length != busExMemorySize) {
      byte[] newArray = new byte[busExMemorySize];
      int copySize = Math.min (busExMemoryArray.length, busExMemorySize);
      if (copySize > 0) {
        System.arraycopy (busExMemoryArray, 0, newArray, 0, copySize);
      }
      if (busExMemoryArray == busHighMemoryArray) {
        busHighMemoryArray = newArray;
      } else if (busExMemoryArray == busLocalMemoryArray) {
        busLocalMemoryArray = newArray;
      }
      busExMemoryArray = newArray;
    }
    //FC2ピンをカットする
    busCutFC2Pin = busRequestCutFC2Pin;
    //メモリマップを作る
    //  すべてのページにデバイスを割り当てること

    //  MMR  メインメモリ
    //  $00000000～$000FFFFF  1MB
    //  $00000000～$001FFFFF  2MB
    //  $00000000～$003FFFFF  4MB
    //  $00000000～$005FFFFF  6MB
    //  $00000000～$007FFFFF  8MB
    //  $00000000～$009FFFFF  10MB
    //  $00000000～$00BFFFFF  12MB
    //  0x00000000  メインメモリ
    busSuper (MemoryMappedDevice.MMD_MMR, 0x00000000, 0x00002000);  //MMR メインメモリ
    busUser ( MemoryMappedDevice.MMD_MMR, 0x00002000, 0x00c00000);  //MMR メインメモリ

    //  GE0  グラフィック画面(512ドット16色ページ0)
    //  $00C00000～$00C7FFFF  512KB
    //  GE1  グラフィック画面(512ドット16色ページ1)
    //  $00C80000～$00CFFFFF  512KB
    //  GE2  グラフィック画面(512ドット16色ページ2)
    //  $00D00000～$00D7FFFF  512KB
    //  GE3  グラフィック画面(512ドット16色ページ3)
    //  $00D80000～$00DFFFFF  512KB
    //  GF0  グラフィック画面(512ドット256色ページ0)
    //  $00C00000～$00C7FFFF  512KB
    //  GF1  グラフィック画面(512ドット256色ページ1)
    //  $00C80000～$00CFFFFF  512KB
    //  GM2  グラフィック画面(メモリモード2)
    //  $00C00000～$00DFFFFF  2MB
    //  GG0  グラフィック画面(512ドット65536色)
    //  $00C00000～$00C7FFFF  512KB
    //  GH0  グラフィック画面(1024ドット16色)
    //  $00C00000～$00DFFFFF  2MB
    //  GI0  グラフィック画面(1024ドット256色)
    //  $00C00000～$00DFFFFF  2MB
    //  GJ0  グラフィック画面(1024ドット65536色)
    //  $00C00000～$00DFFFFF  2MB
    //  0x00c00000  グラフィックVRAM
    busSuper (MemoryMappedDevice.MMD_GE0, 0x00c00000, 0x00c80000);  //GE0 グラフィック画面(512ドット16色ページ0)
    busSuper (MemoryMappedDevice.MMD_GE1, 0x00c80000, 0x00d00000);  //GE1 グラフィック画面(512ドット16色ページ1)
    busSuper (MemoryMappedDevice.MMD_GE2, 0x00d00000, 0x00d80000);  //GE2 グラフィック画面(512ドット16色ページ2)
    busSuper (MemoryMappedDevice.MMD_GE3, 0x00d80000, 0x00e00000);  //GE3 グラフィック画面(512ドット16色ページ3)

    //  TXT  テキスト画面
    //  $00E00000～$00E7FFFF  512KB
    //  0x00e00000  テキストVRAM
    busSuper (MemoryMappedDevice.MMD_TXT, 0x00e00000, 0x00e80000);  //TXT テキスト画面

    //  CRT  CRTコントローラ
    //  $00E80000～$00E81FFF  8KB
    //  0x00e80000  CRTコントローラ
    busSuper (MemoryMappedDevice.MMD_CRT, 0x00e80000, 0x00e82000);  //CRT CRTコントローラ

    //  VCN  ビデオコントローラ
    //  $00E82000～$00E83FFF  8KB
    //  0x00e82000  パレットレジスタ
    //  0x00e82400  ビデオコントローラ
    busSuper (MemoryMappedDevice.MMD_VCN, 0x00e82000, 0x00e84000);  //VCN ビデオコントローラ

    //  DMA  DMAコントローラ
    //  $00E84000～$00E85FFF  8KB
    //  0x00e84000  DMAコントローラ
    busSuper (MemoryMappedDevice.MMD_DMA, 0x00e84000, 0x00e86000);  //DMA DMAコントローラ

    //  SVS  スーパーバイザ領域設定
    //  $00E86000～$00E87FFF  8KB
    //  0x00e86000  スーパーバイザ領域設定
    busSuper (MemoryMappedDevice.MMD_SVS, 0x00e86000, 0x00e88000);  //SVS スーパーバイザ領域設定

    //  MFP  マルチファンクションペリフェラル
    //  $00E88000～$00E89FFF  8KB
    //  0x00e88000  MFP
    busSuper (MemoryMappedDevice.MMD_MFP, 0x00e88000, 0x00e8a000);  //MFP MFP

    //  RTC  リアルタイムクロック
    //  $00E8A000～$00E8BFFF  8KB
    //  0x00e8a000  RTC
    busSuper (MemoryMappedDevice.MMD_RTC_FIRST, 0x00e8a000, 0x00e8c000);  //RTC RTC

    //  PRT  プリンタポート
    //  $00E8C000～$00E8DFFF  8KB
    //  0x00e8c000  プリンタポート
    busSuper (MemoryMappedDevice.MMD_PRN, 0x00e8c000, 0x00e8e000);  //PRN プリンタポート

    //  SYP  システムポート
    //  $00E8E000～$00E8FFFF  8KB
    //  0x00e8e000  システムポート
    busSuper (MemoryMappedDevice.MMD_SYS, 0x00e8e000, 0x00e90000);  //SYS システムポート

    //  OPM  FM音源
    //  $00E90000～$00E91FFF  8KB
    //  0x00e90000  FM音源
    busSuper (MemoryMappedDevice.MMD_OPM, 0x00e90000, 0x00e92000);  //OPM FM音源

    //  PCM  ADPCM音源
    //  $00E92000～$00E93FFF  8KB
    //  0x00e92000  ADPCM音源
    busSuper (MemoryMappedDevice.MMD_PCM, 0x00e92000, 0x00e94000);  //PCM ADPCM音源

    //  FDC  FDコントローラ
    //  $00E94000～$00E95FFF  8KB
    //  0x00e94000  FDC
    busSuper (MemoryMappedDevice.MMD_FDC, 0x00e94000, 0x00e96000);  //FDC FDコントローラ

    //  HDC  SASI HDコントローラ
    //  $00E96000～$00E97FFF  8KB
    //  0x00e96000  HDC SASI HDコントローラ
    //  0x00e96020  SPC 内蔵SCSIプロトコルコントローラ
    busSuper (MemoryMappedDevice.MMD_HDC, 0x00e96000, 0x00e98000);  //HDC SASI HDコントローラ

    //  SCC  シリアルコミュニケーションコントローラ
    //  $00E98000～$00E99FFF  8KB
    //  0x00e98000  SCC
    busSuper (MemoryMappedDevice.MMD_SCC, 0x00e98000, 0x00e9a000);  //SCC SCC

    //  PPI  プログラマブルペリフェラルインタフェイス
    //  $00E9A000～$00E9BFFF  8KB
    //  0x00e9a000  PPI
    busSuper (MemoryMappedDevice.MMD_PPI, 0x00e9a000, 0x00e9c000);  //PPI PPI

    //  IOI  I/O割り込み
    //  $00E9C000～$00E9DFFF  8KB
    //  0x00e9c000  I/O割り込み
    busSuper (MemoryMappedDevice.MMD_IOI, 0x00e9c000, 0x00e9e000);  //IOI I/O割り込み

    //  XB1  拡張ボード領域1
    //  $00E9E000～$00E9FFFF  8KB
    //  0x00e9e000  数値演算プロセッサボード(CZ-6BP1/CZ-6BP1A)
    //  0x00e9e200  ツクモグラフィックアクセラレータPCMボード(TS-6BGA)
    //  0x00e9f000  WINDRV
    //  0x00e9f000  040Excel
    //  0x00e9f020  HFS ホストファイルシステムインタフェイス
    busSuper (MemoryMappedDevice.MMD_XB1, 0x00e9e000, 0x00ea0000);  //XB1 拡張ボード領域1

    //  XSC  拡張SCSI
    //  $00EA0000～$00EA1FFF  8KB
    //  0x00ea0000  拡張SCSI(SCSIボードCZ-6BS1/Mach-2)
    //  0x00ea1ff0  TS-6BS1
    busSuper (MemoryMappedDevice.MMD_NUL, 0x00ea0000, 0x00eae000);  //MemoryMappedDevice.MMD_EXSはSPC.spcReset()で必要なときだけ接続する

    //  XB2  拡張ボード領域2
    //  $00EA2000～$00EAFFFF  56KB
    //  0x00eaf900  FAXボード(CZ-6BC1)
    //  0x00eafa00  MIDIボード(CZ-6BM1)
    //  0x00eafb00  パラレルボード(CZ-6BN1)
    //  0x00eafc00  RS-232Cボード(CZ-6BF1)
    //  0x00eafd00  ユニバーサルI/Oボード(CZ-6BU1)
    //  0x00eafe00  GP-IBボード(CZ-6BG1)
    //  0x00eaff00  スーパーバイザエリア設定
    //  拡張
    //  $00EAFF7F  バンクメモリのページ番号
    busSuper (MemoryMappedDevice.MMD_XB2, 0x00eae000, 0x00eb0000);  //拡張ボード領域2

    //  SPR  スプライト画面
    //  $00EB0000～$00EBFFFF  64KB
    //  0x00eb0000  スプライトレジスタ
    //  0x00eb0800  スプライトコントローラ
    //  0x00eb8000  スプライトPCGエリア
    //  0x00ebc000  スプライトテキストエリア0
    //  0x00ebe000  スプライトテキストエリア1
    busSuper (MemoryMappedDevice.MMD_SPR, 0x00eb0000, 0x00ec0000);  //SPR スプライト画面

    //  XB3  拡張ボード領域3
    //  $00EC0000～$00ECFFFF  64KB
    //  0x00ec0000  ユーザI/Oエリア
    //  0x00ec0000  Awesome
    //  0x00ec0000  Xellent30
    //  0x00ecc000  Mercury
    //  0x00ece000  Neptune
    //  0x00ecf000  Venus-X/030
    busSuper (MemoryMappedDevice.MMD_XB3, 0x00ec0000, 0x00ed0000);  //拡張ボード領域3

    //  SMR  SRAM
    //  $00ED0000～$00ED3FFF  16KB
    //  $00ED0000～$00ED7FFF  32KB。改造したとき
    //  $00ED0000～$00EDFFFF  64KB。改造したとき
    //  0x00ed0000  SRAM
    busSuper (MemoryMappedDevice.MMD_SMR, 0x00ed0000, 0x00ed0000 + 16384);  //SMR SRAM
    busSuper (MemoryMappedDevice.MMD_NUL, 0x00ed0000 + 16384, 0x00ed0000 + 65536);  //空き

    //  XB4  拡張ボード領域4
    //  $00EE0000～$00EFFFFF  128KB
    //  0x00ee0000  GAフレームバッファウインドウ(サブ)
    //  0x00ef0000  GAフレームバッファウインドウ(メイン)
    //  0x00efff00  PSX16550
    //  拡張
    //  $00EE0000～$00EFFFFF  バンクメモリ
    busSuper (MemoryMappedDevice.MMD_XB4, 0x00ee0000, 0x00f00000);  //拡張ボード領域4

    //  CG1  CGROM1
    //  $00F00000～$00F3FFFF  256KB。CGROM作成前
    //  0x00f00000  KNJ16x16フォント(1～8区,非漢字752文字)
    //  0x00f05e00  KNJ16x16フォント(16～47区,第1水準漢字3008文字)
    //  0x00f1d600  KNJ16x16フォント(48～84区,第2水準漢字3478文字)
    //  0x00f3a000  ANK8x8フォント(256文字)
    //  0x00f3a800  ANK8x16フォント(256文字)
    //  0x00f3b800  ANK12x12フォント(256文字)
    //  0x00f3d000  ANK12x24フォント(256文字)
    busSuper (MemoryMappedDevice.MMD_CG1, 0x00f00000, 0x00f40000);  //CG1 CGROM1

    //  CG2  CGROM2
    //  $00F40000～$00FBFFFF  512KB。CGROM作成前
    //  0x00f40000  KNJ24x24フォント(1～8区,非漢字752文字)
    //  0x00f4d380  KNJ24x24フォント(16～47区,第1水準漢字3008文字)
    //  0x00f82180  KNJ24x24フォント(48～84区,第2水準漢字3478文字)
    //  0x00fbf400  [13]ANK6x12フォント(256文字)
    busSuper (MemoryMappedDevice.MMD_CG2, 0x00f40000, 0x00fc0000);  //CG2 CGROM2

    //  ROM  ROM
    //  $00FC0000～$00FFFFFF  256KB
    //  0x00fc0000  [11,12]内蔵SCSI BIOS,[13]内蔵SCSIハンドル
    //  0x00fc0200  [13]ROM Human
    //  0x00fce000  [13]ROM Float
    //  0x00fd3800  [13]ROMデバッガ
    //  0x00fe0000  [10,11,12]ROMデバッガ
    //  0x00fe5000  [10,11,12]ROM Human
    //  0x00ff0000  IPLROM
    //  0x00ffd018  [10]ANK6x12フォント(254文字)
    //  0x00ffd344  [11]ANK6x12フォント(254文字)
    //  0x00ffd45e  [12]ANK6x12フォント(254文字)
    //  0x00ffdc00  [10]ROMディスク
    busSuper (MemoryMappedDevice.MMD_ROM, 0x00fc0000, 0x01000000);  //ROM ROM

    //  HMR  ハイメモリ
    //  $01000000～$01FFFFFF  16MB

    //  LMR  ローカルメモリ
    //  $10000000～$10FFFFFF  16MB
    //  $10000000～$11FFFFFF  32MB
    //  $10000000～$13FFFFFF  64MB
    //  $10000000～$17FFFFFF  128MB
    //  $10000000～$1FFFFFFF  256MB

  }  //busUpdateMemoryMap()

  public static void busReset () {
    if (regSRS != 0) {  //スーパーバイザモード
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap = DataBreakPoint.dbpSuperMap;
      } else {
        busMemoryMap = busSuperMap;
      }
    } else {  //ユーザモード
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap = DataBreakPoint.dbpUserMap;
      } else {
        busMemoryMap = busUserMap;
      }
    }
    //スーパーバイザ領域設定
    busSuper (MemoryMappedDevice.MMD_MMR, 0x00000000, 0x00002000);
    busUser ( MemoryMappedDevice.MMD_MMR, 0x00002000, MainMemory.mmrMemorySizeCurrent);
    if (MainMemory.mmrMemorySizeCurrent < 0x00c00000) {
      busSuper (MemoryMappedDevice.MMD_NUL, MainMemory.mmrMemorySizeCurrent, 0x00c00000);
    }
  }  //busReset()

  //busUser (mmd, motherStartAddress, motherEndAddress)
  //  ユーザ領域にデバイスを割り当てる
  //  motherStartAddress  マザーボード開始アドレス。BUS_PAGE_SIZEで割り切れること
  //  motherEndAddress    マザーボード終了アドレス。BUS_PAGE_SIZEで割り切れること
  public static void busUser (MemoryMappedDevice mmd, int motherStartAddress, int motherEndAddress) {
    int motherStartPage = motherStartAddress >>> BUS_PAGE_BITS;  //マザーボード開始ページ
    int motherEndPage = motherEndAddress >>> BUS_PAGE_BITS;  //マザーボード終了ページ
    if (false &&
        (motherStartPage << BUS_PAGE_BITS != motherStartAddress ||
         motherEndPage << BUS_PAGE_BITS != motherEndAddress)) {  //開始アドレスまたは終了アドレスがページサイズで割り切れない
      System.out.printf ("ERROR: busUser (\"%s\", 0x%08x, 0x%08x)\n", mmd.toString (), motherStartAddress, motherEndAddress);
    }
    int exMemoryStartPage = busExMemoryStart >>> BUS_PAGE_BITS;  //拡張メモリ開始ページ
    int exMemoryEndPage = exMemoryStartPage + (busExMemorySize >>> BUS_PAGE_BITS);  //拡張メモリ終了ページ
    for (int block = 0; block < 1 << 32 - BUS_MOTHER_BITS; block++) {  //ブロック
      int blockStartPage = block << BUS_MOTHER_BITS - BUS_PAGE_BITS;  //ブロック開始ページ
      int startPage = blockStartPage + motherStartPage;  //デバイス開始ページ
      int endPage = blockStartPage + motherEndPage;  //デバイス終了ページ
      for (int page = startPage; page < endPage; page++) {  //デバイスページ
        MemoryMappedDevice superMmd = exMemoryStartPage <= page && page < exMemoryEndPage ? MemoryMappedDevice.MMD_XMM : mmd;
        busUserMap[page] = busSuperMap[page] = superMmd;
        if (InstructionBreakPoint.IBP_ON) {
          if (InstructionBreakPoint.ibpUserMap[page] != MemoryMappedDevice.MMD_IBP) {  //命令ブレークポイントがない
            InstructionBreakPoint.ibpUserMap[page] = superMmd;
          }
          if (InstructionBreakPoint.ibpSuperMap[page] != MemoryMappedDevice.MMD_IBP) {  //命令ブレークポイントがない
            InstructionBreakPoint.ibpSuperMap[page] = superMmd;
          }
        }
        if (DataBreakPoint.DBP_ON) {
          if (DataBreakPoint.dbpUserMap[page] != MemoryMappedDevice.MMD_DBP) {  //データブレークポイントがない
            DataBreakPoint.dbpUserMap[page] = superMmd;
          }
          if (DataBreakPoint.dbpSuperMap[page] != MemoryMappedDevice.MMD_DBP) {  //データブレークポイントがない
            DataBreakPoint.dbpSuperMap[page] = superMmd;
          }
        }
      }
    }
  }  //busUser(MMD,int,int)

  //busSuper (mmd, motherStartAddress, motherEndAddress)
  //  スーパーバイザ領域にデバイスを割り当てる
  //  motherStartAddress  マザーボード開始アドレス。BUS_PAGE_SIZEで割り切れること
  //  motherEndAddress    マザーボード終了アドレス。BUS_PAGE_SIZEで割り切れること
  public static void busSuper (MemoryMappedDevice mmd, int motherStartAddress, int motherEndAddress) {
    int motherStartPage = motherStartAddress >>> BUS_PAGE_BITS;  //マザーボード開始ページ
    int motherEndPage = motherEndAddress >>> BUS_PAGE_BITS;  //マザーボード終了ページ
    if (false &&
        (motherStartPage << BUS_PAGE_BITS != motherStartAddress ||
         motherEndPage << BUS_PAGE_BITS != motherEndAddress)) {  //開始アドレスまたは終了アドレスがページサイズで割り切れない
      System.out.printf ("ERROR: busSuper (\"%s\", 0x%08x, 0x%08x)\n", mmd.toString (), motherStartAddress, motherEndAddress);
    }
    int exMemoryStartPage = busExMemoryStart >>> BUS_PAGE_BITS;  //拡張メモリ開始ページ
    int exMemoryEndPage = exMemoryStartPage + (busExMemorySize >>> BUS_PAGE_BITS);  //拡張メモリ終了ページ
    for (int block = 0; block < 1 << 32 - BUS_MOTHER_BITS; block++) {  //ブロック
      int blockStartPage = block << BUS_MOTHER_BITS - BUS_PAGE_BITS;  //ブロック開始ページ
      int startPage = blockStartPage + motherStartPage;  //デバイス開始ページ
      int endPage = blockStartPage + motherEndPage;  //デバイス終了ページ
      for (int page = startPage; page < endPage; page++) {  //デバイスページ
        boolean isExMemory = exMemoryStartPage <= page && page < exMemoryEndPage;
        MemoryMappedDevice userMmd = isExMemory ? MemoryMappedDevice.MMD_XMM : busCutFC2Pin ? mmd : MemoryMappedDevice.MMD_NUL;
        MemoryMappedDevice superMmd = isExMemory ? MemoryMappedDevice.MMD_XMM : mmd;
        busUserMap[page] = userMmd;
        busSuperMap[page] = superMmd;
        if (InstructionBreakPoint.IBP_ON) {
          if (InstructionBreakPoint.ibpUserMap[page] != MemoryMappedDevice.MMD_IBP) {  //命令ブレークポイントがない
            InstructionBreakPoint.ibpUserMap[page] = userMmd;
          }
          if (InstructionBreakPoint.ibpSuperMap[page] != MemoryMappedDevice.MMD_IBP) {  //命令ブレークポイントがない
            InstructionBreakPoint.ibpSuperMap[page] = superMmd;
          }
        }
        if (DataBreakPoint.DBP_ON) {
          if (DataBreakPoint.dbpUserMap[page] != MemoryMappedDevice.MMD_DBP) {  //データブレークポイントがない
            DataBreakPoint.dbpUserMap[page] = userMmd;
          }
          if (DataBreakPoint.dbpSuperMap[page] != MemoryMappedDevice.MMD_DBP) {  //データブレークポイントがない
            DataBreakPoint.dbpSuperMap[page] = superMmd;
          }
        }
      }
    }
  }  //busSuper(MMD,int,int)

  //d = busPbs (a)
  //  ピークバイト符号拡張
  public static byte busPbs (int a) {
    return busSuperMap[a >>> BUS_PAGE_BITS].mmdPbs (a);
  }  //busPbs(int)

  //d = busPbz (a)
  //  ピークバイトゼロ拡張
  public static int busPbz (int a) {
    return busSuperMap[a >>> BUS_PAGE_BITS].mmdPbz (a);
  }  //busPbz(int)

  //d = busPws (a)
  //  ピークワード符号拡張
  public static int busPws (int a) {
    if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //偶数
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPws (a);
    } else {  //奇数
      int a1 = a + 1;
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPbs (a) << 8 | busSuperMap[a1 >>> BUS_PAGE_BITS].mmdPbz (a1);
    }
  }  //busPws(int)

  //d = busPwse (a)
  //  ピークワード符号拡張(偶数)
  public static int busPwse (int a) {
    return busSuperMap[a >>> BUS_PAGE_BITS].mmdPws (a);
  }  //busPwse(int)

  //d = busPwz (a)
  //  ピークワードゼロ拡張
  public static int busPwz (int a) {
    if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //偶数
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPwz (a);
    } else {  //奇数
      int a1 = a + 1;
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPbz (a) << 8 | busSuperMap[a1 >>> BUS_PAGE_BITS].mmdPbz (a1);
    }
  }  //busPwz(int)

  //d = busPwze (a)
  //  ピークワードゼロ拡張(偶数)
  public static int busPwze (int a) {
    return busSuperMap[a >>> BUS_PAGE_BITS].mmdPwz (a);
  }  //busPwze(int)

  //d = busPls (a)
  //  ピークロング符号拡張
  public static int busPls (int a) {
    if (TEST_BIT_0_SHIFT && TEST_BIT_1_SHIFT ? a << 30 == 0 : (a & 3) == 0) {  //4の倍数
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPls (a);
    } else if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //4の倍数ではない偶数
      int a2 = a + 2;
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPws (a) << 16 | busSuperMap[a2 >>> BUS_PAGE_BITS].mmdPwz (a2);
    } else {  //奇数
      int a1 = a + 1;
      int a3 = a + 3;
      return busSuperMap[a >>> BUS_PAGE_BITS].mmdPbs (a) << 24 | busSuperMap[a1 >>> BUS_PAGE_BITS].mmdPwz (a1) << 8 | busSuperMap[a3 >>> BUS_PAGE_BITS].mmdPbz (a3);
    }
  }  //busPls(int)

  //d = busPlsf (a)
  //  ピークロング符号拡張(4の倍数)
  public static int busPlsf (int a) {
    return busSuperMap[a >>> BUS_PAGE_BITS].mmdPls (a);
  }  //busPlsf(int)

  //d = busPqs (a)
  //  ピーククワッド符号拡張
  public static long busPqs (int a) {
    return (long) busPls (a) << 32 | busPls (a + 4) & 0xffffffffL;
  }  //busPqs(int)

  //d = busRbs (a)
  //  リードバイト符号拡張
  public static byte busRbs (int a) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRbs (a);
    } else {
      return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRbs (a);
    }
  }  //busRbs(int)

  //d = busRbz (a)
  //  リードバイトゼロ拡張
  public static int busRbz (int a) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRbz (a);
    } else {
      return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRbz (a);
    }
  }  //busRbz(int)

  //d = busRws (a)
  //  リードワード符号拡張
  public static int busRws (int a) throws M68kException {
    if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //偶数
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a);
      }
    } else if (mpuIgnoreAddressError) {  //奇数でアドレスエラーを検出しない
      int a1 = a + 1;
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRbs (a) << 8 | DataBreakPoint.dbpMemoryMap[a1 >>> BUS_PAGE_BITS].mmdRbz (a1);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRbs (a) << 8 | busMemoryMap[a1 >>> BUS_PAGE_BITS].mmdRbz (a1);
      }
    } else {  //奇数でアドレスエラーを検出する
      M68kException.m6eNumber = M68kException.M6E_ADDRESS_ERROR;
      M68kException.m6eAddress = a;
      M68kException.m6eDirection = MPU_WR_READ;
      M68kException.m6eSize = MPU_SS_WORD;
      throw M68kException.m6eSignal;
    }
  }  //busRws(int)

  //d = busRwse (a)
  //  リードワード符号拡張(偶数限定)
  //  MOVEM命令の2ワード目以降など、アドレスが偶数であることが分かっている場合
  public static int busRwse (int a) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a);
    } else {
      return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a);
    }
  }  //busRwse(int)

  //d = busRwz (a)
  //  リードワードゼロ拡張
  public static int busRwz (int a) throws M68kException {
    if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //偶数
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRwz (a);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRwz (a);
      }
    } else if (mpuIgnoreAddressError) {  //奇数でアドレスエラーを検出しない
      int a1 = a + 1;
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRbz (a) << 8 | DataBreakPoint.dbpMemoryMap[a1 >>> BUS_PAGE_BITS].mmdRbz (a1);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRbz (a) << 8 | busMemoryMap[a1 >>> BUS_PAGE_BITS].mmdRbz (a1);
      }
    } else {  //奇数でアドレスエラーを検出する
      M68kException.m6eNumber = M68kException.M6E_ADDRESS_ERROR;
      M68kException.m6eAddress = a;
      M68kException.m6eDirection = MPU_WR_READ;
      M68kException.m6eSize = MPU_SS_WORD;
      throw M68kException.m6eSignal;
    }
  }  //busRwz(int)

  //d = busRwze (a)
  //  リードワードゼロ拡張(偶数限定)
  //  MOVEM命令の2ワード目以降など、アドレスが偶数であることが分かっている場合
  public static int busRwze (int a) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRwz (a);
    } else {
      return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRwz (a);
    }
  }  //busRwze(int)

  //d = busRls (a)
  //  リードロング符号拡張
  public static int busRls (int a) throws M68kException {
    if (TEST_BIT_0_SHIFT && TEST_BIT_1_SHIFT ? a << 30 == 0 : (a & 3) == 0) {  //4の倍数
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRls (a);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRls (a);
      }
    } else if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //4の倍数ではない偶数
      int a2 = a + 2;
      if (BUS_SPLIT_UNALIGNED_LONG) {  //常に分割する
        if (DataBreakPoint.DBP_ON) {
          return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a) << 16 | DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS].mmdRwz (a2);
        } else {
          return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a) << 16 | busMemoryMap[a2 >>> BUS_PAGE_BITS].mmdRwz (a2);
        }
      } else {  //デバイスを跨がないとき分割しない
        MemoryMappedDevice mmd;
        MemoryMappedDevice mmd2;
        if (DataBreakPoint.DBP_ON) {
          mmd = DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS];
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = busMemoryMap[a2 >>> BUS_PAGE_BITS];
        }
        return mmd == mmd2 ? mmd.mmdRls (a) : mmd.mmdRws (a) << 16 | mmd2.mmdRwz (a2);  //デバイスを跨がない/デバイスを跨ぐ
      }
    } else if (mpuIgnoreAddressError) {  //奇数でアドレスエラーを検出しない
      int a1 = a + 1;
      int a3 = a + 3;
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRbs (a) << 24 | DataBreakPoint.dbpMemoryMap[a1 >>> BUS_PAGE_BITS].mmdRwz (a1) << 8 | DataBreakPoint.dbpMemoryMap[a3 >>> BUS_PAGE_BITS].mmdRbz (a3);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRbs (a) << 24 | busMemoryMap[a1 >>> BUS_PAGE_BITS].mmdRwz (a1) << 8 | busMemoryMap[a3 >>> BUS_PAGE_BITS].mmdRbz (a3);
      }
    } else {  //奇数でアドレスエラーを検出する
      M68kException.m6eNumber = M68kException.M6E_ADDRESS_ERROR;
      M68kException.m6eAddress = a;
      M68kException.m6eDirection = MPU_WR_READ;
      M68kException.m6eSize = MPU_SS_LONG;
      throw M68kException.m6eSignal;
    }
  }  //busRls(int)

  //d = busRlse (a)
  //  リードロング符号拡張(偶数限定)
  //  MOVEM命令の2ワード目以降など、アドレスが偶数であることが分かっている場合
  public static int busRlse (int a) throws M68kException {
    if (TEST_BIT_0_SHIFT && TEST_BIT_1_SHIFT ? a << 30 == 0 : (a & 3) == 0) {  //4の倍数
      if (DataBreakPoint.DBP_ON) {
        return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRls (a);
      } else {
        return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRls (a);
      }
    } else {  //4の倍数ではない偶数
      int a2 = a + 2;
      if (BUS_SPLIT_UNALIGNED_LONG) {  //常に分割する
        if (DataBreakPoint.DBP_ON) {
          return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a) << 16 | DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS].mmdRwz (a2);
        } else {
          return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRws (a) << 16 | busMemoryMap[a2 >>> BUS_PAGE_BITS].mmdRwz (a2);
        }
      } else {  //デバイスを跨がないとき分割しない
        MemoryMappedDevice mmd;
        MemoryMappedDevice mmd2;
        if (DataBreakPoint.DBP_ON) {
          mmd = DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS];
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = busMemoryMap[a2 >>> BUS_PAGE_BITS];
        }
        return mmd == mmd2 ? mmd.mmdRls (a) : mmd.mmdRws (a) << 16 | mmd2.mmdRwz (a2);  //デバイスを跨がない/デバイスを跨ぐ
      }
    }
  }  //busRlse(int)

  //d = busRlsf (a)
  //  リードロング符号拡張(4の倍数限定)
  //  例外ベクタテーブルなど、アドレスが4の倍数であることが分かっている場合
  public static int busRlsf (int a) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      return DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdRls (a);
    } else {
      return busMemoryMap[a >>> BUS_PAGE_BITS].mmdRls (a);
    }
  }  //busRlsf(int)

  //d = busRqs (a)
  //  リードクワッド符号拡張
  public static long busRqs (int a) throws M68kException {
    return (long) busRls (a) << 32 | busRls (a + 4) & 0xffffffffL;
  }  //busRqs(int)

  //busWb (a, d)
  //  ライトバイト
  public static void busWb (int a, int d) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d);
    } else {
      busMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d);
    }
  }  //busWb(int,int)

  //busWw (a, d)
  //  ライトワード
  public static void busWw (int a, int d) throws M68kException {
    if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //偶数
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d);
      } else {
        busMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d);
      }
    } else if (mpuIgnoreAddressError) {  //奇数でアドレスエラーを検出しない
      int a1 = a + 1;
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d >> 8);
        DataBreakPoint.dbpMemoryMap[a1 >>> BUS_PAGE_BITS].mmdWb (a1, d);
      } else {
        busMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d >> 8);
        busMemoryMap[a1 >>> BUS_PAGE_BITS].mmdWb (a1, d);
      }
    } else {  //奇数でアドレスエラーを検出する
      M68kException.m6eNumber = M68kException.M6E_ADDRESS_ERROR;
      M68kException.m6eAddress = a;
      M68kException.m6eDirection = MPU_WR_WRITE;
      M68kException.m6eSize = MPU_SS_WORD;
      throw M68kException.m6eSignal;
    }
  }  //busWw(int,int)

  //busWwe (a, d)
  //  ライトワード(偶数限定)
  //  MOVEM命令の2ワード目以降など、アドレスが偶数であることが分かっている場合
  public static void busWwe (int a, int d) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d);
    } else {
      busMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d);
    }
  }  //busWwe(int,int)

  //busWl (a, d)
  //  ライトロング
  public static void busWl (int a, int d) throws M68kException {
    if (TEST_BIT_0_SHIFT && TEST_BIT_1_SHIFT ? a << 30 == 0 : (a & 3) == 0) {  //4の倍数
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
      } else {
        busMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
      }
    } else if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //4の倍数ではない偶数
      int a2 = a + 2;
      if (BUS_SPLIT_UNALIGNED_LONG) {  //常に分割する
        if (DataBreakPoint.DBP_ON) {
          DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d >> 16);
          DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS].mmdWw (a2, d);
        } else {
          busMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d >> 16);
          busMemoryMap[a2 >>> BUS_PAGE_BITS].mmdWw (a2, d);
        }
      } else {  //デバイスを跨がないとき分割しない
        MemoryMappedDevice mmd;
        MemoryMappedDevice mmd2;
        if (DataBreakPoint.DBP_ON) {
          mmd = DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS];
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = busMemoryMap[a2 >>> BUS_PAGE_BITS];
        }
        if (mmd == mmd2) {  //デバイスを跨がない
          mmd.mmdWl (a, d);
        } else {  //デバイスを跨ぐ
          mmd.mmdWw (a, d >> 16);
          mmd2.mmdWw (a2, d);
        }
      }
    } else if (mpuIgnoreAddressError) {  //奇数でアドレスエラーを検出しない
      int a1 = a + 1;
      int a3 = a + 3;
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d >> 24);
        DataBreakPoint.dbpMemoryMap[a1 >>> BUS_PAGE_BITS].mmdWw (a1, d >> 8);
        DataBreakPoint.dbpMemoryMap[a3 >>> BUS_PAGE_BITS].mmdWb (a3, d);
      } else {
        busMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d >> 24);
        busMemoryMap[a1 >>> BUS_PAGE_BITS].mmdWw (a1, d >> 8);
        busMemoryMap[a3 >>> BUS_PAGE_BITS].mmdWb (a3, d);
      }
    } else {  //奇数でアドレスエラーを検出する
      M68kException.m6eNumber = M68kException.M6E_ADDRESS_ERROR;
      M68kException.m6eAddress = a;
      M68kException.m6eDirection = MPU_WR_WRITE;
      M68kException.m6eSize = MPU_SS_LONG;
      throw M68kException.m6eSignal;
    }
  }  //busWl(int,int)

  //busWlf (a, d)
  //  ライトロング(4の倍数限定)
  //  例外ベクタテーブルなど、アドレスが4の倍数であることが分かっている場合
  public static void busWlf (int a, int d) throws M68kException {
    if (DataBreakPoint.DBP_ON) {
      DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
    } else {
      busMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
    }
  }  //busWlf(int,int)

  //busWle (a, d)
  //  ライトロング(偶数限定)
  //  MOVEM命令の2ワード目以降など、アドレスが偶数であることが分かっている場合
  public static void busWle (int a, int d) throws M68kException {
    if (TEST_BIT_0_SHIFT && TEST_BIT_1_SHIFT ? a << 30 == 0 : (a & 3) == 0) {  //4の倍数
      if (DataBreakPoint.DBP_ON) {
        DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
      } else {
        busMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
      }
    } else {  //4の倍数ではない偶数
      int a2 = a + 2;
      if (BUS_SPLIT_UNALIGNED_LONG) {  //常に分割する
        if (DataBreakPoint.DBP_ON) {
          DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d >> 16);
          DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS].mmdWw (a2, d);
        } else {
          busMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d >> 16);
          busMemoryMap[a2 >>> BUS_PAGE_BITS].mmdWw (a2, d);
        }
      } else {  //デバイスを跨がないとき分割しない
        MemoryMappedDevice mmd;
        MemoryMappedDevice mmd2;
        if (DataBreakPoint.DBP_ON) {
          mmd = DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = DataBreakPoint.dbpMemoryMap[a2 >>> BUS_PAGE_BITS];
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = busMemoryMap[a2 >>> BUS_PAGE_BITS];
        }
        if (mmd == mmd2) {  //デバイスを跨がない
          mmd.mmdWl (a, d);
        } else {  //デバイスを跨ぐ
          mmd.mmdWw (a, d >> 16);
          mmd2.mmdWw (a2, d);
        }
      }
    }
  }  //busWle(int,int)

  //busWq (a, d)
  //  ライトクワッド
  public static void busWq (int a, long d) throws M68kException {
    busWl (a, (int) (d >>> 32));
    busWl (a + 4, (int) d);
  }  //busWq(int,long)

  //以下は拡張

  //busRbb (a, bb, o, l)
  //  リードバイトバッファ
  public static void busRbb (int a, byte[] bb, int o, int l) throws M68kException {
    if (false) {
      for (int i = 0; i < l; i++) {
        int ai = a + i;
        if (DataBreakPoint.DBP_ON) {
          bb[o + i] = DataBreakPoint.dbpMemoryMap[ai >>> BUS_PAGE_BITS].mmdRbs (ai);
        } else {
          bb[o + i] = busMemoryMap[ai >>> BUS_PAGE_BITS].mmdRbs (ai);
        }
      }
    } else {
      int r = (~a & BUS_PAGE_SIZE - 1) + 1;  //最初のページの残りの長さ。1～BUS_PAGE_SIZE
      while (l > 0) {
        MemoryMappedDevice mmd;
        if (DataBreakPoint.DBP_ON) {
          mmd = DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS];  //ページのデバイス
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];  //ページのデバイス
        }
        int s = l <= r ? l : r;  //このページで転送する長さ
        l -= s;
        if (true) {
          for (s -= 16; s >= 0; s -= 16) {
            bb[o     ] = mmd.mmdRbs (a     );
            bb[o +  1] = mmd.mmdRbs (a +  1);
            bb[o +  2] = mmd.mmdRbs (a +  2);
            bb[o +  3] = mmd.mmdRbs (a +  3);
            bb[o +  4] = mmd.mmdRbs (a +  4);
            bb[o +  5] = mmd.mmdRbs (a +  5);
            bb[o +  6] = mmd.mmdRbs (a +  6);
            bb[o +  7] = mmd.mmdRbs (a +  7);
            bb[o +  8] = mmd.mmdRbs (a +  8);
            bb[o +  9] = mmd.mmdRbs (a +  9);
            bb[o + 10] = mmd.mmdRbs (a + 10);
            bb[o + 11] = mmd.mmdRbs (a + 11);
            bb[o + 12] = mmd.mmdRbs (a + 12);
            bb[o + 13] = mmd.mmdRbs (a + 13);
            bb[o + 14] = mmd.mmdRbs (a + 14);
            bb[o + 15] = mmd.mmdRbs (a + 15);
            a += 16;
            o += 16;
          }
          s += 16;
        }
        for (int i = 0; i < s; i++) {
          bb[o + i] = mmd.mmdRbs (a + i);
        }
        a += s;
        o += s;
        r = BUS_PAGE_SIZE;
      }
    }
  }  //busRbb(int,byte[],int,int)

  //busWbb (a, bb, o, l)
  //  ライトバイトバッファ
  public static void busWbb (int a, byte[] bb, int o, int l) throws M68kException {
    if (false) {
      for (int i = 0; i < l; i++) {
        int ai = a + i;
        if (DataBreakPoint.DBP_ON) {
          DataBreakPoint.dbpMemoryMap[ai >>> BUS_PAGE_BITS].mmdWb (ai, bb[o + i]);
        } else {
          busMemoryMap[ai >>> BUS_PAGE_BITS].mmdWb (ai, bb[o + i]);
        }
      }
    } else {
      int r = (~a & BUS_PAGE_SIZE - 1) + 1;  //最初のページの残りの長さ。1～BUS_PAGE_SIZE
      while (l > 0) {
        MemoryMappedDevice mmd;
        if (DataBreakPoint.DBP_ON) {
          mmd = DataBreakPoint.dbpMemoryMap[a >>> BUS_PAGE_BITS];  //ページのデバイス
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];  //ページのデバイス
        }
        int s = l <= r ? l : r;  //このページで転送する長さ
        l -= s;
        if (true) {
          for (s -= 16; s >= 0; s -= 16) {
            mmd.mmdWb (a     , bb[o     ]);
            mmd.mmdWb (a +  1, bb[o +  1]);
            mmd.mmdWb (a +  2, bb[o +  2]);
            mmd.mmdWb (a +  3, bb[o +  3]);
            mmd.mmdWb (a +  4, bb[o +  4]);
            mmd.mmdWb (a +  5, bb[o +  5]);
            mmd.mmdWb (a +  6, bb[o +  6]);
            mmd.mmdWb (a +  7, bb[o +  7]);
            mmd.mmdWb (a +  8, bb[o +  8]);
            mmd.mmdWb (a +  9, bb[o +  9]);
            mmd.mmdWb (a + 10, bb[o + 10]);
            mmd.mmdWb (a + 11, bb[o + 11]);
            mmd.mmdWb (a + 12, bb[o + 12]);
            mmd.mmdWb (a + 13, bb[o + 13]);
            mmd.mmdWb (a + 14, bb[o + 14]);
            mmd.mmdWb (a + 15, bb[o + 15]);
            a += 16;
            o += 16;
          }
          s += 16;
        }
        for (int i = 0; i < s; i++) {
          mmd.mmdWb (a + i, bb[o + i]);
        }
        a += s;
        o += s;
        r = BUS_PAGE_SIZE;
      }
    }
  }  //busWbb(int,byte[],int,int)

  //busVb (a, d)
  //  ライトバイト(エラーなし)
  public static void busVb (int a, int d) {
    try {
      if (DataBreakPoint.DBP_ON) {
        (regSRS != 0 ? busSuperMap : busUserMap)[a >>> BUS_PAGE_BITS].mmdWb (a, d);
      } else {
        busMemoryMap[a >>> BUS_PAGE_BITS].mmdWb (a, d);
      }
    } catch (M68kException e) {
    }
  }  //busVb(int,int)

  //busVw (a, d)
  //  ライトワード(エラーなし)
  public static void busVw (int a, int d) {
    try {
      if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //偶数
        if (DataBreakPoint.DBP_ON) {
          (regSRS != 0 ? busSuperMap : busUserMap)[a >>> BUS_PAGE_BITS].mmdWw (a, d);
        } else {
          busMemoryMap[a >>> BUS_PAGE_BITS].mmdWw (a, d);
        }
      }
    } catch (M68kException e) {
    }
  }  //busVw(int,int)

  //busVl (a, d)
  //  ライトロング(エラーなし)
  public static void busVl (int a, int d) {
    try {
      if (TEST_BIT_0_SHIFT && TEST_BIT_1_SHIFT ? a << 30 == 0 : (a & 3) == 0) {  //4の倍数
        if (DataBreakPoint.DBP_ON) {
          (regSRS != 0 ? busSuperMap : busUserMap)[a >>> BUS_PAGE_BITS].mmdWl (a, d);
        } else {
          busMemoryMap[a >>> BUS_PAGE_BITS].mmdWl (a, d);
        }
      } else if (TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {  //4の倍数ではない偶数
        int a2 = a + 2;
        MemoryMappedDevice mmd;
        MemoryMappedDevice mmd2;
        if (DataBreakPoint.DBP_ON) {
          mmd = (regSRS != 0 ? busSuperMap : busUserMap)[a >>> BUS_PAGE_BITS];
          mmd2 = (regSRS != 0 ? busSuperMap : busUserMap)[a2 >>> BUS_PAGE_BITS];
        } else {
          mmd = busMemoryMap[a >>> BUS_PAGE_BITS];
          mmd2 = busMemoryMap[a2 >>> BUS_PAGE_BITS];
        }
        if (mmd == mmd2) {  //4の倍数ではない偶数でデバイスを跨がない
          mmd.mmdWl (a, d);
        } else {  //4の倍数ではない偶数でデバイスを跨ぐ
          mmd.mmdWw (a, d >> 16);
          mmd2.mmdWw (a2, d);
        }
      }
    } catch (M68kException e) {
    }
  }  //busVl(int,int)



  //========================================================================================
  //$$GRA グラフィック画面
  //  512ドットのときハードウェアスクロールが常に4ページあるのでメモリモードに関わらず4ビットゼロ拡張で格納する
  //  偶数アドレスの8ビット全部と奇数アドレスの上位4ビットは常に0
  //
  //  512ドット16色
  //                アドレス             アクセス                  格納
  //    GE0  0x00c00000～0x00c7ffff  ............3210  ──  ............3210
  //    GE1  0x00c80000～0x00cfffff  ............7654  ──  ............7654
  //    GE2  0x00d00000～0x00d7ffff  ............ba98  ──  ............ba98
  //    GE3  0x00d80000～0x00dfffff  ............fedc  ──  ............fedc
  //
  //  512ドット256色
  //                アドレス             アクセス                  格納
  //    GF0  0x00c00000～0x00c7ffff  ........76543210  ─┬  ............3210
  //    GF1  0x00c80000～0x00cfffff  ........fedcba98  ┐└  ............7654
  //         0x00d00000～0x00d7ffff                    ├─  ............ba98
  //         0x00d80000～0x00dfffff                    └─  ............fedc
  //
  //  512ドット65536色
  //                アドレス             アクセス                  格納
  //    GG0  0x00c00000～0x00c7ffff  fedcba9876543210  ─┬  ............3210
  //         0x00c80000～0x00cfffff                      ├  ............7654
  //         0x00d00000～0x00d7ffff                      ├  ............ba98
  //         0x00d80000～0x00dfffff                      └  ............fedc
  //
  //  1024ドット16色
  //                アドレス             アクセス                  格納
  //    GH0  0x00c00000～0x00dfffff  ............3210  ──  ............3210
  //



    //graInit ()
    public static void graInit () {
    }



  //========================================================================================
  //$$TXT テキスト画面

  //txtInit ()
  //  テキスト画面の初期化
  public static void txtInit () {
  }  //txtInit()



  //========================================================================================
  //$$SVS スーパーバイザ領域設定
  public static final int SVS_AREASET = 0x00e86001;  //0x00 0x00000000～0x00002000
  //                                                      0x01 0x00000000～0x00004000
  //                                                      0x02 0x00000000～0x00008000
  //                                                      0x04 0x00000000～0x00010000
  //                                                      0x08 0x00000000～0x00020000
  //                                                      0x10 0x00000000～0x00040000
  //                                                      0x20 0x00000000～0x00080000
  //                                                      0x40 0x00000000～0x00100000
  //                                                      0x80 0x00000000～0x00200000

  public static void svsInit () {
  }



  //========================================================================================
  //$$SYS システムポート
  //
  //     アドレス   bit  RW  名前   X68030
  //    0x00e8e001  0-3  RW           13    コントラスト(0=最も暗い,15=最も明るい)
  //    0x00e8e003   3   R             1    TV ON/OFFステータス(0=ON,1=OFF)
  //                      W                 TVリモコン信号
  //                 2   R   FIELD     0
  //                 1   RW  3D-L      0    (3Dスコープ)シャッター左(0=OPEN,1=CLOSE)
  //                 0   RW  3D-R      0    (3Dスコープ)シャッター右(0=OPEN,1=CLOSE)
  //    0x00e8e005  4-0   W                 (カラーイメージユニット(デジタイズテロッパ))画像入力コントロール
  //                                          bit4  IMAGE IN bit17
  //                                          bit3  IMAGE IN bit18
  //                                          bit2  IMAGE IN bit19
  //                                          bit1  IMAGE IN bit20
  //                                          bit0  IMAGE IN bit21
  //    0x00e8e007   3   R             1    キージャックステータス(0=抜かれている,1=差し込まれている)
  //                      W                 キーレディ(0=キーデータ送信禁止,1=キーデータ送信許可)
  //                 2    W            1    1=NMIリセット
  //                 1   RW            0    HRL。詳細はCRTCを参照
  //                 0   RW            0    (現在は使用されていない)解像度LED(0=消灯,1=点灯)
  //    0x00e8e009  7-4   W                 (X68030のみ)ROMアクセスウェイト
  //                3-0   W                 (X68030のみ)RAMアクセスウェイト(0=25MHz,4=16MHz相当,10=10MHz相当)
  //    0x00e8e00b  7-4  R            13    機種(13=MC68030,15=MC68000)
  //                3-0  R            12    動作周波数(12=25MHz,14=16MHz,15=10MHz)
  //    0x00e8e00d  7-0   W                 SRAM WRITE ENABLE(49=SRAM書き込み可)
  //    0x00e8e00f  3-0   W                 フロント電源スイッチがOFFになっているとき0→15→15で電源OFF
  //
  //    未定義のbitはリードすると1、ライトしてもバスエラーは発生しない
  //    アドレスのbit12-4はデコードされない。0x00e8e000～0x00e8ffffの範囲に16バイトのポートが512回繰り返し現れる
  //

  public static boolean sysNMIFlag;  //true=INTERRUPTスイッチが押された

  //sysInit ()
  //  初期化
  public static void sysInit () {
    sysNMIFlag = false;
  }  //sysInit()

  //割り込み受付
  //  コアが割り込み要求を受け付けたときに呼び出す
  //  割り込みベクタ番号を返す
  //  割り込み要求を取り下げる場合は0を返す
  //  オートベクタを使用するデバイスはオートベクタの番号を返すこと
  public static int sysAcknowledge () {
    return M68kException.M6E_LEVEL_7_INTERRUPT_AUTOVECTOR;
  }  //sysAcknowledge()

  //割り込み終了
  //  コアが割り込み処理を終了したときに呼び出す
  //  まだ処理されていない割り込みが残っていたら再度割り込み要求を出す
  public static void sysDone () {
    if (sysNMIFlag) {  //NMIリセットされていない
      mpuIRR |= MPU_SYS_INTERRUPT_MASK;
    }
  }  //sysDone()

  //sysInterrupt ()
  //  INTERRUPTスイッチが押された
  public static void sysInterrupt () {
    sysNMIFlag = true;
    mpuIRR |= MPU_SYS_INTERRUPT_MASK;
  }  //sysInterrupt()

  //sysResetNMI ()
  //  NMIリセット
  public static void sysResetNMI () {
    sysNMIFlag = false;
  }  //sysResetNMI()



  //========================================================================================
  //$$EB2 拡張ボードレベル2割り込み

  public static final int EB2_SPC_REQUEST = 0x4000;  //拡張SCSI
  public static final int EB2_SPC_VECTOR = 0xf6;  //拡張SCSI

  //割り込み要求
  //  0b00010000_00000000  SPC割り込み要求あり
  public static int eb2Request;  //割り込み要求。デバイスは操作しないこと

  //eb2Reset ()
  //  拡張ボードレベル2割り込みをリセットする
  public static void eb2Reset () {
    eb2Request = 0;
  }  //eb2Reset()

  //eb2Interrupt (mask)
  //  割り込み要求
  //  デバイスが割り込みを要求するときに呼び出す
  //  mask  EB2_SPC_VECTOR  拡張SCSI割り込みを要求する
  public static void eb2Interrupt (int mask) {
    eb2Request |= mask;
    mpuIRR |= MPU_EB2_INTERRUPT_MASK;
  }  //eb2Interrupt(int)

  //vector = eb2Acknowledge ()
  //  割り込み受付
  //  コアが割り込み要求を受け付けたときに呼び出す
  //  割り込みベクタ番号を返す
  //  割り込み要求を取り下げる場合は0を返す
  //  オートベクタを使用するデバイスはオートベクタの番号を返すこと
  public static int eb2Acknowledge () {
    if ((eb2Request & EB2_SPC_REQUEST) != 0) {
      eb2Request &= ~EB2_SPC_REQUEST;
      return EB2_SPC_VECTOR;
    }
    return 0;
  }  //eb2Acknowledge()

  //eb2Done ()
  //  割り込み終了
  //  コアが割り込み処理を終了したときに呼び出す
  //  まだ処理されていない割り込みが残っていたら再度割り込み要求を出す
  public static void eb2Done () {
    if (eb2Request != 0) {
      mpuIRR |= MPU_EB2_INTERRUPT_MASK;
    }
  }  //eb2Done()



  //========================================================================================
  //TGA ツクモグラフィックアクセラレータPCMボード
  //  0x00e9e200  ツクモグラフィックアクセラレータPCMボード(TS-6BGA)
  //! 非対応



  //========================================================================================
  //WIN WINDRV
  //  0x00e9f000  040Excel
  //  0x00e9f000  WINDRV
  //! 非対応



  //========================================================================================
  //$$EXS 拡張SCSI
  //  内蔵SCSIと共通



  //========================================================================================
  //$$XB2 拡張ボード領域2
  //  0x00eaf900  FAXボード(CZ-6BC1)
  //  0x00eafa00  MIDIボード(CZ-6BM1)
  //  0x00eafb00  パラレルボード(CZ-6BN1)
  //  0x00eafc00  RS-232Cボード(CZ-6BF1)
  //  0x00eafd00  ユニバーサルI/Oボード(CZ-6BU1)
  //  0x00eafe00  GP-IBボード(CZ-6BG1)
  //  0x00eaff00  スーパーバイザエリア設定
  //
  //  スーパーバイザエリア設定のみ対応

  //スーパーバイザエリア設定ポート
  //  0x00eaff81  bit0  0x00200000～0x0023ffff
  //              bit1  0x00240000～0x0027ffff
  //              bit2  0x00280000～0x002bffff
  //              bit3  0x002c0000～0x002fffff
  //              bit4  0x00300000～0x0033ffff
  //              bit5  0x00340000～0x0037ffff
  //              bit6  0x00380000～0x003bffff
  //              bit7  0x003c0000～0x003fffff
  //  0x00eaff83  bit0  0x00400000～0x0043ffff
  //              bit1  0x00440000～0x0047ffff
  //              bit2  0x00480000～0x004bffff
  //              bit3  0x004c0000～0x004fffff
  //              bit4  0x00500000～0x0053ffff
  //              bit5  0x00540000～0x0057ffff
  //              bit6  0x00580000～0x005bffff
  //              bit7  0x005c0000～0x005fffff
  //  0x00eaff85  bit0  0x00600000～0x0063ffff
  //              bit1  0x00640000～0x0067ffff
  //              bit2  0x00680000～0x006bffff
  //              bit3  0x006c0000～0x006fffff
  //              bit4  0x00700000～0x0073ffff
  //              bit5  0x00740000～0x0077ffff
  //              bit6  0x00780000～0x007bffff
  //              bit7  0x007c0000～0x007fffff
  //  0x00eaff87  bit0  0x00800000～0x0083ffff
  //              bit1  0x00840000～0x0087ffff
  //              bit2  0x00880000～0x008bffff
  //              bit3  0x008c0000～0x008fffff
  //              bit4  0x00900000～0x0093ffff
  //              bit5  0x00940000～0x0097ffff
  //              bit6  0x00980000～0x009bffff
  //              bit7  0x009c0000～0x009fffff
  //  0x00eaff89  bit0  0x00a00000～0x00a3ffff
  //              bit1  0x00a40000～0x00a7ffff
  //              bit2  0x00a80000～0x00abffff
  //              bit3  0x00ac0000～0x00afffff
  //              bit4  0x00b00000～0x00b3ffff
  //              bit5  0x00b40000～0x00b7ffff
  //              bit6  0x00b80000～0x00bbffff
  //              bit7  0x00bc0000～0x00bfffff



  //========================================================================================
  //$$BNK バンクメモリ
  //  $00EAFF7F  バンクメモリのページ番号。0～255
  //  $00EE0000～$00EFFFFF  バンクメモリ

  public static final int BNK_SIZE = 1024 * 1024 * 32;  //サイズ
  public static byte[] bnkMemory;  //メモリの配列
  public static int bnkPageStart;  //ページの先頭のインデックス。ページ番号<<17
  public static boolean bnkOn;  //true=バンクメモリが有効

  public static void bnkInit () {
    bnkMemory = new byte[BNK_SIZE];
    byte[] array = Settings.sgsGetData ("bankdata");
    if (array.length != 0) {
      System.arraycopy (array, 0, bnkMemory, 0, Math.min (array.length, BNK_SIZE));
    }
    bnkPageStart = 0;
    //bnkOn = true;
    bnkOn = false;
  }

  public static void bnkTini () {
    Settings.sgsPutData ("bankdata", bnkMemory, 0, BNK_SIZE);
  }



  //========================================================================================
  //$$SMR SRAM
  //
  //  起動デバイスの強制指定
  //    リセット後最初の1回だけSRAMからの起動デバイスのリードを偽ることでOPT.1キーを押しながら起動や特定のデバイスからの起動が可能になる

  public static int smrSramSizeRequest;  //SRAMのサイズ。16384,32768,65536のいずれか
  public static int smrSramCurrentSize;  //現在のSRAMのサイズ。16384,32768,65536のいずれか

  public static String smrSramName;  //SRAMイメージファイル名

  public static int smrBootDevice;  //起動デバイス。0x00ed0018。0x0000=STD,0x9070～0x9370=FDn,0x8000～0x8f00=HDn,0xa000=ROM,0xb000=RAM,-1=設定しない
  public static int smrROMBootHandle;  //ROM起動ハンドル。0x00ed000c。-1=設定しない
  public static int smrRAMBootAddress;  //RAM起動アドレス。0x00ed0010。-1=設定しない

  public static int smrRepeatDelay;  //リピートディレイ。0x00ed003a。-1=設定しない,0..15=200+100*n(ms)
  public static int smrRepeatInterval;  //リピートインターバル。0x00ed003b。-1=設定しない,0..15=30+5*n^2(ms)
  public static boolean smrWriteEnableOn;  //true=SRAM書き込み可

  //-romdb
  public static boolean smrROMDBOn;  //ROMデバッガ起動フラグ

  public static JMenu smrMenu;  //SRAMメニュー
  public static JMenu smrBootMenu;  //起動デバイスメニュー
  public static JRadioButtonMenuItem smrDefaultMenuItem;

  //smrInit ()
  //  SRAMを初期化する
  public static void smrInit () {

    //-sramsize
    int sizeKB = Settings.sgsGetInt ("sramsize");  //SRAMのサイズ
    smrSramSizeRequest = (sizeKB == 16 ||
                          sizeKB == 32 ||
                          sizeKB == 64 ? sizeKB << 10 : 16 << 10);
    smrSramCurrentSize = smrSramSizeRequest;
    System.out.printf (Multilingual.mlnJapanese ?
                       "SRAM のサイズは %dKB です\n" :
                       "SRAM size is %dKB\n",
                       smrSramSizeRequest >> 10);

    boolean initialized = false;

    //-sram
    smrSramName = "";
    if (!initialized) {
      smrSramName = Settings.sgsGetString ("sram");  //SRAMイメージファイル名
      if (smrSramName.length () != 0) {
        byte[] array = rscGetFile (smrSramName, smrSramSizeRequest);
        if (array != null) {
          System.arraycopy (array, 0, MainMemory.mmrM8, 0x00ed0000, smrSramSizeRequest);
          if (smrSramSizeRequest < 64 << 10) {
            Arrays.fill (MainMemory.mmrM8, 0x00ed0000 + smrSramSizeRequest, 0x00ed0000 + (64 << 10), (byte) 0);  //範囲外をゼロクリアする
          }
          initialized = true;
        }
      }
    }

    //-sramdata
    //  マジックの有無に関係なく復元する
    if (!initialized) {
      byte[] array = Settings.sgsGetData ("sramdata");  //SRAMの内容(gzip+base64)
      if (array.length != 0) {  //復元するデータがある
        System.out.println (Multilingual.mlnJapanese ?
                            "SRAM のデータを復元します" :
                            "SRAM data is restored");
        System.arraycopy (array, 0, MainMemory.mmrM8, 0x00ed0000, Math.min (array.length, smrSramSizeRequest));
        if (array.length < smrSramSizeRequest) {
          Arrays.fill (MainMemory.mmrM8, 0x00ed0000 + array.length, 0x00ed0000 + smrSramSizeRequest, (byte) 0);  //復元されなかった部分をゼロクリアする
        }
        if (smrSramSizeRequest < 64 << 10) {
          Arrays.fill (MainMemory.mmrM8, 0x00ed0000 + smrSramSizeRequest, 0x00ed0000 + (64 << 10), (byte) 0);  //範囲外をゼロクリアする
        }
        initialized = true;
      }
    }

    if (!initialized) {
      System.out.println (Multilingual.mlnJapanese ?
                          "SRAM をゼロクリアします" :
                          "SRAM is zero-cleared");
      Arrays.fill (MainMemory.mmrM8, 0x00ed0000, 0x00ed0000 + (64 << 10), (byte) 0);  //ゼロクリアする
      initialized = true;
    }

    //-keydly
    int repeatDelay = Settings.sgsGetInt ("keydly");
    smrRepeatDelay = -1 <= repeatDelay && repeatDelay <= 15 ? repeatDelay : -1;

    //-keyrep
    int repeatInterval = Settings.sgsGetInt ("keyrep");
    smrRepeatInterval = -1 <= repeatInterval && repeatInterval <= 15 ? repeatInterval : -1;

    //-romdb
    smrROMDBOn = Settings.sgsGetOnOff ("romdb");

    //-boot
    smrParseBootDevice (Settings.sgsGetString ("boot"));

  }  //smrInit

  //smrTini ()
  //  SRAMの後始末
  public static void smrTini () {

    //-sramsize
    Settings.sgsPutInt ("sramsize", smrSramSizeRequest >> 10);

    //-sram
    Settings.sgsPutString ("sram", smrSramName);

    //-sramdata
    //  マジックの有無に関係なく保存する
    //  常に64KB保存する
    Settings.sgsPutData ("sramdata", MainMemory.mmrM8, 0x00ed0000, 64 << 10);

    //-keydly
    Settings.sgsPutInt ("keydly", smrRepeatDelay);  //リピートディレイ

    //-keyrep
    Settings.sgsPutInt ("keyrep", smrRepeatInterval);  //リピートインターバル

    //-romdb
    Settings.sgsPutOnOff ("romdb", smrROMDBOn);

    //-boot
    Settings.sgsPutString ("boot",
                           smrBootDevice == -1 ? "" :
                           smrBootDevice == 0x0000 ? "std" :
                           (smrBootDevice & 0xf000) == 0x9000 ? "fd" + (smrBootDevice >> 8 & 3) :
                           (smrBootDevice & 0xf000) == 0x8000 ? "hd" + (smrBootDevice >> 8 & 15) :
                           smrBootDevice == 0xa000 ?
                           (smrROMBootHandle & ~(7 << 2)) == SPC.SPC_HANDLE_EX ? "sc" + (smrROMBootHandle >> 2 & 7) :
                           smrROMBootHandle == HFS.HFS_BOOT_HANDLE ? "hf" + HFS.hfsBootUnit :
                           String.format ("rom$%08X", smrROMBootHandle) :
                           smrBootDevice == 0xb000 ? String.format ("ram$%08X", smrRAMBootAddress) :
                           "");

  }

  public static void smrMakeMenu () {
    //メニュー
    ActionListener listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        switch (ae.getActionCommand ()) {
        case "Zero-clear":  //SRAM をゼロクリア
          smrClear ();
          break;
        case "Import":  //SRAM をインポート
          smrLoad ();
          break;
        case "Export":  //SRAM をエクスポート
          smrSave ();
          break;
        case "16KB":
          smrSramSizeRequest = 16384;
          break;
        case "32KB":
          smrSramSizeRequest = 32768;
          break;
        case "64KB":
          smrSramSizeRequest = 65536;
          break;
        }
      }
    };
    ButtonGroup sizeGroup = new ButtonGroup ();
    smrMenu =
      ComponentFactory.createMenu (
        "SRAM",
        Multilingual.mlnText (
          ComponentFactory.createMenuItem ("Zero-clear", listener),
          "ja", "ゼロクリア"),
        Multilingual.mlnText (
          ComponentFactory.createMenuItem ("Import", listener),
          "ja", "インポート"),
        Multilingual.mlnText (
          ComponentFactory.createMenuItem ("Export", listener),
          "ja", "エクスポート"),
        ComponentFactory.createHorizontalSeparator (),
        Multilingual.mlnText (
          ComponentFactory.createRadioButtonMenuItem (sizeGroup, smrSramSizeRequest >> 10 == 16, "16KB", listener),
          "ja", "16KB"),
        Multilingual.mlnText (
          ComponentFactory.createRadioButtonMenuItem (sizeGroup, smrSramSizeRequest >> 10 == 32, "32KB", listener),
          "ja", "32KB"),
        Multilingual.mlnText (
          ComponentFactory.createRadioButtonMenuItem (sizeGroup, smrSramSizeRequest >> 10 == 64, "64KB", listener),
          "ja", "64KB")
        );

    //起動デバイスメニュー
    ButtonGroup bootGroup = new ButtonGroup ();
    ActionListener bootListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        String command = ae.getActionCommand ();
        smrParseBootDevice (
          command.startsWith ("FDD ") ? "fd" + command.substring (4) :
          command.startsWith ("SASI ") ? "hd" + command.substring (5) :
          command.startsWith ("SCSI ") ? "sc" + command.substring (5) :
          command.startsWith ("HFS ") ? "hf" + command.substring (4) :
          command);  //Default/STD
        if (smrBootDevice != -1) {  //メニューで起動デバイスが既定以外に設定されたとき
          mpuSavedBootDevice = -1;  //保存されている起動デバイスを消す
          mpuSavedROMBootHandle = -1;
        }
      }
    };
    JMenu bootMenuFDD = ComponentFactory.createMenu ("FDD");
    for (int u = 0; u < FDC.FDC_MAX_UNITS; u++) {
      bootMenuFDD.add (ComponentFactory.createRadioButtonMenuItem (
        bootGroup, smrBootDevice == 0x9070 + (u << 8),
        "FDD " + u, bootListener));
    }
    JMenu bootMenuSASI = ComponentFactory.createMenu ("SASI");
    for (int u = 0; u < 16; u++) {
      bootMenuSASI.add (ComponentFactory.createRadioButtonMenuItem (
        bootGroup, smrBootDevice == 0x8000 + (u << 8),
        "SASI " + u, bootListener));
    }
    JMenu bootMenuSCSI = ComponentFactory.createMenu ("SCSI");
    for (int u = 0; u < 8; u++) {
      bootMenuSCSI.add (ComponentFactory.createRadioButtonMenuItem (
        bootGroup, smrBootDevice == 0xa000 && (smrROMBootHandle == SPC.SPC_HANDLE_EX + (u << 2) ||
                                               smrROMBootHandle == SPC.SPC_HANDLE_IN + (u << 2)), "SCSI " + u, bootListener));
    }
    JMenu bootMenuHFS = ComponentFactory.createMenu ("HFS");
    for (int u = 0; u < HFS.HFS_MAX_UNITS; u++) {
      bootMenuHFS.add (ComponentFactory.createRadioButtonMenuItem (
        bootGroup, smrBootDevice == 0xa000 && smrROMBootHandle == HFS.HFS_BOOT_HANDLE && HFS.hfsBootUnit == u,
        "HFS " + u, bootListener));
    }
    smrBootMenu =
      Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Boot device",
          smrDefaultMenuItem = Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (bootGroup, smrBootDevice == -1, "Default", bootListener), "ja", "既定"),
          ComponentFactory.createRadioButtonMenuItem (bootGroup, smrBootDevice == 0x0000, "STD", bootListener),
          bootMenuFDD,
          bootMenuSASI,
          bootMenuSCSI,
          bootMenuHFS
          ),
        "ja", "起動デバイス");

  }  //smrInit()

  //smrParseBootDevice ()
  //  起動デバイスの設定を読み取る
  //  既定は"default"を用いる
  public static void smrParseBootDevice (String boot) {
    smrBootDevice = -1;  //起動デバイス
    smrROMBootHandle = -1;  //ROM起動ハンドル
    smrRAMBootAddress = -1;  //RAM起動アドレス
    boot = boot.toLowerCase ();
    if (boot.equals ("std")) {  //std
      smrBootDevice = 0x0000;  //STD 起動
    } else if (boot.startsWith ("fd")) {  //fdN
      int u = fmtParseInt (boot, 2, 0, FDC.FDC_MAX_UNITS - 1, FDC.FDC_MAX_UNITS);  //起動ユニット番号
      if (u < FDC.FDC_MAX_UNITS) {
        smrBootDevice = 0x9070 + (u << 8);  //FDD起動
      }
    } else if (boot.startsWith ("hd")) {  //hdN
      int u = fmtParseInt (boot, 2, 0, 15, 16);  //起動ユニット番号
      if (u < 16) {
        smrBootDevice = 0x8000 + (u << 8);  //SASI起動
      }
    } else if (boot.startsWith ("sc")) {  //scN
      int u = fmtParseInt (boot, 2, 0, 7, 8);
      if (u < 8) {
        smrBootDevice = 0xa000;  //ROM起動
        smrROMBootHandle = SPC.SPC_HANDLE_EX + ((u & 7) << 2);  //仮に拡張SCSI起動にしておく。リセットしたとき拡張SCSIがなければ内蔵SCSIに読み替えられる
      }
    } else if (boot.startsWith ("hf")) {  //hfN
      int u = fmtParseInt (boot, 2, 0, HFS.HFS_MAX_UNITS - 1, HFS.HFS_MAX_UNITS);  //起動ユニット番号
      if (u < HFS.HFS_MAX_UNITS) {
        HFS.hfsBootUnit = u;
        smrBootDevice = 0xa000;  //ROM起動
        smrROMBootHandle = HFS.HFS_BOOT_HANDLE;  //IPL起動ハンドル
      }
    } else if (boot.startsWith ("rom$")) {  //rom$X
      int handle = fmtParseIntRadix (boot, 3, 0, 0x00ffffff, 0x01000000, 16);  //起動ハンドル
      if (handle < 0x01000000) {
        smrBootDevice = 0xa000;  //ROM起動
        smrROMBootHandle = handle;
      }
    } else if (boot.startsWith ("ram$")) {  //ram$X
      int handle = fmtParseIntRadix (boot, 3, 0, 0x00ffffff, 0x01000000, 16);  //起動ハンドル
      if (handle < 0x01000000) {
        smrBootDevice = 0xb000;  //RAM起動
        smrRAMBootAddress = handle;
      }
    }
  }  //smrParseBootDevice(String)

  //smrReset ()
  //  SRAMリセット
  //  ここでROMも上書きする
  //  ROMを初期化してから呼び出すこと
  //  SPC.spcReset()よりも後であること
  public static void smrReset () {
    smrWriteEnableOn = false;

    //SRAMの容量を変更する
    {
      smrSramCurrentSize = smrSramSizeRequest;
      busSuper (MemoryMappedDevice.MMD_SMR, 0x00ed0000, 0x00ed0000 + smrSramSizeRequest);  //SMR SRAM
      if (smrSramSizeRequest < 65536) {
        busSuper (MemoryMappedDevice.MMD_NUL, 0x00ed0000 + smrSramSizeRequest, 0x00ed0000 + 65536);  //空き
      }
      System.out.printf (Multilingual.mlnJapanese ?
                         "SRAM の容量は %dKB ($%08X-$%08X) です\n" :
                         "Capacity of SRAM is %dKB ($%08X-$%08X)\n",
                         smrSramCurrentSize >> 10, 0x00ed0000, 0x00ed0000 + smrSramCurrentSize - 1);
    }

    //SCSI起動が選択されたときSCSIポートの有無でROM起動ハンドルを調整する
    //  リセットされたときROM起動ハンドルが内蔵SCSI起動と拡張SCSI起動のどちらを指していてもどちらかからのSCSI起動とみなす
    //  拡張SCSIポートがあるときは拡張SCSI起動に、内蔵SCSIポートがあるときは内蔵SCSI起動に、どちらもなければDefaultに変更する
    //  起動デバイス
    if (smrBootDevice == 0xa000 && ((smrROMBootHandle & -32) == SPC.SPC_HANDLE_EX ||
                                    (smrROMBootHandle & -32) == SPC.SPC_HANDLE_IN)) {  //SCSI起動のとき
      if (SPC.spcSCSIEXOn) {  //拡張SCSIポートがあるとき
        smrROMBootHandle = SPC.SPC_HANDLE_EX | smrROMBootHandle & 31;  //拡張SCSI起動にする
        //MainMemory.mmrWb (0x00ed0070, MainMemory.mmrRbs (0x00ed0070) | 0x08);  //拡張フラグをセットする
      } else if (SPC.spcSCSIINOn) {  //内蔵SCSIポートがあるとき
        smrROMBootHandle = SPC.SPC_HANDLE_IN | smrROMBootHandle & 31;  //内蔵SCSI起動にする
        //MainMemory.mmrWb (0x00ed0070, MainMemory.mmrRbs (0x00ed0070) & ~0x08);  //拡張フラグをクリアする
      } else {  //SCSIポートがない
        smrBootDevice = -1;
        smrROMBootHandle = -1;
        smrDefaultMenuItem.setSelected (true);
      }
    }
    //  ここから再起動
    if (mpuBootDevice == 0xa000 && ((mpuROMBootHandle & -32) == SPC.SPC_HANDLE_EX ||
                                    (mpuROMBootHandle & -32) == SPC.SPC_HANDLE_IN)) {  //SCSI起動のとき
      if (SPC.spcSCSIEXOn) {  //拡張SCSIポートがあるとき
        mpuROMBootHandle = SPC.SPC_HANDLE_EX | mpuROMBootHandle & 31;  //拡張SCSI起動にする
        //MainMemory.mmrWb (0x00ed0070, MainMemory.mmrRbs (0x00ed0070) | 0x08);  //拡張フラグをセットする
      } else if (SPC.spcSCSIINOn) {  //内蔵SCSIポートがあるとき
        mpuROMBootHandle = SPC.SPC_HANDLE_IN | mpuROMBootHandle & 31;  //内蔵SCSI起動にする
        //MainMemory.mmrWb (0x00ed0070, MainMemory.mmrRbs (0x00ed0070) & ~0x08);  //拡張フラグをクリアする
      } else {  //SCSIポートがない
        mpuBootDevice = -1;
        mpuROMBootHandle = -1;
      }
    }

    smrOverride ();

  }  //smrReset()

  //smrOverride ()
  //  SRAMの設定を上書きする
  //  ここから再起動もSRAMに上書きする
  public static void smrOverride () {
    if (MainMemory.mmrRls (0x00ed0000) == 0x82773638 &&  //Ｘ68
        MainMemory.mmrRls (0x00ed0004) == 0x30303057) {  //000W  初期化されている。初期化されていないときに上書きしても意味がない
      int memorySizeAddress = 0x00ed0008;
      int memorySizeOld = MainMemory.mmrRls (memorySizeAddress);
      int memorySizeNew = MainMemory.mmrMemorySizeCurrent;
      if (memorySizeOld != memorySizeNew) {
        MainMemory.mmrWl (memorySizeAddress, memorySizeNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にあるメモリサイズを %dMB から %dMB に変更しました\n" :
                           "Changed the memory size in SRAM from %dMB to %dMB\n",
                           memorySizeOld >> 20,
                           memorySizeNew >> 20);
      }
      int romHandleAddress = 0x00ed000c;
      int romHandleOld = MainMemory.mmrRls (romHandleAddress);
      int romHandleNew = mpuROMBootHandle != -1 ? mpuROMBootHandle : smrROMBootHandle;
      if (romHandleNew != -1 &&
          romHandleOld != romHandleNew) {
        MainMemory.mmrWl (romHandleAddress, romHandleNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にある ROM 起動ハンドルを $%08X から $%08X に変更しました\n" :
                           "Changed the ROM boot handle in SRAM from $%08X to $%08X\n",
                           romHandleOld,
                           romHandleNew);
      }
      int ramAddressAddress = 0x00ed0010;
      int ramAddressOld = MainMemory.mmrRls (ramAddressAddress);
      int ramAddressNew = smrRAMBootAddress;
      if (ramAddressNew != -1 &&
          ramAddressOld != ramAddressNew) {
        MainMemory.mmrWl (ramAddressAddress, ramAddressNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にある RAM 起動アドレスを $%08X から $%08X に変更しました\n" :
                           "Changed the RAM boot address in SRAM from $%08X to $%08X\n",
                           ramAddressOld,
                           ramAddressNew);
      }
      int deviceAddress = 0x00ed0018;
      int deviceOld = MainMemory.mmrRwz (deviceAddress);
      int deviceNew = mpuBootDevice != -1 ? mpuBootDevice : smrBootDevice;
      if (deviceNew != -1 &&
          deviceOld != deviceNew) {
        MainMemory.mmrWw (deviceAddress, deviceNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にある起動デバイスを %s から %s に変更しました\n" :
                           "Changed the boot device in SRAM from %s to %s\n",
                           smrBootDescription (romHandleOld, ramAddressOld, deviceOld),
                           smrBootDescription (romHandleNew, ramAddressNew, deviceNew));
      }
      int repeatDelayAddress = 0x00ed003a;
      int repeatDelayOld = MainMemory.mmrRbz (repeatDelayAddress);
      int repeatDelayNew = smrRepeatDelay;
      if (repeatDelayNew != -1 &&
          repeatDelayOld != repeatDelayNew) {
        MainMemory.mmrWb (repeatDelayAddress, repeatDelayNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にあるキーリピート遅延を %dms から %dms に変更しました\n" :
                           "Changed the key repeat delay in SRAM from %dms to %dms\n",
                           200 + 100 * (repeatDelayOld & 15),
                           200 + 100 * (repeatDelayNew & 15));
      }
      int repeatIntervalAddress = 0x00ed003b;
      int repeatIntervalOld = MainMemory.mmrRbz (repeatIntervalAddress);
      int repeatIntervalNew = smrRepeatInterval;
      if (repeatIntervalNew != -1 &&
          repeatIntervalOld != repeatIntervalNew) {
        MainMemory.mmrWb (repeatIntervalAddress, repeatIntervalNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にあるキーリピート間隔を %dms から %dms に変更しました\n" :
                           "Changed the key repeat interval in SRAM from %dms to %dms\n",
                           30 + 5 * (repeatDelayOld & 15),
                           30 + 5 * (repeatDelayNew & 15));
      }
      int romdbAddress = 0x00ed0058;
      int romdbOld = MainMemory.mmrRbz (romdbAddress);
      int romdbNew = smrROMDBOn ? 0xff : 0x00;
      if (romdbOld != romdbNew) {
        MainMemory.mmrWb (romdbAddress, romdbNew);
        if (romdbNew == 0x00) {
          System.out.printf (Multilingual.mlnJapanese ?
                             "SRAM にある ROM デバッガ起動フラグを OFF にしました\n" :
                             "Turned off the ROM debugger startup flag in SRAM\n");
        } else {
          System.out.printf (Multilingual.mlnJapanese ?
                             "SRAM にある ROM デバッガ起動フラグを ON にしました\n" :
                             "Turned on the ROM debugger startup flag in SRAM\n");
        }
      }
      int hdMaxAddress = 0x00ed005a;
      int hdMaxOld = MainMemory.mmrRbz (hdMaxAddress);
      int hdMaxNew = SPC.spcSCSIINOn ? 0 : HDC.hdcHDMax;
      if (hdMaxOld != hdMaxNew) {
        MainMemory.mmrWb (hdMaxAddress, hdMaxNew);
        System.out.printf (Multilingual.mlnJapanese ?
                           "SRAM にある SASI ハードディスクの最大数を %d から %d に変更しました\n" :
                           "Changed the maximum number of SASI hard disks in SRAM from %d to %d\n",
                           hdMaxOld, hdMaxNew);
      }
    }
    //「ここから再起動」をキャンセルする
    //mpuBootDevice = -1;
    //mpuROMBootHandle = -1;
  }

  //s = smrBootDescription (code, romHandle, ramAddress)
  //  device      $00ED0018  起動デバイス
  //                           $0000  STD  FD→HD→ROM→RAM
  //                           $8xxx  HD
  //                           $9xxx  FD
  //                           $Axxx  ROM  ROM起動ハンドルからROM起動アドレスを取り出して呼び出す
  //                                       ROM起動ハンドルがバスエラーになるときはSTDで起動する
  //                           $Bxxx  RAM  RAM起動アドレスを呼び出す。先頭が$60でなければならない
  //  romHandle   $00ED000C  ROM起動ハンドル
  //  ramAddress  $00ED0010  RAM起動アドレス
  public static String smrBootDescription (int romHandle, int ramAddress, int device) {
    switch (device & 0xf000) {
    case 0x8000:
      return "HD" + (device >> 8 & 15);
    case 0x9000:
      return "2HD" + (device >> 8 & 3);
    case 0xa000:
      return ((romHandle & ~0x1f) == 0x00fc0000 ? "SCSI" + (romHandle >> 2 & 7) :   //内蔵SCSI
              (romHandle & ~0x1f) == 0x00ea0020 ? "SCSI" + (romHandle >> 2 & 7) :   //拡張SCSI
              "ROM$" + fmtHex6 (romHandle));
    case 0xb000:
      return "RAM$" + fmtHex6 (ramAddress);
    default:
      return "STD";
    }
  }

  //smrClear ()
  //  SRAMクリア
  public static void smrClear () {
    if (JOptionPane.showConfirmDialog (
      frmFrame,
      Multilingual.mlnJapanese ? "SRAM をクリアしますか？" : "Do you want to clear SRAM?",
      Multilingual.mlnJapanese ? "確認" : "Confirmation",
      JOptionPane.YES_NO_OPTION,
      JOptionPane.PLAIN_MESSAGE) == JOptionPane.YES_OPTION) {
      Arrays.fill (MainMemory.mmrM8, 0x00ed0000, 0x00ed0000 + 65536, (byte) 0x00);
    }
  }  //smrClear()

  //smrLoad ()
  //  SRAM読み込み
  public static void smrLoad () {
    JFileChooser2 fileChooser = new JFileChooser2 ();
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 upperName.startsWith ("SRAM")));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "SRAM データファイル (SRAM*.*)" :
                "SRAM data files (SRAM*.*)");
      }
    });
    if (fileChooser.showOpenDialog (null) == JFileChooser.APPROVE_OPTION) {
      File file = fileChooser.getSelectedFile ();
      String name = file.getPath ();
      if (!smrLoadData (name)) {  //読み込めなかった
        JOptionPane.showMessageDialog (null,
                                       Multilingual.mlnJapanese ?
                                       name + " のサイズが違います" :
                                       name + " has wrong size");
        return;
      }
    }
  }  //smrLoad()

  //success = smrLoadData (name)
  //  SRAMのイメージファイルを読み込む
  public static boolean smrLoadData (String name) {
    byte[] array = rscGetFile (name, smrSramSizeRequest);
    if (array != null) {  //読み込めた
      System.arraycopy (array, 0, MainMemory.mmrM8, 0x00ed0000, smrSramSizeRequest);  //SRAMにコピーする
      if (smrSramSizeRequest < 65536) {
        Arrays.fill (MainMemory.mmrM8, 0x00ed0000 + smrSramSizeRequest, 0x00ed0000 + 65536, (byte) 0x00);  //空き
      }
      return true;
    }
    return false;
  }

  //smrSave ()
  //  SRAM書き出し
  public static void smrSave () {
    JFileChooser2 fileChooser = new JFileChooser2 ();
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 upperName.startsWith ("SRAM")));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "SRAM データファイル (SRAM*.*)" :
                "SRAM data files (SRAM*.*)");
      }
    });
    if (fileChooser.showSaveDialog (null) == JFileChooser.APPROVE_OPTION) {
      rscPutFile (fileChooser.getSelectedFile ().getPath (), MainMemory.mmrM8, 0x00ed0000, smrSramSizeRequest);
    }
  }  //smrSave()



  //========================================================================================
  //$$FPU FPU

  //FPU/FPCP
  public static ExpressionEvaluator fpuMotherboardCoprocessor;  //マザーボードコプロセッサ
  public static ExpressionEvaluator fpuOnChipFPU;  //on-chip FPU
  public static ExpressionEvaluator fpuBox;  //浮動小数点命令を実行するFPU/FPCP

  //数値演算プロセッサボード
  public static EFPBox fpuCoproboard1;  //数値演算プロセッサボード1
  public static EFPBox fpuCoproboard2;  //数値演算プロセッサボード2

  //浮動小数点レジスタ
  public static EFPBox.EFP[] fpuFPn;

  //FPCR control register
  //  exception enable byte
  public static final int FPU_FPCR_BSUN   = 0b00000000_00000000_10000000_00000000;  //branch/set on unordered
  public static final int FPU_FPCR_SNAN   = 0b00000000_00000000_01000000_00000000;  //signaling not a number
  public static final int FPU_FPCR_OPERR  = 0b00000000_00000000_00100000_00000000;  //operand error
  public static final int FPU_FPCR_OVFL   = 0b00000000_00000000_00010000_00000000;  //overflow
  public static final int FPU_FPCR_UNFL   = 0b00000000_00000000_00001000_00000000;  //underflow
  public static final int FPU_FPCR_DZ     = 0b00000000_00000000_00000100_00000000;  //divide by zero
  public static final int FPU_FPCR_INEX2  = 0b00000000_00000000_00000010_00000000;  //inexact operation
  public static final int FPU_FPCR_INEX1  = 0b00000000_00000000_00000001_00000000;  //inexact decimal input
  //  mode control byte
  //    rounding precision
  public static final int FPU_FPCR_PE     = 0b00000000_00000000_00000000_00000000;  //extended
  public static final int FPU_FPCR_PS     = 0b00000000_00000000_00000000_01000000;  //single
  public static final int FPU_FPCR_PD     = 0b00000000_00000000_00000000_10000000;  //double
  //    rounding mode
  public static final int FPU_FPCR_RN     = 0b00000000_00000000_00000000_00000000;  //to nearest
  public static final int FPU_FPCR_RZ     = 0b00000000_00000000_00000000_00010000;  //toward zero
  public static final int FPU_FPCR_RM     = 0b00000000_00000000_00000000_00100000;  //toward minus infinity
  public static final int FPU_FPCR_RP     = 0b00000000_00000000_00000000_00110000;  //toward plus infinity

  //FPSR status register
  //  condition code byte
  public static final int FPU_FPSR_N         = 0b00001000_00000000_00000000_00000000;  //negative
  public static final int FPU_FPSR_Z         = 0b00000100_00000000_00000000_00000000;  //zero
  public static final int FPU_FPSR_I         = 0b00000010_00000000_00000000_00000000;  //infinity
  public static final int FPU_FPSR_NAN       = 0b00000001_00000000_00000000_00000000;  //not a number or unordered
  //  quotient byte
  public static final int FPU_FPSR_S         = 0b00000000_10000000_00000000_00000000;  //sign of quotient
  public static final int FPU_FPSR_QUOTIENT  = 0b00000000_01111111_00000000_00000000;  //quotient
  //  exception status byte
  public static final int FPU_FPSR_EXC_BSUN  = 0b00000000_00000000_10000000_00000000;  //branch/set on unordered
  public static final int FPU_FPSR_EXC_SNAN  = 0b00000000_00000000_01000000_00000000;  //signaling not a number
  public static final int FPU_FPSR_EXC_OPERR = 0b00000000_00000000_00100000_00000000;  //operand error
  public static final int FPU_FPSR_EXC_OVFL  = 0b00000000_00000000_00010000_00000000;  //overflow
  public static final int FPU_FPSR_EXC_UNFL  = 0b00000000_00000000_00001000_00000000;  //underflow
  public static final int FPU_FPSR_EXC_DZ    = 0b00000000_00000000_00000100_00000000;  //divide by zero
  public static final int FPU_FPSR_EXC_INEX2 = 0b00000000_00000000_00000010_00000000;  //inexact operation
  public static final int FPU_FPSR_EXC_INEX1 = 0b00000000_00000000_00000001_00000000;  //inexact decimal input
  //  accrued exception byte
  public static final int FPU_FPSR_AEXC_IOP  = 0b00000000_00000000_00000000_10000000;  //invalid operation
  public static final int FPU_FPSR_AEXC_OVFL = 0b00000000_00000000_00000000_01000000;  //overflow
  public static final int FPU_FPSR_AEXC_UNFL = 0b00000000_00000000_00000000_00100000;  //underflow
  public static final int FPU_FPSR_AEXC_DZ   = 0b00000000_00000000_00000000_00010000;  //divide by zero
  public static final int FPU_FPSR_AEXC_INEX = 0b00000000_00000000_00000000_00001000;  //inexact

  //  EXCからAEXCへの変換
  //    AEXC_IOP |= EXC_BSUN | EXC_SNAN | EXC_OPERR
  //    AEXC_OVFL |= EXC_OVFL
  //    AEXC_UNFL |= EXC_UNFL & EXC_INEX2
  //    AEXC_DZ |= EXC_DZ
  //    AEXC_INEX |= EXC_OVFL | EXC_INEX2 | EXC_INEX1
  public static final int[] FPU_FPSR_EXC_TO_AEXC = new int[256];

  //コンディション
  //
  //  fpsrのbit27-24
  //    MZIN
  //    0000  0<x
  //    0001  NaN
  //    0010  +Inf
  //    0100  +0
  //    1000  x<0
  //    1010  -Inf
  //    1100  -0
  //
  //  FPU_CCMAP_882[(オペコードまたは拡張ワード&63)<<4|fpsr>>24&15]==trueならば条件成立
  //  FPU_CCMAP_060[(オペコードまたは拡張ワード&63)<<4|fpsr>>24&15]==trueならば条件成立
  //
  //  MC68882とMC68060ではOR,NE,GLE,SNEの条件が異なる

  //MC68882
  //  perl -e "@a=();for$a(0..15){$m=$a>>3&1;$z=$a>>2&1;$n=$a&1;@b=map{$_&1}(0,$z,~($n|$z|$m),$z|~($n|$m),$m&~($n|$z),$z|($m&~$n),~($n|$z),$z|~$n,$n,$n|$z,$n|~($m|$z),$n|($z|~$m),$n|($m&~$z),$n|$z|$m,$n|~$z,1);push@a,@b,@b,@b,@b}for$y(0..63){print'    ';$t=0;for$x(0..15){$c=$a[$x<<6|$y];$t=($t<<1)+$c;print(($c?'T':'F').',');}printf'  //%04x %06b%c',$t,$y,10;}
  //
  //F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,  //N
  //F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //Z
  //F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //NAN
  public static final boolean[] FPU_CCMAP_882 = {
    //                                       cccccc  cc    等式          意味
    //IEEEアウェアテスト
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 000000  F     0             偽
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 000001  EQ    Z             等しい
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 000010  OGT   ~(NAN|Z|N)    比較可能でより大きい
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 000011  OGE   Z|~(NAN|N)    等しいか比較可能でより大きい
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 000100  OLT   N&~(NAN|Z)    比較可能でより小さい
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 000101  OLE   Z|(N&~NAN)    等しいか比較可能でより小さい
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 000110  OGL   ~(NAN|Z)      比較可能で等しくない
    T,F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,  //afaf 000111  OR    Z|~NAN        等しいか比較可能
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 001000  UN    NAN           比較不能
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 001001  UEQ   NAN|Z         比較不能か等しい
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 001010  UGT   NAN|~(N|Z)    比較不能かより大きい
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 001011  UGE   NAN|(Z|~N)    比較不能かより大きいか等しい
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 001100  ULT   NAN|(N&~Z)    比較不能かより小さい
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 001101  ULE   NAN|Z|N       比較不能かより小さいか等しい
    T,T,T,T,F,T,F,T,T,T,T,T,F,T,F,T,  //f5f5 001110  NE    NAN|~Z        比較不能か等しくない
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 001111  T     1             真
    //IEEEノンアウェアテスト
    //  NANがセットされているとき、FPSRのBSUNがセットされ、許可されていれば例外が発生する
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 010000  SF    0             偽(シグナリング)
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 010001  SEQ   Z             等しい(シグナリング)
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 010010  GT    ~(NAN|Z|N)    より大きい
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 010011  GE    Z|~(NAN|N)    より大きいか等しい
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 010100  LT    N&~(NAN|Z)    より小さい
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 010101  LE    Z|(N&~NAN)    より小さいか等しい
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 010110  GL    ~(NAN|Z)      等しくない
    T,F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,  //afaf 010111  GLE   Z|~NAN        より大きいか小さいか等しい
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 011000  NGLE  NAN           GLEでない
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 011001  NGL   NAN|Z         GLでない
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 011010  NLE   NAN|~(N|Z)    LEでない
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 011011  NLT   NAN|(Z|~N)    LTでない
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 011100  NGE   NAN|(N&~Z)    GEでない
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 011101  NGT   NAN|Z|N       GTでない
    T,T,T,T,F,T,F,T,T,T,T,T,F,T,F,T,  //f5f5 011110  SNE   NAN|~Z        等しくない(シグナリング)
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 011111  ST    1             真(シグナリング)
    //
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 100000
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 100001
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 100010
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 100011
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 100100
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 100101
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 100110
    T,F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,  //afaf 100111
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 101000
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 101001
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 101010
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 101011
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 101100
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 101101
    T,T,T,T,F,T,F,T,T,T,T,T,F,T,F,T,  //f5f5 101110
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 101111
    //
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 110000
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 110001
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 110010
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 110011
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 110100
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 110101
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 110110
    T,F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,  //afaf 110111
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 111000
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 111001
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 111010
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 111011
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 111100
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 111101
    T,T,T,T,F,T,F,T,T,T,T,T,F,T,F,T,  //f5f5 111110
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 111111
  };

  //MC68060
  //  perl -e "@a=();for$a(0..15){$m=$a>>3&1;$z=$a>>2&1;$n=$a&1;@b=map{$_&1}(0,$z,~($n|$z|$m),$z|~($n|$m),$m&~($n|$z),$z|($m&~$n),~($n|$z),~$n,$n,$n|$z,$n|~($m|$z),$n|($z|~$m),$n|($m&~$z),$n|$z|$m,~$z,1);push@a,@b,@b,@b,@b}for$y(0..63){print'    ';$t=0;for$x(0..15){$c=$a[$x<<6|$y];$t=($t<<1)+$c;print(($c?'T':'F').',');}printf'  //%04x %06b%c',$t,$y,10;}
  //
  //F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,T,  //N
  //F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //Z
  //F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //NAN
  public static final boolean[] FPU_CCMAP_060 = {
    //                                       cccccc  cc    等式          意味
    //IEEEアウェアテスト
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 000000  F     0             偽
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 000001  EQ    Z             等しい
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 000010  OGT   ~(NAN|Z|N)    比較可能でより大きい
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 000011  OGE   Z|~(NAN|N)    等しいか比較可能でより大きい
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 000100  OLT   N&~(NAN|Z)    比較可能でより小さい
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 000101  OLE   Z|(N&~NAN)    等しいか比較可能でより小さい
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 000110  OGL   ~(NAN|Z)      比較可能で等しくない
    T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,  //aaaa 000111  OR    ~NAN          比較可能
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 001000  UN    NAN           比較不能
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 001001  UEQ   NAN|Z         比較不能か等しい
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 001010  UGT   NAN|~(N|Z)    比較不能かより大きい
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 001011  UGE   NAN|(Z|~N)    比較不能かより大きいか等しい
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 001100  ULT   NAN|(N&~Z)    比較不能かより小さい
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 001101  ULE   NAN|Z|N       比較不能かより小さいか等しい
    T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,  //f0f0 001110  NE    ~Z            等しくない
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 001111  T     1             真
    //IEEEノンアウェアテスト
    //  NANがセットされているとき、FPSRのBSUNがセットされ、許可されていれば例外が発生する
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 010000  SF    0             偽(シグナリング)
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 010001  SEQ   Z             等しい(シグナリング)
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 010010  GT    ~(NAN|Z|N)    より大きい
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 010011  GE    Z|~(NAN|N)    より大きいか等しい
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 010100  LT    N&~(NAN|Z)    より小さい
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 010101  LE    Z|(N&~NAN)    より小さいか等しい
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 010110  GL    ~(NAN|Z)      等しくない
    T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,  //aaaa 010111  GLE   ~NAN          より大きいか小さいか等しい
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 011000  NGLE  NAN           GLEでない
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 011001  NGL   NAN|Z         GLでない
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 011010  NLE   NAN|~(N|Z)    LEでない
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 011011  NLT   NAN|(Z|~N)    LTでない
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 011100  NGE   NAN|(N&~Z)    GEでない
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 011101  NGT   NAN|Z|N       GTでない
    T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,  //f0f0 011110  SNE   ~Z            等しくない(シグナリング)
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 011111  ST    1             真(シグナリング)
    //
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 100000
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 100001
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 100010
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 100011
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 100100
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 100101
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 100110
    T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,  //aaaa 100111
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 101000
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 101001
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 101010
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 101011
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 101100
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 101101
    T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,  //f0f0 101110
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 101111
    //
    F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,  //0000 110000
    F,F,F,F,T,T,T,T,F,F,F,F,T,T,T,T,  //0f0f 110001
    T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,  //a000 110010
    T,F,T,F,T,T,T,T,F,F,F,F,T,T,T,T,  //af0f 110011
    F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,  //00a0 110100
    F,F,F,F,T,T,T,T,T,F,T,F,T,T,T,T,  //0faf 110101
    T,F,T,F,F,F,F,F,T,F,T,F,F,F,F,F,  //a0a0 110110
    T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,  //aaaa 110111
    F,T,F,T,F,T,F,T,F,T,F,T,F,T,F,T,  //5555 111000
    F,T,F,T,T,T,T,T,F,T,F,T,T,T,T,T,  //5f5f 111001
    T,T,T,T,F,T,F,T,F,T,F,T,F,T,F,T,  //f555 111010
    T,T,T,T,T,T,T,T,F,T,F,T,T,T,T,T,  //ff5f 111011
    F,T,F,T,F,T,F,T,T,T,T,T,F,T,F,T,  //55f5 111100
    F,T,F,T,T,T,T,T,T,T,T,T,T,T,T,T,  //5fff 111101
    T,T,T,T,F,F,F,F,T,T,T,T,F,F,F,F,  //f0f0 111110
    T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,  //ffff 111111
  };

  //fpuInit ()
  //  FPUを初期化する
  //  これはmpuInit()から呼ばれる
  public static void fpuInit () {
    for (int i = 0; i < 256; i++) {
      FPU_FPSR_EXC_TO_AEXC[i] = (((i << 8 & (FPU_FPSR_EXC_BSUN | FPU_FPSR_EXC_SNAN | FPU_FPSR_EXC_OPERR)) != 0 ? FPU_FPSR_AEXC_IOP : 0) |
                                 ((i << 8 & FPU_FPSR_EXC_OVFL) != 0 ? FPU_FPSR_AEXC_OVFL : 0) |
                                 ((i << 8 & (FPU_FPSR_EXC_UNFL | FPU_FPSR_EXC_INEX2)) == (FPU_FPSR_EXC_UNFL | FPU_FPSR_EXC_INEX2) ? FPU_FPSR_AEXC_UNFL : 0) |
                                 ((i << 8 & FPU_FPSR_EXC_DZ) != 0 ? FPU_FPSR_AEXC_DZ : 0) |
                                 ((i << 8 & (FPU_FPSR_EXC_OVFL | FPU_FPSR_EXC_INEX2 | FPU_FPSR_EXC_INEX1)) != 0 ? FPU_FPSR_AEXC_INEX : 0));
    }
    //マザーボードコプロセッサ
    fpuMotherboardCoprocessor = new ExpressionEvaluator ();
    //on-chip FPU
    fpuOnChipFPU = new ExpressionEvaluator ();
    //浮動小数点命令を実行するFPU/FPCP
    fpuBox = mpuCoreType <= 3 ? fpuMotherboardCoprocessor : fpuOnChipFPU;
    //浮動小数点レジスタ
    fpuFPn = fpuBox.epbFPn;
    //数値演算プロセッサボード
    fpuCoproboard1 = new EFPBox ();
    fpuCoproboard1.epbSetMC68881 ();
    fpuCoproboard2 = new EFPBox ();
    fpuCoproboard2.epbSetMC68881 ();
  }  //fpuInit()



  //========================================================================================
  //$$DBG デバッガ共通コンポーネント

  public static final boolean DBG_ORI_BYTE_ZERO_D0 = true;  //true=ORI.B #$00,D0(オペコード0x0000)を不当命令とみなす機能を有効にする。暴走をなるべく早く検出することで暴走の原因を特定しやすくする

  public static boolean dbgHexSelected;  //true=16進数が選択されている
  public static int dbgHexValue;  //選択されている16進数の値
  public static int dbgSupervisorMode;  //0=ユーザモード,0以外=スーパーバイザモード
  public static JPopupMenu dbgPopupMenu;  //ポップアップメニュー
  public static JMenu dbgPopupIBPMenu;  //命令ブレークポイントメニュー
  public static SpinnerNumberModel dbgPopupIBPCurrentModel;  //現在値のスピナーモデル
  public static int dbgPopupIBPCurrentValue;  //現在値
  public static SpinnerNumberModel dbgPopupIBPThresholdModel;  //閾値のスピナーモデル
  public static int dbgPopupIBPThresholdValue;  //閾値
  public static JMenuItem dbgPopupIBPClearMenuItem;  //解除
  public static JMenu dbgPopupHexMenu;  //16進数メニュー
  public static JMenuItem dbgPopupDisMenuItem;  //逆アセンブル
  public static JMenuItem dbgPopupMemMenuItem;  //メモリダンプ
  public static JMenuItem dbgPopupCopyMenuItem;  //コピー
  public static JMenuItem dbgPopupSelectAllMenuItem;  //すべて選択
  public static JTextArea dbgPopupTextArea;  //ポップアップメニューを表示したテキストエリア
  public static int dbgEventMask;  //イベントマスク。0でないときチェンジリスナーとキャレットリスナーを無効化
  public static boolean dbgStopOnError;  //true=エラーが発生したときコアを止める
  public static boolean dbgOriByteZeroD0;  //true=ORI.B #$00,D0を不当命令とみなす。普段はOFFにしておくこと

  //共通
  //  sb.append(DBG_SPACES,0,length)でStringBuilderに連続する空白を追加するための配列
  public static final char[] DBG_SPACES = (
    //         11111111112222222222333333333344444444445555555555666666666677777777778
    //12345678901234567890123456789012345678901234567890123456789012345678901234567890
    "                                                                                ").toCharArray ();

  public static final int DBG_DRP_VISIBLE_MASK = 1;  //レジスタウインドウが表示されている
  public static final int DBG_DDP_VISIBLE_MASK = 2;  //逆アセンブルリストウインドウが表示されている
  public static final int DBG_DMP_VISIBLE_MASK = 4;  //メモリダンプウインドウが表示されている
  public static final int DBG_BLG_VISIBLE_MASK = 8;  //分岐ログが表示されている
  public static final int DBG_PFV_VISIBLE_MASK = 16;  //プログラムフロービジュアライザが表示されている
  public static final int DBG_RBP_VISIBLE_MASK = 32;  //ラスタブレークポイントウインドウが表示されている
  public static final int DBG_DBP_VISIBLE_MASK = 64;  //データブレークポイントウインドウが表示されている
  public static final int DBG_SMT_VISIBLE_MASK = 128;  //表示モードテストが表示されている
  public static final int DBG_ATW_VISIBLE_MASK = 256;  //アドレス変換ウインドウが表示されている
  public static final int DBG_PAA_VISIBLE_MASK = 512;  //物理アドレス空間ウインドウが表示されている
  public static final int DBG_RTL_VISIBLE_MASK = 1024;  //ルートポインタリストが表示されている
  public static final int DBG_SPV_VISIBLE_MASK = 2048;  //スプライトパターンビュアが表示されている
  public static final int DBG_ACM_VISIBLE_MASK = 4096;  //アドレス変換キャッシュモニタが表示されている
  public static int dbgVisibleMask;  //表示されているデバッグ関連ウインドウのマスク

  //dbgInit ()
  //  初期化
  public static void dbgInit () {
    dbgVisibleMask = 0;
    dbgHexSelected = false;
    dbgHexValue = 0;
    dbgSupervisorMode = 1;
    dbgPopupMenu = null;
    dbgPopupDisMenuItem = null;
    dbgPopupMemMenuItem = null;
    dbgPopupCopyMenuItem = null;
    dbgPopupSelectAllMenuItem = null;
    dbgPopupIBPMenu = null;
    dbgPopupIBPCurrentModel = null;
    dbgPopupIBPCurrentValue = 0;
    dbgPopupIBPThresholdModel = null;
    dbgPopupIBPThresholdValue = 0;
    dbgPopupHexMenu = null;
    dbgPopupTextArea = null;
    dbgEventMask = 0;
    dbgStopOnError = false;  //ウインドウを表示する前にも必要なのでここで初期化すること
    if (DBG_ORI_BYTE_ZERO_D0) {
      dbgOriByteZeroD0 = false;
    }
  }  //dbgInit()

  //dbgMakePopup ()
  //  デバッグ関連ウインドウの共通コンポーネントを作る
  public static void dbgMakePopup () {

    //ポップアップメニュー
    ActionListener popupActionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        switch (ae.getActionCommand ()) {
        case "Disassemble":
          DisassembleList.ddpBacktraceRecord = -1L;  //分岐レコードの選択を解除する
          DisassembleList.ddpOpen (dbgHexValue, dbgSupervisorMode, false);
          break;
        case "Memory Dump":
          MemoryDumpList.dmpOpen (dbgHexValue, dbgSupervisorMode, false);
          break;
        case "Run to Here":
          if (InstructionBreakPoint.IBP_ON) {
            if (mpuTask == null) {
              InstructionBreakPoint.ibpInstant (DisassembleList.ddpPopupAddress, DisassembleList.ddpSupervisorMode);
              mpuStart ();
            }
          }
          break;
        case "Set Breakpoint":
          if (InstructionBreakPoint.IBP_ON) {
            InstructionBreakPoint.ibpPut (DisassembleList.ddpPopupAddress, DisassembleList.ddpSupervisorMode, dbgPopupIBPCurrentValue, dbgPopupIBPThresholdValue, null);
            DisassembleList.ddpOpen (0, DisassembleList.ddpSupervisorMode, true);
          }
          break;
        case "Clear Breakpoint":
          if (InstructionBreakPoint.IBP_ON) {
            InstructionBreakPoint.ibpRemove (DisassembleList.ddpPopupAddress, DisassembleList.ddpSupervisorMode);
            DisassembleList.ddpOpen (0, DisassembleList.ddpSupervisorMode, true);
          }
          break;
        case "Copy":
          dbgCopy ();
          break;
        case "Select All":
          dbgSelectAll ();
          break;
        }
      }
    };
    dbgPopupMenu = ComponentFactory.createPopupMenu (
      dbgPopupIBPMenu =
      InstructionBreakPoint.IBP_ON ?
      ComponentFactory.createMenu (
        "XXXXXXXX", KeyEvent.VK_UNDEFINED,
        Multilingual.mlnText (ComponentFactory.createMenuItem ("Run to Here", 'R', popupActionListener), "ja", "ここまで実行"),
        ComponentFactory.createHorizontalSeparator (),
        Multilingual.mlnText (ComponentFactory.createMenuItem ("Set Breakpoint", 'S', popupActionListener), "ja", "ブレークポイントを設定"),
        ComponentFactory.createHorizontalBox (
          Box.createHorizontalStrut (7),
          Box.createHorizontalGlue (),
          ComponentFactory.setPreferredSize (
            Multilingual.mlnText (ComponentFactory.createLabel ("current"), "ja", "現在値"),
            60, 16),
          ComponentFactory.createNumberSpinner (dbgPopupIBPCurrentModel = new SpinnerNumberModel (0, 0, 0x7fffffff, 1), 10, new ChangeListener () {
            @Override public void stateChanged (ChangeEvent ce) {
              dbgPopupIBPCurrentValue = dbgPopupIBPCurrentModel.getNumber ().intValue ();
            }
          }),
          Box.createHorizontalGlue ()
          ),
        ComponentFactory.createHorizontalBox (
          Box.createHorizontalStrut (7),
          Box.createHorizontalGlue (),
          ComponentFactory.setPreferredSize (
            Multilingual.mlnText (ComponentFactory.createLabel ("threshold"), "ja", "閾値"),
            60, 16),
          ComponentFactory.createNumberSpinner (dbgPopupIBPThresholdModel = new SpinnerNumberModel (0, 0, 0x7fffffff, 1), 10, new ChangeListener () {
            @Override public void stateChanged (ChangeEvent ce) {
              dbgPopupIBPThresholdValue = dbgPopupIBPThresholdModel.getNumber ().intValue ();
            }
          }),
          Box.createHorizontalGlue ()
          ),
        dbgPopupIBPClearMenuItem =
        Multilingual.mlnText (ComponentFactory.createMenuItem ("Clear Breakpoint", 'C', popupActionListener), "ja", "ブレークポイントを消去")
        ) :
      null,
      dbgPopupHexMenu =
      ComponentFactory.createMenu (
        "XXXXXXXX", KeyEvent.VK_UNDEFINED,
        dbgPopupDisMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Disassemble", 'D', popupActionListener), "ja", "逆アセンブル"),
        dbgPopupMemMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Memory Dump", 'M', popupActionListener), "ja", "メモリダンプ")
        ),
      dbgPopupCopyMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Copy", 'C', popupActionListener), "ja", "コピー"),
      dbgPopupSelectAllMenuItem = Multilingual.mlnText (ComponentFactory.createMenuItem ("Select All", 'A', popupActionListener), "ja", "すべて選択")
      );

  }  //dbgMakePopup()

  //dbgShowPopup (me, textArea, dis)
  //  ポップアップメニューを表示する
  public static void dbgShowPopup (MouseEvent me, JTextArea textArea, boolean dis) {
    dbgEventMask++;
    int x = me.getX ();
    int y = me.getY ();
    //int p = textArea.viewToModel (me.getPoint ());  //クリックされた位置。viewToModel2Dは9から
    int p = textArea.viewToModel2D (me.getPoint ());  //クリックされた位置。viewToModel2Dは9から
    DisassembleList.ddpPopupAddress = -1;  //クリックされた行のアドレス
    if (dis) {
      int i = Arrays.binarySearch (DisassembleList.ddpSplitArray, 1, DisassembleList.ddpItemCount, p + 1);
      i = (i >> 31 ^ i) - 1;  //クリックされた項目の番号
      DisassembleList.ddpPopupAddress = DisassembleList.ddpAddressArray[i];  //クリックされた項目の先頭アドレス
    }
    int start = textArea.getSelectionStart ();  //選択範囲の開始位置
    int end = textArea.getSelectionEnd ();  //選択範囲の終了位置
    String text = textArea.getText ();  //テキスト全体
    int length = text.length ();  //テキスト全体の長さ
    if ((start == end ||  //選択範囲がないか
         p < start || end <= p) &&  //選択範囲の外側がクリックされて
        0 <= p && p < length && isWord (text.charAt (p))) {  //クリックされた位置に単語があるとき
      //クリックされた位置にある単語を選択する
      for (start = p; 0 < start && isWord (text.charAt (start - 1)); start--) {
      }
      for (end = p + 1; end < length && isWord (text.charAt (end)); end++) {
      }
      textArea.select (start, end);
    }
    dbgHexSelected = false;
    if (start < end) {  //選択範囲があるとき
      textArea.requestFocusInWindow ();  //フォーカスがないと選択範囲が見えない
      //選択範囲にある16進数の文字を取り出す
      //  以下の条件を加える
      //    選択範囲に16進数以外の単語の文字がないこと
      //    選択範囲に16進数の文字が9文字以上ないこと
      //    16進数の文字が偶数文字ずつの塊になっていること
      dbgHexValue = 0;
      int n = 0;
      for (int i = start; i < end; i++) {
        int t;
        if ((t = Character.digit (text.charAt (i), 16)) >= 0) {  //16進数の文字
          dbgHexValue = dbgHexValue << 4 | t;
          if (n >= 8 ||  //選択範囲に16進数の文字が9文字以上ある
              i + 1 >= end || (t = Character.digit (text.charAt (i + 1), 16)) < 0) {  //16進数の文字が偶数文字ずつの塊になっていない
            n = 0;
            break;
          }
          dbgHexValue = dbgHexValue << 4 | t;
          n += 2;
          i++;
        } else if (isWord (text.charAt (i))) {  //16進数以外の単語の文字
          n = 0;
          break;
        }
      }
      dbgHexSelected = n > 0;
      try {
        //Rectangle r = textArea.modelToView (start).getBounds ();  //modelToView2Dは9から
        Rectangle r = textArea.modelToView2D (start).getBounds ();  //modelToView2Dは9から
        //Rectangle s = textArea.modelToView (end - 1).getBounds ();  //modelToView2Dは9から
        Rectangle s = textArea.modelToView2D (end - 1).getBounds ();  //modelToView2Dは9から
        if (r.y == s.y) {  //選択範囲が1行だけのとき
          //選択範囲を隠してしまわないようにポップアップを選択範囲の下側に表示する
          y = r.y + r.height;
        }
      } catch (BadLocationException ble) {
      }
    }
    //逆アセンブルリストでコアが止まっていて選択範囲がなくてクリックされた行のアドレスがわかるとき命令ブレークポイントメニューが有効
    if (InstructionBreakPoint.IBP_ON) {
      if (dis && mpuTask == null && DisassembleList.ddpPopupAddress != -1) {
        ComponentFactory.setText (dbgPopupIBPMenu, fmtHex8 (DisassembleList.ddpPopupAddress));
        TreeMap<Integer,InstructionBreakPoint.InstructionBreakRecord> pointTable = DisassembleList.ddpSupervisorMode != 0 ? InstructionBreakPoint.ibpSuperPointTable : InstructionBreakPoint.ibpUserPointTable;
        InstructionBreakPoint.InstructionBreakRecord r = pointTable.get (DisassembleList.ddpPopupAddress);
        if (r != null) {  //命令ブレークポイントがあるとき
          dbgPopupIBPCurrentModel.setValue (Integer.valueOf (dbgPopupIBPCurrentValue = r.ibrValue));  //現在値
          dbgPopupIBPThresholdModel.setValue (Integer.valueOf (dbgPopupIBPThresholdValue = r.ibrThreshold));  //閾値
          dbgPopupIBPClearMenuItem.setEnabled (true);  //消去できる
        } else {  //命令ブレークポイントがないとき
          dbgPopupIBPCurrentModel.setValue (Integer.valueOf (dbgPopupIBPCurrentValue = 0));  //現在値
          dbgPopupIBPThresholdModel.setValue (Integer.valueOf (dbgPopupIBPThresholdValue = 0));  //閾値
          dbgPopupIBPClearMenuItem.setEnabled (false);  //消去できない
        }
        ComponentFactory.setVisible (dbgPopupIBPMenu, true);
      } else {
        ComponentFactory.setVisible (dbgPopupIBPMenu, false);
      }
    }
    //16進数が選択されていれば16進数メニューが有効
    if (dbgHexSelected) {
      ComponentFactory.setText (dbgPopupHexMenu, fmtHex8 (dbgHexValue));
      ComponentFactory.setVisible (dbgPopupHexMenu, true);
    } else {
      ComponentFactory.setVisible (dbgPopupHexMenu, false);
    }
    //選択範囲があればコピーが有効
    ComponentFactory.setEnabled (dbgPopupCopyMenuItem, clpClipboard != null && start < end);
    //クリップボードがあればすべて選択が有効
    ComponentFactory.setEnabled (dbgPopupSelectAllMenuItem, clpClipboard != null);
    //ポップアップメニューを表示する
    dbgPopupTextArea = textArea;
    dbgPopupMenu.show (textArea, x, y);
    dbgEventMask--;
  }  //dbgShowPopup(MouseEvent,JTextArea,boolean)

  public static boolean isWord (char c) {
    return '0' <= c && c <= '9' || 'A' <= c && c <= 'Z' || 'a' <= c && c <= 'z' || c == '_';
  }  //isWord(char)

  //dbgCopy ()
  //  コピー
  public static void dbgCopy () {
    if (clpClipboard != null) {
      //選択範囲の文字列をコピーする
      String selectedText = dbgPopupTextArea.getSelectedText ();
      if (selectedText != null) {
        clpClipboardString = selectedText;
        try {
          clpClipboard.setContents (clpStringContents, clpClipboardOwner);
          clpIsClipboardOwner = true;  //自分がコピーした
        } catch (Exception e) {
          return;
        }
      }
    }
  }  //dbgCopy()

  //dbgSelectAll ()
  //  すべて選択
  public static void dbgSelectAll () {
    if (clpClipboard != null) {
      //すべて選択する
      dbgEventMask++;
      dbgPopupTextArea.selectAll ();
      dbgPopupTextArea.requestFocusInWindow ();
      dbgEventMask--;
    }
  }  //dbgSelectAll()

  //dbgUpdate ()
  //  デバッグウインドウを更新する
  //  コアのrun()の末尾でdbgVisibleMask!=0のとき呼び出す
  public static void dbgUpdate () {
    if ((dbgVisibleMask & DBG_DRP_VISIBLE_MASK) != 0) {
      RegisterList.drpUpdate ();  //レジスタウインドウを更新する
    }
    if (ProgramFlowVisualizer.PFV_ON) {
      if ((dbgVisibleMask & DBG_PFV_VISIBLE_MASK) != 0) {
        if (ProgramFlowVisualizer.pfvTimer == 0) {
          ProgramFlowVisualizer.pfvUpdate ();  //プログラムフロービジュアライザを更新する
        } else {
          ProgramFlowVisualizer.pfvTimer--;
        }
      }
    }
    if (RasterBreakPoint.RBP_ON) {
      if ((dbgVisibleMask & DBG_RBP_VISIBLE_MASK) != 0) {
        if (RasterBreakPoint.rbpTimer == 0) {
          RasterBreakPoint.rbpUpdateFrame ();  //ラスタブレークポイントウインドウを更新する
        } else {
          RasterBreakPoint.rbpTimer--;
        }
      }
    }
    if (ScreenModeTest.SMT_ON) {
      if ((dbgVisibleMask & DBG_SMT_VISIBLE_MASK) != 0) {
        if (ScreenModeTest.smtTimer == 0) {
          ScreenModeTest.smtUpdateFrame ();  //表示モードテストウインドウを更新する
        } else {
          ScreenModeTest.smtTimer--;
        }
      }
    }
    if (RootPointerList.RTL_ON) {
      if ((dbgVisibleMask & DBG_RTL_VISIBLE_MASK) != 0) {
        if (RootPointerList.rtlTimer == 0) {
          RootPointerList.rtlTimer = RootPointerList.RTL_INTERVAL - 1;
          RootPointerList.rtlUpdateFrame ();  //ルートポインタリストを更新する
        } else {
          RootPointerList.rtlTimer--;
        }
      }
    }
    if (SpritePatternViewer.SPV_ON) {
      if ((dbgVisibleMask & DBG_SPV_VISIBLE_MASK) != 0) {
        if (SpritePatternViewer.spvTimer == 0) {
          SpritePatternViewer.spvTimer = SpritePatternViewer.SPV_INTERVAL - 1;
          SpritePatternViewer.spvUpdateFrame ();  //スプライトパターンビュアを更新する
        } else {
          SpritePatternViewer.spvTimer--;
        }
      }
    }
    if (ATCMonitor.ACM_ON) {
      if ((dbgVisibleMask & DBG_ACM_VISIBLE_MASK) != 0) {
        if (ATCMonitor.acmTimer == 0) {
          ATCMonitor.acmTimer = ATCMonitor.ACM_INTERVAL - 1;
          ATCMonitor.acmUpdateFrame ();  //アドレス変換キャッシュモニタを更新する
        } else {
          ATCMonitor.acmTimer--;
        }
      }
    }
  }  //dbgUpdate()

  //dbgDoStopOnError ()
  //  エラーで停止する
  //  エラーを検出して例外スタックフレームを構築した後にdbgStopOnErrorならば呼び出してコアを停止させる
  //
  //  Human68kの_BUS_ERRの中では停止させない
  //  human302のシステムディスクで起動した場合、
  //  SCSIボード、RS-232Cボード、MIDIボードのテストで_BUS_ERRが呼び出されてバスエラーが発生する
  //    bus error on reading from 00EA0044 at 0000E2F4
  //    bus error on reading from 00EAFC04 at 0002D04A
  //    bus error on reading from 00EAFC14 at 0002D04A
  //    bus error on reading from 00EAFA01 at 0005CD54
  //  _BUS_ERRはレベル0で入ったDOSコールの番号を更新しないので、_BUS_ERRの中かどうかはpcで判断する
  //    0x0000e342 <= pc0 && pc0 < 0x0000e3b6  human200/human201の_BUS_ERR
  //    0x0000e3c8 <= pc0 && pc0 < 0x0000e43c  human202の_BUS_ERR
  //    0x0000e1a8 <= pc0 && pc0 < 0x0000e21c  human203の_BUS_ERR
  //    0x0000e256 <= pc0 && pc0 < 0x0000e2ca  human215の_BUS_ERR
  //    0x0000e174 <= pc0 && pc0 < 0x0000e1e8  human301の_BUS_ERR
  //    0x0000e28a <= pc0 && pc0 < 0x0000e2fe  human302の_BUS_ERR
  public static boolean dbgDoStopOnError () {
    String message = (
      M68kException.m6eNumber < 0 ?
      fmtHex8 (new StringBuilder ("breaked").append (" at "), regPC0).toString () :
      M68kException.m6eNumber <= M68kException.M6E_ADDRESS_ERROR ?
      fmtHex8 (fmtHex8 (new StringBuilder ("ERROR: ").append (M68kException.M6E_ERROR_NAME[M68kException.m6eNumber])
                        .append (M68kException.m6eDirection == 0 ? " on writing to " : " on reading from "), M68kException.m6eAddress)
               .append (" at "), regPC0).toString () :
      fmtHex8 (new StringBuilder (M68kException.M6E_ERROR_NAME[M68kException.m6eNumber])
               .append (" at "), regPC0).toString ()
      );
    System.out.println (message);
    if (!(M68kException.m6eNumber == M68kException.M6E_ACCESS_FAULT &&
          0x0000e100 <= regPC0 && regPC0 < 0x0000e500)) {  //_BUS_ERRの中で発生したバスエラーでないとき
      mpuStop (message);
      return true;
    }
    return false;
  }  //dbgDoStopOnError()

  //dbgDoubleBusFault ()
  //  ダブルバスフォルト
  public static void dbgDoubleBusFault () {
    String message =
      fmtHex8 (fmtHex8 (new StringBuilder ("FATAL ERROR: ").append (M68kException.M6E_ERROR_NAME[M68kException.m6eNumber])
                        .append (M68kException.m6eDirection == 0 ? " on writing to " : " on reading from "), M68kException.m6eAddress)
               .append (" at "), regPC0).toString ();
    System.out.println (message);
    mpuStop (message);
  }  //dbgDoubleBusFault()



  //========================================================================================
  //$$RSC ResourceFile

  public static final HashMap<String,byte[]> rscResourceCache = new HashMap<String,byte[]> ();

  //array = rscGetResource (name, sizes...)
  //  リソースファイルを読み込む。null=読み込めなかった
  //  ファイル名をコンマで区切る機能はない。XEiJ.jarの中なので必要ない
  //  *.zipや*.gzは非対応。XEiJ.jarの中なので必要ない
  //  キャッシュした配列をそのまま返すので書き換えないように注意すること
  public static byte[] rscGetResource (String name, int... sizes) {
    byte[] array = rscResourceCache.get (name);  //キャッシュから出す
    if (array != null) {  //読み込み済み
      return array;
    }
    array = new byte[1024 * 64];  //最初は64KB
    int size = 0;  //これまでに読み込んだ長さ
    try (BufferedInputStream bis = new BufferedInputStream (XEiJ.class.getResourceAsStream (name))) {
      for (;;) {
        if (size == array.length) {  //配列が一杯
          byte[] newArray = new byte[array.length * 2];  //2倍に伸ばした配列を作る
          System.arraycopy (array, 0, newArray, 0, size);  //読み込んだ部分を新しい配列に移す
          array = newArray;  //新しい配列に移行する
        }
        int step = bis.read (array, size, array.length - size);  //続きを読み込む。今回読み込んだ長さ
        if (step == -1) {  //終わり
          break;
        }
        size += step;  //これまでに読み込んだ長さ
      }
      if (size < array.length) {  //配列が長すぎる
        byte[] newArray = new byte[size];  //切り詰めた配列を作る
        System.arraycopy (array, 0, newArray, 0, size);  //読み込んだ部分を新しい配列に移す
        array = newArray;  //新しい配列に移行する
      }
      boolean fit = sizes.length == 0;
      if (!fit) {
        for (int i = 0; i < sizes.length; i++) {
          if (size == sizes[i]) {  //サイズが合っている
            fit = true;
            break;
          }
        }
      }
      if (fit) {
        System.out.println (Multilingual.mlnJapanese ?
                            name + " を読み込みました" :
                            name + " was read");
        rscResourceCache.put (name, array);  //キャッシュに入れる
        return array;  //try-with-resourcesなのでここでクローズされる
      }
      System.out.println (Multilingual.mlnJapanese ?
                          name + " のサイズが違います" :
                          name + " has wrong size");
      return null;  //try-with-resourcesなのでここでクローズされる
    } catch (IOException ioe) {
    }
    //try-with-resourcesなのでここでクローズされる
    System.out.println (Multilingual.mlnJapanese ?
                        name + " を読み込めません" :
                        name + " cannot be read");
    return null;
  }

  //string = rscGetResourceText (name, charset)
  //  リソースファイルからテキストを読み込む
  public static String rscGetResourceText (String name) {
    return rscGetResourceText (name, "UTF-8");
  }
  public static String rscGetResourceText (String name, String charset) {
    byte[] array = rscGetResource (name);
    if (name != null) {
      try {
        return new String (array, charset);
      } catch (UnsupportedEncodingException uee) {
      }
    }
    return "";
  }

  public static final Pattern RSC_ZIP_SEPARATOR = Pattern.compile ("(?<=\\.(?:jar|zip))(?:/|\\\\)(?=.)", Pattern.CASE_INSENSITIVE);
  public static String rscLastFileName = null;  //最後に読み込んだファイル名

  //array = rscGetFile (names, sizes...)
  //  ファイルを読み込む。null=読み込めなかった
  //  コンマで区切られたファイルを順に探して最初に読み込めたものを返す
  //  *.zip/entryを指定するとエントリを読み込む
  //  *.gzを指定すると解凍する
  public static byte[] rscGetFile (String names, int... sizes) {
    for (String name : names.split (",")) {  //コンマで区切る
      name = name.trim ();  //前後の空白を削除する
      if (name.length () == 0 || name.equalsIgnoreCase ("none")) {
        continue;
      }
      String[] zipSplittedName = RSC_ZIP_SEPARATOR.split (name, 2);  //*.zip/entryを*.zipとentryに分ける
      InputStream is = null;
      if (zipSplittedName.length < 2) {  //*.zip/entryではない
        File file = new File (name);
        if (file.isFile ()) {  //ファイルがある
          try {
            is = new FileInputStream (file);  //ファイルを開く
          } catch (IOException ioe) {
          }
        } else {  //ファイルがない
          System.out.println (Multilingual.mlnJapanese ?
                              name + " がありません" :
                              name + " does not exist");
          continue;
        }
      } else {  //*.zip/entry
        String zipName = zipSplittedName[0];  //*.zip
        String entryName = zipSplittedName[1];  //entry
        if (new File (zipName).isFile ()) {  //ファイルがある
          try {
            ZipFile zipFile = new ZipFile (zipName);  //ファイル
            ZipEntry zipEntry = zipFile.getEntry (entryName);  //エントリ
            if (zipEntry != null) {  //エントリがある
              is = zipFile.getInputStream (zipEntry);  //エントリを開く
            } else {  //エントリがない
              System.out.println (Multilingual.mlnJapanese ?
                                  zipName + " に " + zipEntry + " がありません" :
                                  zipName + " does not include " + zipEntry);
            }
          } catch (IOException ioe) {
          }
        } else {  //ファイルがない
          System.out.println (Multilingual.mlnJapanese ?
                              zipName + " がありません" :
                              zipName + " does not exist");
          continue;
        }
      }
      if (is != null) {  //開けた
        try {
          is = new BufferedInputStream (is);
          if (name.toLowerCase ().endsWith (".gz")) {  //*.gz
            is = new GZIPInputStream (is);  //展開する
          }
          byte[] array = new byte[1024 * 64];  //最初は64KB
          int size = 0;  //これまでに読み込んだ長さ
          for (;;) {
            if (size == array.length) {  //配列が一杯
              byte[] newArray = new byte[array.length * 2];  //2倍に伸ばした配列を作る
              System.arraycopy (array, 0, newArray, 0, size);  //読み込んだ部分を新しい配列に移す
              array = newArray;  //新しい配列に移行する
            }
            int step = is.read (array, size, array.length - size);  //続きを読み込む。今回読み込んだ長さ
            if (step == -1) {  //終わり
              break;
            }
            size += step;  //これまでに読み込んだ長さ
          }
          is.close ();  //ここはtry-with-resourcesではないので明示的にクローズすること
          is = null;
          if (size < array.length) {  //配列が長すぎる
            byte[] newArray = new byte[size];  //切り詰めた配列を作る
            System.arraycopy (array, 0, newArray, 0, size);  //読み込んだ部分を新しい配列に移す
            array = newArray;  //新しい配列に移行する
          }
          boolean fit = sizes.length == 0;
          if (!fit) {
            for (int i = 0; i < sizes.length; i++) {
              if (size == sizes[i]) {  //サイズが合っている
                fit = true;
                break;
              }
            }
          }
          if (fit) {
            System.out.println (Multilingual.mlnJapanese ?
                                name + " を読み込みました" :
                                name + " was read");
            rscLastFileName = name;  //最後に読み込んだファイル名
            return array;  //配列を返す
          }
          System.out.println (Multilingual.mlnJapanese ?
                              name + " のサイズが違います" :
                              name + " has wrong size");
          continue;
        } catch (IOException ioe) {
        }
        if (is != null) {
          try {
            is.close ();  //ここはtry-with-resourcesではないので明示的にクローズすること
            is = null;
          } catch (IOException ioe) {
          }
        }
      }  //if 開けた
      System.out.println (Multilingual.mlnJapanese ?
                          name + " を読み込めません" :
                          name + " cannot be read");
    }  //for name
    //読み込めなかった
    //ファイル名が指定されたときはメッセージが表示される
    //ファイル名が指定されなかったときはエラーメッセージも表示されない
    return null;
  }

  //string = rscGetTextFile (name)
  //string = rscGetTextFile (name, charset)
  //  テキストファイルを読み込む
  public static String rscGetTextFile (String name) {
    return rscGetTextFile (name, "UTF-8");
  }
  public static String rscGetTextFile (String name, String charset) {
    byte[] array = rscGetFile (name);
    if (array != null) {
      try {
        return new String (array, charset);
      } catch (UnsupportedEncodingException uee) {
      }
    }
    return "";
  }

  //mask = rscShowError (message, mask)
  //  ファイル操作に失敗したときに表示するダイアログ
  public static final int RSC_A_MASK = 1;  //中止
  public static final int RSC_R_MASK = 2;  //再実行
  public static final int RSC_I_MASK = 4;  //無視
  public static final String RSC_A_EN = "Abort";
  public static final String RSC_R_EN = "Retry";
  public static final String RSC_I_EN = "Ignore";
  public static final String RSC_A_JA = "中止";
  public static final String RSC_R_JA = "再実行";
  public static final String RSC_I_JA = "無視";
  public static final String[][] RSC_EN_OPTIONS = {
    { RSC_A_EN                     },
    { RSC_A_EN                     },
    {           RSC_R_EN           },
    { RSC_A_EN, RSC_R_EN           },
    {                     RSC_I_EN },
    { RSC_A_EN,           RSC_I_EN },
    {           RSC_R_EN, RSC_I_EN },
    { RSC_A_EN, RSC_R_EN, RSC_I_EN },
  };
  public static final String[][] RSC_JA_OPTIONS = {
    { RSC_A_JA                     },
    { RSC_A_JA                     },
    {           RSC_R_JA           },
    { RSC_A_JA, RSC_R_JA           },
    {                     RSC_I_JA },
    { RSC_A_JA,           RSC_I_JA },
    {           RSC_R_JA, RSC_I_JA },
    { RSC_A_JA, RSC_R_JA, RSC_I_JA },
  };
  public static int rscShowError (String message, int mask) {
    System.out.println (message);
    mask &= RSC_A_MASK | RSC_R_MASK | RSC_I_MASK;
    if (mask == 0) {
      mask = RSC_A_MASK;
    }
    String[] options = (Multilingual.mlnJapanese ? RSC_JA_OPTIONS : RSC_EN_OPTIONS)[mask];
    int def = Integer.numberOfTrailingZeros (mask);  //デフォルトの選択肢。0,1,2。中止、再実行、無視の順で最初に見つかったもの
    int bit = JOptionPane.showOptionDialog (
      null,  //parentComponent
      message,  //message
      Multilingual.mlnJapanese ? "ファイル操作エラー" : "File operation error",  //title
      JOptionPane.YES_NO_CANCEL_OPTION,  //optionType
      JOptionPane.ERROR_MESSAGE,  //messageType
      null,  //icon
      options,  //options
      options[def]);  //initialValue
    if (bit == JOptionPane.CLOSED_OPTION) {  //閉じた。-1
      bit = def;
    }
    return 1 << bit;
  }

  //success = rscPutTextFile (name, string)
  //  テキストファイルに書き出す
  //  同じ内容のファイルが既にあるときは何もしない
  //  *.zip/entryは非対応
  public static boolean rscPutTextFile (String name, String s) {
    try {
      return rscPutFile (name, s.getBytes ("UTF-8"));
    } catch (UnsupportedEncodingException uee) {
    }
    return false;
  }

  //success = rscPutFile (name, array)
  //success = rscPutFile (name, array, offset, length)
  //success = rscPutFile (name, array, offset, length, longLength2)
  //  ファイルに書き出す
  //  nameにarray[offset..offset+length-1]とlongLength2-length個の0を書き出す
  //  親ディレクトリがなければ作る
  //  同じ内容のファイルが既にあるときは更新しない
  //  *.zip/entryは非対応
  public static boolean rscPutFile (String name, byte[] array) {
    return rscPutFile (name, array, 0, array.length, (long) array.length);
  }
  public static boolean rscPutFile (String name, byte[] array, int offset, int length) {
    return rscPutFile (name, array, offset, length, (long) length);
  }
  public static boolean rscPutFile (String name, byte[] array, int offset, int length, long longLength2) {
    if (RSC_ZIP_SEPARATOR.matcher (name).matches ()) {  //*.zip/entry
      //*.zip/entryのとき確実に弾かないとファイルを破壊するおそれがある
      return false;
    }
    File file = new File (name);
    boolean fileExists = file.isFile ();  //true=同じ名前のファイルがある
    if (fileExists && file.length () == longLength2) {  //同じ名前で同じ長さのファイルがある
    comparison:
      {
        try (BufferedInputStream bis = new BufferedInputStream (new FileInputStream (file))) {
          byte[] buffer = new byte[(int) Math.min (Math.max ((long) length, longLength2 - (long) length), (long) (1024 * 1024))];  //最大1MBずつ読み込んで比較する
          int position = 0;
          while (position < length) {
            int step = bis.read (buffer, 0, Math.min (buffer.length, length - position));
            if (step == -1) {  //足りない。長さを確認してから始めたのだから途中で終わるはずがない
              break comparison;
            }
            int offsetPosition = offset + position;
            for (int i = 0; i < step; i++) {
              if (buffer[i] != array[offsetPosition + i]) {  //一致しない
                break comparison;
              }
            }
            position += step;
          }
          long longPosition2 = (long) length;
          while (longPosition2 < longLength2) {
            int step = bis.read (buffer, 0, (int) Math.min ((long) buffer.length, longLength2 - longPosition2));
            if (step == -1) {  //足りない。長さを確認してから始めたのだから途中で終わるはずがない
              break comparison;
            }
            for (int i = 0; i < step; i++) {
              if (buffer[i] != 0) {  //一致しない
                break comparison;
              }
            }
            longPosition2 += (long) step;
          }
          return true;  //一致した
        } catch (IOException ioe) {
        }
      }  //match
    }  //if 同じ名前で同じ長さのファイルがある
    String nameTmp = name + ".tmp";
    File fileTmp = new File (nameTmp);
    String nameBak = name + ".bak";
    File fileBak = new File (nameBak);
  retry:
    for (;;) {
      File parentDirectory = file.getParentFile ();  //親ディレクトリ
      if (parentDirectory != null && !parentDirectory.isDirectory ()) {  //親ディレクトリがない
        String parentName = parentDirectory.getPath ();
        if (parentDirectory.mkdirs ()) {  //親ディレクトリを作る。必要ならば遡って作る。作れた
          System.out.println (Multilingual.mlnJapanese ?
                              parentName + " を作りました" :
                              parentName + " was created");
        } else {  //作れなかった
          switch (rscShowError (Multilingual.mlnJapanese ?
                                parentName + " を作れません" :
                                parentName + " cannot be created",
                                RSC_A_MASK | RSC_R_MASK | RSC_I_MASK)) {
          case RSC_A_MASK:
            break retry;
          case RSC_R_MASK:
            continue retry;
          }
        }
      }
      if (fileTmp.isFile ()) {  //name.tmpがある。前回異常終了して残骸が残っている場合など
        if (!fileTmp.delete ()) {  //name.tmpを削除する。削除できない
          switch (rscShowError (Multilingual.mlnJapanese ?
                                nameTmp + " を削除できません" :
                                nameTmp + " cannot be deleted",
                                RSC_A_MASK | RSC_R_MASK | RSC_I_MASK)) {
          case RSC_A_MASK:
            break retry;
          case RSC_R_MASK:
            continue retry;
          }
        }
      }
      try (OutputStream os = name.toLowerCase ().endsWith (".gz") ?
           new GZIPOutputStream (new BufferedOutputStream (new FileOutputStream (fileTmp))) {
             {
               //def.setLevel (Deflater.BEST_COMPRESSION);
               def.setLevel (Deflater.DEFAULT_COMPRESSION);
               //def.setLevel (Deflater.BEST_SPEED);
             }
           } :
           new BufferedOutputStream (new FileOutputStream (fileTmp))) {  //name.tmpに書き出す
        //array[offset..offset+length-1]を書き出す
        os.write (array, offset, length);
        //longLength2-length個の0を書き出す
        //  RandomAccessFile.setLength()は拡張部分の内容が定義されていないので使えない
        if ((long) length < longLength2) {
          byte[] buffer = new byte[(int) Math.min (longLength2 - (long) length, (long) (1024 * 1024))];  //最大1MBずつ書き出す
          Arrays.fill (buffer, 0, buffer.length, (byte) 0);  //念の為
          long longPosition2 = (long) length;
          while (longPosition2 < longLength2) {
            int step = (int) Math.min ((long) buffer.length, longLength2 - longPosition2);
            os.write (buffer, 0, step);
            longPosition2 += (long) step;
          }
        }
      } catch (IOException ioe) {
        switch (rscShowError (Multilingual.mlnJapanese ?
                              nameTmp + " に書き出せません" :
                              nameTmp + " cannot be written",
                              RSC_A_MASK | RSC_R_MASK | RSC_I_MASK)) {
        case RSC_A_MASK:
          break retry;
        case RSC_R_MASK:
          continue retry;
        }
      }
      if (fileExists && file.isFile ()) {  //同じ名前で内容が異なるファイルがある。リトライでなくなった可能性があるので確認し直す
        if (fileBak.isFile ()) {  //name.bakがある
          if (!fileBak.delete ()) {  //name.bakを削除する。削除できない
            switch (rscShowError (Multilingual.mlnJapanese ?
                                  nameBak + " を削除できません" :
                                  nameBak + " cannot be deleted",
                                  RSC_A_MASK | RSC_R_MASK | RSC_I_MASK)) {
            case RSC_A_MASK:
              break retry;
            case RSC_R_MASK:
              continue retry;
            }
          }
        }
        if (!file.renameTo (fileBak)) {  //nameをname.bakにリネームする。リネームできない
          switch (rscShowError (Multilingual.mlnJapanese ?
                                name + " を " + nameBak + " にリネームできません" :
                                name + " cannot be renamed to " + nameBak,
                                RSC_A_MASK | RSC_R_MASK | RSC_I_MASK)) {
          case RSC_A_MASK:
            break retry;
          case RSC_R_MASK:
            continue retry;
          }
        }
      }
      if (fileTmp.renameTo (file)) {  //name.tmpをnameにリネームする。リネームできた
        if (fileExists) {  //同じ名前のファイルがあった
          System.out.println (Multilingual.mlnJapanese ?
                              name + " を更新しました" :
                              name + " was updated");
        } else {  //同じ名前のファイルがなかった
          System.out.println (Multilingual.mlnJapanese ?
                              name + " を作りました" :
                              name + " was created");
        }
        return true;  //成功
      } else {  //リネームできない
        switch (rscShowError (Multilingual.mlnJapanese ?
                              nameTmp + " を " + name + " にリネームできません" :
                              nameTmp + " cannot be renamed to " + name,
                              RSC_A_MASK | RSC_R_MASK | RSC_I_MASK)) {
        case RSC_A_MASK:
          break retry;
        case RSC_R_MASK:
          continue retry;
        }
      }
      break;
    }  //retry
    if (fileExists) {  //同じ名前のファイルがあった
      System.out.println (Multilingual.mlnJapanese ?
                          name + " を更新できません" :
                          name + " cannot be updated");
    } else {  //同じ名前のファイルがなかった
      System.out.println (Multilingual.mlnJapanese ?
                          name + " を作れません" :
                          name + " cannot be created");
    }
    return false;  //失敗
  }



  //========================================================================================
  //$$ISM InputStream

  public static final Pattern ISM_ZIP_SEPARATOR = Pattern.compile ("(?<=\\.(?:jar|zip))(?:/|\\\\)(?=.)", Pattern.CASE_INSENSITIVE);

  //in = ismOpen (name)
  //  InputStreamを開く
  //  InputStreamを返す
  //    失敗したときはnullを返す
  //  ZIPファイルの中のファイルを指定できる
  //    ZIPファイルの中のファイルはZipInputStreamで開く
  //    ZIPファイルの中のファイル名は{ZIPファイル名}/{ZIPファイルの中のファイル名}で指定する
  //    JARファイルもZIPファイルとして開くことができる
  //  GZIPで圧縮されているファイルを指定できる
  //    GZIPで圧縮されているファイルはGZIPInputStreamで開く
  public static InputStream ismOpen (String name) {
    InputStream in = null;
    in = ismOpen (name, false);  //ファイルを開く
    if (in == null && name.indexOf ('/') < 0 && name.indexOf ('\\') < 0) {  //ファイルがないとき
      in = ismOpen (name, true);  //リソースを開く
    }
    return in;
  }  //ismOpen(String)
  public static InputStream ismOpen (String name, boolean useGetResource) {
    boolean gzipped = name.toLowerCase ().endsWith (".gz");  //true=GZIPファイルが指定された
    String[] zipSplittedName = ISM_ZIP_SEPARATOR.split (name, 2);  //ZIPファイル名とZIPファイルの中のファイル名に分ける
    String fileName = zipSplittedName[0];  //通常のファイル名またはZIPファイル名
    String zipEntryName = zipSplittedName.length < 2 ? null : zipSplittedName[1];  //ZIPファイルの中のファイル名
    InputStream in = null;
    try {
      if (useGetResource) {  //getResourceを使うとき
        if (false) {
          URL url = XEiJ.class.getResource (fileName);
          if (url != null) {  //ファイルがある
            in = url.openStream ();
          }
        } else {
          in = XEiJ.class.getResourceAsStream (fileName);
        }
      } else {
        File file = new File (fileName);
        if (file.exists ()) {  //ファイルがある
          in = new FileInputStream (file);
        }
      }
      if (in != null && zipEntryName != null) {  //ZIPファイルの中のファイルが指定されたとき
        ZipInputStream zin = new ZipInputStream (in);
        in = null;
        ZipEntry entry;
        while ((entry = zin.getNextEntry ()) != null) {  //指定されたファイル名のエントリを探す
          if (zipEntryName.equals (entry.getName ())) {  //エントリが見つかった
            in = zin;
            break;
          }
        }
        if (in == null) {
          System.out.println (Multilingual.mlnJapanese ? fileName + " の中に " + zipEntryName + " がありません" :
                              zipEntryName + " does not exist in " + fileName);
        }
      }
      if (in != null && gzipped) {  //GZIPで圧縮されたファイルが指定されたとき
        in = new GZIPInputStream (in);
      }
      if (in != null) {
        System.out.println (Multilingual.mlnJapanese ? (useGetResource ? "リソースファイル " : "ファイル ") + name + " を読み込みます" :
                            (useGetResource ? "Reading resource file " : "Reading file ") + name);
        return new BufferedInputStream (in);
      }
    } catch (Exception ioe) {
      if (prgVerbose) {
        prgPrintStackTraceOf (ioe);
      }
    }
    System.out.println (Multilingual.mlnJapanese ? (useGetResource ? "リソースファイル " : "ファイル ") + name + " が見つかりません" :
                        (useGetResource ? "Resource file " : "File ") + name + " is not found");
    return null;  //失敗
  }  //ismOpen(String,boolean)

  //k = ismRead (in, bb, o, l)
  //  InputStreamからバイトバッファに読み込む
  //  読み込んだ長さを返す
  //    エラーのときは-1を返す
  //  指定されたサイズまたはファイルの末尾まで読み込む
  //    k=in.read(bb,o,l)は1回で指定されたサイズを読み込めるとは限らない
  //  ブロックされる可能性があるのでコアの動作中にコアのスレッドから呼ばないほうがよい
  public static int ismRead (InputStream in, byte[] bb, int o, int l) {
    try {
      int k = 0;
      while (k < l) {
        int t = in.read (bb, o + k, l - k);
        if (t < 0) {
          break;
        }
        k += t;
      }
      return k;
    } catch (IOException ioe) {
      if (prgVerbose) {
        prgPrintStackTraceOf (ioe);
      }
    }
    return -1;
  }  //ismRead(InputStream,byte[],int,int)

  //k = ismSkip (in, l)
  //  InputStreamを読み飛ばす
  //  読み飛ばした長さを返す
  //    エラーのときは-1を返す
  //  指定されたサイズまたはファイルの末尾まで読み飛ばす
  //    k=in.skip(l)は1回で指定されたサイズを読み飛ばせるとは限らない
  //  ブロックされる可能性があるのでコアの動作中にコアのスレッドから呼ばないほうがよい
  public static int ismSkip (InputStream in, int l) {
    try {
      int k = 0;
      while (k < l) {
        //skip(long)はファイルの末尾でなくても0を返す可能性があるのでファイルの末尾の判定はread()で行う
        //skip(long)する前に毎回read()しないとskip()がファイルの末尾で止まらなくなるらしい
        if (in.read () < 0) {
          break;
        }
        k++;
        if (k < l) {
          int t = (int) in.skip ((long) (l - k));
          if (t < 0) {
            break;
          }
          k += t;
        }
      }
      return k;
    } catch (IOException ioe) {
      if (prgVerbose) {
        prgPrintStackTraceOf (ioe);
      }
    }
    return -1;
  }  //ismSkip(InputStream,int)

  //ismClose (in)
  //  InputStreamを閉じる
  //  in==nullのときは何もしない
  //  in.close()でIOExceptionを無視するだけ
  public static void ismClose (InputStream in) {
    try {
      if (in != null) {
        in.close ();
      }
    } catch (IOException ioe) {
      if (prgVerbose) {
        prgPrintStackTraceOf (ioe);
      }
    }
  }  //ismClose(InputStream)

  //length = ismLength (name, maxLength)
  //  ファイルの長さを数える
  //  ZIPファイルの中のファイルを指定できる
  //  GZIPで圧縮されているファイルを指定できる
  //  -1  ファイルがない
  public static int ismLength (String name, int maxLength) {
    int length;
    InputStream in = ismOpen (name);
    if (in == null) {  //ファイルがない
      length = -1;
    } else {  //ファイルがある
      length = ismSkip (in, maxLength);
      ismClose (in);
    }
    return length;
  }  //ismLength(String,int)

  //success = ismLoad (bb, o, l, names)
  //  ファイルからバイトバッファに読み込む
  //  ファイル名を,で区切って複数指定できる
  //    先頭から順に指定されたサイズまで読み込めるファイルを探す
  //    1つでも読み込むことができればその時点で成功、1つも読み込めなければ失敗
  //  成功したときtrueを返す
  //  ZIPファイルの中のファイルを指定できる
  //    ZIPファイルの中のファイルはZipInputStreamで開く
  //    ZIPファイルの中のファイル名は{ZIPファイル名}/{ZIPファイルの中のファイル名}で指定する
  //    JARファイルもZIPファイルとして開くことができる
  //  GZIPで圧縮されているファイルを指定できる
  //    GZIPで圧縮されているファイルはGZIPInputStreamで開く
  //  ブロックされることがあるのでコアの動作中にコアのスレッドから呼ばないほうがよい
  public static boolean ismLoad (byte[] bb, int o, int l, String names) {
    for (String name : names.split (",")) {  //先頭から順に
      if (name.length () != 0) {  //ファイル名が指定されているとき
        InputStream in = ismOpen (name);  //開く
        if (in != null) {  //開けたら
          int k = ismRead (in, bb, o, l);  //読み込んで
          ismClose (in);  //閉じる
          if (k == l) {  //指定されたサイズまで読み込めたら
            return true;  //成功
          }
        }
      }
    }
    return false;  //1つも読み込めなかったので失敗
  }  //ismLoad(byte[],int,int,String)

  //success = ismSave (bb, offset, length, path, verbose)
  //  バイトバッファからファイルに書き出す
  //  出力範囲がバッファに収まっているとき
  //    ファイルが既に存在していてファイルサイズと内容が一致しているときは更新しない
  //  出力範囲がバッファに収まっていないとき
  //    バッファからはみ出した部分をゼロクリアする
  //    RandomAccessFileのsetLengthを使うとファイルサイズを簡単に変更できるが、ファイルを伸ばしたときに書き込まれる内容が仕様で定義されていない
  //    同じ手順で同じ内容のファイルができない可能性があるのは困るので明示的に0を書き込んでクリアする
  public static boolean ismSave (byte[] bb, int offset, long length, String path, boolean verbose) {
    if (ISM_ZIP_SEPARATOR.split (path, 2).length != 1) {  //ZIPファイルの中のファイル
      if (verbose) {
        JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? path + " に書き出せません" : "Cannot write " + path);
      }
      return false;
    }
    long step = 0;  //一度にゼロクリアする長さ。0=バッファに収まっている
    byte[] zz = null;  //ゼロクリア用の配列
    long pointer = (long) (bb.length - offset);  //バッファから出力できる長さ
    if (pointer < length) {  //バッファに収まっていない
      step = Math.min (1024L * 512, length - pointer);  //一度にゼロクリアする長さ。最大512KB
      zz = new byte[(int) step];  //ゼロクリア用の配列
      Arrays.fill (zz, (byte) 0);
    }
    //ファイル
    File file = new File (path);  //ファイル
    //ファイルが既に存在しているときはファイルサイズと内容が一致しているかどうか確認する
    if (step == 0 &&  //バッファに収まっている
        file.exists () && file.length () == length) {  //ファイルサイズが一致している
      //ファイルを読み込む
      if (length == 0L) {  //ファイルサイズが0で一致しているので成功
        return true;
      }
      InputStream in = ismOpen (path);
      if (in != null) {
        int l = (int) length;  //バッファに収まっているのだからintの範囲内
        byte[] tt = new byte[l];
        int k = ismRead (in, tt, 0, l);
        ismClose (in);
        if (k == l &&
            Arrays.equals (tt, bb.length == l ? bb : Arrays.copyOfRange (bb, offset, offset + l))) {  //内容が一致している
          return true;  //更新する必要がないので成功
        }
      }
    }  //check
    //*.tmpと*.bak
    String pathTmp = path + ".tmp";  //*.tmp
    String pathBak = path + ".bak";  //*.bak
    File fileTmp = new File (pathTmp);  //*.tmp
    File fileBak = new File (pathBak);  //*.bak
    //*.tmpを削除する
    if (fileTmp.exists ()) {  //*.tmpがあるとき
      if (!fileTmp.delete ()) {  //*.tmpを削除する
        if (verbose) {
          JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? pathTmp + " を削除できません" : "Cannot delete " + pathTmp);
        }
        return false;
      }
    }
    //*.tmpに書き出す
    try (OutputStream out = path.toLowerCase ().endsWith (".gz") ?  //pathの末尾が".gz"のときpathTmpの末尾は".gz.tmp"であることに注意
         new GZIPOutputStream (new BufferedOutputStream (new FileOutputStream (fileTmp))) {
           {
             //def.setLevel (Deflater.BEST_COMPRESSION);
             def.setLevel (Deflater.DEFAULT_COMPRESSION);
             //def.setLevel (Deflater.BEST_SPEED);
           }
         } :
         new BufferedOutputStream (new FileOutputStream (fileTmp))) {  //try-with-resourcesは1.7から
      if (step == 0) {  //バッファに収まっている
        out.write (bb, offset, (int) length);  //OutputStreamのwriteの返り値はvoid。エラーが出なければ1回で最後まで書き出される
      } else {  //バッファに収まっていない
        out.write (bb, offset, (int) pointer);  //バッファから出力できる範囲
        for (; pointer < length; pointer += step) {
          out.write (zz, 0, (int) Math.min (step, length - pointer));  //バッファから出力できない範囲
        }
      }
    } catch (IOException ioe) {
      if (verbose) {
        prgPrintStackTraceOf (ioe);
        JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? pathTmp + " に書き出せません" : "Cannot write " + pathTmp);
      }
      return false;
    }
    //ファイルを*.bakにリネームする
    //  javaのFileのrenameToはPerlと違って上書きしてくれないので明示的に削除またはリネームする必要がある
    if (file.exists ()) {  //ファイルがあるとき
      if (fileBak.exists ()) {  //*.bakがあるとき
        if (!fileBak.delete ()) {  //*.bakを削除する
          if (verbose) {
            JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? pathBak + " を削除できません" : "Cannot delete " + pathBak);
          }
          return false;
        }
      }
      if (!file.renameTo (fileBak)) {  //ファイルを*.bakにリネームする
        if (verbose) {
          JOptionPane.showMessageDialog (
            null, Multilingual.mlnJapanese ? path + " を " + pathBak + " にリネームできません" : "Cannot rename " + path + " to " + pathBak);
        }
        return false;
      }
    }
    //*.tmpをファイルにリネームする
    //  javaのFileのrenameToはPerlと違って上書きしてくれないので明示的に削除またはリネームする必要がある
    if (!fileTmp.renameTo (file)) {  //*.tmpをファイルにリネームする
      if (verbose) {
        JOptionPane.showMessageDialog (
          null, Multilingual.mlnJapanese ? pathTmp + " を " + path + " にリネームできません" : "Cannot rename " + pathTmp + " to " + path);
      }
      return false;
    }
    return true;
  }  //ismSave(byte[],int,long,String,boolean)



  //========================================================================================
  //$$FMT フォーマット変換
  //  Formatterは遅いので自前で展開する

  public static final char[] FMT_TEMP = new char[32];

  //--------------------------------------------------------------------------------
  //2進数変換
  //  ainNは'.'と'*'、binNは'0'と'1'に変換する
  //
  //  x          00 01
  //  x<<2       00 04
  //  x<<2&4     00 04
  //  x<<2&4^46  2e 2a
  //              .  *

  public static final char[] FMT_AIN4_BASE = ".......*..*...**.*...*.*.**..****...*..**.*.*.****..**.****.****".toCharArray ();
  public static final char[] FMT_BIN4_BASE = "0000000100100011010001010110011110001001101010111100110111101111".toCharArray ();

  //fmtAin4 (a, o, x)
  //fmtBin4 (a, o, x)
  //s = fmtAin4 (x)
  //s = fmtBin4 (x)
  //sb = fmtAin4 (sb, x)
  //sb = fmtBin4 (sb, x)
  //  4桁2進数変換
  public static void fmtAin4 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>  1 & 4 ^ 46);
    a[o +  1] = (char) (x       & 4 ^ 46);
    a[o +  2] = (char) (x <<  1 & 4 ^ 46);
    a[o +  3] = (char) (x <<  2 & 4 ^ 46);
  }  //fmtAin4(char[],int,int)
  public static void fmtBin4 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>>  3 & 1 | 48);
    a[o +  1] = (char) (x >>>  2 & 1 | 48);
    a[o +  2] = (char) (x >>>  1 & 1 | 48);
    a[o +  3] = (char) (x        & 1 | 48);
  }  //fmtBin4(char[],int,int)
  public static String fmtAin4 (int x) {
    return String.valueOf (FMT_AIN4_BASE, (x & 15) << 2, 4);
  }  //fmtAin4(int)
  public static String fmtBin4 (int x) {
    return String.valueOf (FMT_BIN4_BASE, (x & 15) << 2, 4);
  }  //fmtBin4(int)
  public static StringBuilder fmtAin4 (StringBuilder sb, int x) {
    return sb.append (FMT_AIN4_BASE, (x & 15) << 2, 6);
  }  //fmtAin4(StringBuilder,int)
  public static StringBuilder fmtBin4 (StringBuilder sb, int x) {
    return sb.append (FMT_BIN4_BASE, (x & 15) << 2, 6);
  }  //fmtBin4(StringBuilder,int)

  //fmtAin6 (a, o, x)
  //fmtBin6 (a, o, x)
  //s = fmtAin6 (x)
  //s = fmtBin6 (x)
  //sb = fmtAin6 (sb, x)
  //sb = fmtBin6 (sb, x)
  //  6桁2進数変換
  public static void fmtAin6 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>  3 & 4 ^ 46);
    a[o +  1] = (char) (x >>  2 & 4 ^ 46);
    a[o +  2] = (char) (x >>  1 & 4 ^ 46);
    a[o +  3] = (char) (x       & 4 ^ 46);
    a[o +  4] = (char) (x <<  1 & 4 ^ 46);
    a[o +  5] = (char) (x <<  2 & 4 ^ 46);
  }  //fmtAin6(char[],int,int)
  public static void fmtBin6 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>>  5 & 1 | 48);
    a[o +  1] = (char) (x >>>  4 & 1 | 48);
    a[o +  2] = (char) (x >>>  3 & 1 | 48);
    a[o +  3] = (char) (x >>>  2 & 1 | 48);
    a[o +  4] = (char) (x >>>  1 & 1 | 48);
    a[o +  5] = (char) (x        & 1 | 48);
  }  //fmtBin6(char[],int,int)
  public static String fmtAin6 (int x) {
    FMT_TEMP[ 0] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x       & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x <<  2 & 4 ^ 46);
    return String.valueOf (FMT_TEMP, 0, 6);
  }  //fmtAin6(int)
  public static String fmtBin6 (int x) {
    FMT_TEMP[ 0] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x        & 1 | 48);
    return String.valueOf (FMT_TEMP, 0, 6);
  }  //fmtBin6(int)
  public static StringBuilder fmtAin6 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x       & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x <<  2 & 4 ^ 46);
    return sb.append (FMT_TEMP, 0, 6);
  }  //fmtAin6(StringBuilder,int)
  public static StringBuilder fmtBin6 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x        & 1 | 48);
    return sb.append (FMT_TEMP, 0, 6);
  }  //fmtBin6(StringBuilder,int)

  //fmtAin8 (a, o, x)
  //fmtBin8 (a, o, x)
  //s = fmtAin8 (x)
  //s = fmtBin8 (x)
  //sb = fmtAin8 (sb, x)
  //sb = fmtBin8 (sb, x)
  //  8桁2進数変換
  public static void fmtAin8 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>  5 & 4 ^ 46);
    a[o +  1] = (char) (x >>  4 & 4 ^ 46);
    a[o +  2] = (char) (x >>  3 & 4 ^ 46);
    a[o +  3] = (char) (x >>  2 & 4 ^ 46);
    a[o +  4] = (char) (x >>  1 & 4 ^ 46);
    a[o +  5] = (char) (x       & 4 ^ 46);
    a[o +  6] = (char) (x <<  1 & 4 ^ 46);
    a[o +  7] = (char) (x <<  2 & 4 ^ 46);
  }  //fmtAin8(char[],int,int)
  public static void fmtBin8 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>>  7 & 1 | 48);
    a[o +  1] = (char) (x >>>  6 & 1 | 48);
    a[o +  2] = (char) (x >>>  5 & 1 | 48);
    a[o +  3] = (char) (x >>>  4 & 1 | 48);
    a[o +  4] = (char) (x >>>  3 & 1 | 48);
    a[o +  5] = (char) (x >>>  2 & 1 | 48);
    a[o +  6] = (char) (x >>>  1 & 1 | 48);
    a[o +  7] = (char) (x        & 1 | 48);
  }  //fmtBin8(char[],int,int)
  public static String fmtAin8 (int x) {
    FMT_TEMP[ 0] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x       & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x <<  2 & 4 ^ 46);
    return String.valueOf (FMT_TEMP, 0, 8);
  }  //fmtAin8(int)
  public static String fmtBin8 (int x) {
    FMT_TEMP[ 0] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x        & 1 | 48);
    return String.valueOf (FMT_TEMP, 0, 8);
  }  //fmtBin8(int)
  public static StringBuilder fmtAin8 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x       & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x <<  2 & 4 ^ 46);
    return sb.append (FMT_TEMP, 0, 8);
  }  //fmtAin8(StringBuilder,int)
  public static StringBuilder fmtBin8 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x        & 1 | 48);
    return sb.append (FMT_TEMP, 0, 8);
  }  //fmtBin8(StringBuilder,int)

  //fmtAin12 (a, o, x)
  //fmtBin12 (a, o, x)
  //s = fmtAin12 (x)
  //s = fmtBin12 (x)
  //sb = fmtAin12 (sb, x)
  //sb = fmtBin12 (sb, x)
  //  12桁2進数変換
  public static void fmtAin12 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>  9 & 4 ^ 46);
    a[o +  1] = (char) (x >>  8 & 4 ^ 46);
    a[o +  2] = (char) (x >>  7 & 4 ^ 46);
    a[o +  3] = (char) (x >>  6 & 4 ^ 46);
    a[o +  4] = (char) (x >>  5 & 4 ^ 46);
    a[o +  5] = (char) (x >>  4 & 4 ^ 46);
    a[o +  6] = (char) (x >>  3 & 4 ^ 46);
    a[o +  7] = (char) (x >>  2 & 4 ^ 46);
    a[o +  8] = (char) (x >>  1 & 4 ^ 46);
    a[o +  9] = (char) (x       & 4 ^ 46);
    a[o + 10] = (char) (x <<  1 & 4 ^ 46);
    a[o + 11] = (char) (x <<  2 & 4 ^ 46);
  }  //fmtAin12(char[],int,int)
  public static void fmtBin12 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>> 11 & 1 | 48);
    a[o +  1] = (char) (x >>> 10 & 1 | 48);
    a[o +  2] = (char) (x >>>  9 & 1 | 48);
    a[o +  3] = (char) (x >>>  8 & 1 | 48);
    a[o +  4] = (char) (x >>>  7 & 1 | 48);
    a[o +  5] = (char) (x >>>  6 & 1 | 48);
    a[o +  6] = (char) (x >>>  5 & 1 | 48);
    a[o +  7] = (char) (x >>>  4 & 1 | 48);
    a[o +  8] = (char) (x >>>  3 & 1 | 48);
    a[o +  9] = (char) (x >>>  2 & 1 | 48);
    a[o + 10] = (char) (x >>>  1 & 1 | 48);
    a[o + 11] = (char) (x        & 1 | 48);
  }  //fmtBin12(char[],int,int)
  public static String fmtAin12 (int x) {
    FMT_TEMP[ 0] = (char) (x >>  9 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >>  8 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >>  7 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >>  6 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[ 8] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[ 9] = (char) (x       & 4 ^ 46);
    FMT_TEMP[10] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[11] = (char) (x <<  2 & 4 ^ 46);
    return String.valueOf (FMT_TEMP, 0, 12);
  }  //fmtAin12(int)
  public static String fmtBin12 (int x) {
    FMT_TEMP[ 0] = (char) (x >>> 11 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>> 10 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>>  9 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>>  8 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[ 8] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[ 9] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[10] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[11] = (char) (x        & 1 | 48);
    return String.valueOf (FMT_TEMP, 0, 12);
  }  //fmtBin12(int)
  public static StringBuilder fmtAin12 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>  9 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >>  8 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >>  7 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >>  6 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[ 8] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[ 9] = (char) (x       & 4 ^ 46);
    FMT_TEMP[10] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[11] = (char) (x <<  2 & 4 ^ 46);
    return sb.append (FMT_TEMP, 0, 12);
  }  //fmtAin12(StringBuilder,int)
  public static StringBuilder fmtBin12 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>> 11 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>> 10 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>>  9 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>>  8 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[ 8] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[ 9] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[10] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[11] = (char) (x        & 1 | 48);
    return sb.append (FMT_TEMP, 0, 12);
  }  //fmtBin12(StringBuilder,int)

  //fmtAin16 (a, o, x)
  //fmtBin16 (a, o, x)
  //s = fmtAin16 (x)
  //s = fmtBin16 (x)
  //sb = fmtAin16 (sb, x)
  //sb = fmtBin16 (sb, x)
  //  16桁2進数変換
  public static void fmtAin16 (char[] a, int o, int x) {
    a[o     ] = (char) (x >> 13 & 4 ^ 46);
    a[o +  1] = (char) (x >> 12 & 4 ^ 46);
    a[o +  2] = (char) (x >> 11 & 4 ^ 46);
    a[o +  3] = (char) (x >> 10 & 4 ^ 46);
    a[o +  4] = (char) (x >>  9 & 4 ^ 46);
    a[o +  5] = (char) (x >>  8 & 4 ^ 46);
    a[o +  6] = (char) (x >>  7 & 4 ^ 46);
    a[o +  7] = (char) (x >>  6 & 4 ^ 46);
    a[o +  8] = (char) (x >>  5 & 4 ^ 46);
    a[o +  9] = (char) (x >>  4 & 4 ^ 46);
    a[o + 10] = (char) (x >>  3 & 4 ^ 46);
    a[o + 11] = (char) (x >>  2 & 4 ^ 46);
    a[o + 12] = (char) (x >>  1 & 4 ^ 46);
    a[o + 13] = (char) (x       & 4 ^ 46);
    a[o + 14] = (char) (x <<  1 & 4 ^ 46);
    a[o + 15] = (char) (x <<  2 & 4 ^ 46);
  }  //fmtAin16(char[],int,int)
  public static void fmtBin16 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>> 15 & 1 | 48);
    a[o +  1] = (char) (x >>> 14 & 1 | 48);
    a[o +  2] = (char) (x >>> 13 & 1 | 48);
    a[o +  3] = (char) (x >>> 12 & 1 | 48);
    a[o +  4] = (char) (x >>> 11 & 1 | 48);
    a[o +  5] = (char) (x >>> 10 & 1 | 48);
    a[o +  6] = (char) (x >>>  9 & 1 | 48);
    a[o +  7] = (char) (x >>>  8 & 1 | 48);
    a[o +  8] = (char) (x >>>  7 & 1 | 48);
    a[o +  9] = (char) (x >>>  6 & 1 | 48);
    a[o + 10] = (char) (x >>>  5 & 1 | 48);
    a[o + 11] = (char) (x >>>  4 & 1 | 48);
    a[o + 12] = (char) (x >>>  3 & 1 | 48);
    a[o + 13] = (char) (x >>>  2 & 1 | 48);
    a[o + 14] = (char) (x >>>  1 & 1 | 48);
    a[o + 15] = (char) (x        & 1 | 48);
  }  //fmtBin16(char[],int,int)
  public static String fmtAin16 (int x) {
    FMT_TEMP[ 0] = (char) (x >> 13 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >> 12 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >> 11 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >> 10 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >>  9 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x >>  8 & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x >>  7 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x >>  6 & 4 ^ 46);
    FMT_TEMP[ 8] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[ 9] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[10] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[11] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[12] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[13] = (char) (x       & 4 ^ 46);
    FMT_TEMP[14] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[15] = (char) (x <<  2 & 4 ^ 46);
    return String.valueOf (FMT_TEMP, 0, 16);
  }  //fmtAin16(int)
  public static String fmtBin16 (int x) {
    FMT_TEMP[ 0] = (char) (x >>> 15 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>> 14 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>> 13 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>> 12 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>> 11 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>> 10 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>>  9 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x >>>  8 & 1 | 48);
    FMT_TEMP[ 8] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[ 9] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[10] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[11] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[12] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[13] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[14] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[15] = (char) (x        & 1 | 48);
    return String.valueOf (FMT_TEMP, 0, 16);
  }  //fmtBin16(int)
  public static StringBuilder fmtAin16 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >> 13 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >> 12 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >> 11 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >> 10 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >>  9 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x >>  8 & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x >>  7 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x >>  6 & 4 ^ 46);
    FMT_TEMP[ 8] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[ 9] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[10] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[11] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[12] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[13] = (char) (x       & 4 ^ 46);
    FMT_TEMP[14] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[15] = (char) (x <<  2 & 4 ^ 46);
    return sb.append (FMT_TEMP, 0, 16);
  }  //fmtAin16(StringBuilder,int)
  public static StringBuilder fmtBin16 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>> 15 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>> 14 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>> 13 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>> 12 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>> 11 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>> 10 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>>  9 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x >>>  8 & 1 | 48);
    FMT_TEMP[ 8] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[ 9] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[10] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[11] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[12] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[13] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[14] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[15] = (char) (x        & 1 | 48);
    return sb.append (FMT_TEMP, 0, 16);
  }  //fmtBin16(StringBuilder,int)

  //fmtAin24 (a, o, x)
  //fmtBin24 (a, o, x)
  //s = fmtAin24 (x)
  //s = fmtBin24 (x)
  //sb = fmtAin24 (sb, x)
  //sb = fmtBin24 (sb, x)
  //  24桁2進数変換
  public static void fmtAin24 (char[] a, int o, int x) {
    a[o     ] = (char) (x >> 21 & 4 ^ 46);
    a[o +  1] = (char) (x >> 20 & 4 ^ 46);
    a[o +  2] = (char) (x >> 19 & 4 ^ 46);
    a[o +  3] = (char) (x >> 18 & 4 ^ 46);
    a[o +  4] = (char) (x >> 17 & 4 ^ 46);
    a[o +  5] = (char) (x >> 16 & 4 ^ 46);
    a[o +  6] = (char) (x >> 15 & 4 ^ 46);
    a[o +  7] = (char) (x >> 14 & 4 ^ 46);
    a[o +  8] = (char) (x >> 13 & 4 ^ 46);
    a[o +  9] = (char) (x >> 12 & 4 ^ 46);
    a[o + 10] = (char) (x >> 11 & 4 ^ 46);
    a[o + 11] = (char) (x >> 10 & 4 ^ 46);
    a[o + 12] = (char) (x >>  9 & 4 ^ 46);
    a[o + 13] = (char) (x >>  8 & 4 ^ 46);
    a[o + 14] = (char) (x >>  7 & 4 ^ 46);
    a[o + 15] = (char) (x >>  6 & 4 ^ 46);
    a[o + 16] = (char) (x >>  5 & 4 ^ 46);
    a[o + 17] = (char) (x >>  4 & 4 ^ 46);
    a[o + 18] = (char) (x >>  3 & 4 ^ 46);
    a[o + 19] = (char) (x >>  2 & 4 ^ 46);
    a[o + 20] = (char) (x >>  1 & 4 ^ 46);
    a[o + 21] = (char) (x       & 4 ^ 46);
    a[o + 22] = (char) (x <<  1 & 4 ^ 46);
    a[o + 23] = (char) (x <<  2 & 4 ^ 46);
  }  //fmtAin24(char[],int,int)
  public static void fmtBin24 (char[] a, int o, int x) {
    a[o     ] = (char) (x >>> 23 & 1 | 48);
    a[o +  1] = (char) (x >>> 22 & 1 | 48);
    a[o +  2] = (char) (x >>> 21 & 1 | 48);
    a[o +  3] = (char) (x >>> 20 & 1 | 48);
    a[o +  4] = (char) (x >>> 19 & 1 | 48);
    a[o +  5] = (char) (x >>> 18 & 1 | 48);
    a[o +  6] = (char) (x >>> 17 & 1 | 48);
    a[o +  7] = (char) (x >>> 16 & 1 | 48);
    a[o +  8] = (char) (x >>> 15 & 1 | 48);
    a[o +  9] = (char) (x >>> 14 & 1 | 48);
    a[o + 10] = (char) (x >>> 13 & 1 | 48);
    a[o + 11] = (char) (x >>> 12 & 1 | 48);
    a[o + 12] = (char) (x >>> 11 & 1 | 48);
    a[o + 13] = (char) (x >>> 10 & 1 | 48);
    a[o + 14] = (char) (x >>>  9 & 1 | 48);
    a[o + 15] = (char) (x >>>  8 & 1 | 48);
    a[o + 16] = (char) (x >>>  7 & 1 | 48);
    a[o + 17] = (char) (x >>>  6 & 1 | 48);
    a[o + 18] = (char) (x >>>  5 & 1 | 48);
    a[o + 19] = (char) (x >>>  4 & 1 | 48);
    a[o + 20] = (char) (x >>>  3 & 1 | 48);
    a[o + 21] = (char) (x >>>  2 & 1 | 48);
    a[o + 22] = (char) (x >>>  1 & 1 | 48);
    a[o + 23] = (char) (x        & 1 | 48);
  }  //fmtBin24(char[],int,int)
  public static String fmtAin24 (int x) {
    FMT_TEMP[ 0] = (char) (x >> 21 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >> 20 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >> 19 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >> 18 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >> 17 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x >> 16 & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x >> 15 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x >> 14 & 4 ^ 46);
    FMT_TEMP[ 8] = (char) (x >> 13 & 4 ^ 46);
    FMT_TEMP[ 9] = (char) (x >> 12 & 4 ^ 46);
    FMT_TEMP[10] = (char) (x >> 11 & 4 ^ 46);
    FMT_TEMP[11] = (char) (x >> 10 & 4 ^ 46);
    FMT_TEMP[12] = (char) (x >>  9 & 4 ^ 46);
    FMT_TEMP[13] = (char) (x >>  8 & 4 ^ 46);
    FMT_TEMP[14] = (char) (x >>  7 & 4 ^ 46);
    FMT_TEMP[15] = (char) (x >>  6 & 4 ^ 46);
    FMT_TEMP[16] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[17] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[18] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[19] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[20] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[21] = (char) (x       & 4 ^ 46);
    FMT_TEMP[22] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[23] = (char) (x <<  2 & 4 ^ 46);
    return String.valueOf (FMT_TEMP, 0, 24);
  }  //fmtAin24(int)
  public static String fmtBin24 (int x) {
    FMT_TEMP[ 0] = (char) (x >>> 23 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>> 22 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>> 21 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>> 20 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>> 19 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>> 18 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>> 17 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x >>> 16 & 1 | 48);
    FMT_TEMP[ 8] = (char) (x >>> 15 & 1 | 48);
    FMT_TEMP[ 9] = (char) (x >>> 14 & 1 | 48);
    FMT_TEMP[10] = (char) (x >>> 13 & 1 | 48);
    FMT_TEMP[11] = (char) (x >>> 12 & 1 | 48);
    FMT_TEMP[12] = (char) (x >>> 11 & 1 | 48);
    FMT_TEMP[13] = (char) (x >>> 10 & 1 | 48);
    FMT_TEMP[14] = (char) (x >>>  9 & 1 | 48);
    FMT_TEMP[15] = (char) (x >>>  8 & 1 | 48);
    FMT_TEMP[16] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[17] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[18] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[19] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[20] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[21] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[22] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[23] = (char) (x        & 1 | 48);
    return String.valueOf (FMT_TEMP, 0, 24);
  }  //fmtBin24(int)
  public static StringBuilder fmtAin24 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >> 21 & 4 ^ 46);
    FMT_TEMP[ 1] = (char) (x >> 20 & 4 ^ 46);
    FMT_TEMP[ 2] = (char) (x >> 19 & 4 ^ 46);
    FMT_TEMP[ 3] = (char) (x >> 18 & 4 ^ 46);
    FMT_TEMP[ 4] = (char) (x >> 17 & 4 ^ 46);
    FMT_TEMP[ 5] = (char) (x >> 16 & 4 ^ 46);
    FMT_TEMP[ 6] = (char) (x >> 15 & 4 ^ 46);
    FMT_TEMP[ 7] = (char) (x >> 14 & 4 ^ 46);
    FMT_TEMP[ 8] = (char) (x >> 13 & 4 ^ 46);
    FMT_TEMP[ 9] = (char) (x >> 12 & 4 ^ 46);
    FMT_TEMP[10] = (char) (x >> 11 & 4 ^ 46);
    FMT_TEMP[11] = (char) (x >> 10 & 4 ^ 46);
    FMT_TEMP[12] = (char) (x >>  9 & 4 ^ 46);
    FMT_TEMP[13] = (char) (x >>  8 & 4 ^ 46);
    FMT_TEMP[14] = (char) (x >>  7 & 4 ^ 46);
    FMT_TEMP[15] = (char) (x >>  6 & 4 ^ 46);
    FMT_TEMP[16] = (char) (x >>  5 & 4 ^ 46);
    FMT_TEMP[17] = (char) (x >>  4 & 4 ^ 46);
    FMT_TEMP[18] = (char) (x >>  3 & 4 ^ 46);
    FMT_TEMP[19] = (char) (x >>  2 & 4 ^ 46);
    FMT_TEMP[20] = (char) (x >>  1 & 4 ^ 46);
    FMT_TEMP[21] = (char) (x       & 4 ^ 46);
    FMT_TEMP[22] = (char) (x <<  1 & 4 ^ 46);
    FMT_TEMP[23] = (char) (x <<  2 & 4 ^ 46);
    return sb.append (FMT_TEMP, 0, 24);
  }  //fmtAin24(StringBuilder,int)
  public static StringBuilder fmtBin24 (StringBuilder sb, int x) {
    FMT_TEMP[ 0] = (char) (x >>> 23 & 1 | 48);
    FMT_TEMP[ 1] = (char) (x >>> 22 & 1 | 48);
    FMT_TEMP[ 2] = (char) (x >>> 21 & 1 | 48);
    FMT_TEMP[ 3] = (char) (x >>> 20 & 1 | 48);
    FMT_TEMP[ 4] = (char) (x >>> 19 & 1 | 48);
    FMT_TEMP[ 5] = (char) (x >>> 18 & 1 | 48);
    FMT_TEMP[ 6] = (char) (x >>> 17 & 1 | 48);
    FMT_TEMP[ 7] = (char) (x >>> 16 & 1 | 48);
    FMT_TEMP[ 8] = (char) (x >>> 15 & 1 | 48);
    FMT_TEMP[ 9] = (char) (x >>> 14 & 1 | 48);
    FMT_TEMP[10] = (char) (x >>> 13 & 1 | 48);
    FMT_TEMP[11] = (char) (x >>> 12 & 1 | 48);
    FMT_TEMP[12] = (char) (x >>> 11 & 1 | 48);
    FMT_TEMP[13] = (char) (x >>> 10 & 1 | 48);
    FMT_TEMP[14] = (char) (x >>>  9 & 1 | 48);
    FMT_TEMP[15] = (char) (x >>>  8 & 1 | 48);
    FMT_TEMP[16] = (char) (x >>>  7 & 1 | 48);
    FMT_TEMP[17] = (char) (x >>>  6 & 1 | 48);
    FMT_TEMP[18] = (char) (x >>>  5 & 1 | 48);
    FMT_TEMP[19] = (char) (x >>>  4 & 1 | 48);
    FMT_TEMP[20] = (char) (x >>>  3 & 1 | 48);
    FMT_TEMP[21] = (char) (x >>>  2 & 1 | 48);
    FMT_TEMP[22] = (char) (x >>>  1 & 1 | 48);
    FMT_TEMP[23] = (char) (x        & 1 | 48);
    return sb.append (FMT_TEMP, 0, 24);
  }  //fmtBin24(StringBuilder,int)

  //--------------------------------------------------------------------------------
  //16進数変換
  //
  //     x             00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f
  //   9-x             09 08 07 06 05 04 03 02 01 00 ff fe fd fc fb fa
  //   9-x>>4          00 00 00 00 00 00 00 00 00 00 ff ff ff ff ff ff
  //   9-x>>4&7        00 00 00 00 00 00 00 00 00 00 07 07 07 07 07 07
  //   9-x>>4&7|48     30 30 30 30 30 30 30 30 30 30 37 37 37 37 37 37
  //  (9-x>>4&7|48)+x  30 31 32 33 34 35 36 37 38 39 41 42 43 44 45 46
  //                    0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
  //
  //     x              00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f
  //   9-x              09 08 07 06 05 04 03 02 01 00 ff fe fd fc fb fa
  //   9-x>>4           00 00 00 00 00 00 00 00 00 00 ff ff ff ff ff ff
  //   9-x>>4&39        00 00 00 00 00 00 00 00 00 00 27 27 27 27 27 27
  //  (9-x>>4&39)+48    30 30 30 30 30 30 30 30 30 30 57 57 57 57 57 57
  //  (9-x>>4&39)+48+x  30 31 32 33 34 35 36 37 38 39 61 62 63 64 65 66
  //                     0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f
  //
  //             c            30 31 32 33 34 35 36 37 38 39 41 42 43 44 45 46 61 62 63 64 65 66
  //          64-c            10 0f 0e 0d 0c 0b 0a 09 08 07 ff fe fd fc fb fa df de dd dc db da
  //          64-c>>8         00 00 00 00 00 00 00 00 00 00 ff ff ff ff ff ff ff ff ff ff ff ff
  //          64-c>>8&39      00 00 00 00 00 00 00 00 00 00 27 27 27 27 27 27 27 27 27 27 27 27
  //          64-c>>8&39|48   30 30 30 30 30 30 30 30 30 30 57 57 57 57 57 57 57 57 57 57 57 57
  //   c|32                   30 31 32 33 34 35 36 37 38 39 61 62 63 64 65 66 61 62 63 64 65 66
  //  (c|32)-(64-c>>8&39|48)  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f 0a 0b 0c 0d 0e 0f

  //c = fmtHexc (x)
  //a = fmtHex1 (a, o, x)
  //s = fmtHex1 (x)
  //sb = fmtHex1 (sb, x)
  //  1桁16進数変換
  public static char fmtHexc (int x) {
    x &= 15;
    return (char) ((9 - x >> 4 & 7 | 48) + x);
  }  //fmtHexc(int)
  public static void fmtHex1 (char[] a, int o, int x) {
    x &= 15;
    a[o] = (char) ((9 - x >> 4 & 7 | 48) + x);
  }  //fmtHex1(char[],int,int)
  public static String fmtHex1 (int x) {
    x &= 15;
    return Character.toString ((char) ((9 - x >> 4 & 7 | 48) + x));
  }  //fmtHex1(int)
  public static StringBuilder fmtHex1 (StringBuilder sb, int x) {
    x &= 15;
    return sb.append ((char) ((9 - x >> 4 & 7 | 48) + x));
  }  //fmtHex1(StringBuilder,int)

  //fmtHex2 (a, o, x)
  //s = fmtHex2 (x)
  //sb = fmtHex2 (sb, x)
  //  2桁16進数変換
  //  byte用
  public static void fmtHex2 (char[] a, int o, int x) {
    int x0 = x        & 15;
    int x1 = x >>>  4 & 15;
    a[o    ] = (char) ((9 - x1 >> 4 & 7 | 48) + x1);
    a[o + 1] = (char) ((9 - x0 >> 4 & 7 | 48) + x0);
  }  //fmtHex2(char[],int,int)
  public static String fmtHex2 (int x) {
    //fmtHex2 (FMT_TEMP, 0, x);
    int x0 = x        & 15;
    int x1 = x >>>  4 & 15;
    FMT_TEMP[0] = (char) ((9 - x1 >> 4 & 7 | 48) + x1);
    FMT_TEMP[1] = (char) ((9 - x0 >> 4 & 7 | 48) + x0);
    return String.valueOf (FMT_TEMP, 0, 2);
  }  //fmtHex2(int)
  public static StringBuilder fmtHex2 (StringBuilder sb, int x) {
    int x0 = x        & 15;
    int x1 = x >>>  4 & 15;
    return (sb.
            append ((char) ((9 - x1 >> 4 & 7 | 48) + x1)).
            append ((char) ((9 - x0 >> 4 & 7 | 48) + x0)));
  }  //fmtHex2(StringBuilder,int)

  //fmtHex4 (a, o, x)
  //s = fmtHex4 (x)
  //sb = fmtHex4 (sb, x)
  //  4桁16進数変換
  //  word用
  public static void fmtHex4 (char[] a, int o, int x) {
    int t;
    t = (char) x >>> 12;
    a[o    ] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    a[o + 1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    a[o + 2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    a[o + 3] = (char) ((9 - t >> 4 & 7 | 48) + t);
  }  //fmtHex4(char[],int,int)
  public static String fmtHex4 (int x) {
    //fmtHex4 (FMT_TEMP, 0, x);
    int t;
    t = (char) x >>> 12;
    FMT_TEMP[0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    FMT_TEMP[1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    FMT_TEMP[2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    FMT_TEMP[3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return String.valueOf (FMT_TEMP, 0, 4);
  }  //fmtHex4(int)
  public static StringBuilder fmtHex4 (StringBuilder sb, int x) {
    //fmtHex4 (FMT_TEMP, 0, x);
    int t;
    t = (char) x >>> 12;
    FMT_TEMP[0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    FMT_TEMP[1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    FMT_TEMP[2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    FMT_TEMP[3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return sb.append (FMT_TEMP, 0, 4);
  }  //fmtHex4(StringBuilder,int)

  //fmtHex6 (a, o, x)
  //s = fmtHex6 (x)
  //sb = fmtHex6 (sb, x)
  //  6桁16進数変換
  //  rgb用
  public static void fmtHex6 (char[] a, int o, int x) {
    int t;
    t =        x >>> 20 & 15;
    a[o    ] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 16 & 15;
    a[o + 1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) x >>> 12;
    a[o + 2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    a[o + 3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    a[o + 4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    a[o + 5] = (char) ((9 - t >> 4 & 7 | 48) + t);
  }  //fmtHex6(char[],int,int)
  public static String fmtHex6 (int x) {
    //fmtHex6 (FMT_TEMP, 0, x);
    int t;
    t =        x >>> 20 & 15;
    FMT_TEMP[0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 16 & 15;
    FMT_TEMP[1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) x >>> 12;
    FMT_TEMP[2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    FMT_TEMP[3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    FMT_TEMP[4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    FMT_TEMP[5] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return String.valueOf (FMT_TEMP, 0, 6);
  }  //fmtHex6(int)
  public static StringBuilder fmtHex6 (StringBuilder sb, int x) {
    //fmtHex6 (FMT_TEMP, 0, x);
    int t;
    t =        x >>> 20 & 15;
    FMT_TEMP[0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 16 & 15;
    FMT_TEMP[1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) x >>> 12;
    FMT_TEMP[2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    FMT_TEMP[3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    FMT_TEMP[4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    FMT_TEMP[5] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return sb.append (FMT_TEMP, 0, 6);
  }  //fmtHex6(StringBuilder,int)

  //fmtHex8 (a, o, x)
  //s = fmtHex8 (x)
  //sb = fmtHex8 (sb, x)
  //  8桁16進数変換
  //  argb,long用
  public static void fmtHex8 (char[] a, int o, int x) {
    int t;
    t =        x >>> 28;
    a[o    ] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 24 & 15;
    a[o + 1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 20 & 15;
    a[o + 2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 16 & 15;
    a[o + 3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) x >>> 12;
    a[o + 4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    a[o + 5] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    a[o + 6] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    a[o + 7] = (char) ((9 - t >> 4 & 7 | 48) + t);
  }  //fmtHex8(char[],int,int)
  public static String fmtHex8 (int x) {
    //fmtHex8 (FMT_TEMP, 0, x);
    int t;
    t =        x >>> 28;
    FMT_TEMP[0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 24 & 15;
    FMT_TEMP[1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 20 & 15;
    FMT_TEMP[2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 16 & 15;
    FMT_TEMP[3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) x >>> 12;
    FMT_TEMP[4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    FMT_TEMP[5] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    FMT_TEMP[6] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    FMT_TEMP[7] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return String.valueOf (FMT_TEMP, 0, 8);
  }  //fmtHex8(int)
  public static StringBuilder fmtHex8 (StringBuilder sb, int x) {
    //fmtHex8 (FMT_TEMP, 0, x);
    int t;
    t =        x >>> 28;
    FMT_TEMP[0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 24 & 15;
    FMT_TEMP[1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 20 & 15;
    FMT_TEMP[2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>> 16 & 15;
    FMT_TEMP[3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) x >>> 12;
    FMT_TEMP[4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  8 & 15;
    FMT_TEMP[5] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x >>>  4 & 15;
    FMT_TEMP[6] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        x        & 15;
    FMT_TEMP[7] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return sb.append (FMT_TEMP, 0, 8);
  }  //fmtHex8(StringBuilder,int)

  public static StringBuilder fmtHex16 (StringBuilder sb, long x) {
    //fmtHex16 (FMT_TEMP, 0, x);
    int s, t;
    s = (int) (x >>> 32);
    t =        s >>> 28;
    FMT_TEMP[ 0] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>> 24 & 15;
    FMT_TEMP[ 1] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>> 20 & 15;
    FMT_TEMP[ 2] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>> 16 & 15;
    FMT_TEMP[ 3] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) s >>> 12;
    FMT_TEMP[ 4] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>>  8 & 15;
    FMT_TEMP[ 5] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>>  4 & 15;
    FMT_TEMP[ 6] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s        & 15;
    FMT_TEMP[ 7] = (char) ((9 - t >> 4 & 7 | 48) + t);
    s = (int)  x;
    t =        s >>> 28;
    FMT_TEMP[ 8] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>> 24 & 15;
    FMT_TEMP[ 9] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>> 20 & 15;
    FMT_TEMP[10] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>> 16 & 15;
    FMT_TEMP[11] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t = (char) s >>> 12;
    FMT_TEMP[12] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>>  8 & 15;
    FMT_TEMP[13] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s >>>  4 & 15;
    FMT_TEMP[14] = (char) ((9 - t >> 4 & 7 | 48) + t);
    t =        s        & 15;
    FMT_TEMP[15] = (char) ((9 - t >> 4 & 7 | 48) + t);
    return sb.append (FMT_TEMP, 0, 16);
  }  //fmtHex16(StringBuilder,long)

  //--------------------------------------------------------------------------------
  //10進数変換
  //  除算は遅いので逆数乗算で商と余りを求めて下位から充填する
  //  x/yを計算する代わりにceil(pow(2,n)/y)を掛けてから右にnビットシフトする
  //  m=((1<<n)+y-1)/yとおくと0<=x<=((((m-1)/(m*y-(1<<n))+1)<<n)-1)/mの範囲でx/yはm*x>>nに置き換えられる
  //    >perl -e "use GMP::Mpz qw(:all);$y=mpz(10);for($n=mpz(0);$n<=31;$n++){$m=((1<<$n)+$y-1)/$y;$l=(((($m-1)/($m*$y-(1<<$n))+1)<<$n)-1)/$m;printf'x*%s>>%s (0<=x<=%s)%c',$m,$n,$l,10;}"
  //    x*1>>0 (0<=x<=0)
  //    x*1>>1 (0<=x<=1)
  //    x*1>>2 (0<=x<=3)
  //    x*1>>3 (0<=x<=7)
  //    x*2>>4 (0<=x<=7)
  //    x*4>>5 (0<=x<=7)
  //    x*7>>6 (0<=x<=18)
  //    x*13>>7 (0<=x<=68)
  //    x*26>>8 (0<=x<=68)
  //    x*52>>9 (0<=x<=68)
  //    x*103>>10 (0<=x<=178)  2桁
  //    x*205>>11 (0<=x<=1028)  3桁
  //    x*410>>12 (0<=x<=1028)
  //    x*820>>13 (0<=x<=1028)
  //    x*1639>>14 (0<=x<=2738)
  //    x*3277>>15 (0<=x<=16388)  4桁
  //    x*6554>>16 (0<=x<=16388)
  //    x*13108>>17 (0<=x<=16388)
  //    x*26215>>18 (0<=x<=43698)
  //    x*52429>>19 (0<=x<=262148)  5桁。ここからlong
  //    x*104858>>20 (0<=x<=262148)
  //    x*209716>>21 (0<=x<=262148)
  //    x*419431>>22 (0<=x<=699058)
  //    x*838861>>23 (0<=x<=4194308)  6桁
  //    x*1677722>>24 (0<=x<=4194308)
  //    x*3355444>>25 (0<=x<=4194308)
  //    x*6710887>>26 (0<=x<=11184818)  7桁
  //    x*13421773>>27 (0<=x<=67108868)
  //    x*26843546>>28 (0<=x<=67108868)
  //    x*53687092>>29 (0<=x<=67108868)
  //    x*107374183>>30 (0<=x<=178956978)  8桁
  //    x*214748365>>31 (0<=x<=1073741828)  9桁
  //
  //  検算
  //    >perl -e "use GMP::Mpz qw(:all);$y=mpz(10);for($n=mpz(0);$n<=23;$n++){$m=((1<<$n)+$y-1)/$y;$l=(((($m-1)/($m*$y-(1<<$n))+1)<<$n)-1)/$m;for($x=mpz(0);$x<=$l+1;$x++){$t=$m*$x>>$n;$z=$x/$y;if($z!=$t){printf'n=%d,m=%d,l=%d,x=%d,y=%d,z=%d,t=%d%c',$n,$m,$l,$x,$y,$z,$t,10;}}}"
  //    n=0,m=1,l=0,x=1,y=10,z=0,t=1
  //    n=1,m=1,l=1,x=2,y=10,z=0,t=1
  //    n=2,m=1,l=3,x=4,y=10,z=0,t=1
  //    n=3,m=1,l=7,x=8,y=10,z=0,t=1
  //    n=4,m=2,l=7,x=8,y=10,z=0,t=1
  //    n=5,m=4,l=7,x=8,y=10,z=0,t=1
  //    n=6,m=7,l=18,x=19,y=10,z=1,t=2
  //    n=7,m=13,l=68,x=69,y=10,z=6,t=7
  //    n=8,m=26,l=68,x=69,y=10,z=6,t=7
  //    n=9,m=52,l=68,x=69,y=10,z=6,t=7
  //    n=10,m=103,l=178,x=179,y=10,z=17,t=18
  //    n=11,m=205,l=1028,x=1029,y=10,z=102,t=103
  //    n=12,m=410,l=1028,x=1029,y=10,z=102,t=103
  //    n=13,m=820,l=1028,x=1029,y=10,z=102,t=103
  //    n=14,m=1639,l=2738,x=2739,y=10,z=273,t=274
  //    n=15,m=3277,l=16388,x=16389,y=10,z=1638,t=1639
  //    n=16,m=6554,l=16388,x=16389,y=10,z=1638,t=1639
  //    n=17,m=13108,l=16388,x=16389,y=10,z=1638,t=1639
  //    n=18,m=26215,l=43698,x=43699,y=10,z=4369,t=4370
  //    n=19,m=52429,l=262148,x=262149,y=10,z=26214,t=26215
  //    n=20,m=104858,l=262148,x=262149,y=10,z=26214,t=26215
  //    n=21,m=209716,l=262148,x=262149,y=10,z=26214,t=26215
  //    n=22,m=419431,l=699058,x=699059,y=10,z=69905,t=69906
  //    n=23,m=838861,l=4194308,x=4194309,y=10,z=419430,t=419431

  //  4桁まではあらかじめテーブルに展開しておく
  public static final int[] FMT_BCD4 = new int[10000];
  public static final int[] FMT_DCB4 = new int[65536];

  //--------------------------------------------------------------------------------
  //fmtInit ()
  //  初期化
  public static void fmtInit () {
    Arrays.fill (FMT_DCB4, -1);
    int i = 0;
    int x = 0;
    for (int a = 0; a < 10; a++) {
      for (int b = 0; b < 10; b++) {
        for (int c = 0; c < 10; c++) {
          FMT_DCB4[FMT_BCD4[i    ] = x    ] = i;
          FMT_DCB4[FMT_BCD4[i + 1] = x + 1] = i + 1;
          FMT_DCB4[FMT_BCD4[i + 2] = x + 2] = i + 2;
          FMT_DCB4[FMT_BCD4[i + 3] = x + 3] = i + 3;
          FMT_DCB4[FMT_BCD4[i + 4] = x + 4] = i + 4;
          FMT_DCB4[FMT_BCD4[i + 5] = x + 5] = i + 5;
          FMT_DCB4[FMT_BCD4[i + 6] = x + 6] = i + 6;
          FMT_DCB4[FMT_BCD4[i + 7] = x + 7] = i + 7;
          FMT_DCB4[FMT_BCD4[i + 8] = x + 8] = i + 8;
          FMT_DCB4[FMT_BCD4[i + 9] = x + 9] = i + 9;
          i += 10;
          x += 1 << 4;
        }
        x += 6 << 4;
      }
      x += 6 << 8;
    }
  }  //fmtInit()

  //y = fmtBcd4 (x)
  //  xを0～9999にクリッピングしてから4桁のBCDに変換する
  public static int fmtBcd4 (int x) {
    //x = Math.max (0, Math.min (9999, x));
    //perl optdiv.pl 9999 10
    //  x/10==x*3277>>>15 (0<=x<=16388) [9999*3277==32766723]
    //int t = x * 3277 >> 15;  //x/10
    //int y = x - t * 10;  //1の位
    //x = t * 3277 >> 15;  //x/100
    //y |= t - x * 10 << 4;  //10の位
    //t = x * 3277 >> 15;  //x/1000
    //return t << 12 | x - t * 10 << 8 | y;  //1000の位,100の位
    return FMT_BCD4[Math.max (0, Math.min (9999, x))];
  }  //fmtBcd4(int)

  //y = fmtBcd8 (x)
  //  xを0～99999999にクリッピングしてから8桁のBCDに変換する
  public static int fmtBcd8 (int x) {
    x = Math.max (0, Math.min (99999999, x));
    //perl optdiv.pl 99999999 10000
    //  x/10000==x*109951163>>>40 (0<=x<=494389998) [99999999*109951163==10995116190048837]
    int q = (int) ((long) x * 109951163L >>> 40);  //x/10000。1.6ns@2000000000
    //int q = x / 10000;  //2.0ns@2000000000
    return FMT_BCD4[q] << 16 | FMT_BCD4[x - 10000 * q];
  }  //fmtBcd8(int)

  //y = fmtBcd12 (x)
  //  xを0～999999999999Lにクリッピングしてから12桁のBCDに変換する
  public static long fmtBcd12 (long x) {
    x = Math.max (0L, Math.min (999999999999L, x));
    int q = (int) ((double) x / 100000000.0);  //(int) (x / 100000000L);
    int r = (int) (x - 100000000L * q);
    //perl optdiv.pl 99999999 10000
    //  x/10000==x*109951163>>>40 (0<=x<=494389998) [99999999*109951163==10995116190048837]
    int rq = (int) ((long) r * 109951163L >>> 40);  //r/10000
    //int rq = r / 10000;
    return (long) FMT_BCD4[q] << 32 | 0xffffffffL & (FMT_BCD4[rq] << 16 | FMT_BCD4[r - 10000 * rq]);
  }  //fmtBcd12(long)

  //y = fmtBcd16 (x)
  //  xを0～9999999999999999Lにクリッピングしてから16桁のBCDに変換する
  public static long fmtBcd16 (long x) {
    x = Math.max (0L, Math.min (9999999999999999L, x));
    int q = x <= (1L << 53) ? (int) ((double) x / 100000000.0) : (int) (x / 100000000L);
    int r = (int) (x - 100000000L * q);
    //perl optdiv.pl 99999999 10000
    //  x/10000==x*109951163>>>40 (0<=x<=494389998) [99999999*109951163==10995116190048837]
    int qq = (int) ((long) q * 109951163L >>> 40);  //q/10000
    //int qq = q / 10000;
    //perl optdiv.pl 99999999 10000
    //  x/10000==x*109951163>>>40 (0<=x<=494389998) [99999999*109951163==10995116190048837]
    int rq = (int) ((long) r * 109951163L >>> 40);  //r/10000
    //int rq = r / 10000;
    return (long) (FMT_BCD4[qq] << 16 | FMT_BCD4[q - 10000 * qq]) << 32 | 0xffffffffL & (FMT_BCD4[rq] << 16 | FMT_BCD4[r - 10000 * rq]);
  }  //fmtBcd16(long)

  //--------------------------------------------------------------------------------
  //o = fmtCA02u (a, o, x)
  //sb = fmtSB02u (sb, x)
  //  %02u
  //  2桁10進数変換(符号なし,ゼロサプレスなし)
  public static int fmtCA02u (char[] a, int o, int x) {
    if (x < 0 || 99 < x) {
      x = 99;
    }
    x = FMT_BCD4[x];
    a[o    ] = (char) ('0' | x >>> 4);
    a[o + 1] = (char) ('0' | x        & 15);
    return o + 2;
  }  //fmtCA02u(char[],int,int)
  public static StringBuilder fmtSB02u (StringBuilder sb, int x) {
    return sb.append (FMT_TEMP, 0, fmtCA02u (FMT_TEMP, 0, x));
  }  //fmtSB02u(StringBuilder,int)

  //o = fmtCA2u (a, o, x)
  //sb = fmtSB2u (sb, x)
  //  %2u
  //  2桁10進数変換(符号なし,ゼロサプレスあり)
  public static int fmtCA2u (char[] a, int o, int x) {
    if (x < 0 || 99 < x) {
      x = 99;
    }
    x = FMT_BCD4[x];
    if (x <= 0x000f) {  //1桁
      a[o++] = (char) ('0' | x);
    } else {  //2桁
      a[o++] = (char) ('0' | x >>>  4);
      a[o++] = (char) ('0' | x        & 15);
    }
    return o;
  }  //fmtCA2u(char[],int,int)
  public static StringBuilder fmtSB2u (StringBuilder sb, int x) {
    return sb.append (FMT_TEMP, 0, fmtCA2u (FMT_TEMP, 0, x));
  }  //fmtSB2u(StringBuilder,int)

  //o = fmtCA04u (a, o, x)
  //sb = fmtSB04u (sb, x)
  //  %04u
  //  4桁10進数変換(符号なし,ゼロサプレスなし)
  public static int fmtCA04u (char[] a, int o, int x) {
    if (x < 0 || 9999 < x) {
      x = 9999;
    }
    x = FMT_BCD4[x];
    a[o    ] = (char) ('0' | x >>> 12);
    a[o + 1] = (char) ('0' | x >>>  8 & 15);
    a[o + 2] = (char) ('0' | x >>>  4 & 15);
    a[o + 3] = (char) ('0' | x        & 15);
    return o + 4;
  }  //fmtCA04u(char[],int,int)
  public static StringBuilder fmtSB04u (StringBuilder sb, int x) {
    return sb.append (FMT_TEMP, 0, fmtCA04u (FMT_TEMP, 0, x));
  }  //fmtSB04u(StringBuilder,int)

  //o = fmtCA4u (a, o, x)
  //sb = fmtSB4u (sb, x)
  //  %4u
  //  4桁10進数変換(符号なし,ゼロサプレスあり)
  public static int fmtCA4u (char[] a, int o, int x) {
    if (x < 0 || 9999 < x) {
      x = 9999;
    }
    x = FMT_BCD4[x];
    if (x <= 0x000f) {  //1桁
      a[o++] = (char) ('0' | x);
    } else if (x <= 0x00ff) {  //2桁
      a[o++] = (char) ('0' | x >>>  4);
      a[o++] = (char) ('0' | x        & 15);
    } else if (x <= 0x0fff) {  //3桁
      a[o++] = (char) ('0' | x >>>  8);
      a[o++] = (char) ('0' | x >>>  4 & 15);
      a[o++] = (char) ('0' | x        & 15);
    } else {  //4桁
      a[o++] = (char) ('0' | x >>> 12);
      a[o++] = (char) ('0' | x >>>  8 & 15);
      a[o++] = (char) ('0' | x >>>  4 & 15);
      a[o++] = (char) ('0' | x        & 15);
    }
    return o;
  }  //fmtCA4u(char[],int,int)
  public static StringBuilder fmtSB4u (StringBuilder sb, int x) {
    return sb.append (FMT_TEMP, 0, fmtCA4u (FMT_TEMP, 0, x));
  }  //fmtSB4u(StringBuilder,int)

  //o = fmtCA08u (a, o, x)
  //sb = fmtSB08u (sb, x)
  //  %08u
  //  8桁10進数変換(符号なし,ゼロサプレスなし)
  public static int fmtCA08u (char[] a, int o, int x) {
    if (x < 0 || 99999999 < x) {
      x = 99999999;
    }
    //perl optdiv.pl 99999999 10000
    //  x/10000==x*109951163>>>40 (0<=x<=494389998) [99999999*109951163==10995116190048837]
    int h = (int) ((long) x * 109951163L >>> 40);  //x/10000
    return fmtCA04u (a, fmtCA04u (a, o, h), x - h * 10000);
  }  //fmtCA08u(char[],int,int)
  public static StringBuilder fmtSB08u (StringBuilder sb, int x) {
    return sb.append (FMT_TEMP, 0, fmtCA08u (FMT_TEMP, 0, x));
  }  //fmtSB08u(StringBuilder,int)

  //o = fmtCA8u (a, o, x)
  //sb = fmtSB8u (sb, x)
  //  %8u
  //  8桁10進数変換(符号なし,ゼロサプレスあり)
  public static int fmtCA8u (char[] a, int o, int x) {
    if (x < 0 || 99999999 < x) {
      x = 99999999;
    }
    if (x <= 9999) {  //1～4桁
      return fmtCA4u (a, o, x);
    } else {  //5～8桁
      //perl optdiv.pl 99999999 10000
      //  x/10000==x*109951163>>>40 (0<=x<=494389998) [99999999*109951163==10995116190048837]
      int h = (int) ((long) x * 109951163L >>> 40);  //x/10000
      return fmtCA04u (a, fmtCA4u (a, o, h), x - h * 10000);
    }
  }  //fmtCA8u(char[],int,int)
  public static StringBuilder fmtSB8u (StringBuilder sb, int x) {
    return sb.append (FMT_TEMP, 0, fmtCA8u (FMT_TEMP, 0, x));
  }  //fmtSB8u(StringBuilder,int)

  //o = fmtCAd (a, o, x)
  //sb = fmtSBd (sb, x)
  //  %d
  //  10進数変換(符号あり,ゼロサプレスあり)
  public static int fmtCAd (char[] a, int o, long x) {
    if (x < 0L) {
      x = -x;
      a[o++] = '-';
    }
    if (x <= 99999999L) {  //1～8桁
      return fmtCA8u (a, o, (int) x);
    } else if (x <= 9999999999999999L) {  //9～16桁
      long h = x / 100000000L;
      return fmtCA08u (a, fmtCA8u (a, o, (int) h), (int) (x - h * 100000000L));
    } else {  //17～19桁
      long hh = x / 10000000000000000L;
      x -= hh * 10000000000000000L;
      long h = x / 100000000L;
      return fmtCA08u (a, fmtCA08u (a, fmtCA4u (a, o, (int) hh), (int) h), (int) (x - h * 100000000L));
    }
  }  //fmtCAd(char[],int,long)
  public static StringBuilder fmtSBd (StringBuilder sb, long x) {
    return sb.append (FMT_TEMP, 0, fmtCAd (FMT_TEMP, 0, x));
  }  //fmtSBd(StringBuilder,long)

  //o = fmtCAnd (a, o, n, x)
  //sb = fmtSBnd (sb, n, x)
  //  %*d
  //  n桁10進数変換(符号あり,ゼロサプレスあり)
  //  n桁に収まらないとき右側にはみ出すのでバッファのサイズに注意
  public static int fmtCAnd (char[] a, int o, int n, long x) {
    int t = fmtCAd (a, o, x);  //現在の末尾
    n += o;  //必要な末尾
    if (t < n) {  //余っている
      int i = n;
      while (o < t) {  //右から順に右にずらす
        a[--i] = a[--t];
      }
      while (o < i) {  //左にできた隙間を' 'で埋める
        a[--i] = ' ';
      }
      t = n;
    }
    return t;
  }  //fmtnu(char[],int,int,long)
  public static StringBuilder fmtSBnd (StringBuilder sb, int n, int x) {
    return sb.append (FMT_TEMP, 0, fmtCAnd (FMT_TEMP, 0, n, x));
  }  //fmtSBnu(StringBuilder,int,long)

  //--------------------------------------------------------------------------------
  //10進数文字列解析

  //x = fmtParseInt (s, i, min, max, err)
  //  文字列sのインデックスiから基数10で整数xを読み取る
  //x = fmtParseIntRadix (s, i, min, max, err, radix)
  //  文字列sのインデックスiから基数radixで整数xを読み取る
  //  基数radixは2,8,10,16のいずれかに限る
  //  1文字も読み取れないかmin<=x&&x<=maxでないときはerrを返す
  //  先頭の空白を読み飛ばす
  //  数値の先頭の'$'は16進数の強制指定とみなす
  //  数値の後のゴミは無視する
  public static int fmtParseInt (String s, int i, int min, int max, int err) {
    return fmtParseIntRadix (s, i, min, max, err, 10);
  }  //fmtParseInt(String,int,int,int,int)
  public static int fmtParseIntRadix (String s, int i, int min, int max, int err, int radix) {
    if (s == null) {
      return err;
    }
    int l = s.length ();
    int c = i < l ? s.charAt (i++) : -1;
    //空白を読み飛ばす
    while (c == ' ' || c == '\t') {
      c = i < l ? s.charAt (i++) : -1;
    }
    //符号を読み取る
    int n = 0;
    if (c == '+') {
      c = i < l ? s.charAt (i++) : -1;
    } else if (c == '-') {
      n = 1;
      c = i < l ? s.charAt (i++) : -1;
    }
    //基数を読み取る
    //        2進数の範囲        8進数の範囲       10進数の範囲        16進数の範囲
    //  +    0x3fffffff*2+1     0x0fffffff*8+7     214748364*10+7     0x07ffffff*16+15
    //  -  -(0x40000000*2+0)  -(0x10000000*8+0)  -(214748364*10+8)  -(0x08000000*16+ 0)
    int o;
    int p;
    if (c == '$') {  //16進数
      o = 0x07ffffff + n;
      p = 15 + n & 15;
      radix = 16;
      c = i < l ? s.charAt (i++) : -1;
    } else if (radix == 16) {  //16進数
      o = 0x07ffffff + n;
      p = 15 + n & 15;
    } else if (radix == 8) {  //8進数
      o = 0x0fffffff + n;
      p = 7 + n & 7;
    } else if (radix == 2) {  //2進数
      o = 0x3fffffff + n;
      p = 1 + n & 1;
    } else {  //10進数
      o = 214748364;
      p = 7 + n;
      radix = 10;
    }
    //数値を読み取る
    int x = Character.digit (c, radix);
    if (x < 0) {
      return err;
    }
    c = i < l ? Character.digit (s.charAt (i++), radix) : -1;
    while (c >= 0) {
      int t = x - o;
      if (t > 0 || t == 0 && c > p) {
        return err;
      }
      x = x * radix + c;
      c = i < l ? Character.digit (s.charAt (i++), radix) : -1;
    }
    if (n != 0) {
      x = -x;
    }
    return min <= x && x <= max ? x : err;
  }  //fmtParseIntRadix(String,int,int,int,int,int)



  //========================================================================================
  //$$MAT 数学関数

  //x = matMax3 (x1, x2, x3)
  //x = matMax4 (x1, x2, x3, x4)
  //x = matMax5 (x1, x2, x3, x4, x5)
  //  最大値
  public static long matMax3 (long x1, long x2, long x3) {
    return Math.max (Math.max (x1, x2), x3);
  }  //matMax3(long,long,long)
  public static long matMax4 (long x1, long x2, long x3, long x4) {
    return Math.max (Math.max (x1, x2), Math.max (x3, x4));
  }  //matMax4(long,long,long,long)
  public static long matMax5 (long x1, long x2, long x3, long x4, long x5) {
    return Math.max (Math.max (Math.max (x1, x2), Math.max (x3, x4)), x5);
  }  //matMax5(long,long,long,long,long)

  //x = matMin3 (x1, x2, x3)
  //x = matMin4 (x1, x2, x3, x4)
  //x = matMin5 (x1, x2, x3, x4, x5)
  //  最小値
  public static long matMin3 (long x1, long x2, long x3) {
    return Math.min (Math.min (x1, x2), x3);
  }  //matMin3(long,long,long)
  public static long matMin4 (long x1, long x2, long x3, long x4) {
    return Math.min (Math.min (x1, x2), Math.min (x3, x4));
  }  //matMin4(long,long,long,long)
  public static long matMin5 (long x1, long x2, long x3, long x4, long x5) {
    return Math.min (Math.min (Math.min (x1, x2), Math.min (x3, x4)), x5);
  }  //matMin5(long,long,long,long,long)



  //========================================================================================
  //$$STR 文字列

  //s = encodeUTF8 (s)
  //  UTF-8変換
  //  00000000 00000000 00000000 0xxxxxxx => 00000000 00000000 00000000 0xxxxxxx
  //  00000000 00000000 00000xxx xxyyyyyy => 00000000 00000000 110xxxxx 10yyyyyy
  //  00000000 00000000 xxxxyyyy yyzzzzzz => 00000000 1110xxxx 10yyyyyy 10zzzzzz
  //  00000000 000xxxyy yyyyzzzz zzxxxxxx => 11110xxx 10yyyyyy 10zzzzzz 10xxxxxx
  public static String strEncodeUTF8 (String s) {
    StringBuilder sb = new StringBuilder ();
    int l = s.length ();
    for (int i = 0; i < l; i++) {
      int u = s.charAt (i);
      if (0xd800 <= u && u <= 0xdbff && i + 1 < l) {
        int v = s.charAt (i + 1);
        if (0xdc00 <= v && v <= 0xdfff) {  //surrogate pair
          u = 0x10000 + ((u & 0x3ff) << 10) + (v & 0x3ff);
          i++;
        }
      }
      if ((u & 0xffffff80) == 0) {  //7bit
        sb.append ((char) u);
      } else if ((u & 0xfffff800) == 0) {  //11bit
        u = (0x0000c080 |
             (u & 0x000007c0) << 2 |
             (u & 0x0000003f));
        sb.append ((char) (u >> 8)).append ((char) (u & 0xff));
      } else if ((u & 0xffff0000) == 0 && !(0xd800 <= u && u <= 0xdfff)) {  //16bit except broken surrogate pair
        u = (0x00e08080 |
             (u & 0x0000f000) << 4 |
             (u & 0x00000fc0) << 2 |
             (u & 0x0000003f));
        sb.append ((char) (u >> 16)).append ((char) ((u >> 8) & 0xff)).append ((char) (u & 0xff));
      } else if ((u & 0xffe00000) == 0) {  //21bit
        u = (0xf0808080 |
             (u & 0x001c0000) << 6 |
             (u & 0x0003f000) << 4 |
             (u & 0x00000fc0) << 2 |
             (u & 0x0000003f));
        sb.append ((char) ((u >> 24) & 0xff)).append ((char) ((u >> 16) & 0xff)).append ((char) ((u >> 8) & 0xff)).append ((char) (u & 0xff));
      } else {  //out of range or broken surrogate pair
        sb.append ((char) 0xef).append ((char) 0xbf).append ((char) 0xbd);  //U+FFFD REPLACEMENT CHARACTER
      }
    }
    return sb.toString ();
  }  //encodeUTF8(String)

  //s = decodeUTF8 (s)
  //  UTF-8逆変換
  //  00000000 00000000 00000000 0xxxxxxx => 00000000 00000000 00000000 0xxxxxxx
  //  00000000 00000000 110xxxxx 10yyyyyy => 00000000 00000000 00000xxx xxyyyyyy
  //  00000000 1110xxxx 10yyyyyy 10zzzzzz => 00000000 00000000 xxxxyyyy yyzzzzzz
  //  11110xxx 10yyyyyy 10zzzzzz 10xxxxxx => 00000000 000xxxyy yyyyzzzz zzxxxxxx
  public static String strDecodeUTF8 (String s) {
    StringBuilder sb = new StringBuilder ();
    int l = s.length ();
    for (int i = 0; i < l; i++) {
      int c = s.charAt (i) & 0xff;
      for (int k = ((c & 0x80) == 0x00 ? 0 :  //0xxxxxxx 7bit
                    (c & 0xe0) == 0xc0 ? 1 :  //110xxxxx 11bit
                    (c & 0xf0) == 0xe0 ? 2 :  //1110xxxx 16bit
                    (c & 0xf8) == 0xf0 ? 3 :  //11110xxx 21bit
                    -1);  //not supported
           --k >= 0; ) {
        c = c << 8 | (i + 1 < l ? s.charAt (++i) & 0xff : 0);
      }
      int u = ((c & 0xffffff80) == 0x00000000 ? c :
               (c & 0xffffe0c0) == 0x0000c080 ? ((c & 0x00001f00) >> 2 |
                                                 (c & 0x0000003f)) :
               (c & 0xfff0c0c0) == 0x00e08080 ? ((c & 0x000f0000) >> 4 |
                                                 (c & 0x00003f00) >> 2 |
                                                 (c & 0x0000003f)) :
               (c & 0xf8c0c0c0) == 0xf0808080 ? ((c & 0x07000000) >> 6 |
                                                 (c & 0x003f0000) >> 4 |
                                                 (c & 0x00003f00) >> 2 |
                                                 (c & 0x0000003f)) :
               0xfffd);  //U+FFFD REPLACEMENT CHARACTER
      if (u <= 0x0000ffff) {
        sb.append (0xd800 <= u && u <= 0xdfff ? '\ufffd' :  //U+FFFD REPLACEMENT CHARACTER
                   (char) u);
      } else if (u <= 0x0010ffff) {
        u -= 0x000010000;
        sb.append ((char) (0xd800 + ((u >> 10) & 0x3ff))).append ((char) (0xdc00 + (u & 0x3ff)));
      }
    }
    return sb.toString ();
  }  //decodeUTF8(String)

  //uri = strEncodeURI (s)
  //  URI変換
  //  UTF-8変換を行ってからRFC3986のPercent-Encodingを行う
  //  フォームの送信に使用されるapplication/x-www-form-urlencodedではない。" "は"+"ではなく"%20"に変換される
  public static final int[] IsURIChar = {  //URIに使える文字。RFC3986のUnreserved Characters。[-.0-9A-Z_a-z~]
    //00000000 00000000 11111111 11111111
    //01234567 89abcdef 01234567 89abcdef
    0b00000000_00000000_00000000_00000000,  //0x00..0x1f
    0b00000000_00000110_11111111_11000000,  //0x20..0x3f [-.0-9]
    0b01111111_11111111_11111111_11100001,  //0x40..0x5f [A-Z_]
    0b01111111_11111111_11111111_11100010,  //0x60..0x7f [a-z~]
  };
  public static String strEncodeURI (String s) {
    s = strEncodeUTF8 (s);  //UTF-8変換
    StringBuilder sb = new StringBuilder ();
    int l = s.length ();
    for (int i = 0; i < l; i++) {
      int c = s.charAt (i);
      if (c < 0x80 && IsURIChar[c >> 5] << c < 0) {  //URIに使える文字
        sb.append ((char) c);
      } else {
        fmtHex2 (sb.append ('%'), c);
      }
    }
    return sb.toString ();
  }  //encodeURI(String)

  //s = strDecodeURI (s)
  //  URI逆変換
  //  RFC3986のPercent-Encodingの逆変換を行ってからUTF-8逆変換を行う
  //  フォームの送信に使用されるapplication/x-www-form-urlencodedではない。"+"は" "に変換されない
  public static final byte[] strIsHexChar = {  //16進数に使えるASCII文字。[0-9A-Fa-f]
    // 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f 10 11 12 13 14 15 16 17 18 19 1a 1b 1c 1d 1e 1f
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  //0x00..0x1f
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1,  //0x20..0x3f [0-9]
    -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  //0x40..0x5f [A-F]
    -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  //0x60..0x7f [a-f]
  };
  public static String strDecodeURI (String s) {
    StringBuilder sb = new StringBuilder ();
    int l = s.length ();
    for (int i = 0; i < l; i++) {
      int c = s.charAt (i);
      if (c == '%' && i + 2 < l) {
        int d = s.charAt (i + 1);
        int e = s.charAt (i + 2);
        if (d < 0x80 && (d = strIsHexChar[d]) >= 0 &&
            e < 0x80 && (e = strIsHexChar[e]) >= 0) {
          sb.append ((char) (d << 4 | e));
        } else {
          sb.append ((char) c);
        }
      } else {
        sb.append ((char) c);
      }
    }
    return sb.toString ();
  }  //decodeURI(String)



  //========================================================================================
  //$$IMG イメージ

  //image = createImage (width, height, pattern, rgb, ...)
  //  イメージを作る
  public static BufferedImage createImage (int width, int height, String pattern, int... rgbs) {
    BufferedImage image = new BufferedImage (width, height, BufferedImage.TYPE_INT_RGB);
    int[] bitmap = ((DataBufferInt) image.getRaster ().getDataBuffer ()).getData ();
    int length = width * height;
    for (int i = 0; i < length; i++) {
      char c = pattern.charAt (i);
      bitmap[i] = rgbs[c < '0' ? 0 : Character.digit (c, 16)];
    }
    return image;
  }  //createImage(int,int,String,int...)

  //icon = createImageIcon (width, height, pattern, rgb, ...)
  //  イメージアイコンを作る
  public static ImageIcon createImageIcon (int width, int height, String pattern, int... rgbs) {
    return new ImageIcon (createImage (width, height, pattern, rgbs));
  }  //createImageIcon(int,int,String,int...)

  //paint = createTexturePaint (width, height, pattern, rgb, ...)
  //  テクスチャペイントを作る
  public static TexturePaint createTexturePaint (int width, int height, String pattern, int... rgbs) {
    return new TexturePaint (createImage (width, height, pattern, rgbs), new Rectangle (0, 0, width, height));
  }  //createTexturePaint(int,int,String,int...)

  //image = loadImage (name)
  //  イメージを読み込む
  public static BufferedImage loadImage (String name) {
    BufferedImage image = null;
    try {
      image = ImageIO.read (new File (name));
    } catch (Exception e) {
    }
    return image;
  }  //loadImage(String)

  //sucess = saveImage (image, name)
  //sucess = saveImage (image, name, quality)
  //  イメージを書き出す
  public static boolean saveImage (BufferedImage image, String name) {
    return saveImage (image, name, 0.75F);
  }  //saveImage(BufferedImage,String)
  public static boolean saveImage (BufferedImage image, String name, float quality) {
    int index = name.lastIndexOf (".");
    if (index < 0) {  //拡張子がない
      return false;
    }
    if (name.substring (index).equalsIgnoreCase (".ico")) {  //アイコンファイルの作成
      return saveIcon (name, image);
    }
    Iterator<ImageWriter> iterator = ImageIO.getImageWritersBySuffix (name.substring (index + 1));  //拡張子に対応するImageWriterがないときは空のIteratorを返す
    if (!iterator.hasNext ()) {  //拡張子に対応するImageWriterがない
      return false;
    }
    ImageWriter imageWriter = iterator.next ();
    ImageWriteParam imageWriteParam = imageWriter.getDefaultWriteParam ();
    if (imageWriteParam.canWriteCompressed ()) {
      imageWriteParam.setCompressionMode (ImageWriteParam.MODE_EXPLICIT);
      imageWriteParam.setCompressionQuality (quality);
    }
    try {
      File file = new File (name);
      ImageOutputStream imageOutputStream = ImageIO.createImageOutputStream (file);
      imageWriter.setOutput (imageOutputStream);
      imageWriter.write (null, new IIOImage (image, null, null), imageWriteParam);
      imageOutputStream.close ();
    } catch (Exception e) {
      //e.printStackTrace ();
      return false;
    }
    return true;
  }  //saveImage(BufferedImage,String,float)



  //========================================================================================
  //$$ICO アイコンファイル
  //  サイズの異なる複数のアイコンを格納することができる
  //  ワードサイズとロングサイズのデータはすべてリトルエンディアン
  //
  //  アイコンファイル(.ico)
  //    ファイルヘッダ
  //    エントリデータ0
  //           :
  //    エントリデータn-1
  //    イメージデータ0
  //           :
  //    イメージデータn-1
  //
  //  ファイルヘッダ
  //    0000  .iw  0=予約
  //    0002  .iw  1=アイコン
  //    0004  .iw  n=アイコン数
  //    0006
  //
  //  エントリデータ
  //    0000  .b   幅
  //    0001  .b   高さ
  //    0002  .b   色数(0=256以上)
  //    0003  .b   0=予約
  //    0004  .iw  1=プレーン数
  //    0006  .iw  ピクセル毎のビット数
  //    0008  .il  イメージデータの長さ
  //    000c  .il  ファイルヘッダからイメージデータまでのオフセット
  //    0010
  //
  //  イメージデータ
  //    イメージヘッダ
  //    パレットテーブル
  //    パターンデータ
  //    マスクデータ
  //
  //  イメージヘッダ
  //    0000  .il  40=イメージヘッダの長さ
  //    0004  .il  幅
  //    0008  .il  高さ*2
  //    000c  .iw  1=プレーン数
  //    000e  .iw  ピクセル毎のビット数
  //    0010  .il  0=無圧縮
  //    0014  .il  0=画像データサイズ(省略)
  //    0018  .il  0=横解像度(省略)
  //    001c  .il  0=縦解像度(省略)
  //    0020  .il  p=パレット数
  //    0024  .il  0=重要なパレットのインデックス(省略)
  //    0028
  //
  //  パレットテーブル
  //    0000  .il  パレット0(BGR0)
  //                   :
  //          .il  パレットp-1(BGR0)
  //    4*p
  //
  //  パターンデータ
  //    ピクセルの順序は左下から右上
  //    バイト内のビットの順序は上位から下位
  //    1ラインのデータ長は4の倍数
  //
  //  マスクデータ
  //    0=描画,1=透過
  //    ピクセルの順序は左下から右上
  //    バイト内のビットの順序は上位から下位
  //    1ラインのデータ長は4の倍数

  //success = saveIcon (name, image, ...);
  //  アイコンファイルを出力する
  public static boolean saveIcon (String fileName, BufferedImage... arrayImage) {
    int iconCount = arrayImage.length;
    int[][] arrayPaletTable = new int[iconCount][];  //パレットテーブル
    int[] arrayPaletCount = new int[iconCount];  //パレット数。0=パレットを使わない
    int[] arrayPixelBits = new int[iconCount];  //ピクセル毎のビット数。24,8,4,2,1のいずれか
    int[] arrayPatternLineSize = new int[iconCount];  //パターンデータの1ラインのバイト数
    int[] arrayMaskLineSize = new int[iconCount];  //マスクデータの1ラインのバイト数
    int[] arrayImageSize = new int[iconCount];  //イメージデータの長さ
    int[] arrayImageOffset = new int[iconCount];  //ファイルヘッダからイメージデータまでのオフセット
    int fileSize = 6 + 16 * iconCount;
    for (int iconNumber = 0; iconNumber < iconCount; iconNumber++) {
      BufferedImage image = arrayImage[iconNumber];
      int width = image.getWidth ();
      int height = image.getHeight ();
      //パレットテーブルを作る
      int[] paletTable = new int[256];
      int paletCount = 0;
    countPalet:
      for (int y = height - 1; y >= 0; y--) {
        for (int x = 0; x < width; x++) {
          int rgb = image.getRGB (x, y);
          if (rgb >>> 24 != 0xff) {  //alphaが0xffでなければred,green,blueを無視して透過色とみなす
            continue;
          }
          int l = 0;
          int r = paletCount;
          while (l < r) {
            int m = l + r >> 1;
            if (paletTable[m] < rgb) {
              l = m + 1;
            } else {
              r = m;
            }
          }
          if (l == paletCount || paletTable[l] != rgb) {  //新しい色
            if (paletCount == 256) {  //色数が多すぎる
              paletCount = 0;
              break countPalet;
            }
            for (int i = paletCount; i > l; i--) {
              paletTable[i] = paletTable[i - 1];
            }
            paletTable[l] = rgb;
            paletCount++;
          }
        }  //for x
      }  //for y
      int pixelBits = (paletCount == 0 ? 24 :
                       paletCount > 16 ? 8 :
                       paletCount > 4 ? 4 :
                       paletCount > 2 ? 2 :
                       1);
      int patternLineSize = pixelBits * width + 31 >> 5 << 2;
      int maskLineSize = width + 31 >> 5 << 2;
      int imageSize = 40 + 4 * paletCount + patternLineSize * height + maskLineSize * height;
      arrayPaletTable[iconNumber] = paletTable;
      arrayPaletCount[iconNumber] = paletCount;
      arrayPixelBits[iconNumber] = pixelBits;
      arrayPatternLineSize[iconNumber] = patternLineSize;
      arrayMaskLineSize[iconNumber] = maskLineSize;
      arrayImageSize[iconNumber] = imageSize;
      arrayImageOffset[iconNumber] = fileSize;
      fileSize += imageSize;
    }  //for iconNumber
    byte[] bb = new byte[fileSize];
    //ファイルヘッダ
    ByteArray.byaWiw (bb, 0, 0);
    ByteArray.byaWiw (bb, 2, 1);
    ByteArray.byaWiw (bb, 4, iconCount);
    for (int iconNumber = 0; iconNumber < iconCount; iconNumber++) {
      BufferedImage image = arrayImage[iconNumber];
      int width = image.getWidth ();
      int height = image.getHeight ();
      int[] paletTable = arrayPaletTable[iconNumber];
      int paletCount = arrayPaletCount[iconNumber];
      int pixelBits = arrayPixelBits[iconNumber];
      int patternLineSize = arrayPatternLineSize[iconNumber];
      int maskLineSize = arrayMaskLineSize[iconNumber];
      int imageSize = arrayImageSize[iconNumber];
      int imageOffset = arrayImageOffset[iconNumber];
      //エントリデータ
      int o = 6 + 16 * iconNumber;
      ByteArray.byaWb (bb, o, width);
      ByteArray.byaWb (bb, o + 1, height);
      ByteArray.byaWb (bb, o + 2, paletCount);
      ByteArray.byaWb (bb, o + 3, 0);
      ByteArray.byaWiw (bb, o + 4, 1);
      ByteArray.byaWiw (bb, o + 6, pixelBits);
      ByteArray.byaWil (bb, o + 8, imageSize);
      ByteArray.byaWil (bb, o + 12, imageOffset);
      //イメージヘッダ
      o = imageOffset;
      ByteArray.byaWil (bb, o, 40);
      ByteArray.byaWil (bb, o + 4, width);
      ByteArray.byaWil (bb, o + 8, height * 2);
      ByteArray.byaWiw (bb, o + 12, 1);
      ByteArray.byaWiw (bb, o + 14, pixelBits);
      ByteArray.byaWil (bb, o + 16, 0);
      ByteArray.byaWil (bb, o + 20, 0);
      ByteArray.byaWil (bb, o + 24, 0);
      ByteArray.byaWil (bb, o + 28, 0);
      ByteArray.byaWil (bb, o + 32, paletCount);
      ByteArray.byaWil (bb, o + 36, 0);
      //パレットテーブル
      o += 40;
      for (int i = 0; i < paletCount; i++) {
        ByteArray.byaWil (bb, o, paletTable[i] & 0x00ffffff);
        o += 4;
      }
      //パターンデータ
      for (int y = height - 1; y >= 0; y--) {
        for (int x = 0; x < width; x++) {
          int rgb = image.getRGB (x, y);
          if (rgb >>> 24 != 0xff) {  //alphaが0xffでなければred,green,blueを無視して透過色とみなす
            continue;
          }
          if (pixelBits == 24) {  //パレットなし
            bb[o + 3 * x] = (byte) rgb;  //blue
            bb[o + 3 * x + 1] = (byte) (rgb >> 8);  //green
            bb[o + 3 * x + 2] = (byte) (rgb >> 16);  //red
            continue;
          }
          int l = 0;
          int r = paletCount;
          while (l < r) {
            int m = l + r >> 1;
            if (paletTable[m] < rgb) {
              l = m + 1;
            } else {
              r = m;
            }
          }
          if (l != 0) {
            if (pixelBits == 8) {
              bb[o + x] = (byte) l;
            } else if (pixelBits == 4) {
              bb[o + (x >> 1)] |= (byte) (l << ((~x & 1) << 2));
            } else if (pixelBits == 2) {
              bb[o + (x >> 2)] |= (byte) (l << ((~x & 3) << 1));
            } else {
              bb[o + (x >> 3)] |= (byte) (l << (~x & 7));
            }
          }
        }  //for x
        o += patternLineSize;
      }  //for y
      //マスクデータ
      for (int y = height - 1; y >= 0; y--) {
        for (int x = 0; x < width; x++) {
          int rgb = image.getRGB (x, y);
          if (rgb >>> 24 != 0xff) {  //alphaが0xffでなければred,green,blueを無視して透過色とみなす
            bb[o + (x >> 3)] |= (byte) (1 << (~x & 7));
          }
        }
        o += maskLineSize;
      }
    }  //for iconNumber
    return rscPutFile (fileName, bb, 0, fileSize);
  }  //saveIcon(String,BufferedImage...)



}  //class XEiJ



